# -*- coding: iso-8859-1 -*-

__doc__ = """ This routine compares graphically the
statistics (bias and std dev) of surf validation files produced by
hofx or validation runs . It can be used to verify the assimilation
performances

Input: numpy files with validation statistics for a single experiment
Output: png plots of vertically and latitude separated bias and std dev
"""

__author__ = "Emanuele Emili - April 2012"

import os
os.environ['QT_QPA_PLATFORM']='offscreen'
import numpy as np
import seaborn as sns
from matplotlib import pylab as plt
from matplotlib.ticker import AutoMinorLocator, MaxNLocator
# from taylorDiagram import TaylorDiagram
from sys import version_info
import glob
from datetime import datetime, timedelta
import matplotlib.colors as mcolors
from mpl_toolkits.basemap import Basemap
import copy

def PlotTaylor(fig,data,plotcfg,subplotn):
    
    stdrefs = data['stdo'][0]
    samples = [[triade[0],triade[1],triade[2]] for triade in zip(data['stdm'],data['corr'],plotcfg['labels'])]
    colors = plt.matplotlib.cm.Dark2(numpy.linspace(0,1,len(data['stdm'])))
    # markers = ["o","v","D","p","P","s"]
    
    #plt.grid(False,linestyle='--')
    dia = TaylorDiagram(stdrefs, fig=fig, rect=subplotn,
                        label='Reference',srange=(0.2, 2.2))
    #dia._ax.grid(True,linestyle='--')
    # Add samples to Taylor diagram
    for i,(stddev,corrcoef,name) in enumerate(samples):
        dia.add_sample(stddev, corrcoef,
                       # marker=markers[i], ms=10, ls='',
                       #mfc='k', mec='k', # B&W
                       mfc=colors[i], mec=colors[i], # Colors
                       label=name,zorder=10+i)
        
    # Add RMS contours, and label them
    contours = dia.add_contours(levels=5, colors='0.5') # 5 levels
    dia.ax.clabel(contours, inline=1, fontsize=10, fmt='%.1f')
    # Tricky: ax is the polar ax (used for plots), _ax is the
    # container (used for layout)
    if 'title' in plotcfg: dia._ax.set_title(plotcfg['title'])
    
    return dia

def setmap(ax):

    m = Basemap(projection='cyl',resolution='l')
    m.drawmapboundary()
    m.fillcontinents()
    ulx = -26.0
    uly = 72.0
    lrx = 46.0
    lry = 28.0
    xmin, ymin = m(ulx, lry)
    xmax, ymax = m(lrx, uly)
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    return m
    
def ReadTimeAverages():
    
    line, hline, hhline = '','',''
    for col in columns:

        fig = plt.figure(figsize=figsize,dpi=200)
        markersize = 3

        # Find the appropriate file
        for nf,fdir in enumerate(dirlist):
            allavg = np.ma.empty((0,),dtype=np.float32)
            allstd = np.ma.empty((0,),dtype=np.float32)
            allrmse = np.ma.empty((0,),dtype=np.float32)

            # delta = datetime.strptime(dates[nf].split('_')[0],'%Y%m%dT%H')-datetime.strptime(cyclerefdate,'%Y%m%dT%H')
            lncycle = 0

            for ic in range(1,ncycl+1):

                if skipdays and ic in (8,9): continue

                lfdir = fdir.replace('cycle_%04i' % (1),'cycle_%04i' % (ic))
                ldate = '%s_%s' % ((datetime.strptime(dates[nf].split('_')[0],'%Y%m%dT%H')+dtcycl*(ic-1)).strftime('%Y%m%dT%H'),\
                    (datetime.strptime(dates[nf].split('_')[1],'%Y%m%dT%H')+dtcycl*(ic-1)).strftime('%Y%m%dT%H'))
                
                # print('Searching for %s%s_AvgConcMap_%s.npz' % \
                #                              (instrtype,col.replace('.',''),ldate))

                filelist = sorted(glob.glob(\
                                os.path.join(lfdir,domain,statstype[nf],\
                                             "%s%s_AvgConcMap_%s.npz" % \
                                             (instrtype,col.replace('.',''),ldate))))
                if len(filelist) != 1 :
                    print("Anomaly: did not find a unique filename for %s %s %s in %s" % (statstype[nf],col,ldate,lfdir))
                    pass
                else:
                    lncycle += 1
                    lfile = filelist[0]
                    data = np.load(lfile,allow_pickle=True)
                
                    avg = np.ma.masked_values(data['avg'],data['missing_value'])
                    std = np.ma.masked_values(data['std'],data['missing_value'])
                    avg = np.ma.masked_equal(avg,1e20)
                    std = np.ma.masked_equal(std,1e20)
                    rmse = np.ma.sqrt(avg**2+std**2)

                    if 'PercentOf' in statstype[nf]:
                        avg*=100
                        std*=100
                        rmse*=100
                    else:
                        avg*=1e9
                        std*=1e9
                        rmse*=1e9

                    if ic == 1:
                        cyclavg, cyclstd, cyclrmse, nobs = avg.compressed(), std.compressed(), rmse.compressed(), np.ones(avg.count())
                        lonref, latref = data['lon'][~avg.mask][0], data['lat'][~avg.mask][0]
                        refpos = [ (lat,lon) for lat, lon in zip(lonref,latref) ]
                    else:
                        # Find common sites in the latest dataset, sum them up and increase the obs counter only for those. Later I might include also
                        # sites that are not already there on the first day. But I don't care much.
                        newpos = [ (lat,lon) for lat, lon in zip(data['lon'][~avg.mask][0],data['lat'][~avg.mask][0])]
                        indexnew = []
                        indexref = []
                        for iloc,loc in enumerate(refpos):
                            if loc in newpos:
                                indexref.append(iloc)
                                indexnew.append(newpos.index(loc))

                        cyclavg[indexref] = cyclavg[indexref]+avg.compressed()[indexnew]
                        cyclstd[indexref] = cyclstd[indexref]+std.compressed()[indexnew]
                        cyclrmse[indexref] = cyclrmse[indexref]+rmse.compressed()[indexnew]
                        nobs[indexref] += 1

                # Save the first dataset for later comparison of two experiments
                if nf == 0:
                    refavg = copy.deepcopy(cyclavg)
                    refstd = copy.deepcopy(cyclstd)
                    refrmse = copy.deepcopy(cyclrmse)

            if 'PercentOf' in statstype[0]:
                units = '%'
                maxrmse, minrmse = 150, 20
                maxstd, minstd = 100, 20
                maxbias, minbias = 80, -80
                # maxrmse, minrmse = 40, 0
                # maxstd, minstd = 40, 0
                # maxbias, minbias = 60, -60
            else:
                units = 'ppbv'
                maxrmse, minrmse = 30, 0
                maxstd, minstd = 20, 0
                maxbias, minbias = -20, 20

            if (nobs < lncycle).sum()/nobs.size*100 > 10:
                print('Percentage of sites with less than ncycles: %4.3f' % ((nobs < lncycle).sum()/nobs.size*100))

            z = cyclavg/nobs
            z = np.ma.masked_where(nobs < lncycle*0.75, z)
            ax = fig.add_subplot(3,4,nf+1)
            m = setmap(ax)
            sc = m.scatter(lonref, latref,markersize,marker='o', \
                   c=z, cmap="seismic",\
                   latlon=True,zorder=10,vmin=minbias,vmax=maxbias)
            plt.colorbar(sc,orientation='horizontal')
            ax.set_title('%s %s Bias N=%i \nMed: %.1f Min: %.1f Max: %.1f (%s)' % (instrtype,col,z.count(),\
            np.ma.median(z),\
            z.min(),z.max(),\
            units))

            z = cyclstd/nobs
            z = np.ma.masked_where(nobs < lncycle*0.75, z)
            ax = fig.add_subplot(3,4,4+nf+1)
            m = setmap(ax)
            sc = m.scatter(lonref, latref,markersize,marker='o', \
                   c=z, cmap="YlOrRd",\
                   latlon=True,zorder=10,vmin=minstd,vmax=maxstd)
            plt.colorbar(sc,orientation='horizontal')
            ax.set_title('%s %s Std N=%i \nMed: %.1f Min: %.1f Max: %.1f (%s)' % (instrtype,col,z.count(),\
            np.ma.median(z),\
            z.min(),z.max(),\
            units))

            z = cyclrmse/nobs
            z = np.ma.masked_where(nobs < lncycle*0.75, z)
            ax = fig.add_subplot(3,4,8+nf+1)
            m = setmap(ax)
            sc = m.scatter(lonref, latref,markersize,marker='o', \
                   c=z, cmap="YlOrRd",\
                   latlon=True,zorder=10,vmin=minrmse,vmax=maxrmse)
            plt.colorbar(sc,orientation='horizontal')
            ax.set_title('%s %s Rmse N=%i \nMed: %.1f Min: %.1f Max: %.1f (%s)' % (instrtype,col,z.count(),\
            np.ma.median(z),\
            z.min(),z.max(),\
            units))

        if 'PercentOf' in statstype[0]:
            units = '%'
            maxrmse, minrmse = 50, -50
            maxstd, minstd = 50, -50
            maxbias, minbias = 50, -50
        else:
            units = 'ppbv'
            maxrmse, minrmse = 8, -8
            maxstd, minstd = 8, -8
            maxbias, minbias = 8, -8

        hline = hline + '%s & ' % col
        hhline = hhline + 'Bias gain & Std gain & Rmse gain (%s) & ' % (units)
        z = np.ma.abs(cyclavg/nobs)-np.ma.abs(refavg/nobs)
        z = np.ma.masked_where(nobs < lncycle*0.75, z)
        ax = fig.add_subplot(3,4,3)
        m = setmap(ax)
        sc = m.scatter(lonref, latref,markersize,marker='o', \
               c=z, cmap="seismic",\
               latlon=True,zorder=10,vmin=minbias,vmax=maxbias)
        plt.colorbar(sc,orientation='horizontal')
        ax.set_title('%s %s Bias Gain N=%i \nMed: %.1f Min: %.1f Max: %.1f (%s)' % (instrtype,col,z.count(),\
        np.ma.median(z),\
        z.min(),z.max(),\
        units))
        line = line + '%.1f & ' % (np.ma.median(z))

        ax = fig.add_subplot(3,4,4)
        ax.set_title("%s %s Median: %.1f" % (instrtype,col,np.ma.median(z)))
        ax.hist(z,bins=20,density=True,range=(minbias,maxbias))
        ax.vlines(np.ma.median(z),0,ymax=ax.get_ylim()[1],linewidths=5,colors='r')

        z = (cyclstd-refstd)/nobs
        z = np.ma.masked_where(nobs < lncycle*0.75, z)
        ax = fig.add_subplot(3,4,7)
        m = setmap(ax)
        sc = m.scatter(lonref, latref,markersize,marker='o', \
               c=z, cmap="seismic",\
               latlon=True,zorder=10,vmin=minstd,vmax=maxstd)
        plt.colorbar(sc,orientation='horizontal')
        ax.set_title('%s %s Std Gain N=%i \nMed: %.1f Min: %.1f Max: %.1f (%s)' % (instrtype,col,z.count(),\
        np.ma.median(z),\
        z.min(),z.max(),\
        units))
        line = line + '%.1f & ' % (np.ma.median(z))

        ax = fig.add_subplot(3,4,8)
        ax.set_title("%s %s Median: %.1f" % (instrtype,col,np.ma.median(z)))
        ax.hist(z,bins=20,density=True,range=(minstd,maxstd))
        ax.vlines(np.ma.median(z),0,ymax=ax.get_ylim()[1],linewidths=5,colors='r')

        z = (cyclrmse-refrmse)/nobs
        z = np.ma.masked_where(nobs < lncycle*0.75, z)
        ax = fig.add_subplot(3,4,11)
        m = setmap(ax)
        sc = m.scatter(lonref, latref,markersize,marker='o', \
               c=z, cmap="seismic",\
               latlon=True,zorder=10,vmin=minrmse,vmax=maxrmse)
        plt.colorbar(sc,orientation='horizontal')
        ax.set_title('%s %s Rmse Gain N=%i \nMed: %.1f Min: %.1f Max: %.1f (%s)' % (instrtype,col,z.count(),\
        np.ma.median(z),\
        z.min(),z.max(),\
        units))
        line = line + '%.1f & ' % (np.ma.median(z))

        ax = fig.add_subplot(3,4,12)
        ax.set_title("%s %s Median: %.1f" % (instrtype,col,np.ma.median(z)))
        ax.hist(z,bins=20,density=True,range=(minrmse,maxrmse))
        ax.vlines(np.ma.median(z),0,ymax=ax.get_ylim()[1],linewidths=5,colors='r')

        fig.savefig(os.path.join(outputdir,fname+'_'+col.replace('.','_')))

        plt.close()

    print(hline)
    print(hhline)
    print(line)

if __name__ == "__main__":
    
    sns.set() # Apply seaborn beautificator
        
    dirlist = []

    # OSSE

    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_35/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_37/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_42/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_43/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_44/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_00/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/perturbed_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_31/model_error_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/model_error_truth/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_00/3dvar/FINAL_FORECAST/hofx/obs_plots')

    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/perturbed_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_006/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_010/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_001/cycle_0001/model_error_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_001/cycle_0001/model_error_truth/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_004/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_003/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    
    # Real exp
    dirlist.append('/scratch/globc/emili/seeds/real/re_001/cycle_0001/forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/re_001/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/3dvar_01/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/3dvar_02/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    dirlist.append('/scratch/globc/emili/seeds/real/re_045/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real_test/long_03/cycle_0001/forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real_test/envar_014/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real_test/envar_015/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real_test/long_03/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real_test/envar_015/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real_test/envar_007/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real_test/envar_008/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real_test/envar_013/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/envar_20/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/envar_19/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/envar_11/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/full_test/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/full_test/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/simu/exp_59/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/simu/exp_61/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/full_test/cycle_0001/model_error_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/full_test/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/full_test/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')

    # Long exp

    # dirlist.append('/scratch/globc/emili/seeds/long_05/cycle_0001/forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_06/cycle_0001/forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_02/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_02/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_05/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_06/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_05/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_06/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_05/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_06/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_summer_allVal/cycle_0001/forecast/obs_plots')

    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_spring_allVal/cycle_0001/forecast/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_04/cycle_0001/forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_04/cycle_0001/model_error_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_04/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_05/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_07/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_05/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_07/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')

    # dirlist.append('/scratch/globc/emili/seeds/long_02/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/model_error_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')

    # Ensembles
    # ensize = 30
    # for ie in range(1,ensize+1): dirlist.append('/scratch/globc/jonville/seeds/ensembles/ens_emispertread_eps0/cycle_0001/ensemble/E%04i/obs_plots' % (ie))
    # for ie in range(1,ensize+1): dirlist.append('/scratch/globc/jonville/seeds/ensembles/ens_spring_eps0/cycle_0001/ensemble/E%04i/obs_plots' % (ie))
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_summer/cycle_0001/forecast/obs_plots')
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_summer/cycle_0001/forecast/obs_plots')
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_spring/cycle_0001/forecast/obs_plots')
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_spring/cycle_0001/forecast/obs_plots')

    # ensize = 40
    # for ie in range(1,ensize+1): dirlist.append('/archive/pae/emili/seeds/ensembles/ens_05/cycle_0001/ensemble/E%04i/obs_plots' % (ie))
    # dirlist.append('/archive/pae/emili/seeds/simu/exp_00/forecast/obs_plots')

    linenames = []
    # linenames.append('truth')
    # linenames.append('truth')
    # linenames.append('free_forecast')
    linenames.append('forecast')
    # linenames.append('3dvar_long')
    # linenames.append('3dvar_01')
    # linenames.append('3dvar')
    linenames.append('envar_045')
    # linenames.append('envar_007')
    # linenames.append('3dvar')
    # linenames.append('analysis')
    # linenames.append('model_error_forecast')
    # linenames.append('3dvar_final_forecast')
    # linenames.append('envar_final_forecast_nobc')
    # linenames.append('final_forecast_corrected')
    # linenames.append('3d_analysis')
    # linenames.append('perturbed_forecast_30')

    # linenames.append('model_error_forecast')
    # linenames.append('model_error_truth')
    # linenames.append('final_forecast')
    # linenames.append('final_forecast_corrected_007')

    # for ie in range(1,ensize+1): linenames.append('E%04i' % (ie))
    # linenames.append('control')
    # linenames.append('obs')

    statstype = []
    statstype.append("Misfit_Hx_dry_PercentOfObs")
    statstype.append("Misfit_Hx_dry_PercentOfObs")
    # for ie in range(1,ensize+1): statstype.append("Misfit_Hx_dry_Absolute")
    # statstype.append("Misfit_Hx_dry_Absolute")
    # statstype.append("Misfit_Hx_dry_Absolute")
    # for ie in range(1,ensize+1): statstype.append("Misfit_Hx_dry_PercentOfObs")
    # statstype.append("Misfit_Hx_dry_PercentOfObs")
    # for ie in range(1,ensize+1): statstype.append("Hx_dry")
    # statstype.append("Hx_dry")
    # statstype.append("Y")

    # statstype.append("Hx_dry")
    # statstype.append("Hx_dry")
    # statstype.append("Hx_dry")
    # # statstype.append("Hx_dry")
    # statstype.append("Hx_dry")
    # statstype.append("Hx_dry")

    dates = []
    # # dates.append('20190619T00_20190630T23')
    # # dates.append('20100704T00_20100706T23')
    # dates.append('20190619T00_20190621T23')
    dates.append('20190619T00_20190619T23')
    dates.append('20190619T00_20190619T23')
    # dates.append('20190619T00_20190619T23')
    # # dates.append('20190619T00_20190619T23')
    # # dates.append('20190619T00_20190630T23')
    # # dates.append('20190619T00_20190619T23')
    # dates.append('20190620T00_20190621T23')
    # dates.append('20190620T00_20190621T23')    
    # dates.append('20190619T00_20190619T23')
    # dates.append('20190620T00_20190621T23') 

    # dates.append('20190620T00_20190621T23')
    # dates.append('20190401T00_20190412T23')
    # dates.append('20190401T00_20190401T23')
    # dates.append('20190401T00_20190403T23')
    # dates.append('20190401T00_20190401T23')
    # dates.append('20190402T00_20190403T23')
    # dates.append('20190402T00_20190403T23')
    # dates.append('20190401T00_20190401T23')
    # dates.append('20190402T00_20190403T23')

    # for ie in range(1,ensize+1): dates.append('20190619T00_20190630T23')
    # dates.append('20190619T00_20190630T23')
    # dates.append('20190619T00_20190630T23')
    # for ie in range(1,ensize+1): dates.append('20190401T00_20190412T23')
    # dates.append('20190401T00_20190412T23')
    # dates.append('20190401T00_20190412T23')
    # for ie in range(1,ensize+1): dates.append('20100704T00_20100706T23')
    # dates.append('20100704T00_20100706T23')

    columns = []

    columns.append('O3')
    columns.append('NO2')
    columns.append('PM10')
    columns.append('PM2.5')
    # columns.append('CO')

    colors = ['tab:blue','tab:orange','tab:red','tab:purple','tab:green','tab:brown','tab:olive']
    # cm = plt.get_cmap('viridis')
    # colors = [cm(1.*i/(ensize)) for i in range(ensize)]
    # colors.append('black')
    # colors.append('darkred')
    ncycl = 1
    dtcycl = timedelta(days=1)
    cyclerefdate = '20190619T00'
    skipdays = False
    # cyclerefdate = '20190401T00'
    # skipdays = True

    instrtype = 'AIRBASE_Val_'
    # instrtype = 'AIRBASE_'

    # outputdir = "/archive/pae/emili/seeds/long/exp_03/validation"
    # outputdir = "/archive/pae/emili/seeds/simu/comparisons/validation"
    # outputdir = "/archive/pae/emili/seeds/osse/comparisons/validation"
    # outputdir = "/scratch/globc/emili/seeds/comparisons/validation/envar-bugged"
    # outputdir = "/scratch/globc/emili/seeds/comparisons/validation/long"
    outputdir = "/scratch/globc/emili/seeds/comparisons/validation/real"
    # outputdir = "/scratch/globc/emili/seeds/comparisons/validation/osse"

    if not os.path.exists(outputdir) : os.makedirs(outputdir)

    os.chdir(outputdir)

    # domain = "GEMS02"
    domain = "MACC01"
    
    # expname = 'spring'
    # expname = 'absmisfit'
    # expname = 'summer'
    expname = ''
     # It contains the relevant instrument and date info
    
    # figsize = (15,30)
    figsize = (22,18)
    # figsize = (10,6)

    outfilenames = [string for string in linenames]

    fname = outfilenames[0]
    for string in outfilenames[1:]: fname = fname+'_'+string 
    fname = fname + '_maps'
    if expname != '': fname = fname + '_' + expname
    
    ReadTimeAverages()
        
