# -*- coding: iso-8859-1 -*-

__doc__ = """ This routine compares graphically the
statistics (bias and std dev) of surf validation files produced by
hofx or validation runs . It can be used to verify the assimilation
performances

Input: numpy files with validation statistics for a single experiment
Output: png plots of vertically and latitude separated bias and std dev
"""

__author__ = "Emanuele Emili - April 2012"

import os
os.environ['QT_QPA_PLATFORM']='offscreen'
import numpy as np
import seaborn as sns
from matplotlib import pylab as plt
from matplotlib.ticker import AutoMinorLocator, MaxNLocator
# from taylorDiagram import TaylorDiagram
from sys import version_info
import glob
from datetime import datetime, timedelta
import matplotlib.colors as mcolors
import subprocess

def PlotBars (fig,xpos,x,dictionary):
    width = dictionary['width']
    plt.title(dictionary['title'])
    plt.ylabel(dictionary['ylabel'])
    #print xpos, x
    plt.bar(xpos,x,width,color=plotformat['color'],label=dictionary['label'])
    # add some
    ax = fig.gca()
    ax.set_ylabel(dictionary['ylabel'])
    if len(filelist) == 2: ax.set_xticks([0.2,0.8])
    if len(filelist) == 3: ax.set_xticks([0.3,1.1])
    ax.set_xticklabels(dictionary['xtickslabels'])
    ax.set_ylim(dictionary['yrange'][0],dictionary['yrange'][1])
    ax.set_xlim(dictionary['xrange'][0],dictionary['xrange'][1])

def PlotTaylor(fig,data,plotcfg,subplotn):
    
    stdrefs = data['stdo'][0]
    samples = [[triade[0],triade[1],triade[2]] for triade in zip(data['stdm'],data['corr'],plotcfg['labels'])]
    colors = plt.matplotlib.cm.Dark2(numpy.linspace(0,1,len(data['stdm'])))
    # markers = ["o","v","D","p","P","s"]
    
    #plt.grid(False,linestyle='--')
    dia = TaylorDiagram(stdrefs, fig=fig, rect=subplotn,
                        label='Reference',srange=(0.2, 2.2))
    #dia._ax.grid(True,linestyle='--')
    # Add samples to Taylor diagram
    for i,(stddev,corrcoef,name) in enumerate(samples):
        dia.add_sample(stddev, corrcoef,
                       # marker=markers[i], ms=10, ls='',
                       #mfc='k', mec='k', # B&W
                       mfc=colors[i], mec=colors[i], # Colors
                       label=name,zorder=10+i)
        
    # Add RMS contours, and label them
    contours = dia.add_contours(levels=5, colors='0.5') # 5 levels
    dia.ax.clabel(contours, inline=1, fontsize=10, fmt='%.1f')
    # Tricky: ax is the polar ax (used for plots), _ax is the
    # container (used for layout)
    if 'title' in plotcfg: dia._ax.set_title(plotcfg['title'])
    
    return dia

def ReadEnsTimeSeries():

    for col in columns:
        
        fig = plt.figure(figsize=figsize,dpi=200)
        plt.subplots_adjust(hspace=0.1)

        # plt.subplot(3,1,1)
        plt.subplot(1,1,1)
        fig.autofmt_xdate()
        ax1 = fig.gca()
        ax1.set_xlabel('Date')
        # ax1.set_title(col+' avg')
        ax1.set_title(col+' bias')

        # plt.subplot(3,1,2)
        # fig.autofmt_xdate()
        # ax2 = fig.gca()
        # ax2.set_xlabel('Date')
        # ax2.set_title(col+' std')

        # plt.subplot(3,1,3)
        # fig.autofmt_xdate()
        # ax3 = fig.gca()
        # ax3.set_xlabel('Date')
        # ax3.set_title(col+' rmse')


        # Find the appropriate file
        for nf,fdir in enumerate(dirlist):
            allavg = np.ma.empty((0,),dtype=np.float32)
            allstd = np.ma.empty((0,),dtype=np.float32)
            allrmse = np.ma.empty((0,),dtype=np.float32)
            lw = 1
            # if nf > len(dirlist)-3: lw = 4
            if nf > len(dirlist)-2: lw = 4

            for ic in range(1,ncycl+1):

                fdir = fdir.replace('cycle_%04i' % (ic-1 if ic > 1 else ic),'cycle_%04i' % (ic))
                ldate = '%s_%s' % ((datetime.strptime(dates[nf].split('_')[0],'%Y%m%dT%H')+dtcycl*(ic-1)).strftime('%Y%m%dT%H'),\
                    (datetime.strptime(dates[nf].split('_')[1],'%Y%m%dT%H')+dtcycl*(ic-1)).strftime('%Y%m%dT%H'))
                filelist = sorted(glob.glob(\
                                os.path.join(fdir,domain,statstype[nf],\
                                             "%s%s_ConcTimeSeries_%s.npz" % \
                                             (instrtype,col.replace('.',''),ldate))))
                if len(filelist) != 1 :
                    print("Anomaly: did not find a unique filename for %s %s %s in %s" % (statstype[nf],col,ldate,fdir))
                    pass
                else:
                    lfile = filelist[0]
                    data = np.load(lfile,allow_pickle=True)
                
                    avg = np.ma.masked_values(data['avg'],data['missing_value'])
                    std = np.ma.masked_values(data['std'],data['missing_value'])
                    avg = np.ma.masked_equal(avg,1e20)
                    std = np.ma.masked_equal(std,1e20)
                    rmse = np.ma.sqrt(avg**2+std**2)

                    if 'PercentOf' in statstype[nf]:
                        avg*=100
                        std*=100
                        rmse*=100
                    else:
                        avg*=1e9
                        std*=1e9
                        rmse*=1e9

                    allavg = np.ma.concatenate((allavg,avg))
                    allstd = np.ma.concatenate((allstd,std))
                    allrmse = np.ma.concatenate((allrmse,rmse))
                    if ic < ncycl:
                        # ax1.plot(data['time'],avg,lw=lw,color=colors[nf])
                        # ax2.plot(data['time'],std,lw=lw,color=colors[nf])
                        # ax3.plot(data['time'],rmse,lw=lw,color=colors[nf])
                        ax1.plot(data['time'],avg,lw=lw,color=colors[nf])
                    else:
                        # ax1.plot(data['time'],avg,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(allavg)))
                        # ax2.plot(data['time'],std,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(allstd)))
                        # ax3.plot(data['time'],rmse,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(allrmse)))
                        ax1.plot(data['time'],avg,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(allrmse)))
        
        if 'PercentOf' in statstype[0]:
            ax1.set_ylim(-100,100)
            ax2.set_ylim(0,120)
            ax3.set_ylim(0,120)
        # ax1.legend()
        # ax2.legend()
        plt.savefig(os.path.join(outputdir,fname+'_'+col.replace('.','_')))
        plt.close()
    
def ReadTimeSeries():
    
    combine_plots = True

    for col in columns:
        
        # fig = plt.figure(figsize=figsize,dpi=200)
        # plt.subplots_adjust(hspace=0.1)

        # plt.subplot(3,1,1)
        # # plt.subplot(1,1,1)
        # fig.autofmt_xdate()
        # ax1 = fig.gca()
        # ax1.set_xlabel('Date')
        # # ax1.set_title(col+' avg')
        # ax1.set_title(col+' bias')

        # plt.subplot(3,1,2)
        # fig.autofmt_xdate()
        # ax2 = fig.gca()
        # ax2.set_xlabel('Date')
        # ax2.set_title(col+' std')

        # plt.subplot(3,1,3)
        # fig.autofmt_xdate()
        # ax3 = fig.gca()
        # ax3.set_xlabel('Date')
        # ax3.set_title(col+' rmse')

        fig, (ax1, ax2, ax3) = plt.subplots(3,1,figsize=figsize,dpi=200)
        plt.subplots_adjust(hspace=0.1)

        fig.autofmt_xdate()
        # ax1.set_xlabel('Date')
        # ax1.set_title(col+' avg')
        ax1.set_title(col+' bias')
        # ax2.set_xlabel('Date')
        ax2.set_title(col+' std')
        ax3.set_xlabel('Date')
        ax3.set_title(col+' rmse')

        if pltcycleavg:
            fig2, (ax4, ax5, ax6) = plt.subplots(3,1,figsize=figsize,dpi=200)
            plt.subplots_adjust(hspace=0.1)

            # ax4.set_xlabel('Date')
            # ax1.set_title(col+' avg')
            ax4.set_title(col+' bias')
            # ax5.set_xlabel('Date')
            ax5.set_title(col+' std')
            ax6.set_xlabel('Hour')
            ax6.set_title(col+' rmse')

        # Find the appropriate file
        for nf,fdir in enumerate(dirlist):
            allavg = np.ma.empty((0,),dtype=np.float32)
            allstd = np.ma.empty((0,),dtype=np.float32)
            allrmse = np.ma.empty((0,),dtype=np.float32)
            lw = 1
            # if nf > len(dirlist)-1: lw = 4
            # if nf > len(dirlist)-2: lw = 4

            delta = datetime.strptime(dates[nf].split('_')[0],'%Y%m%dT%H')-datetime.strptime(cyclerefdate,'%Y%m%dT%H')

            lncycle = 0
            for ic in range(1,ncycl+1):

                # Skip a bad cycle
                print('Cycle %02i' % (ic))
                if skipdays and ic in (8,9): continue

                # lfdir = fdir.replace('cycle_%04i' % (ic-1 if ic > 1 else ic),'cycle_%04i' % (ic))
                lfdir = fdir.replace('cycle_%04i' % (1),'cycle_%04i' % (ic))
                ldate = '%s_%s' % ((datetime.strptime(dates[nf].split('_')[0],'%Y%m%dT%H')+dtcycl*(ic-1)).strftime('%Y%m%dT%H'),\
                    (datetime.strptime(dates[nf].split('_')[1],'%Y%m%dT%H')+dtcycl*(ic-1)).strftime('%Y%m%dT%H'))
                
                # print(ldate)
                filelist = sorted(glob.glob(\
                                os.path.join(lfdir,domain,statstype[nf],\
                                             "%s%s_ConcTimeSeries_%s.npz" % \
                                             (instrtype,col.replace('.',''),ldate))))
                if len(filelist) != 1 :
                    print("Anomaly: did not find a unique filename for %s %s %s in %s" % (statstype[nf],col,ldate,lfdir))
                    pass
                else:
                    lncycle += 1
                    lfile = filelist[0]
                    data = np.load(lfile,allow_pickle=True)
                
                    avg = np.ma.masked_values(data['avg'],data['missing_value'])
                    std = np.ma.masked_values(data['std'],data['missing_value'])
                    avg = np.ma.masked_equal(avg,1e20)
                    std = np.ma.masked_equal(std,1e20)

                    # Exclude the unbiased timesteps (24H) from the plots
                    if 'final_forecast/FORECAST/bias_correction' in lfile:
                        # avg[0], std[0] = np.ma.masked, np.ma.masked
                        avg[23], std[23] = np.ma.masked, np.ma.masked

                    rmse = np.ma.sqrt(avg**2+std**2)

                    if 'PercentOf' in statstype[nf]:
                        avg*=100
                        std*=100
                        rmse*=100
                    else:
                        avg*=1e9
                        std*=1e9
                        rmse*=1e9

                    if ic == 1:
                        cyclavg, cyclstd, cyclrmse, npoints = np.ma.array(avg), np.ma.array(std), np.ma.array(rmse), (~avg.mask).astype(int)
                    else:
                        cyclavg = cyclavg+avg.filled(fill_value=0)
                        cyclstd = cyclstd+std.filled(fill_value=0)
                        cyclrmse = cyclrmse+rmse.filled(fill_value=0)
                        npoints = npoints + (~avg.mask).astype(int)

                    allavg = np.ma.concatenate((allavg,avg))
                    allstd = np.ma.concatenate((allstd,std))
                    allrmse = np.ma.concatenate((allrmse,rmse))

                    if ic < ncycl:
                        ax1.plot(data['time'],avg,lw=lw,color=colors[nf])
                        ax2.plot(data['time'],std,lw=lw,color=colors[nf])
                        ax3.plot(data['time'],rmse,lw=lw,color=colors[nf])
                        # ax1.plot(data['time'],avg,lw=lw,color=colors[nf])
                    else:
                        ax1.plot(data['time'],avg,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(allavg)))
                        ax2.plot(data['time'],std,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(allstd)))
                        ax3.plot(data['time'],rmse,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(allrmse)))
                        # ax1.plot(data['time'],avg,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(allrmse)))
                        if pltcycleavg:
                            deltah = int(delta.total_seconds() // 3600)
                            # print(cyclavg)
                            # print(range(deltah,deltah+len(cyclavg)))
                            ax4.plot(range(deltah,deltah+len(cyclavg)),cyclavg/npoints,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(cyclavg/npoints)))
                            ax5.plot(range(deltah,deltah+len(cyclavg)),cyclstd/npoints,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(cyclstd/npoints)))
                            ax6.plot(range(deltah,deltah+len(cyclavg)),cyclrmse/npoints,color=colors[nf],lw=lw,label=linenames[nf]+' (%4.2e)' % (np.ma.mean(cyclrmse/npoints)))
        
        if 'PercentOf' in statstype[0]:
            ax1.set_ylim(-75,75)
            ax2.set_ylim(0,150)
            ax3.set_ylim(0,150)
            # ax2.set_ylim(0,80)
            # ax3.set_ylim(0,80)
            ax1.set_ylabel('%')
            ax2.set_ylabel('%')
            ax3.set_ylabel('%')

        legloc = 'upper right'
        if col in ('NO2','PM2.5'): 
            legloc = 'lower right'

        ax1.legend(loc=legloc)
        ax2.legend(loc=legloc)
        ax3.legend(loc=legloc)
        fig.savefig(os.path.join(outputdir,fname+'_'+col.replace('.','_')))

        if pltcycleavg:
            if 'PercentOf' in statstype[0]:
                ax4.set_ylim(-75,75)
                ax5.set_ylim(0,150)
                ax6.set_ylim(0,150)
                # ax5.set_ylim(0,80)
                # ax6.set_ylim(0,80)
                ax4.set_ylabel('%')
                ax5.set_ylabel('%')
                ax6.set_ylabel('%')
            ax4.legend(loc=legloc)
            ax5.legend(loc=legloc)
            ax6.legend(loc=legloc)
            fig2.savefig(os.path.join(outputdir,fname+'_cycleavg_'+col.replace('.','_')))

        plt.close()

    # Combine all maps in a single page if needed
    if combine_plots:

        combine_cmd = 'montage -mode concatenate -tile 4x1 {'

        for isp,spc in enumerate(columns):
            if isp > 3: break # limited to 4 species for the moment
            combine_cmd = combine_cmd + os.path.join(outputdir,fname+'_'+spc.replace('.','_')+'.png')
            if isp < len(columns)-1: combine_cmd=combine_cmd+','

        combine_cmd=combine_cmd+'} '+os.path.join(outputdir,fname+'_allspc.png')
        print(combine_cmd)
        subprocess.run(combine_cmd, shell=True)

        if pltcycleavg:

            combine_cmd = 'montage -mode concatenate -tile 4x1 {'

            for isp,spc in enumerate(columns):
                if isp > 3: break # limited to 4 species for the moment
                combine_cmd = combine_cmd + os.path.join(outputdir,fname+'_cycleavg_'+spc.replace('.','_')+'.png')
                if isp < len(columns)-1: combine_cmd=combine_cmd+','

            combine_cmd=combine_cmd+'} '+os.path.join(outputdir,fname+'_cycleavg_allspc.png')
            print(combine_cmd)
            subprocess.run(combine_cmd, shell=True)

if __name__ == "__main__":
    
    sns.set() # Apply seaborn beautificator
        
    dirlist = []

    # OSSE

    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_00/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/perturbed_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_31/model_error_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/model_error_truth/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_30/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/osse/exp_00/3dvar/FINAL_FORECAST/hofx/obs_plots')

    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/perturbed_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_010/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_011/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_012/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_013/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_014/cycle_0001/envar/hofx/obs_plots')
    # # dirlist.append('/scratch/globc/emili/seeds/osse/exp_001/cycle_0001/model_error_forecast/FORECAST/hofx/obs_plots')
    # # dirlist.append('/scratch/globc/emili/seeds/osse/exp_001/cycle_0001/model_error_truth/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_005/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_003/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/osse/exp_003/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')

    # Real exp
    dirlist.append('/scratch/globc/emili/seeds/real/re_001/cycle_0001/forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/re_001/cycle_0001/envar/hofx/obs_plots')
    dirlist.append('/scratch/globc/emili/seeds/real/re_001/cycle_0001/envar/hofx/obs_plots')
    dirlist.append('/scratch/globc/emili/seeds/real/re_043/cycle_0001/envar/hofx/obs_plots')
    dirlist.append('/scratch/globc/emili/seeds/real/re_044/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/re_022/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/re_023/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/envar_20/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/envar_19/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/envar_11/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/full_test/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/3dvar_01/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    dirlist.append('/scratch/globc/emili/seeds/real/3dvar_02/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/full_test/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/simu/exp_59/envar/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/simu/exp_61/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/full_test/cycle_0001/model_error_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/re_001/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/real/re_001/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')

    # Long exp
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_summer_allVal/cycle_0001/forecast/obs_plots')

    # dirlist.append('/scratch/globc/emili/seeds/long_07/cycle_0001/forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_07/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_07/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_07/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_01/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')

    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_spring_allVal/cycle_0001/forecast/obs_plots')

    # dirlist.append('/scratch/globc/emili/seeds/long_06/cycle_0001/forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_06/cycle_0001/envar/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_06/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_06/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_02/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/globc/emili/seeds/long_02/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')

    # dirlist.append('/scratch/globc/emili/seeds/long_02/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/3dvar/FINAL_FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/3dvar/ANALYSIS/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/model_error_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/final_forecast/FORECAST/hofx/obs_plots')
    # dirlist.append('/scratch/ae/emili/seeds/long/exp_03/cycle_0001/final_forecast/FORECAST/bias_correction/hofx/obs_plots')

    # Ensembles
    # ensize = 30
    # for ie in range(1,ensize+1): dirlist.append('/scratch/globc/jonville/seeds/ensembles/ens_emispertread_eps0/cycle_0001/ensemble/E%04i/obs_plots' % (ie))
    # for ie in range(1,ensize+1): dirlist.append('/scratch/globc/jonville/seeds/ensembles/ens_spring_eps0/cycle_0001/ensemble/E%04i/obs_plots' % (ie))
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_summer/cycle_0001/forecast/obs_plots')
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_summer/cycle_0001/forecast/obs_plots')
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_spring/cycle_0001/forecast/obs_plots')
    # dirlist.append('/scratch/globc/jonville/seeds/ensembles/cams01_30d_spring/cycle_0001/forecast/obs_plots')

    # ensize = 40
    # for ie in range(1,ensize+1): dirlist.append('/archive/pae/emili/seeds/ensembles/ens_05/cycle_0001/ensemble/E%04i/obs_plots' % (ie))
    # dirlist.append('/archive/pae/emili/seeds/simu/exp_00/forecast/obs_plots')

    linenames = []
    # linenames.append('truth')
    # linenames.append('truth')
    # linenames.append('free_forecast')
    linenames.append('forecast')
    # linenames.append('envar_007')
    linenames.append('envar_001')
    linenames.append('envar_043')
    linenames.append('envar_044')
    # linenames.append('envar_022')
    # linenames.append('envar_023')
    # linenames.append('envar_abs_mferr')
    # linenames.append('envar_abs_pcterr')
    # linenames.append('envar_log_mferr')
    # linenames.append('envar_log_pcterr')
    # linenames.append('envar_log16')
    # linenames.append('3dvar_01')
    # linenames.append('3dvar_02')
    # linenames.append('3dvar_relerr')
    # linenames.append('3dvar_ff')
    # linenames.append('envar_74')
    # linenames.append('envar_30members')
    # linenames.append('forecast')
    # linenames.append('envar_72')
    # linenames.append('analysis')
    # linenames.append('model_error_forecast')

    # linenames.append('final_forecast')
    # linenames.append('final_forecast_corr')
    linenames.append('3dvar_02')
    # linenames.append('3d_final_forecast')

    # linenames.append('perturbed_forecast_30')

    # linenames.append('model_error_forecast')
    # linenames.append('model_error_truth')
    # linenames.append('final_forecast')
    # linenames.append('final_forecast_corrected')

    # for ie in range(1,ensize+1): linenames.append('E%04i' % (ie))
    # linenames.append('control')
    # linenames.append('obs')

    statstype = []
    statstype.append("Misfit_Hx_dry_PercentOfObs")
    statstype.append("Misfit_Hx_dry_PercentOfObs")
    statstype.append("Misfit_Hx_dry_PercentOfObs")
    statstype.append("Misfit_Hx_dry_PercentOfObs")
    statstype.append("Misfit_Hx_dry_PercentOfObs")
    # statstype.append("Misfit_Hx_dry_PercentOfObs")
    # statstype.append("Misfit_Hx_dry_PercentOfObs")
    # statstype.append("Misfit_Hx_dry_PercentOfObs")
    # for ie in range(1,ensize+1): statstype.append("Misfit_Hx_dry_Absolute")
    # statstype.append("Misfit_Hx_dry_Absolute")
    # statstype.append("Misfit_Hx_dry_Absolute")
    # statstype.append("Misfit_Hx_dry_Absolute")
    # statstype.append("Misfit_Hx_dry_Absolute")
    # statstype.append("Misfit_Hx_dry_Absolute")
    # statstype.append("Misfit_Hx_dry_Absolute")
    # for ie in range(1,ensize+1): statstype.append("Misfit_Hx_dry_PercentOfObs")
    # statstype.append("Misfit_Hx_dry_PercentOfObs")
    # for ie in range(1,ensize+1): statstype.append("Hx_dry")
    # statstype.append("Hx_dry")
    # statstype.append("Y")

    # statstype.append("Hx_dry")
    # statstype.append("Hx_dry")
    # statstype.append("Hx_dry")
    # # statstype.append("Hx_dry")
    # statstype.append("Hx_dry")
    # statstype.append("Hx_dry")

    dates = []

    # dates.append('20190619T00_20190621T23')
    # dates.append('20190619T00_20190619T23')
    # dates.append('20190620T00_20190621T23')
    # dates.append('20190619T00_20190619T23')
    # dates.append('20190620T00_20190621T23')

    # dates.append('20190619T00_20190619T23')
    # dates.append('20190619T00_20190619T23')
    # # dates.append('20190619T00_20190630T23')

    # Long exp
    dates.append('20190619T00_20190621T23')
    dates.append('20190619T00_20190619T23')
    dates.append('20190619T00_20190619T23')
    dates.append('20190619T00_20190619T23')
    dates.append('20190619T00_20190619T23')
    # dates.append('20190619T00_20190619T23')
    # dates.append('20190619T00_20190619T23')
    # dates.append('20190620T00_20190621T23')
    # dates.append('20190620T00_20190621T23')    
    # dates.append('20190619T00_20190619T23')
    # dates.append('20190620T00_20190621T23')

    # dates.append('20190401T00_20190403T23')
    # dates.append('20190401T00_20190401T23')
    # dates.append('20190402T00_20190403T23')
    # dates.append('20190402T00_20190403T23')
    # dates.append('20190401T00_20190401T23')
    # dates.append('20190402T00_20190403T23')

    # for ie in range(1,ensize+1): dates.append('20190619T00_20190630T23')
    # dates.append('20190619T00_20190630T23')
    # dates.append('20190619T00_20190630T23')
    # for ie in range(1,ensize+1): dates.append('20190401T00_20190412T23')
    # dates.append('20190401T00_20190412T23')
    # dates.append('20190401T00_20190412T23')
    # for ie in range(1,ensize+1): dates.append('20100704T00_20100706T23')
    # dates.append('20100704T00_20100706T23')

    columns = []

    columns.append('O3')
    columns.append('NO2')
    columns.append('PM10')
    columns.append('PM2.5')
    # columns.append('CO')

    colors = ['tab:blue','tab:orange','tab:red','tab:purple','tab:green','tab:brown','tab:olive']
    # cm = plt.get_cmap('viridis')
    # colors = [cm(1.*i/(ensize)) for i in range(ensize)]
    # colors.append('black')
    # colors.append('darkred')
    ncycl = 1
    dtcycl = timedelta(days=1)
    pltcycleavg = False
    cyclerefdate = '20190619T00'
    skipdays = False
    # cyclerefdate = '20190401T00'
    # skipdays = True

    instrtype = 'AIRBASE_Val_'
    # instrtype = 'AIRBASE_'

    # outputdir = "/archive/pae/emili/seeds/long/exp_03/validation"
    # outputdir = "/archive/pae/emili/seeds/simu/comparisons/validation"
    # outputdir = "/archive/pae/emili/seeds/osse/comparisons/validation"
    # outputdir = "/scratch/globc/emili/seeds/comparisons/validation/osse"
    outputdir = "/scratch/globc/emili/seeds/comparisons/validation/real"
    # outputdir = "/scratch/globc/emili/seeds/comparisons/validation/long"

    if not os.path.exists(outputdir) : os.makedirs(outputdir)

    os.chdir(outputdir)

    # domain = "GEMS02"
    domain = "MACC01"
    
    expname = ''
    # expname = 'summer'
    # expname = 'absmisfit'
    # expname = 'summer_absmisfit'
    # expname = 'spring'
     # It contains the relevant instrument and date info
    
    # figsize = (15,30)
    figsize = (10,18)
    # figsize = (10,6)

    outfilenames = [string for string in linenames]

    fname = outfilenames[0]
    for string in outfilenames[1:]: fname = fname+'_'+string 
    if expname != '': fname = fname + '_' + expname
    
    ReadTimeSeries()
        