# -*- coding: utf-8 -*-
"""
Class handling the various ecflow steps for assimilation experiments 
"""

import os
# from ruamel.yaml import YAML
import yaml
from yaml import CLoader as Loader, CDumper as Dumper
import isodate.isodatetime as isodatetime
import isodate.isoduration as isodur
from Jedi_task import Jedi_task, ListMocageOutput, GetForecast, GetEnsemble, PrepareJob
from aq_steps import machines_config
import math
from datetime import timedelta
import itertools
import numpy as np

class Dirac(Jedi_task):

  actions = Jedi_task.actions + ['finalize']

  def __init__(self, **kwargs):

    super().__init__(**kwargs)

  @classmethod
  def py_setup_script(cls):

    kwargs = cls.load_ecfvars()

    # Load the dirac namelist here (my version, not the complete oops one)
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)
    # Remember that expname needs to come through ecflow because this method does not know self
    # and does not use the global experiment namelist! I can come back on this decision if I find 
    # out that a global namelist avoids a lot of duplication (e.g. output ensemble -> input dirac 
    # must be specified twice). In that case remove all the static ecflow arguments in favor
    # of the namelist usage.
    ensname, ensdir = GetEnsemble(cfg,kwargs)
    fctname, bkgdir = GetForecast(cfg,kwargs)

    outdir = os.path.join(ensdir,'dirac')
    dirac = cfg["increment in"]
    locdatadir = os.path.join(kwargs["exppath"],kwargs["expname"],'cycle_%04i' % (int(kwargs["cycle"])),'localization') #Default localization files
    if 'path' in cfg["localization"]: locdatadir = cfg["localization"]["path"]
    ens_size = None
    if 'ensemble size' in cfg: ens_size = cfg['ensemble size']

    if not os.path.exists(outdir): os.makedirs(outdir)

    print('>> Initializing dirac test: namelist creation...')

# Basic structure
    outyaml = {'initial condition':{}}

    statevars = cfg["state variables"]

    dateini = isodatetime.parse_datetime(kwargs['dateini'])
    dateend = isodatetime.parse_datetime(kwargs['dateend'])

    dtwinlen = dateend - dateini
    dtsubwin = isodur.parse_duration(cfg['subwindow'])
    nsub = int(dtwinlen/dtsubwin)+1
    dates =  [dateini+t*dtsubwin for t in range(nsub)]

    outyaml['geometry'] = cfg["grid"]

    initcond = outyaml['initial condition']
    initcond['states'] = ListMocageOutput(bkgdir,dates,statevars,outyaml['geometry']['domname'])

    outyaml['background error'] = {
        'covariance model': 'ensemble',
        # 'randomization size': 5000,    
        'localization':{
            'localization method' : 'SABER',
            'communication mode' : 'fast',
            'saber central block' : {
                'saber block name': 'BUMP_NICAS',
                'bump':{
                    'io':{
                        'data directory':locdatadir,
                        'files prefix':'loc_1pe'
                    },
                    'drivers':{
                        'multivariate strategy': cfg['localization']['multivariate'],
                        'read local nicas': True
                    },
                    'nicas':{
                        'common localization weights':[{'row variables':statevars,'column variables':statevars,'value':cfg['localization']['mc']}]
                    }
                }
            }
        },
        'members' : ListMocageOutput(ensdir,dates,statevars,outyaml['geometry']['domname'],mode='ensemble',ensize=ens_size)}

    if 'time decay' in cfg['localization']:
        outyaml['background error']['localization']['time decay'] = cfg['localization']['time decay']

    # Older way of specifing multivariate localization
    # if 'localization' in outyaml['background error'] and cfg['localization']['multivariate'] == 'common_weighted':
    #     # outyaml['background error']['localization']['saber block']['bump']['loc_wgt'] = {'common':0.5}
    #     # Construct the multivariate localization
    #     pair_order_list = list(itertools.combinations(statevars,2))
    #     outyaml['background error']['localization']['saber central block']['bump']['loc_wgt'] = {el[0]+'-'+el[1]:0.5 for el in pair_order_list}

    dateout = dateini+cfg['increment subwindow']*dtsubwin # Can we choose one date freely?
    outyaml['dirac'] = {
        'date': dateout.isoformat()+'Z',
        'londir': [float(di.split(',')[0]) for di in dirac],    
        'latdir': [float(di.split(',')[1]) for di in dirac],    
        'levdir': [int(di.split(',')[2]) for di in dirac],
        'var':    [di.split(',')[3] for di in dirac]}

    outyaml['output dirac'] = {
        'datadir': outdir,
        'fileprefix': 'D4'+outyaml['geometry']['domname']+r"_%id%"
    }

    outyaml['output variance'] = {
        'datadir': outdir,
        'fileprefix': 'V4'+outyaml['geometry']['domname']
    }

    print(yaml.dump(outyaml))

    with open(kwargs['jedinamelist'],'w') as f:
        print(yaml.dump(outyaml),file=f)

    nodes = 1
    queue = 'interactive'
    hostname = 'nemopae2'

    if 'computing' in cfg and 'run' in cfg['computing']:
      hostname = cfg['computing']['run']['machine']
      if 'nodes' in cfg['computing']['run']: nodes = cfg['computing']['run']['nodes']
      if 'queue' in cfg['computing']['run']: queue = cfg['computing']['run']['queue']

    rmode = machines_config(hostname)['mode']

    OpenMP = True
    OpenMPperMPI = 1

    cores_per_node = machines_config(hostname)['queues'][queue]['cores_per_node']

    total_res = nodes * cores_per_node

    res_per_slot = math.floor(total_res / nsub)

    if OpenMP and OpenMPperMPI > 1:
       ompthreads = min(OpenMPperMPI, res_per_slot)
    else:
       ompthreads = 1

    # Force all slots in one node
    # if nodes == 1: 
    #     mpiprocs = nsub
    #     ompthreads = 1
    #     mpi_per_slot = 1
    # else:
    mpi_per_slot = math.floor(res_per_slot/ompthreads)
    mpiprocs = int(nsub * mpi_per_slot)

    print('mpiprocs ',mpiprocs,'mpi per slot ',mpi_per_slot, 'ompthreads ',ompthreads)
    print('available resources ',total_res)
    print('used resources ',mpiprocs * ompthreads)

    script = PrepareJob(exe=os.path.join(cfg['oops path'],'aq_dirac.x'),namelist=kwargs['jedinamelist'],
        mpiprocs=mpiprocs,ompthreads=ompthreads,
        mode=rmode,nodes=nodes,machine=hostname,
        queue=queue,time=timedelta(hours=1),
        jobout=kwargs['jobout'],
        errout=kwargs['joberr']
        )
    
    with open(kwargs['jobname'],'w') as f:
        print(script,file=f)
    os.chmod(kwargs['jobname'], 0o755)

  @classmethod
  def py_finalize_script(cls):

    import cdms2 as cdms
    import MV2 as MV
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm

    # From dirac output files compute correlations and plot them
    kwargs = cls.load_ecfvars()

    # Load the oops dirac namelist here
    with open(kwargs['jedinamelist']) as file:
        diracyaml = yaml.load(file,Loader=Loader)

    # Load the user dirac namelist here
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)

    if 'postproc namelist' not in cfg: return

    ensname, ensdir = GetEnsemble(cfg,kwargs)
    indir = diracyaml['output dirac']['datadir']
    indirvar = os.path.join(ensdir,'variance')
    outdir = os.path.join(indir,'plots')
    if not os.path.exists(outdir): os.makedirs(outdir)

    incr = diracyaml['dirac'] # attention: only one dirac should be used to study correlations
    drsp = incr['var'][0]
    drlon = incr['londir'][0]
    drlat = incr['latdir'][0]
    drlev = incr['levdir'][0]
    drtime = isodatetime.parse_datetime(incr['date'])

    covprefix = 'D4'+cfg['grid']['domname']+'_ensemble'
    # varprefix = 'EV'+cfg['grid']['domname']
    varprefix = 'V4'+cfg['grid']['domname']
    corrprefix = 'EC'+cfg['grid']['domname']
    dateini = isodatetime.parse_datetime(kwargs['dateini'])
    dateend = isodatetime.parse_datetime(kwargs['dateend'])
    dtwinlen = dateend - dateini
    dtsubwin = isodur.parse_duration(cfg['subwindow'])
    nsub = int(dtwinlen/dtsubwin)+1
    dates =  [dateini+t*dtsubwin for t in range(nsub)]

    # Read the variance at the dirac point
    # This provides the exact normalization, consider restoring it, but needs variance output available
    # fidvar = cdms.open(os.path.join(indirvar,varprefix+'+%s.nc' % (drtime.strftime('%Y%m%d%H')))) 
    fidvar = cdms.open(os.path.join(indir,varprefix+'+%s.nc' % (drtime.strftime('%Y%m%d%H'))))
    drvar = fidvar(drsp,lat=drlat,lon=drlon,lev=drlev,squeeze=1)
    fidvar.close()

    for plot in cfg['postproc namelist']['plots']:

        lcfg = cfg['postproc namelist']['plots'][plot]

        if plot == 'vars':

            # Read appropriate covariance and variance files
            fidcov = cdms.open(os.path.join(indir,covprefix+'+%s.nc' % (dates[lcfg['times'][0]].strftime('%Y%m%d%H'))))
            # fidvar = cdms.open(os.path.join(indirvar,varprefix+'+%s.nc' % (dates[lcfg['times'][0]].strftime('%Y%m%d%H'))))
            fidvar = cdms.open(os.path.join(indir,varprefix+'+%s.nc' % (dates[lcfg['times'][0]].strftime('%Y%m%d%H'))))

            # Compute correlation
            corr = np.zeros(len(cfg['state variables']))
            for isp,sp in enumerate(cfg['state variables']):
                corr[isp] = fidcov(sp,lat=lcfg['latitudes'][0],
                    lon=lcfg['longitudes'][0],lev=lcfg['levels'][0]) / \
                    drvar**0.5 / \
                    fidvar(sp,lat=lcfg['latitudes'][0],
                    lon=lcfg['longitudes'][0],lev=lcfg['levels'][0])**0.5
            
            fidcov.close()
            fidvar.close()

            # Print
            print('Chemical correlations at time %s latitude %4.2f longitude %4.2f and level %i:' % 
                (dates[lcfg['times'][0]].strftime('%Y%m%d%H'),lcfg['latitudes'][0],lcfg['longitudes'][0],lcfg['levels'][0]))
            for isp,sp in enumerate(cfg['state variables']):
                print('%s-%s : %.3f' % (drsp,sp,corr[isp]))

        elif plot == 'time':

            corr = np.zeros(len(dates))
            for it,date in enumerate(dates):
                # Read appropriate covariance and variance files
                fidcov = cdms.open(os.path.join(indir,covprefix+'+%s.nc' % (date.strftime('%Y%m%d%H'))))
                # fidvar = cdms.open(os.path.join(indirvar,varprefix+'+%s.nc' % (date.strftime('%Y%m%d%H'))))
                fidvar = cdms.open(os.path.join(indir,varprefix+'+%s.nc' % (date.strftime('%Y%m%d%H'))))

                corr[it] = fidcov(lcfg['variables'][0],lat=lcfg['latitudes'][0],
                    lon=lcfg['longitudes'][0],lev=lcfg['levels'][0]) / \
                    drvar**0.5 / \
                    fidvar(lcfg['variables'][0],lat=lcfg['latitudes'][0],
                    lon=lcfg['longitudes'][0],lev=lcfg['levels'][0])**0.5

                fidcov.close()
                fidvar.close()

            # Print
            print('Temporal correlations of %s with %s at latitude %4.2f longitude %4.2f and level %i:' % 
                (drsp,lcfg['variables'][0],lcfg['latitudes'][0],lcfg['longitudes'][0],lcfg['levels'][0]))
            for it,date in enumerate(dates):
                print('%s : %.3f' % (date.strftime('%Y%m%d%H'),corr[it]))
        
        elif plot == 'time-vars':
            
            corr = np.zeros((len(dates),len(cfg['state variables'])))
            for it,date in enumerate(dates):
                # Read appropriate covariance and variance files
                fidcov = cdms.open(os.path.join(indir,covprefix+'+%s.nc' % (date.strftime('%Y%m%d%H'))))
                # fidvar = cdms.open(os.path.join(indirvar,varprefix+'+%s.nc' % (date.strftime('%Y%m%d%H'))))
                fidvar = cdms.open(os.path.join(indir,varprefix+'+%s.nc' % (date.strftime('%Y%m%d%H'))))

                for isp,sp in enumerate(cfg['state variables']):
                    corr[it,isp] = fidcov(sp,lat=lcfg['latitudes'][0],
                        lon=lcfg['longitudes'][0],lev=lcfg['levels'][0]) / \
                        drvar**0.5 / \
                        fidvar(sp,lat=lcfg['latitudes'][0],
                        lon=lcfg['longitudes'][0],lev=lcfg['levels'][0])**0.5
                    print('%s %s-%s: %.3f' % (date.strftime('%Y%m%d%H'),sp,drsp,corr[it,isp]))

                fidcov.close()
                fidvar.close()

            # Save plot
            fig = plt.figure(figsize=(10,10))
    
            ax = plt.subplot(1,1,1)
            fig.autofmt_xdate()
            im=ax.imshow(corr.transpose(), cmap=cm.RdBu_r, vmax=1, vmin=-1, interpolation='nearest')
            plt.title('Variables-time correlation')
            plt.colorbar(im)
            # plt.xticks(np.arange(nv),variables,rotation=27)
            plt.yticks(np.arange(len(cfg['state variables'])),cfg['state variables'],rotation=27)

            fig.savefig(os.path.join(outdir,'time-vars-correlation.png'))
            plt.close(fig)

        elif plot == 'space':

            # Read appropriate covariance and variance files
            fidcov = cdms.open(os.path.join(indir,covprefix+'+%s.nc' % (dates[lcfg['times'][0]].strftime('%Y%m%d%H'))))
            # fidvar = cdms.open(os.path.join(indirvar,varprefix+'+%s.nc' % (dates[lcfg['times'][0]].strftime('%Y%m%d%H'))))
            fidvar = cdms.open(os.path.join(indir,varprefix+'+%s.nc' % (dates[lcfg['times'][0]].strftime('%Y%m%d%H'))))
            fidout = cdms.open(os.path.join(outdir,corrprefix+'+%s.nc' % (dates[lcfg['times'][0]].strftime('%Y%m%d%H'))),'w')
            # Compute correlation
            for isp,sp in enumerate(cfg['state variables']):
                corr = MV.array(fidcov(sp) / drvar**0.5 / fidvar(sp)**0.5, \
                    id = sp, \
                    axes = fidcov(sp).getAxisList())
                fidout.write(corr)
            
            fidcov.close()
            fidvar.close()
            fidout.close()

            continue
        
        else:
            continue









