# -*- coding: utf-8 -*-
"""
Class handling an ensemble of forecasts
"""

import os
import yaml
from yaml import CLoader as Loader, CDumper as Dumper
import shutil
import json
import datetime as dt
from aq_steps import Step
from Ensemble_member import Ensemble_member
from Forecast import get_mocage_param
import subprocess

class Ensemble(Step):

  # Variables that transit through ecflow into all ensemble tasks
  ecfvars = Step.ecfvars + ['namelist']
  actions = Step.actions + ['setup','finalize']

  def __init__(self, **kwargs):

    super().__init__(**kwargs)

    self.size = self.namelist['size']
    self.start_seed = self.namelist['perturbations']['start_seed']
    # Init all ensemble members here
    self.members = [Ensemble_member(expname=self.expname,stepname='ensemble_member',member=i+1,
        start_date=kwargs['start_date'],duration=kwargs['duration'],suite_path=self.suite_path,
        seed=self.start_seed+i,namelist=self.namelist,create_ecf=False,create_nml=False)
        for i in range(self.size)]
    # Set the members namelist files to the ensemble one (otherwise it would be ensemble_forecast.json name)
    for i in range(self.size): self.members[i].namelist_file = self.namelist_file
    self.__class__.create_ecf_scripts(path=os.path.join(self.suite_path,'tasks'))
    self.save_namelist()

  def next(self, dt, **kwargs):

    super().next(dt)
    # Increment all members seed by a large amount (code valid up to 1000 members)
    for i in range(self.size): self.members[i].next(dt,seed_incr=1000)

  @classmethod
  def py_setup_script(cls):

    kwargs = cls.load_ecfvars()

    expnamelist = kwargs["namelist"]
    expname = kwargs["expname"]

    with open(expnamelist) as file:
      cfg = yaml.load(file,Loader=Loader)

    workspace = os.path.join(cfg["workspace path"],expname,"ensemble")
    outdir = kwargs['outpath']

    if not os.path.exists(workspace): os.makedirs(workspace)

    shutil.copyfile(expnamelist,os.path.join(workspace,os.path.basename(expnamelist)))

    if not os.path.exists(outdir): os.makedirs(outdir)

    shutil.copyfile(expnamelist,os.path.join(outdir,os.path.basename(expnamelist)))

    # Check wether the mocage simulation is in assim_mode, if so modify the number
    # of windows according to the duration of the forecast.
    refjobname = os.path.join(cfg["mocage path"],"jobs/mocage/rundir", \
      cfg["reference job"])
    val = get_mocage_param(refjobname,'CASSIM')
    if int(val) == 1:
      refassimnml = refjobname.replace('jobdaimon_','assimilation_namelist_')
      refassimnml = refassimnml.replace('_L.sh','.json')
      with open(refassimnml) as file:
        assimcfg = json.load(file)
      assimcfg['Assimilation']['Cycling']['NbOfWindows'] = \
        int((dt.datetime.fromisoformat(kwargs['dateend']) - dt.datetime.fromisoformat(kwargs['dateini'])) / \
         dt.timedelta(seconds=assimcfg['Assimilation']['Cycling']['WindowsLength']))
      with open(refassimnml,"w") as file:
        json.dump(assimcfg,file,indent=4)

  @classmethod
  def py_finalize_script(cls):

    kwargs = cls.load_ecfvars()
    
    expnamelist = kwargs["namelist"]
    exppath = kwargs["exppath"]
    expname = kwargs["expname"]
    dateend = kwargs["dateend"]

    with open(expnamelist) as file:
      cfg = yaml.load(file,Loader=Loader)

    workspace = os.path.join(cfg["workspace path"],expname,"ensemble")
    outdir = kwargs['outpath']
    n_modele = int(cfg["size"])

    for n in range(n_modele):
      workdir = os.path.join(workspace,"E%04i" % (n+1))
      if os.path.exists(workdir):
        shutil.rmtree(workdir)
        print(workdir + ' succesfully removed')

    if os.path.exists(outdir):
      for n in range(n_modele):
        outmemberdir = "E%04i"%(n+1)
        datestr = dateend[:4]+dateend[5:7]+dateend[8:10]+dateend[11:13]  
        if os.path.exists(os.path.join(outdir,outmemberdir,'HMGEMS02+' + datestr)):
          os.remove(os.path.join(outdir,outmemberdir,'HMGEMS02+' + datestr))
          print(outmemberdir+'/HMGEMS02+' + datestr + ' succesfully removed')
        if os.path.exists(os.path.join(outdir,outmemberdir,'HMGLOB22+' + datestr)):
          os.remove(os.path.join(outdir,outmemberdir,'HMGLOB22+' + datestr))
          print(outmemberdir+'/HMGLOB22+' + datestr+' succesfully removed')

      # Add write mode to expdir for the unix group
      cmd="chmod -R g+w {}".format(os.path.join(exppath,expname))
      subprocess.run(cmd,shell=True)

  def setup(self, ecflowbr, **kwargs):

    super().setup(ecflowbr)
    self.ecfstep.add_variable("namelist",self.namelist_file)

  def run(self, **kwargs):

    # The run could actually just init and finalize ensemble members for postprocessor purposes
    first_member = self.namelist['start from member'] # The member index starts from 1 in namelists
    totpeople = self.size-(first_member-1)
    nparallel = self.namelist['parallelism']['num parallel jobs']
    nchildren_per_fam = [int(totpeople/nparallel) for i in range(nparallel)]
    for i in range(totpeople%nparallel): nchildren_per_fam[i]+=1

    #Fix the parallel/sequential order based on the max number of parallel run
    nperson = first_member
    postp_condition = ''
    print('Creating ensemble of ',totpeople,' members')
    print('starting from member ', self.members[first_member-1].member, ' with seed ', self.members[first_member-1].seed)

    #Create parallel families (pfam) on the ecflow branch setup previously
    for i in range(nparallel):
        if nchildren_per_fam[i] > 0: # Avoids empty ecflow branches
          pfam = self.ecfstep.add_family("pfam_%04i" % (i+1)).add_trigger("%s_setup eq complete" % (type(self).__name__))
          #Create sequential members (children) inside a single family
          for nchild in range(nchildren_per_fam[i]):
              sfam = pfam.add_family("child_%04i" % (nchild+1))
              #Add trigger for checking the completion of previous member if it's not the first child
              if nchild > 0 : sfam.add_trigger("child_%04i eq complete" % (nchild))
              #create_ensemble_member(nperson,nchild,pfam,seed)
              self.members[nperson-1].set_run_end(sfam) # Python is zero index based , nperson starts from 1
              nperson+=1
          if i < nparallel-1:
              if nchildren_per_fam[i]>0 and nchildren_per_fam[i+1]==0:
                  postp_condition = postp_condition+"pfam_%04i eq complete" % (i+1)
              elif nchildren_per_fam[i]>0:
                  postp_condition = postp_condition+"pfam_%04i eq complete and " % (i+1)
          if i == nparallel-1:
              if nchildren_per_fam[i]>0:
                  postp_condition = postp_condition+"pfam_%04i eq complete" % (i+1)
    self.ppcond = postp_condition

  def finalize(self, **kwargs):

    if 'finalize' in self.todo:

      task = self.ecfstep.add_task(type(self).__name__+'_finalize')
      task.add_variable("action",'finalize').add_variable("class",type(self).__name__).add_trigger(self.ppcond)

  @classmethod
  def create_ecf_scripts(cls, **kwargs):

    # Ensemble setup script (the run is not a shell command, but made of ecflow families, no ecf needed)
    cls.ecf_script_setup(**kwargs)
    cls.ecf_script_finalize(**kwargs)
    Ensemble_member.create_ecf_scripts(**kwargs)
