# -*- coding: utf-8 -*-
"""
Class handling a single member of an ensemble of forecasts with MOCAGE 
"""

import os
import sys
import shutil
import yaml
from yaml import CLoader as Loader, CDumper as Dumper
import json
import isodate.isodatetime as isodatetime
import isodate.isoduration as isodur
from Forecast import Forecast, create_mocage_job, get_mocage_param, update_mocage_assim, setup_mocage_restart

class Ensemble_member(Forecast):

  # Variables that transit through ecflow into all ensemble tasks
  ecfvars = Forecast.ecfvars + ['seed','member']

  def __init__(self, **kwargs):

    super().__init__(**kwargs)

    # Argument based init is used by the class Ensemble
    if 'member' in kwargs: 
      self.member = '%04i' % (kwargs['member'])
    else:
      self.member = 'fcst' # This key is used if the ensemble member is only one single perturbed forecast
    if 'seed' in kwargs: 
      self.seed = kwargs['seed']
    else:
      self.seed = int(self.namelist['perturbations']['start_seed'])

  def next(self, dt, **kwargs):

    super().next(dt)
    if 'seed_incr' in kwargs:
      self.seed += kwargs['seed_incr']
    else:
      self.seed += 1

  @classmethod
  def py_setup_script(cls):

    # Retrieve variables from ecflow environment
    kwargs = cls.load_ecfvars()

    # Local variables
    member = kwargs['member']

    # Load the ensemble namelist here
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)

    workspace = os.path.join(cfg["workspace path"],kwargs['expname'])

    if member == 'fcst':
      outdir = kwargs['outpath']
      member_wkdir = os.path.join(workspace,kwargs['taskname'])
    else:
      outdir = os.path.join(kwargs['outpath'],"E%s" % (member))
      member_wkdir = os.path.join(workspace,"ensemble","E%s" % (member))

    refjobfile = os.path.join(cfg["mocage path"],'jobs/mocage/rundir',cfg["reference job"])

    print('>> Initializing mocage forecast: namelist creation...')
    print('>> Initializing ensemble member n. ',member,' with seed: ',kwargs['seed'])
    # Update the seed in the ensemble namelist to init the model at execution
    cfg["perturbations"]["start_seed"] = int(kwargs['seed'])

    datebg = kwargs['datebegin']
    dateini = kwargs['dateini']
    dateend = kwargs['dateend']

    dateinidt = isodatetime.parse_datetime(dateini)
    dateenddt = isodatetime.parse_datetime(dateend)
    nh = (dateenddt-dateinidt).total_seconds()/3600.

    cfg["date begin"] = datebg[:4]+datebg[5:7]+datebg[8:10]+datebg[11:13] 
    # date begin is used only for inhisto pert to avoid perturbing multiple times the initial condition
    # at each mocage execution (only useful if a job splitting is used).
    cfg["ensemble member"] = member # Not used yet, but kept for traking output

    mocage_dict  = {
                    'DATEIN':dateinidt.strftime('%Y%m%d%H'),
                    'DATEOUT':dateenddt.strftime('%Y%m%d%H'),
                    'JOBNAME':os.path.basename(kwargs['jobname']),
                    'NPACKS':'12',
                    '--time':'%02i:%02i:00' % (int(nh/12),int((nh % 12)/12.*60)), # On kraken 2 nodes 1 sim day = 1:30 hour approx
                    'FTPDIR': member_wkdir,
                    'SINGLEDIROUT': outdir
                   }
    if "hm frequency" in cfg: mocage_dict['NOUT'] = cfg["hm frequency"]

    use_meteo_ensemble = cfg["perturbations"]["external"]["meteo"]["activate"]
    #In case of meteo ensemble used as forcing redirect to different FM
    if use_meteo_ensemble :
        print('>> Using meteo ensemble')
        if member != 'fcst' and int(member) > 10:
          print('!!!The meteo ensemble is undersized wrt the MOCAGE one!!!')
          sys.exit(-1)
        meteo_ens_path = cfg["perturbations"]["external"]["meteo"]["input path"]
        mocage_dict['DIR_FM']=os.path.join(meteo_ens_path,dateini[:4],'E%03i' % (int(member)))

    # Check wether the mocage simulation is in assim_mode, if so modify the number
    # of windows according to the duration of the forecast.
    cassim = bool(get_mocage_param(refjobfile,'CASSIM'))
    
    # Avoid updating the assimilation namelist for all ensemble members (disk access conflict)
    if cassim and (member == 'fcst' or int(member) == 1):  update_mocage_assim(refjobfile,kwargs,cfg)

    if cassim and "assimilation" in cfg and cfg["assimilation"]: 
      # The assimilation needs longer run time
      mocage_dict['--time'] = '%02i:%02i:00' % (int(nh/8),int((nh % 8)/8.*60))
      if "final forecast" in cfg and cfg["final forecast"]["run"]:
        fflength = isodur.parse_duration(cfg["final forecast"]["length"])
        nhff = fflength.total_seconds()/3600.
        # Increase the total run time accounting the final forecast
        mocage_dict['--time'] = '%02i:%02i:00' % (int(nh/8)+int(nhff/16),
          min(int((nh % 8)/8.*60)+int((nhff % 16)/16.*60),59))
        # Update the mocage job final date
        mocage_dict['DATEOUT'] = (dateenddt+fflength).strftime('%Y%m%d%H')

    print('>> Preparing working dir in ', member_wkdir)
    if not os.path.exists(workspace): os.makedirs(workspace)
    if os.path.exists(member_wkdir): shutil.rmtree(member_wkdir)
    os.mkdir(member_wkdir)

    # Update the input for the restart if it's not the first simulated period
    if int(kwargs['cycle']) > 1: 
      cfg['depends on'] = {
        'forecast': {'cycle':-1},
        'envar': {'cycle':-1,'grid':'MACC01'} # Improve this, where can it go?
      }

    if cassim and "observations path" in cfg: mocage_dict['DIR_AS'] = cfg["observations path"]

    if "depends on" in cfg: 
      if 'init species' in cfg and 'envar' in cfg['depends on']: 
        cfg['depends on']['envar']['init species'] = cfg['init species']
      setup_mocage_restart(dateinidt,cfg,mocage_dict,member_wkdir,kwargs)

    print('>> Modifying MOCAGE job...')
    create_mocage_job(refjobfile,kwargs['jobname'],mocage_dict)

    # Create local copy of the ensemble namelist for the member (MOCAGE uses json namelists)
    member_namelist = os.path.join(member_wkdir,"ensemble_namelist.json")
    with open(member_namelist,"w") as file:
      json.dump(cfg,file,indent=4)

  def setup(self, ecflowbr, **kwargs):

    super().setup(ecflowbr)
    # Replace the forecast job name with member information
    jobname = os.path.join(self.namelist['mocage path'],'jobs/mocage/rundir',
      self.namelist['reference job']).replace('_L.sh',
      '_%s_forecast_member_%s_cycle_%04i_L.sh' % 
      (self.expname,self.member,self.cycle),1)
    self.ecfstep.add_variable("jobname",jobname)
    self.ecfstep.add_variable("seed",self.seed)
    self.ecfstep.add_variable("member",(self.member))
