# -*- coding: utf-8 -*-
"""
Class handling the various ecflow steps for assimilation experiments 
"""

import os
import yaml
from yaml import CLoader as Loader, CDumper as Dumper
# from ruamel.yaml import YAML
import isodate.isodatetime as isodatetime
import isodate.isoduration as isodur
from datetime import timedelta
from Jedi_task import Jedi_task, ListMocageOutput, GetForecast, GetEnsemble, PrepareJob
from aq_steps import machines_config
import subprocess

class Ensemble_variance(Jedi_task):

  actions = Jedi_task.actions + ['finalize']

  def __init__(self, **kwargs):

    super().__init__(**kwargs)

  @classmethod
  def py_setup_script(cls):

    kwargs = cls.load_ecfvars()

    # Load the variational namelist here (my version, not the complete oops one)
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)
    # Remember that expname needs to come through ecflow because this method does not know self
    # and does not use the global experiment namelist! I can come back on this decision if I find 
    # out that a global namelist avoids a lot of duplication (e.g. output ensemble -> input dirac 
    # must be specified twice). In that case remove all the static ecflow arguments in favor
    # of the namelist usage.
    ensname, ensdir = GetEnsemble(cfg,kwargs)
    fctname, bkgdir = GetForecast(cfg,kwargs)

    outdir = os.path.join(ensdir,'variance')

    if not os.path.exists(outdir): os.makedirs(outdir)

    print('>> Initializing Ensemble Variance computation: namelist creation...')

    # Basic structure
    outyaml = {'background':{},'ensemble':{},'variance output':{}}

    outyaml['geometry'] = cfg["grid"]

    dateini = isodatetime.parse_datetime(kwargs['dateini'])
    dateend = isodatetime.parse_datetime(kwargs['dateend'])

    chemsp = cfg['output variables']
    output = outyaml['variance output']
    output['datadir'] = outdir
    output['fileprefix'] = 'EV'+outyaml['geometry']['domname']
    ensemble = outyaml['ensemble']
    ensemble['output variables'] = cfg['output variables']

    dtwinlen = dateend - dateini
    dtsubwin = isodur.parse_duration(cfg['subwindow'])
    nsub = int(dtwinlen/dtsubwin)+1

    dates =  [dateini+t*dtsubwin for t in range(nsub)]

    nodes = 1
    queue = 'interactive'
    hostname = 'nemopae2'
    mpiprocs = 6
    ompthreads = 4

    if 'computing' in cfg and 'run' in cfg['computing']:
      hostname = cfg['computing']['run']['machine']
      if 'nodes' in cfg['computing']['run']: nodes = cfg['computing']['run']['nodes']
      if 'queue' in cfg['computing']['run']: queue = cfg['computing']['run']['queue']
      mpiprocs = machines_config(hostname)['queues'][queue]['cores_per_node']/ompthreads
    
    rmode = machines_config(hostname)['mode']
    
    for it,date in enumerate(dates):

        outyaml['background'] = ListMocageOutput(bkgdir,[date],chemsp,outyaml['geometry']['domname'])
        ensemble['members'] = ListMocageOutput(ensdir,[date],chemsp,outyaml['geometry']['domname'],mode='ensemble')

        if 'transform' in cfg:
            bkg = outyaml['background'] 
            bkg['transform'] = cfg['transform']
            for member in ensemble['members']:
                member['transform'] = cfg['transform']

        print(yaml.dump(outyaml))

        with open(kwargs['jedinamelist']+'-step_%02i.yaml' % (it),'w') as f:
            print(yaml.dump(outyaml),file=f)

    script = PrepareJob(exe=os.path.join(cfg['oops path'],'aq_ens_variance.x'),namelist=kwargs['jedinamelist'],
        nsteps=len(dates),
        mode=rmode,nodes=nodes,machine=hostname,
        mpiprocs=mpiprocs,ompthreads=ompthreads,
        queue=queue,time=timedelta(hours=6),
        jobout=kwargs['jobout'],
        errout=kwargs['joberr'])
    
    with open(kwargs['jobname'],'w') as f:
        print(script,file=f)
    os.chmod(kwargs['jobname'], 0o755)

  @classmethod
  def py_finalize_script(cls):

    import daimonpp as plotter

    kwargs = cls.load_ecfvars()

    # Load the variational namelist here (my version, not the oops one)
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)

    if 'postproc namelist' in cfg:
        
        pltnamelist = cfg['postproc namelist']
        with open(pltnamelist) as file:
            postcfg = yaml.load(file,Loader=Loader)

        ensname, ensdir = GetEnsemble(cfg,kwargs)
        outdir = os.path.join(ensdir,"variance")

        # Set plt namelist elements according to the experiment name and dates
        postcfg['expname'] = kwargs['expname']
        datefmt = '%Y%m%d%H'
        postcfg['starttime'] = isodatetime.parse_datetime(kwargs['dateini']).strftime(datefmt) 
        postcfg['endtime'] = isodatetime.parse_datetime(kwargs['dateend']).strftime(datefmt)

        # Plots here
        if "plots" in postcfg:

            if 'modelflds' in postcfg['plots']:

                if 'hfields' in postcfg['plots']['modelflds']:

                    plotter.plot_hfields(postcfg,outdir,ftype='EV',rtype='std')

                if 'vfields' in postcfg['plots']['modelflds']:

                    plotter.plot_vfields(postcfg,outdir,ftype='EV',rtype='std')

        # Add write mode to variance directory for the unix group
        cmd="chmod -R g+w {}".format(outdir)
        subprocess.run(cmd,shell=True)

