# -*- coding: utf-8 -*-
"""
Class handling the the forecast step with MOCAGE 
"""

import os
import sys
import stat
import yaml
from yaml import CLoader as Loader, CDumper as Dumper
import json
import shutil
import datetime as dt
import isodate.isodatetime as isodatetime
import isodate.isoduration as isodur
from aq_steps import Step
from netCDF4 import Dataset
import subprocess
from aq_steps import machines_config

class Forecast(Step):

  # Variables that transit through ecflow into all ensemble tasks
  ecfvars = Step.ecfvars + ['namelist','jobname','jobout','joberr']
  actions = Step.actions + ['setup','run','finalize']

  def __init__(self, **kwargs):

    super().__init__(**kwargs)

    # Save needed ecf scripts and namelists (the condition are needed because of the ensemble init)
    create_ecf, create_nml = True, True
    if 'create_ecf' in kwargs: create_ecf = kwargs['create_ecf']
    if 'create_nml' in kwargs: create_nml = kwargs['create_nml']
    if create_ecf: self.__class__.create_ecf_scripts(path=os.path.join(self.suite_path,'tasks'))
    if create_nml: self.save_namelist()

  @classmethod
  def ecf_script_run(cls, **kwargs):

    ll = '%include <head.h>\n'
    ll+='echo "Forecast Run"\n'
    ll+=r"""if [ "%jobmode%" = "interactive" ]"""
    ll+='\nthen\n'
    ll+=r"""ssh %runhostname% '%jobname% 1> %jobout% 2> %joberr%'"""
    ll+='\nelse\n'
    # The very ugly perl script that submit the job on nemo is here. Probably need one for each machine.
    ll+=r"""ssh_out=$(ssh %runhostname% 'jobid=$(/usr/bin/sbatch %jobname% | awk "{print \$4}")&& while [ $? -eq 0 ]; do sleep 5s&& squeue -j $jobid | grep $jobid > /dev/null; done; if grep "files stored in:" slurm-${jobid}.out >out.tmp 2>&1; then echo ${jobid} finished OK && exit 0; else echo ${jobid} did not finish OK && exit 22; fi')"""
    ll+='\necho $ssh_out\n'
    ll+='\nfi\n'
    ll+= '%include <tail.h>\n'
    with open(os.path.join(kwargs['path'],cls.__name__+'_run.ecf'),'w') as f:
      f.writelines(ll)

  @classmethod
  def py_setup_script(cls):

    kwargs = cls.load_ecfvars()

    # Load the forecast namelist here
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)

    workspace = os.path.join(cfg["workspace path"],kwargs['expname'])
    workdir = os.path.join(cfg["workspace path"],kwargs['expname'],kwargs['taskname'])
    outdir = kwargs['outpath']
    refjobfile = os.path.join(cfg["mocage path"],'jobs/mocage/rundir',cfg["reference job"])

    print('>> Initializing mocage forecast: namelist creation...')

    dateini = kwargs['dateini']
    dateend = kwargs['dateend']
    dateinidt = isodatetime.parse_datetime(dateini)
    dateenddt = isodatetime.parse_datetime(dateend)
    nh = (dateenddt-dateinidt).total_seconds()/3600.

    mocage_dict  = {
                    'DATEIN':dateinidt.strftime('%Y%m%d%H'),
                    'DATEOUT':dateenddt.strftime('%Y%m%d%H'),
                    'JOBNAME':os.path.basename(kwargs['jobname']),
                    'NPACKS':'30',
                    'NOUT':'24',
                    '--time':'%02i:%02i:00' % (int(nh/16),int((nh % 16)/16.*60)), # On kraken 2 nodes 1 sim day = 1:30 hour approx
                    'FTPDIR': workdir,
                    'SINGLEDIROUT': outdir
                   }

    # Check wether the mocage simulation is in assim_mode, if so modify the number
    # of windows according to the duration of the forecast.
    cassim = bool(get_mocage_param(refjobfile,'CASSIM'))

    if cassim: update_mocage_assim(refjobfile,kwargs,cfg)

    if cassim and "assimilation" in cfg and cfg["assimilation"]: 
      # The assimilation needs longer run time (1 sim day = 3 hours approx)
      mocage_dict['--time'] = '%02i:%02i:00' % (int(nh/8),int((nh % 8)/8.*60))
      # Set the final forecast if needed
      if "final forecast" in cfg and cfg["final forecast"]["run"]:
        fflength = isodur.parse_duration(cfg["final forecast"]["length"])
        nhff = fflength.total_seconds()/3600.
        # Increase the total run time accounting the final forecast
        mocage_dict['--time'] = '%02i:%02i:00' % (int(nh/8)+int(nhff/16),
          min(int((nh % 8)/8.*60)+int((nhff % 16)/16.*60),59))
        # Update the mocage job final date
        mocage_dict['DATEOUT'] = (dateenddt+fflength).strftime('%Y%m%d%H')

    print('>> Preparing working dir in ', workdir)
    if not os.path.exists(workspace): os.makedirs(workspace)
    if os.path.exists(workdir): shutil.rmtree(workdir)
    os.mkdir(workdir)

    # Update the restart from the previous cycle if it's not the first simulated period
    if int(kwargs['cycle']) > 1 and kwargs['taskname'] in ('forecast','model_error_forecast','3dvar'):
      # 3dvar is initialized with 3dvar analysis from last cycle
      if kwargs['taskname'] == '3dvar':
        cfg['depends on'] = {
          '3dvar': {'cycle':-1}
        }
      # forecast, model_error_forecast are initialized with forecast corrected by envar from last cycle
      else:
        cfg['depends on'] = {
          'forecast': {'cycle':-1},
          'envar': {'cycle':-1,'grid':'MACC01'} # Improve this, where can it go?
        }

    if cassim and "observations path" in cfg: mocage_dict['DIR_AS'] = cfg["observations path"]

    if "depends on" in cfg: 
      if 'init species' in cfg and 'envar' in cfg['depends on']: 
        cfg['depends on']['envar']['init species'] = cfg['init species']
      setup_mocage_restart(dateinidt,cfg,mocage_dict,workdir,kwargs)

    print('>> Modifying MOCAGE job...')
    create_mocage_job(refjobfile,kwargs['jobname'],mocage_dict)

  def setup(self, ecflowbr, **kwargs):

    super().setup(ecflowbr)
    if 'computing' in self.namelist and 'run' in self.namelist['computing']:
      runhostname = self.namelist['computing']['run']['machine']
    else:
      runhostname = 'laptop'
    jobname = os.path.join(self.namelist['mocage path'],'jobs/mocage/rundir',
      self.namelist['reference job']).replace('_L.sh','_%s_%s_cycle_%04i_L.sh' % 
      (self.expname,self.taskname,self.cycle),1)
    self.ecfstep.add_variable('runhostname',runhostname)
    self.ecfstep.add_variable("jobname",jobname)
    self.ecfstep.add_variable('jobmode',machines_config(runhostname)['mode'])

    if 'basename' in self.__dict__:
      jobpath = os.path.join(self.suite_path,'jobs',self.basename,self.taskname)
    else:
      jobpath = os.path.join(self.suite_path,'jobs',self.taskname)

    if not os.path.exists(jobpath): os.makedirs(jobpath)

    jobout = os.path.join(jobpath,'%s-cycle_%04i.out' % (self.taskname,self.cycle))
    self.ecfstep.add_variable("jobout",jobout)
    joberr = os.path.join(jobpath,'%s-cycle_%04i.err' % (self.taskname,self.cycle))
    self.ecfstep.add_variable("joberr",joberr)

    self.ecfstep.add_variable("namelist",self.namelist_file)

  @classmethod
  def py_finalize_script(cls):

    import daimonpp as postp

    kwargs = cls.load_ecfvars()

    # Load the variational namelist here (my version, not the oops one)
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)

    pltnamelist = cfg['postproc namelist']
    with open(pltnamelist) as file:
        postcfg = yaml.load(file,Loader=Loader)

    starttime = isodatetime.parse_datetime(kwargs['dateini']).strftime('%Y%m%d%H')
    endtime = isodatetime.parse_datetime(kwargs['dateend']).strftime('%Y%m%d%H')
    # Set plt namelist elements according to the experiment name and dates
    postcfg['expname'] = kwargs['expname']+'_'+kwargs['taskname']
    postcfg['starttime'] = starttime
    postcfg['endtime'] = endtime
    outdir = kwargs['outpath']

    if "postproc options" in cfg:
      if "length" in cfg["postproc options"]: postcfg['endtime'] = (isodatetime.parse_datetime(kwargs['dateini'])+ \
        isodur.parse_duration(cfg["postproc options"]["length"])).strftime('%Y%m%d%H')

    # In case it is an ensemble member adapt the input/output folder 
    # (not very nice but I do not want to recopy the finalize script in the ensemble_member class)
    if kwargs['taskname'] == 'ensemble_member': 
      outdir = os.path.join(outdir,"E%s" % (kwargs['member']))

    print('Forecast postprocessor working in ',outdir)

    # Check wether the mocage simulation is in validation / assimilation mode
    # cassim indicates either validation mode or assimilation. In case of assimilation only the analysis will be
    # postprocessed
    try:
      cassim = bool(get_mocage_param(kwargs['jobname'],'CASSIM'))
    except:
      print('Could not find the MOCAGE forecast job %s' % (kwargs['jobname']))
      cassim = False
      
    if cassim and "assimilation" in cfg and cfg["assimilation"]: 
      analysis = True
      subfolder = "ANALYSIS"
      obstypes = ('Hx_bkg','Hx_ana')
    else:
      analysis = False
      subfolder = "FORECAST"
      obstypes = ('Hx_dry',)

    # Renaming of daimon output
    if cassim:
        print("Linking forecast member files...")
        postp.linknc('f',outdir,domains=postcfg['links']['domains'])
        if analysis: 
          postp.linknc('a',outdir,domains=postcfg['links']['domains'])
          if "final forecast" in cfg and cfg["final forecast"]["run"]:
            postp.linknc('ff',outdir,domains=postcfg['links']['domains'])

        with open(os.path.join(outdir,'assimilation_namelist.json')) as file:
          assimcfg = json.load(file)

    # Plots here
    if "plots" in postcfg:

        if 'modelflds' in postcfg['plots']:

            outdirtmp = outdir
            if cassim: outdirtmp = os.path.join(outdir,subfolder)

            if 'hfields' in postcfg['plots']['modelflds']:

                postp.plot_hfields(postcfg,outdirtmp,rtype='fct')

            if 'vfields' in postcfg['plots']['modelflds']:

                postp.plot_vfields(postcfg,outdirtmp,rtype='fct')

        if cassim and 'increments' in postcfg['plots']:

            if 'hfields' in postcfg['plots']['increments']:

                postp.plot_hincrem(postcfg,outdir,Method=assimcfg['Assimilation']['Method'],
                  WindowsLength=assimcfg['Assimilation']['Cycling']['WindowsLength'])

            if 'vfields' in postcfg['plots']['increments']:

                postp.plot_vincrem(postcfg,outdir,Method=assimcfg['Assimilation']['Method'],
                  WindowsLength=assimcfg['Assimilation']['Cycling']['WindowsLength'])

        if cassim and 'obs' in postcfg['plots']:

            subwindow = assimcfg['Assimilation']['Cycling']['WindowsLength']
            
            # Misfits
            postp.plot_hstat_snapshots(postcfg,outdir,WindowsLength=subwindow,ObsType=obstypes)
                    
            postp.plot_hstat_tser(postcfg,outdir,ObsType=obstypes)

    # Obs generator (for OSSEs)
    if cassim and 'make obs' in cfg and cfg["make obs"]["activate"]:

      cfg["make obs"]['starttime'] = starttime
      cfg["make obs"]['endtime'] = endtime

      postp.generate_obs(cfg["make obs"],outdir)

    # Compare with other forecasts
    if "compare to" in cfg:

      for name,exp in cfg["compare to"].items():

        if "path" in exp:
          expdir = exp["path"]
        else:
          expdir = os.path.join(kwargs["exppath"],kwargs['expname'],'cycle_%04i' % (int(kwargs['cycle'])),name)
        if name in ("forecast","truth"): expdir = os.path.join(expdir,'FORECAST')  # can do better than hardcoded like this?
        outdirtmp = outdir
        if cassim: outdirtmp = os.path.join(outdir,subfolder)
        f2type = 'HM'
        freq=1
        if name in ('envar',) : 
          f2type = 'A4'
          freq=1
        postp.fields_diff(outdirtmp,expdir,
            basename=name,
            starttime=starttime,endtime=endtime,
            f2type=f2type,freq=freq,
            fldslist=postcfg['plots']['modelflds']['species'],
            domains=postcfg['plots']['modelflds']['domains'],
            )

        if "plots" in postcfg:

          if 'modelflds' in postcfg['plots']:

            if 'hfields' in postcfg['plots']['modelflds']:

              postp.plot_hfields(postcfg,os.path.join(outdirtmp,'fields_diff',name),ftype='HD',rtype='fct')

            if 'vfields' in postcfg['plots']['modelflds']:

              postp.plot_vfields(postcfg,os.path.join(outdirtmp,'fields_diff',name),ftype='HD',rtype='fct')

    if "correct with" in cfg:

        name = list(cfg["correct with"].keys())[0]
        exp = cfg["correct with"][name]
        if "path" in exp:
          expdir = exp["path"]
        else:
          expdir = os.path.join(kwargs["exppath"],kwargs['expname'],'cycle_%04i' % (int(kwargs['cycle'])),name)
        expdir = os.path.join(expdir,'FORECAST/fields_diff/envar') # can do better than hardcoded like this?
        outdirtmp = outdir
        if cassim: outdirtmp = os.path.join(outdir,subfolder)
        f2type = 'HD'
        postp.fields_diff(outdirtmp,expdir,
            shift={'date':isodatetime.parse_datetime(kwargs['dateini'])-dt.timedelta(hours=24)},
            starttime=starttime,endtime=endtime,freq=1,
            f2type=f2type,ftypeout='HC',subfolder='bias_correction',
            domains=postcfg['plots']['modelflds']['domains'])

        if "plots" in postcfg:

          postcfg['expname'] += '_biascorr'

          if 'modelflds' in postcfg['plots']:

            if 'hfields' in postcfg['plots']['modelflds']:

              postp.plot_hfields(postcfg,os.path.join(outdirtmp,"bias_correction"),ftype='HC',rtype='fct')

            if 'vfields' in postcfg['plots']['modelflds']:

              postp.plot_vfields(postcfg,os.path.join(outdirtmp,"bias_correction"),ftype='HC',rtype='fct')

        if "compare to" in cfg:

          for name,exp in cfg["compare to"].items():

            if "path" in exp:
              expdir = exp["path"]
            else:
              expdir = os.path.join(kwargs["exppath"],kwargs['expname'],'cycle_%04i' % (int(kwargs['cycle'])),name)
            if name in ("forecast","truth"): expdir = os.path.join(expdir,'FORECAST')  # can do better than hardcoded like this?

            outdirtmp = os.path.join(outdirtmp,"bias_correction")
            f2type = 'HM'
            if name in ('envar',) : f2type = 'A4'
            postp.fields_diff(outdirtmp,expdir,basename=name,
                starttime=starttime,endtime=endtime,
                ftype='HC',
                f2type=f2type,
                domains=postcfg['plots']['modelflds']['domains'])

            if "plots" in postcfg:

              if 'modelflds' in postcfg['plots']:

                if 'hfields' in postcfg['plots']['modelflds']:

                  postp.plot_hfields(postcfg,os.path.join(outdirtmp,'fields_diff',name),ftype='HD',rtype='fct')

                if 'vfields' in postcfg['plots']['modelflds']:

                  postp.plot_vfields(postcfg,os.path.join(outdirtmp,'fields_diff',name),ftype='HD',rtype='fct')

    # Add write mode to expdir for the unix group
    cmd="chmod -R g+w {}".format(os.path.join(kwargs["exppath"],kwargs["expname"]))
    subprocess.run(cmd,shell=True)

# A modifier for MOCAGE shell jobs
def create_mocage_job(refjobfile,newjobfile,dictionary):
  #Read reference job (date,obs,dir)
  f=open(refjobfile,'r')
  refjob=f.read()
  f.close()

  #Replace selected parameters
  for key in dictionary:
    line = [l for l in refjob.splitlines() if key+'=' in l]
    if len(line) > 1: 
      # Detect and remove lines that contain the key only but within a comment
      line = [l for l in line if ('#' in l and l.find(key) < l.find('#')) or ('#' not in l) or ('#SBATCH' in l)]
      if len(line) > 1:
        print('ERROR: more than one line contain the same keyword. Check your job.')
        print(line)
        sys.exit()
    line = line[0]
    print(line)
    print('Changing '+key+' to '+str(dictionary[key]))
    if '#SBATCH' in line:
      cmd = line.split('#SBATCH ',1)[1]
      newline = ''
      lkey = cmd.split('=',1)[0].strip()
      lval = cmd.split('=',1)[1].strip()
      newline += lkey+'='
      if lkey == key: 
        newline+=str(dictionary[key])
      else:
        newline+=lval
      newline='#SBATCH '+newline
    else: 
      linecmd = line.split('#',1)[0]
      newline = ''
      for cmd in linecmd.split(';'):
        lkey = cmd.split('=',1)[0].strip()
        lval = cmd.split('=',1)[1].strip()
        newline += lkey+'='
        if lkey == key: 
          newline+=str(dictionary[key])
        else:
          newline+=lval
        newline+=' ; '
      newline = newline[:-3]
      if '#' in line: newline+='    # '+line.split('#',1)[1]
    refjob = refjob.replace(line,newline,1)   

  #Write new jobfile (replace if one with same name exists)
  if os.path.exists(newjobfile): os.remove(newjobfile)
  f=open(newjobfile,'w')
  f.write(refjob)
  f.close()
  #Set executing permissions
  os.chmod(newjobfile,stat.S_IRWXU)
  return

def get_mocage_param(refjobfile,key):

  f=open(refjobfile,'r')
  refjob=f.read()
  f.close()

  line = [l for l in refjob.splitlines() if key+'=' in l]
  if len(line) > 1: 
    # Detect and remove lines that contain the key but only within a comment
    line = [l for l in line if ('#' in l and l.find(key) < l.find('#')) or ('#' not in l)]
    if len(line) > 1:
      print('ERROR: more than one line contain the same keyword. Check your job.')
      print(line)
      sys.exit()

  line = line[0]
  print(line)
  # print('Changing '+key+' to '+dictionary[key])
  linecmd = line.split('#',1)[0]
  # newline = ''
  for cmd in linecmd.split(';'):
    lval = cmd.split('=',1)[1].strip()

  return lval

def update_mocage_assim(refjobname,kwargs,cfg):

  refassimnml = refjobname.replace('jobdaimon_','assimilation_namelist_')
  refassimnml = refassimnml.replace('_L.sh','.json')
  # It would be better if the namelist name was modifiable through MOCAGE namelist. It is not of course today.
  # newassimnml = refassimnml.replace('.json',kwargs['expname']+'_'+kwargs['taskname']+'.json')
  with open(refassimnml) as file:
    assimcfg = json.load(file)
  # Update the number of assimilation windows
  assimcfg['Assimilation']['Cycling']['NbOfWindows'] = \
    int((dt.datetime.fromisoformat(kwargs['dateend']) - dt.datetime.fromisoformat(kwargs['dateini'])) / \
     dt.timedelta(seconds=assimcfg['Assimilation']['Cycling']['WindowsLength']))
  # Update the final forecast option
  if "final forecast" in cfg:
    assimcfg['Assimilation']['Cycling']['FinalForecast'] = cfg['final forecast']['run']
  # The original namelist is replaced, not optimal. 
  # Parallel executions with same NAMEXP and different namelists not possible.
  with open(refassimnml,"w") as file:
    json.dump(assimcfg,file,indent=4)
  # mocage_dict['assimilation_namelist'] = newassimnml

def setup_mocage_restart(dateini,cfg,mocage_dict,workdir,kwargs):
    
  possible_restarts = ('truth','forecast','perturbed_forecast','3dvar')

  # Set MOCAGE restart
  el = [k for k in possible_restarts if k in cfg["depends on"]]
  if len(el) > 0:
    print('Setting up MOCAGE forecast restart folder')
    if len(el) != 1:
      print("Error: the forecast restart must be either:")
      print(possible_restarts)
      sys.exit(1)
    else:
      fctname = el[0]
      if "path" in cfg["depends on"][fctname]:
        pathhm = cfg["depends on"][fctname]["path"]
      else:
        cycle = int(kwargs['cycle'])
        if 'cycle' in cfg["depends on"][fctname]: cycle+=cfg["depends on"][fctname]["cycle"]
        pathhm = os.path.join(kwargs["exppath"],kwargs["expname"],'cycle_%04i' % cycle,fctname)
      mocage_dict['DIR_R1'] = pathhm
      if not os.path.exists(pathhm):
        sys.exit("Error: the init path does not exists: %s" % (pathhm))

  # Set assimilation increments
  possible_analyses = ('envar',)

  el = [k for k in possible_analyses if k in cfg["depends on"]]
  if len(el) > 0:
    print('Link assimilation increments in workdir...')
    if len(el) != 1:
      print("Error: the increments must be either:")
      print(possible_analyses)
      sys.exit(1)
    else:
      anlname = el[0]
      if "path" in cfg["depends on"][anlname]:
        patha = cfg["depends on"][anlname]["path"]
      else:
        cycle = int(kwargs['cycle'])
        if 'cycle' in cfg["depends on"][anlname]: cycle+=cfg["depends on"][anlname]["cycle"]
        patha = os.path.join(kwargs["exppath"],kwargs["expname"],'cycle_%04i' % cycle,anlname)
      src = os.path.join(patha,'A4'+cfg["depends on"][anlname]["grid"]+'+'+dateini.strftime('%Y%m%d%H')+'.nc')
      dst = os.path.join(workdir,'A4'+cfg["depends on"][anlname]["grid"]+'+'+dateini.strftime('%Y%m%d%H')+'.nc')
      print('Prepare init file...')
      if not os.path.exists(src):
        sys.exit("Error: the analysis file does not exists: %s" % (src))
      if 'init species' in cfg["depends on"][anlname]:
        # Store only wanted species for the model restart
        with Dataset(src) as srcds, Dataset(dst, 'w') as dstds:
          copy_netcdf_vars(srcds,dstds,cfg["depends on"][anlname]['init species'])
      else:
        os.symlink(src,dst)

def copy_netcdf_vars(src, dst, varlist):

        dst.setncatts(src.__dict__)

        for name, dimension in src.dimensions.items():
            if name not in dst.dimensions.keys():
                dst.createDimension(name, (len(dimension) if not dimension.isunlimited() else None))
                x = dst.createVariable(name, src.variables[name].datatype, 
                    src.variables[name].dimensions, fill_value=None)
                dst[name][:] = src[name][:]

        for name, variable in src.variables.items():
            if name not in dst.variables.keys() and (name in varlist):
                fill_value = None
                if hasattr(src[name], "_FillValue"): fill_value = src[name]._FillValue
                x = dst.createVariable(name, variable.datatype, 
                    variable.dimensions, fill_value=fill_value)
                # copy variable attributes
                dst[name].setncatts({k: src[name].getncattr(k) 
                    for k in src[name].ncattrs() if k not in ["_FillValue"]})
                dst[name][:] = src[name][:]
