# -*- coding: utf-8 -*-
"""
Class handling the various ecflow steps for assimilation experiments 
"""

import os
import sys
import glob
import yaml
from yaml import CLoader as Loader, CDumper as Dumper
import isodate.isodatetime as isodatetime
import isodate.isoduration as isodur
from datetime import timedelta
from Jedi_task import Jedi_task, ListMocageOutput, GetForecast, PrepareJob

class Hofx(Jedi_task):

  actions = Jedi_task.actions + ['finalize']

  def __init__(self, **kwargs):

    super().__init__(**kwargs)

  @classmethod
  def py_setup_script(cls):

    kwargs = cls.load_ecfvars()

    # Load the variational namelist here (my version, not the complete oops one)
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)
    # Remember that expname needs to come through ecflow because this method does not know self
    # and does not use the global experiment namelist! I can come back on this decision if I find 
    # out that a global namelist avoids a lot of duplication (e.g. output ensemble -> input dirac 
    # must be specified twice). In that case remove all the static ecflow arguments in favor
    # of the namelist usage.
    fctname, bkgdir = GetForecast(cfg,kwargs)

    obsdir = cfg["observations path"]

    runtype = "fct"
    if "state type" in cfg: runtype = cfg["state type"]
    print(fctname, bkgdir, runtype)
    subpath = ''
    if runtype in ('val','fct'):
      subpath = 'FORECAST'
    if runtype in ('3dvar',):
      subpath = 'ANALYSIS'
    if runtype in ('final forecast',):
      subpath = 'FINAL_FORECAST'
    outdir = os.path.join(bkgdir,subpath,'hofx')

    if not os.path.exists(outdir): os.makedirs(outdir)

    print('>> Initializing Hofx: namelist creation...')

    # Basic structure
    outyaml = {}

    dateini = isodatetime.parse_datetime(kwargs['dateini'])
    dateend = isodatetime.parse_datetime(kwargs['dateend'])

    dtwinlen = dateend - dateini

    outyaml['window length'] = cfg['subwindow']

    outyaml['geometry'] = cfg['grid']

    chemsp = cfg['state variables']

    dtsubwin = isodur.parse_duration(cfg['subwindow'])
    nsub = int(dtwinlen/dtsubwin)+1
    # To be consistent with 4denvar compare to obs that end up in the previous 1hour window.
    # We end up having an empty first window if using files that do not contain obs for the previous day.
    # Is it valid also for windows longer than 1h and different obs slots? I don't know.
    dates =  [dateini+t*dtsubwin for t in range(nsub)]

    # Build ntimes Hofx3d namelists
    for it,date in enumerate(dates):

        outyaml['window begin'] = (date-isodur.parse_duration('PT30M')).isoformat()+'Z'

        obsdatastr = date.strftime('%Y%m%d') # To keep compatibility with MOCAGE HDATs
        # A fix since obs for midnight are set at 23:30, so we need to use previous date for the filename
        if date.hour == 0: obsdatastr = (date-timedelta(days=1)).strftime('%Y%m%d')
        obsfilename = os.path.join(obsdir,'HDAT+'+obsdatastr+'.h5')
        if not os.path.exists(obsfilename): os.sys.exit("%s not found" % (obsfilename))
        
        outyaml['state'] = ListMocageOutput(bkgdir,[date],chemsp,outyaml['geometry']['domname'],runtype=runtype)

        outyaml['state']['date'] = date.isoformat()+'Z'

        outyaml['observations'] = {'observers':[]}

        for obs in cfg['observations']:
            obsdict = {
                'obs operator':obs['obs operator'],
                'obs space':{
                    'obsdatain':{'engine':{'obsfile':obsfilename}},
                    'obsdataout':{'engine':{'obsfile':os.path.join(outdir,'HSTAT+'+dateini.strftime('%Y%m%d')+'_%i.h5' % (it))}},
                    'obs type':obs['obs space']['obs type'],
                    'instr name':obs['obs space']['instr name'],
                    }
                }
            # Optional elements
            # if 'obs filters' in obs: obsdict['obs filters'] = obs['obs filters']

            outyaml['observations']['observers'].append(obsdict)

        print(yaml.dump(outyaml))

        with open(kwargs['jedinamelist']+'-step_%02i.yaml' % (it),'w') as f:
            print(yaml.dump(outyaml),file=f)

    script = PrepareJob(exe=os.path.join(cfg['oops path'],'aq_hofx3d.x'),namelist=kwargs['jedinamelist'],
        nsteps=len(dates),mpiprocs=6,
        jobout=kwargs['jobout'],
        errout=kwargs['joberr'])
    with open(kwargs['jobname'],'w') as f:
        print(script,file=f)
    os.chmod(kwargs['jobname'], 0o755)

  @classmethod
  def py_finalize_script(cls):

    import daimonpp as postp

    kwargs = cls.load_ecfvars()

    # Load the variational namelist here (my version, not the oops one)
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)

    pltnamelist = cfg['postproc namelist']
    with open(pltnamelist) as file:
        postcfg = yaml.load(file,Loader=Loader)

    starttime = isodatetime.parse_datetime(kwargs['dateini']).strftime('%Y%m%d%H')
    endtime = isodatetime.parse_datetime(kwargs['dateend']).strftime('%Y%m%d%H')
    # Set plt namelist elements according to the experiment name and dates
    postcfg['expname'] = kwargs['expname']+'_'+kwargs['taskname']
    postcfg['starttime'] = starttime
    postcfg['endtime'] = endtime

    fctname, bkgdir = GetForecast(cfg,kwargs)

    runtype = "fct"
    if "state type" in cfg: runtype = cfg["state type"]
    subpath = ''
    if runtype in ('val','fct'):
      subpath = 'FORECAST'
    if runtype in ('3dvar',):
      subpath = 'ANALYSIS'
    if runtype in ('final forecast',):
      subpath = 'FINAL_FORECAST'
    outdir = os.path.join(bkgdir,subpath,'hofx')

    # Plots here
    if "plots" in postcfg:

        if 'obs' in postcfg['plots']:

            subwindow = isodur.parse_duration(cfg['subwindow'])
            # Misfits and innovation
            postp.plot_hstat_snapshots(postcfg,outdir,origin='aq4d',WindowsLength=subwindow.total_seconds(),ObsType=["Hx_dry"])
                    
            postp.plot_hstat_tser(postcfg,outdir,origin='aq4d',WindowsLength=subwindow.total_seconds(),ObsType=["Hx_dry"])
