# -*- coding: utf-8 -*-
"""
Class handling the various ecflow steps for assimilation experiments 
"""

import os
import sys
import glob
from aq_steps import Step, machines_config

class Jedi_task(Step):
  
  ecfvars = Step.ecfvars + ['namelist','jedinamelist','jobname','jobout','joberr']
  actions = Step.actions + ['setup','run']

  def __init__(self, **kwargs):

    super().__init__(**kwargs)

    # Save needed ecf scripts and namelists
    self.__class__.create_ecf_scripts(path=os.path.join(self.suite_path,'tasks'))
    # Or create here the specific namelist
    self.save_namelist()

  @classmethod
  def ecf_script_run(cls, **kwargs):

    ll = '%include <head.h>\n'

    ll+=r"""if [ "%jobmode%" = "interactive" ]"""
    ll+='\nthen\n'
    ll+=r"""ssh %runhostname% '%jobname% 1> %jobout% 2> %joberr%'"""
    ll+='\nelse\n'
    ll+=r"""ssh_out=$(ssh %runhostname% 'jobid=$(/usr/bin/sbatch %jobname% | awk "{print \$4}")&& while [ $? -eq 0 ]; do sleep 5s&& squeue -j $jobid | grep $jobid > /dev/null; done; if grep "OOPS Ending" %jobout% >out.tmp 2>&1; then echo ${jobid} finished OK && exit 0; else echo ${jobid} did not finish OK && exit 22; fi')"""
    ll+='\nfi\n'

    ll+='\n'
    ll+= '%include <tail.h>\n'
    with open(os.path.join(kwargs['path'],cls.__name__+'_run.ecf'),'w') as f:
      f.writelines(ll)

  def setup(self, ecflowbr, **kwargs):

    super().setup(ecflowbr)
    if 'computing' in self.namelist and 'run' in self.namelist['computing']:
      runhostname = self.namelist['computing']['run']['machine']
    else:
      runhostname = 'kglobc1'
    self.ecfstep.add_variable('runhostname',runhostname)
    self.ecfstep.add_variable('jobmode',machines_config(runhostname)['mode'])
    self.ecfstep.add_variable("namelist",self.namelist_file)

    if 'basename' in self.__dict__:
      nmlpath = os.path.join(self.suite_path,'namelists',self.basename,self.taskname)
      jobpath = os.path.join(self.suite_path,'jobs',self.basename,self.taskname)
    else:
      nmlpath = os.path.join(self.suite_path,'namelists',self.taskname)
      jobpath = os.path.join(self.suite_path,'jobs',self.taskname)

    if not os.path.exists(nmlpath): os.makedirs(nmlpath)
    if not os.path.exists(jobpath): os.makedirs(jobpath)

    # This is the name of the actual oops namelist
    jedinamelist = os.path.join(nmlpath,'%s-cycle_%04i.yaml' % (self.taskname,self.cycle))
    self.ecfstep.add_variable('jedinamelist',jedinamelist)
    jobname = os.path.join(jobpath,'%s-cycle_%04i.sh' % (self.taskname,self.cycle))
    self.ecfstep.add_variable("jobname",jobname)
    jobout = os.path.join(jobpath,'%s-cycle_%04i.out' % (self.taskname,self.cycle))
    self.ecfstep.add_variable("jobout",jobout)
    joberr = os.path.join(jobpath,'%s-cycle_%04i.err' % (self.taskname,self.cycle))
    self.ecfstep.add_variable("joberr",joberr)

def PrepareJob(**kwargs):

    ompthreads = 1
    nsteps = 1
    if 'ompthreads' in kwargs: ompthreads = kwargs['ompthreads']
    if 'nsteps' in kwargs: nsteps = kwargs['nsteps']
    lines = '#!/bin/bash\n'
    if 'mode' in kwargs and kwargs['mode'] == 'batch':
      # lines+= '#SBATCH -A ae\n'
      lines+= '#SBATCH -J envar\n'
      lines+= '#SBATCH --nodes=%i\n' % kwargs['nodes']
      lines+= '#SBATCH --partition=%s\n' % kwargs['queue']
      lines+= '#SBATCH --qos=special\n'
      lines+= '#SBATCH --time=%02i:%02i:%02i\n' % (kwargs['time'].total_seconds()/3600,0,0)
      lines+= '#SBATCH --mail-user=emili@cerfacs.fr\n'
      lines+= '#SBATCH --mail-type=END\n'
      lines+= '#SBATCH --output=%s\n' % (kwargs['jobout'])
      lines+= '#SBATCH --error=%s\n' % (kwargs['errout'])
    
    lines+= 'source loadaq.sh\n'
    lines+= 'export I_MPI_PIN_DOMAIN=omp\n'
    
    if 'machine' in kwargs and kwargs['machine'] in ('kraken',):
      lines+= 'export KMP_DETERMINISTIC_REDUCTION=true\n'
      lines+= 'export ROMIO_HINTS=/data/softs/Modules/modulefiles/mpi/intelmpi/romio-hints\n'
      lines+= 'export I_MPI_HYDRA_BOOTSTRAP=ssh\n'
      mpidist = '-np %i --ppn %i' % (kwargs['mpiprocs'],18) # Check if this makes sense
      # mpidist = '-np %i' % (kwargs['mpiprocs'])
    else:
      mpidist = '-np %i' % (kwargs['mpiprocs'])

    lines+= 'export OMP_NUM_THREADS=%i\n' % ompthreads
    if nsteps > 1:
      for it in range(kwargs['nsteps']):
        lines+= 'mpirun %s %s %s\n' % (mpidist,kwargs['exe'],kwargs['namelist']+'-step_%02i.yaml' % (it))
    else:
      lines+= 'mpirun %s %s %s' % (mpidist,kwargs['exe'],kwargs['namelist'])

    return lines

# Build a list of ntime states, optionally an ensemble of ntime states
def ListMocageOutput(path,dates,statevars,grid,**kwargs):

  import daimonpp as postp

  mode = 'single'
  runtype = 'fct'
  if 'mode' in kwargs: mode = kwargs['mode']
  if 'runtype' in kwargs: runtype = kwargs['runtype']

  if not os.path.exists(path):
    print("Error: the following background/ensemble path does not exists")
    print(path)
    sys.exit("Model/ensemble path not found")

  prefix = 'HM'
  if runtype == 'corrected': prefix = 'HC'
  if runtype == 'envar': prefix = 'A4'

  if mode == 'ensemble':
    # Handle different type of output files from mocage (ensembles of 3dvar analysis are not treated yet)
    if runtype in ('val','fct'):
        ensdirs = glob.glob(os.path.join(path,'E[0-9]*'))
        ensdirs.sort()
        # Link the wanted states first to avoid the strange mocage naming (based on dates)
        for d,rep in enumerate(ensdirs):
          postp.linknc('f',rep,dates=dates,domains=[grid])
        if 'ensize' in kwargs: ensdirs = ensdirs[:kwargs['ensize']]
    else:
        ensize = len(glob.glob(os.path.join(path,prefix+grid+'+'+dates[0].strftime('%Y%m%d%H')+'+E[0-9]*')))
        if 'ensize' in kwargs: ensize = min(ensize,kwargs['ensize'])
      
    members = []

    if runtype in ('val','fct'):
      for d,rep in enumerate(ensdirs):
        if len(dates) > 1:
          members.append({'states':[]})
          for date in dates:
            filename = os.path.join(rep,'FORECAST',prefix+grid+'+'+date.strftime('%Y%m%d%H')+'.nc')
            if not os.path.exists(filename): sys.exit("File path not found")
            members[d]['states'].append({'date':date.isoformat()+'Z','state variables':statevars,'filename':filename})
        else:
          filename = os.path.join(rep,'FORECAST',prefix+grid+'+'+dates[0].strftime('%Y%m%d%H')+'.nc')
          if not os.path.exists(filename): sys.exit("File path not found")
          members.append({'date':dates[0].isoformat()+'Z','state variables':statevars,'filename':filename})
    else:
      for d in range(ensize):
        if len(dates) > 1:
          members.append({'states':[]})
          for date in dates:
            suffix = '+E%04i' % (d+1)
            filename = os.path.join(path,prefix+grid+'+'+date.strftime('%Y%m%d%H')+suffix+'.nc')
            if not os.path.exists(filename): sys.exit("File path not found")
            members[d]['states'].append({'date':date.isoformat()+'Z','state variables':statevars,'filename':filename})
        else:
          suffix = '+E%04i' % (d+1)
          filename = os.path.join(path,prefix+grid+'+'+dates[0].strftime('%Y%m%d%H')+suffix+'.nc')
          if not os.path.exists(filename): sys.exit("File path not found")
          members.append({'date':dates[0].isoformat()+'Z','state variables':statevars,'filename':filename})

    return members

  else:

    subpath = ''
    if runtype in ('val','fct'):
      postp.linknc('f',path,dates=dates,domains=[grid])
      subpath = 'FORECAST'
    if runtype in ('3dvar',):
      postp.linknc('a',path,dates=dates,domains=[grid])
      subpath = 'ANALYSIS'
    if runtype in ('final forecast',):
      postp.linknc('ff',path,dates=dates,domains=[grid])
      subpath = 'FINAL_FORECAST'

    if len(dates) > 1:
      states = []
      for date in dates:
        filename = os.path.join(path,subpath,prefix+grid+'+'+date.strftime('%Y%m%d%H')+'.nc')
        if not os.path.exists(filename): sys.exit("File %s not found" % (filename))
        states.append({'date':date.isoformat()+'Z','state variables':statevars,'filename':filename})
    else:
      filename = os.path.join(path,subpath,prefix+grid+'+'+dates[0].strftime('%Y%m%d%H')+'.nc')
      if not os.path.exists(filename): sys.exit("File %s not found" % (filename))
      states = {'date':dates[0].isoformat()+'Z','state variables':statevars,'filename':filename}

    return states

# Return the forecast path and name based on the dependence dictionary
def GetForecast(cfg,kwargs):
    
  # The forecast can be instructed through an explicit dependance 
  if "depends on" in cfg:
    possible_forecasts = ('forecast','perturbed_forecast','final_forecast')
    el = [k for k in possible_forecasts if k in cfg["depends on"]]
    if len(el) > 0:
      if len(el) != 1:
        print("Error: the forecast restart must be either:")
        print(possible_forecasts)
        sys.exit("Model path not found")
      else:
        fctname = el[0]
        if "path" in cfg["depends on"][fctname]: 
          fctdir = cfg["depends on"][fctname]["path"]
        else:
          cycle = int(kwargs['cycle'])
          if 'cycle' in cfg["depends on"][fctname]: cycle+=cfg["depends on"][fctname]["cycle"]
          fctdir = os.path.join(kwargs["exppath"],kwargs["expname"],'cycle_%04i' % cycle,fctname)
  # , if not, the path is extracted from the parent variable
  else:
    fctname = "forecast"  # not really necessary?
    fctdir = kwargs["outpath"]

  # An ugly patch to treat special cases
  if "state type" in cfg and cfg["state type"] == "corrected": 
    fctdir = os.path.join(fctdir,"FORECAST","bias_correction")

  if not os.path.exists(fctdir): sys.exit("Model path %s not found" % (fctdir))

  return fctname, fctdir

# Return the ensemble path and name based on the dependence dictionary
def GetEnsemble(cfg,kwargs):

  # The forecast can be instructed through an explicit dependance 
  if "depends on" in cfg and "ensemble" in cfg["depends on"]:
    if "path" in cfg["depends on"]["ensemble"]: 
      ensdir = cfg["depends on"]["ensemble"]["path"]
    else:
      cycle = int(kwargs['cycle'])
      if 'cycle' in cfg["depends on"]["ensemble"]: cycle+=cfg["depends on"]["ensemble"]["cycle"]
      ensdir = os.path.join(kwargs["exppath"],kwargs["expname"],'cycle_%04i' % cycle,"ensemble")
  # , if not, the path is extracted from the parent variable
  else:
    ensdir = kwargs["outpath"]

  if not os.path.exists(ensdir): sys.exit("Ensemble path %s not found" % (ensdir))

  return "ensemble", ensdir