# -*- coding: utf-8 -*-
"""
Class handling the various ecflow steps for assimilation experiments 
"""

import os
import yaml
import math
from yaml import CLoader as Loader, CDumper as Dumper
import isodate.isodatetime as isodatetime
import isodate.isoduration as isodur
from datetime import timedelta
from Jedi_task import Jedi_task, ListMocageOutput, GetForecast, GetEnsemble, PrepareJob
from aq_steps import machines_config
import itertools

class Variational(Jedi_task):

  actions = Jedi_task.actions + ['finalize']

  def __init__(self, **kwargs):

    super().__init__(**kwargs)

  @classmethod
  def py_setup_script(cls):

    kwargs = cls.load_ecfvars()

    # Load the variational namelist here (my version, not the complete oops one)
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)
    # Remember that expname needs to come through ecflow because this method does not know self
    # and does not use the global experiment namelist! I can come back on this decision if I find 
    # out that a global namelist avoids a lot of duplication (e.g. output ensemble -> input dirac 
    # must be specified twice). In that case remove all the static ecflow arguments in favor
    # of the namelist usage.
    ensname, ensdir = GetEnsemble(cfg,kwargs)
    fctname, bkgdir = GetForecast(cfg,kwargs)

    outdir = kwargs['outpath']
    locdatadir = os.path.join(kwargs["exppath"],kwargs["expname"],'cycle_%04i' % (int(kwargs["cycle"])),'localization') #Default localization files
    if 'path' in cfg["localization"]: locdatadir = cfg["localization"]["path"]
    obsdir = cfg["observations path"]
    cftype = '4D'
    ens_size = None
    if 'ensemble size' in cfg: ens_size = cfg['ensemble size']
    if 'type' in cfg['cost function']: cftype = cfg['cost function']['type']

    if not os.path.exists(outdir): os.makedirs(outdir)

    print('>> Initializing 4DEnsVar: namelist creation...')

    # Basic structure
    outyaml = {'cost function':{},'variational':{},'final':{},'output':{}}

    dateini = isodatetime.parse_datetime(kwargs['dateini'])
    dateend = isodatetime.parse_datetime(kwargs['dateend'])

    costf = outyaml['cost function']

    dtwinlen = dateend - dateini
    chemsp = cfg['cost function']['analysis variables']

    if cftype == '3D': 
        costf['cost type'] = '3D-Var'
        dtwinlen = isodur.parse_duration(cfg['cost function']['subwindow'])
        costf['window begin'] = (dateini+dtwinlen/2.0).isoformat()+'Z'
        costf['window length'] = cfg['cost function']['subwindow'] # 3D-Var offline mode makes sense only on a small window
    else:
        costf['cost type'] = '4D-Ens-Var'
        costf['window begin'] = dateini.isoformat()+'Z'
        costf['window length'] = isodur.duration_isoformat(dtwinlen)
        costf['subwindow'] = cfg['cost function']['subwindow']

    costf['analysis variables'] = chemsp
    costf['jb evaluation'] = False

    costf['geometry'] = cfg["grid"]
    
    # Possible log transformation of variables
    if 'transform' in cfg['cost function']:
        transformdic = {}
        transformlist = []
        if 'all' in cfg['cost function']['transform']:
            for sp in chemsp:
                transformdic[sp] = cfg['cost function']['transform']['all']
            ldict = cfg['cost function']['transform']['all']
            ldict['variables'] = chemsp
            transformlist.append(ldict)
        else:
            transformdic = cfg['cost function']['transform']
            for sp in transformdic:
                ldict = cfg['cost function']['transform'][sp]
                ldict['variables'] = [sp]
                transformlist.append(ldict)

    if cftype == '3D':
        dates =  [dateini+dtwinlen]
        costf['background'] = ListMocageOutput(bkgdir,dates,chemsp,costf['geometry']['domname'])
        if 'transform' in cfg['cost function']:
            state = costf['background']
            state['transforms'] = transformlist
    else:
        dtsubwin = isodur.parse_duration(costf['subwindow'])
        nsub = int(dtwinlen/dtsubwin)+1
        dates =  [dateini+t*dtsubwin for t in range(nsub)]
        costf['background'] = {'states': ListMocageOutput(bkgdir,dates,chemsp,costf['geometry']['domname'])}
        if 'transform' in cfg['cost function']:
            for state in costf['background']['states']:
                state['transforms'] = transformlist

    costf['background error'] = {
        'covariance model': 'ensemble',
        'localization':{
            'localization method' : 'SABER',
            'communication mode' : 'fast',
            'saber central block' : {
                'saber block name': 'BUMP_NICAS',
                'bump':{
                    'io':{
                        'data directory':locdatadir,
                        'files prefix':'loc_1pe'
                    },
                    'drivers':{
                        'multivariate strategy': cfg['localization']['multivariate'],
                        'read local nicas': True
                    },
                    'nicas':{
                        'common localization weights':[{'row variables':chemsp,'column variables':chemsp,'value':cfg['localization']['mc']}]
                    }
                }
            }
        },
        'members' : ListMocageOutput(ensdir,dates,chemsp,costf['geometry']['domname'],mode='ensemble',ensize=ens_size)}

    if 'time decay' in cfg['localization']:
        costf['background error']['localization']['time decay'] = cfg['localization']['time decay']

    # Older way of specifing multivariate localization
    # if cfg['localization']['multivariate'] == 'common_weighted':
    #     # costf['background error']['localization']['saber block']['bump']['loc_wgt'] = {'common':0.5}
    #     # Construct the multivariate localization
    #     pair_order_list = list(itertools.combinations(chemsp,2))
    #     costf['background error']['localization']['saber central block']['bump']['loc_wgt'] = {el[0]+'-'+el[1]:0.5 for el in pair_order_list}

    if 'transform' in cfg['cost function']:
        for member in costf['background error']['members']:
            if cftype == '3D':
                member['transforms'] = transformlist
            else:
                for state in member['states']:
                    state['transforms'] = transformlist

    obsdatastr = dateini.strftime('%Y%m%d')
    obsdataoutstr = dateini.strftime('%Y%m%d')
    if cftype == '3D': obsdataoutstr = (dateini+dtwinlen).strftime('%Y%m%d%H')

    costf['observations'] = {'observers':[]}

    for obs in cfg['observations']:
        obsdict = {
            'obs error':{'covariance model':'diagonal'},
            'obs operator':obs['obs operator'],
            'obs space':{
                'obsdatain':{'engine':{'obsfile':os.path.join(obsdir,'HDAT+'+obsdatastr+'.h5')}},
                'obsdataout':{'engine':{'obsfile':os.path.join(outdir,'HSTAT+'+obsdataoutstr+'.h5')}},
                'obs type':obs['obs space']['obs type'],
                'instr name':obs['obs space']['instr name'],
                'obserrors':obs['obs space']['obserrors']
                }
            }
        # Optional elements
        if 'obs filters' in obs: obsdict['obs filters'] = obs['obs filters']
        if 'transform' in cfg['cost function'] and obs['obs operator']['mod var'] in transformdic:
            obsdict['obs space']['transform'] = transformdic[obs['obs operator']['mod var']]

        costf['observations']['observers'].append(obsdict)

    var = outyaml['variational']

    var['minimizer'] = {'algorithm':'RPLanczos'}
    var['iterations'] = []
    var['iterations'].append({
        'diagnostics':{'departures':'ombg'},
        'gradient norm reduction':1.0e-10,
        'ninner':cfg['cost function']['ninner'],
        'geometry':cfg["grid"],
        'test':'on'
        })

    final = outyaml['final']

    final['diagnostics'] = {
        'departures':'oman'
        }
    if cftype == '4D': final['prints'] = {'frequency':costf['subwindow']}

    output = outyaml['output']

    output['datadir'] = outdir
    output['fileprefix'] = 'A4'+cfg["grid"]['domname']

    yamlstr = yaml.dump(outyaml)
    
    #Substitute 'NO': with NO: if present, because it crashes eckit parser
    if 'NO' in chemsp: yamlstr = yamlstr.replace("'NO':","NO:")

    print(yamlstr)

    with open(kwargs['jedinamelist'],'w') as f:
        print(yamlstr,file=f)

    # Make also a copy of the namelist in the output folder
    with open(os.path.join(outdir,os.path.basename(kwargs['jedinamelist'])),'w') as f:
        print(yamlstr,file=f)

    # Prepare the job execution script and store it
    if 'type' in cfg['cost function'] and cfg['cost function']['type'] == '3D':
        nsub = 1
    else:
        nsub = int((dateend - dateini)/isodur.parse_duration(cfg['cost function']['subwindow']))+1

    nodes = 1
    queue = 'interactive'
    hostname = 'kglobc1'

    if 'computing' in cfg and 'run' in cfg['computing']:
      hostname = cfg['computing']['run']['machine']
      if 'nodes' in cfg['computing']['run']: nodes = cfg['computing']['run']['nodes']
      if 'queue' in cfg['computing']['run']: queue = cfg['computing']['run']['queue']

    rmode = machines_config(hostname)['mode']

    OpenMP = True
    OpenMPperMPI = 1

    cores_per_node = machines_config(hostname)['queues'][queue]['cores_per_node']

    total_res = nodes * cores_per_node

    res_per_slot = math.floor(total_res / nsub)

    if OpenMP and OpenMPperMPI > 1:
       ompthreads = min(OpenMPperMPI, res_per_slot)
    else:
       ompthreads = 1

    # Force all slots in one node
    # if nodes == 1: 
    #     mpiprocs = nsub
    #     ompthreads = 1
    #     mpi_per_slot = 1
    # else:
    mpi_per_slot = math.floor(res_per_slot/ompthreads)
    mpiprocs = int(nsub * mpi_per_slot)

    print('mpiprocs ',mpiprocs,'mpi per slot ',mpi_per_slot, 'ompthreads ',ompthreads)
    print('available resources ',total_res)
    print('used resources ',mpiprocs * ompthreads)

    script = PrepareJob(exe=os.path.join(cfg['oops path'],'aq_var.x'),namelist=kwargs['jedinamelist'],
        mpiprocs=mpiprocs,ompthreads=ompthreads,
        mode=rmode,nodes=nodes,machine=hostname,
        queue=queue,time=timedelta(hours=3),
        jobout=kwargs['jobout'],
        errout=kwargs['joberr']
        )
    with open(kwargs['jobname'],'w') as f:
        print(script,file=f)
    os.chmod(kwargs['jobname'], 0o755)

  @classmethod
  def py_finalize_script(cls):

    import daimonpp as postp

    kwargs = cls.load_ecfvars()

    # Load the variational namelist here (my version, not the oops one)
    with open(kwargs['namelist']) as file:
        cfg = yaml.load(file,Loader=Loader)

    if 'postproc namelist' in cfg:
        
        pltnamelist = cfg['postproc namelist']
        with open(pltnamelist) as file:
            postcfg = yaml.load(file,Loader=Loader)

        cftype = '4D'
        hdatorig = 'aq4d'
        if 'type' in cfg['cost function']: cftype = cfg['cost function']['type']
        if cftype == '3D': hdatorig = 'aq3d'

        starttime = isodatetime.parse_datetime(kwargs['dateini']).strftime('%Y%m%d%H')
        endtime = isodatetime.parse_datetime(kwargs['dateend']).strftime('%Y%m%d%H')
        if cftype == '3D':
            dtwinlen = isodur.parse_duration(cfg['cost function']['subwindow'])
            starttime = (isodatetime.parse_datetime(kwargs['dateini'])+dtwinlen).strftime('%Y%m%d%H')
            endtime = (isodatetime.parse_datetime(kwargs['dateini'])+dtwinlen).strftime('%Y%m%d%H')
        # Set plt namelist elements according to the experiment name and dates
        postcfg['expname'] = kwargs['expname']+'_'+kwargs['taskname']
        postcfg['starttime'] = starttime
        postcfg['endtime'] = endtime
        outdir = kwargs['outpath']

        # Plots here
        if "plots" in postcfg:

            if 'costfunction' in postcfg['plots'] and postcfg['plots']['costfunction']:

                inputfile = kwargs['jobout']
                PlotCostFunc(inputfile,outdir)

            if 'modelflds' in postcfg['plots']:

                if 'hfields' in postcfg['plots']['modelflds']:

                    postp.plot_hfields(postcfg,outdir,ftype='A4',rtype='anl')

                if 'vfields' in postcfg['plots']['modelflds']:

                    postp.plot_vfields(postcfg,outdir,ftype='A4',rtype='anl')

            if 'increments' in postcfg['plots']:

                fctname, bkgdir = GetForecast(cfg,kwargs)

                fctpath = os.path.join(bkgdir,'FORECAST') # Can be conditional on some key? Now is almost always FORECAST

                if 'hfields' in postcfg['plots']['increments']:

                    postp.plot_hfields(postcfg,outdir,fctpath,ftype='A4',f2type='HM',rtype='incr')

                if 'vfields' in postcfg['plots']['increments']:

                    postp.plot_vfields(postcfg,outdir,fctpath,ftype='A4',f2type='HM',rtype='incr')

            if 'obs' in postcfg['plots']:

                subwindow = isodur.parse_duration(cfg['cost function']['subwindow'])
                # Misfits and innovation
                postp.plot_hstat_snapshots(postcfg,outdir,origin=hdatorig,WindowsLength=subwindow.total_seconds())
                        
                postp.plot_hstat_tser(postcfg,outdir,origin=hdatorig,WindowsLength=subwindow.total_seconds())

        # Compare with other forecasts
        if "compare to" in cfg:

          for name,exp in cfg["compare to"].items():

            if "path" in exp:
              expdir = exp["path"]
            else:
              expdir = os.path.join(kwargs["exppath"],kwargs['expname'],'cycle_%04i' % (int(kwargs['cycle'])),name)
            if name in ("forecast","truth"): expdir = os.path.join(expdir,'FORECAST')  # can do better than hardcoded like this?

            postp.fields_diff(outdir,expdir,basename=name,starttime=starttime,endtime=endtime,domains=[cfg['grid']['domname']],
                ftype='A4',f2type='HM')

            if "plots" in postcfg:

              if 'modelflds' in postcfg['plots']:

                if 'hfields' in postcfg['plots']['modelflds']:

                  postp.plot_hfields(postcfg,os.path.join(outdir,'fields_diff',name),ftype='HD',rtype='fct')

                if 'vfields' in postcfg['plots']['modelflds']:

                  postp.plot_vfields(postcfg,os.path.join(outdir,'fields_diff',name),ftype='HD',rtype='fct')

def PlotCostFunc(infile,outputdir):

    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt

    Jb = [0.0]
    Jo = []
    Jo_sp = {}
    gr_red = []
    nr_red = []
    init = True
    gmres = False
    with open(infile) as fi:
      for line in fi:
        if ('Quadratic cost function: Jb' in line) and \
           'Test' not in line:
          Jb.append(float(line.split()[-1]))
        if 'CostJo   : Nonlinear Jo(' in line:
          sp = line.split()[7][3:-1]
          pr = 'Initial:' if init else 'Final  :'
          print(pr," ".join(line.split()[4:]))
          if sp not in Jo_sp:
            Jo_sp[sp] = [line.split()[8][:-1]]
          else:
            Jo_sp[sp].append(line.split()[8][:-1])
        if 'Quadratic cost function: JoJc' in line:
          Jo.append(float(line.split()[-1]))
          init = False
        if 'GMRESR end of iteration' in line:
          gmres = True
        if 'Gradient reduction' in line:
          if not gmres:
            gr_red.append(float(line.split()[-1]))
        if 'Norm reduction' in line:
          if not gmres:
            nr_red.append(float(line.split()[-1]))

    sum = 0.
    for sp in Jo_sp:
      sum += float(Jo_sp[sp][0])
    Jo = [sum]+Jo
    sum = 0.
    for sp in Jo_sp:
      sum += float(Jo_sp[sp][-1])
    #AP Jo.append(sum)  
    Jofin = sum

    J = [b+o for b,o in zip(Jb, Jo)]
    print('J  :',J)
    print('Jb :',Jb)
    print('Jo :',Jo)
    print('Estimated final Jo {}'.format(Jofin))

    fig = plt.figure()
    fig.set_size_inches(8,10)
    gs = fig.add_gridspec(5, hspace=0.5)
    ax = gs.subplots()

    ax[0].plot(J)
    ax[0].title.set_text('J')
    ax[1].plot(Jb)
    ax[1].title.set_text('Jb')
    ax[2].plot(Jo)
    ax[2].scatter(x=len(J)-1,y=Jofin,c='r')
    ax[2].title.set_text('Jo')
    ax[3].plot(gr_red)
    ax[3].title.set_text('Gradient reduction')
    ax[4].plot(nr_red)
    ax[4].title.set_text('Norm reduction')

    outputdir = os.path.join(outputdir,'cost_function')
    if not os.path.exists(outputdir): os.makedirs(outputdir)

    fig.savefig(os.path.join(outputdir,'CostFunction'))
    plt.close(fig)

