# -*- coding: utf-8 -*-
"""
Class handling the various ecflow steps for assimilation experiments 
"""

import os
import sys
import yaml

# Each step object correspond to a single ecflow task (which includes init and finalize)
class Step(object):

  counter = 0  
  # Variables that transit through ecflow into all tasks
  ecfvars = ['expname','exppath','cycle','taskname','datebegin','dateini','dateend','outpath']
  actions = []

  def __init__(self, **kwargs):

    if 'cycle' in kwargs:
      self.cycle = kwargs['cycle']
    else:
      self.cycle = 1

    self.datebegin = kwargs['start_date'] # This remains fixed all time, it is the timeseries start time
    self.dateini = kwargs['start_date'] # This increases with cycles, it is the step init time
    self.duration = kwargs['duration']
    self.dateend = self.dateini + self.duration # This increases with cycles, it is the step end time
    self.expname = kwargs['expname']
    self.taskname = kwargs['stepname']
    self.suite_path = kwargs['suite_path']
    # Not all tasks need a specified output path
    if 'output_path' in kwargs: self.output_path = os.path.join(kwargs['output_path'],self.expname,
      'cycle_%04i' % (self.cycle),self.taskname)
    # Namelist from input
    self.namelist = kwargs['namelist'] # This is the task namelist section
    # An optional basename can be used for namelists naming or other things (used for postprocessors now)
    if 'basename' in kwargs: self.basename = kwargs['basename']
    if 'basename' in self.__dict__:
      self.namelist_file = os.path.join(self.suite_path,'namelists','%s_%s_namelist.yaml' % (self.basename,self.taskname.lower()))
    else:
      self.namelist_file = os.path.join(self.suite_path,'namelists','%s_namelist.yaml' % (self.taskname.lower()))
    self.todo = self.namelist['todo']
    self.postproc_steps = []
    self.has_postproc_steps = False

    Step.counter+=1

  def next(self, dt, **kwargs):

    self.cycle+=1
    self.dateini+=dt
    self.dateend+=dt

    if 'output_path' in self.__dict__: self.output_path = os.path.join(os.sep.join(self.output_path.split(os.sep)[:-2]),
      'cycle_%04i' % (self.cycle),self.taskname)

    if self.has_postproc_steps:
        for pp in self.postproc_steps: pp.next(dt)

  def setup(self, ecflowbr, **kwargs):

    self.ecfstep = ecflowbr.add_family(self.taskname)
    trigger = ''
    if 'depends on' in self.namelist:
      # Loop on all dependencies and include potential triggers
      for el in self.namelist['depends on']:
        if 'path' not in self.namelist['depends on'][el]:
          if len(trigger) > 0: trigger+=' and '
          if el not in [node.name() for node in ecflowbr.get_all_nodes()]:
            trigger+='/%s/cycle_%04i/' % (self.expname,self.cycle)+el+' eq complete'
            # trigger+='%s/' % (ecflowbr.get_parent().get_parent().get_abs_node_path())+el+' eq complete'
          else:
            trigger+=el+' eq complete'
      
    if 'wait for' in self.namelist:
      # This trigger does not depend on the 'new' attibute but is used only to avoid multiple runs on single nodes
      for el in self.namelist['wait for']:
        if len(trigger) > 0: trigger+=' and ' 
        trigger+=el+' eq complete'

    if len(trigger) > 0: self.ecfstep.add_trigger(trigger)

    self.ecfstep.add_variable("datebegin",self.datebegin.isoformat())
    self.ecfstep.add_variable("dateini",self.dateini.isoformat())
    self.ecfstep.add_variable("dateend",self.dateend.isoformat())
    self.ecfstep.add_variable("cycle",self.cycle)
    self.ecfstep.add_variable("taskname",self.taskname)
    if 'output_path' in self.__dict__: self.ecfstep.add_variable("outpath",self.output_path)
    if 'run' in self.todo:
      task = self.ecfstep.add_task(type(self).__name__+'_setup')
      task.add_variable("action",'setup').add_variable("class",type(self).__name__) # Needed only if using python scripts within the task
      return task

  def run(self, **kwargs):

    if 'run' in self.todo:
      task = self.ecfstep.add_task(type(self).__name__+'_run').add_trigger("%s_setup eq complete" % (type(self).__name__))
      task.add_variable("action",'run').add_variable("class",type(self).__name__) # Needed only if using python scripts within the task
      return task
    else:
      return None

  def finalize(self, **kwargs):

    if 'finalize' in self.todo:
      task = self.ecfstep.add_task(type(self).__name__+'_finalize')
      if 'run' in self.namelist['todo']: 
        task.add_trigger("%s_run eq complete" % (type(self).__name__))
      # else:
      #   task.add_trigger("%s_setup eq complete" % (type(self).__name__))
      task.add_variable("action",'finalize').add_variable("class",type(self).__name__) # Needed only if using python scripts within the task
      return task
    else:
      return None

  def set_run_end(self, ecflowbr):

      self.setup(ecflowbr)
      self.run()
      self.finalize()

      if self.has_postproc_steps:
        ppfam = self.ecfstep.add_family("post_process")
        if 'finalize' in self.namelist['todo']: 
          ppfam.add_trigger("%s_finalize eq complete" % (type(self).__name__))
        elif 'run' in self.namelist['todo'] and 'finalize' not in self.namelist['todo']: 
          ppfam.add_trigger("%s_run eq complete" % (type(self).__name__))
        else:
          ppfam.add_trigger("%s_setup eq complete" % (type(self).__name__))
        for pp in self.postproc_steps:
          pp.setup(ppfam)
          pp.run()
          pp.finalize()

  def save_namelist(self, **kwargs):
    with open(self.namelist_file,"w") as file:
      yaml.dump(self.namelist,file,indent=4)

  def attach_postproc_step(self, other):

    self.postproc_steps.append(other)
    self.has_postproc_steps = True

  # A generic call to a python setup task (overload this with class specific actions where needed)
  @classmethod
  def ecf_script_setup(cls, **kwargs):

    ll = ecf_script_py(cls)
    with open(os.path.join(kwargs['path'],cls.__name__+'_setup'+'.ecf'),'w') as f:
      f.writelines(ll)

  # A generic call to a python run task (overload this with class specific actions where needed)
  @classmethod
  def ecf_script_run(cls, **kwargs):

    ll = ecf_script_py(cls)
    with open(os.path.join(kwargs['path'],cls.__name__+'_run'+'.ecf'),'w') as f:
      f.writelines(ll)

  # A generic call to a python finalize task (overload this with class specific actions where needed)
  @classmethod
  def ecf_script_finalize(cls, **kwargs):

    ll = ecf_script_py(cls)
    with open(os.path.join(kwargs['path'],cls.__name__+'_finalize'+'.ecf'),'w') as f:
      f.writelines(ll)

  @classmethod
  def create_ecf_scripts(cls, **kwargs):
    
    for action in cls.actions: getattr(cls, 'ecf_script_'+action)(**kwargs)

  @classmethod
  def load_ecfvars(cls):

    # Retrieve ecf variables from ecflow environment
    kwargs = {}
    for var in cls.ecfvars:
      val = os.getenv(var)
      if val is None:
        print('Error: %s not initialized' % (var))
        sys.exit(-1)
      else:
        kwargs[var] = val

    return kwargs

def ecf_script_py(cls):

    ll = '%include <head.h>\n'
    ll+='export %s=%%%s%%\n' % ('action','action')
    ll+='export %s=%%%s%%\n' % ('class','class')
    for var in cls.ecfvars:
      ll+='export %s=%%%s%%\n' % (var,var)
    
    ll+=r"""if [[ "%action%" == "finalize" && `hostname` == kraken* ]]"""
    ll+='\nthen\n'
    ll+='echo Finalize Task in interactive mode on a specific queue/node\n'
    ll+='echo "#!/bin/ksh" > stepmf$$.sh\n'
    ll+='echo "source $HOME/loadpy.sh" >> stepmf$$.sh\n'
    ll+='echo "export %s=%%%s%%" >> stepmf$$.sh\n' % ('action','action')
    ll+='echo "export %s=%%%s%%" >> stepmf$$.sh\n' % ('class','class')
    for var in cls.ecfvars:
      ll+='echo "export %s=%%%s%%" >> stepmf$$.sh\n' % (var,var)
    ll+='echo "%PYTHON% %ECF_INCLUDE%/aq_task.py" >> stepmf$$.sh\n'

    # Plot on a specific queue
    ll+='srun -p prodshared --time 01:00:00 sh ./stepmf$$.sh'

    ll+='\nelse\n'
    ll+='source %PYTHON_ENV_LOAD%\n'
    ll+='%PYTHON% %ECF_INCLUDE%/aq_task.py\n'
    ll+='fi\n'
    ll+= '%include <tail.h>\n'
    return ll

def machines_config(hostname):

  cfg = {
    # Nemo switched off in Feb 2023, kept for historical and sentimental reasons
    # 'nemopae2':{'mode':'interactive',
    #             'queues':{
    #               'interactive':{'cores_per_node':24}
    #               }
    #             },
    # 'nemopae1':{'mode':'interactive',               
    #             'queues':{
    #               'interactive':{'cores_per_node':24}
    #               }
    #             },
    # 'nemo':{'mode':'batch',
    #         'queues':{
    #             'prod':{'cores_per_node':24},
    #             'glo':{'cores_per_node':24}
    #             }
    #         },
    'kraken':{'mode':'batch',
              'queues':{
                  'prod':{'cores_per_node':36},
                  'icelake':{'cores_per_node':76},
                  'bigmem2':{'cores_per_node':36}
                  }
              },
    'kglobc1':{'mode':'interactive',
               'queues':{
                  'interactive':{'cores_per_node':64}
                  }
              },
    'laptop':{'mode':'interactive'
              }
  }

  return cfg[hostname]
