# -*- coding: utf-8 -*-
"""
	ECFLOW suite definition for running AQ experiments
"""

import os
import sys
import shutil
import yaml
from yaml import CLoader as Loader, CDumper as Dumper
import ecflow
import datetime
# import isodate.isodatetime as isodatetime
import isodate.isoduration as isodur
import copy
from Forecast import Forecast
from Ensemble import Ensemble
from Ensemble_member import Ensemble_member
from Dirac import Dirac
from Variational import Variational
from Hofx import Hofx
from Ensemble_variance import Ensemble_variance
from Localization import Localization
from argparse import ArgumentParser, RawDescriptionHelpFormatter

def create_suite(suite, suitepath, **kwargs):

  suite.add_variable('expname',kwargs['experiment name'])
  suite.add_variable('exppath',kwargs['output path'])

  # Loop on multiple cycles of forecasts and assimilation
  tstart = datetime.datetime.fromisoformat(kwargs['start date'])
  t = copy.deepcopy(tstart)
  dt = isodur.parse_duration(kwargs['cycle length'])
  icstart = 1
  if 'start from cycle' in kwargs: icstart = kwargs['start from cycle']
  ic = 1

  # Local useful variables
  steps = kwargs['steps']

  # Initialize all the step objects (the suite path is for namelists and tasks save, can get rid of that?)
  for step in steps:

    stepnml = kwargs[step]
    stepdict = {'expname':kwargs['experiment name'],'start_date':tstart,'duration':dt, \
      'namelist':stepnml,'suite_path':suitepath,'stepname':step,
      'output_path':kwargs['output path']}
    # Overwrite the step start time and default length if needed by the task
    if 'delay' in stepnml: stepdict['start_date'] = tstart + isodur.parse_duration(stepnml['delay'])
    if 'length' in stepnml: stepdict['duration'] = isodur.parse_duration(stepnml['length'])

    # Init step object
    locals()[step] = globals()[stepnml["class"]](**stepdict)

    # Add possible postprocessor tasks (belonging to step classes) to the step 
    # (the postprocessor namelist can be given within the step or in the global namelist)
    # The postprocessor here inherits time shifts from the step and normally uses the same output path of the step
    if 'post processors' in stepnml:
        for pp in stepnml['post processors']:
            postprocnl = stepnml['post processors'][pp]
            postprocdict = {'expname':kwargs['experiment name'],'start_date':tstart,'duration':dt, \
                'namelist':postprocnl,'suite_path':suitepath,'stepname':pp}
            # The postprocessors can have their own delays/length attributes, they are summed to the steps ones
            if 'delay' in stepnml: postprocdict['start_date'] += isodur.parse_duration(stepnml['delay'])
            if 'length' in stepnml: postprocdict['duration'] = isodur.parse_duration(stepnml['length'])
            if 'delay' in postprocnl: postprocdict['start_date'] += isodur.parse_duration(postprocnl['delay'])
            if 'length' in postprocnl: postprocdict['duration'] = isodur.parse_duration(postprocnl['length'])
            # The postprocessor taskname inherits also from the mother step naming to avoid filenaming conflicts
            postprocdict['basename'] = step
            # Init postprocessor object and attach to the step
            poststep = globals()[postprocnl["class"]](**postprocdict)
            locals()[step].attach_postproc_step(poststep)

  while t < tstart+dt*kwargs['number of cycles']:

    if ic >= icstart: 
      cycle = suite.add_family("cycle_%04i" % ic)
      for step in steps: locals()[step].set_run_end(cycle)

    if ic > icstart: cycle.add_trigger("cycle_%04i eq complete" % (ic-1))

    ic += 1
    t += dt

    for step in steps: locals()[step].next(dt)

if __name__ == "__main__":

  parser = ArgumentParser(
    formatter_class=RawDescriptionHelpFormatter,
    description=('''\n\
    Prepare the ecflow suite for SEEDS experiments.
    E.g. ./aq_suite.py 
    '''))
  parser.add_argument(
    "-n", "--namelist", dest="namelist",
    help="YAML experiment namelist")
  parser.add_argument(
    "-p", "--prepare", dest="prep",
    action='store_true',
    help="Prepare the ecf suite")
  parser.add_argument(
    "-np", "--noprepare", dest="prep",
    action='store_false',
    help="Do not prepare the ecf suite")
  parser.add_argument(
    "-l", "--load", dest="load",
    action='store_true',
    help="Load the ecf suite")
  parser.add_argument(
    "-nl", "--noload", dest="load",
    action='store_false',
    help="Do not load the ecf suite")
  parser.add_argument(
    "-r", "--run", dest="launch",
    action='store_true',
    help="Launch the ecf suite")
  parser.add_argument(
    "-nr", "--norun", dest="launch",
    action='store_false',
    help="Do not launch the ecf suite")
  parser.set_defaults(prep=True,load=True,launch=False)

  opt = parser.parse_args()

  with open(opt.namelist) as file:
    namelist = yaml.load(file,Loader=Loader)

  scriptsdir = os.getcwd()
  ecflowpath = os.path.join(os.path.expanduser('~'),'aq-logs')
  suitename = namelist['experiment name']
  suitepath = os.path.join(ecflowpath,'experiments',suitename)
  suitedef = os.path.join(suitepath,suitename+'.def')

  if opt.prep:

    if not os.path.exists(suitepath): 
      os.makedirs(suitepath)
      shutil.copyfile(os.path.join(scriptsdir,'head.h'), os.path.join(suitepath,'head.h'))
      shutil.copyfile(os.path.join(scriptsdir,'tail.h'), os.path.join(suitepath,'tail.h'))

    # Create directories to save ecf task files and task specific namelists for later usage
    taskspath = os.path.join(suitepath,'tasks')
    if not os.path.exists(taskspath): os.mkdir(taskspath)
    nmlistpath = os.path.join(suitepath,'namelists')
    if not os.path.exists(nmlistpath): os.mkdir(nmlistpath)
    jobspath = os.path.join(suitepath,'jobs')
    if not os.path.exists(jobspath): os.mkdir(jobspath)

    with open(os.path.join(nmlistpath,'experiment_namelist.yaml'),"w") as file:
      yaml.dump(namelist,file,indent=4)

    print("Creating suite definition")
    defs = ecflow.Defs()
    suite = defs.add_suite(suitename)

    suite.add_variable("ECF_INSTALL_PATH", "/home/eemili/miniconda3/envs/ecflow")
    suite.add_variable("PYTHON_ENV_LOAD","/home/eemili/loadpy.sh")
    suite.add_variable("ECF_INCLUDE", scriptsdir)
    suite.add_variable("ECF_HOME", suitepath)
    suite.add_variable("ECF_FILES", os.path.join(suitepath,"tasks"))
    suite.add_variable("PYTHON","python")
    suite.add_variable("ECF_TRIES", 1)

    create_suite(suite,suitepath,**namelist)

    # print(defs)
    print("Checking job creation: .ecf -> .job0")
    print(defs.check_job_creation())

    print("Saving definition to file '%s.def'" % (suitename))
    defs.save_as_defs(suitedef)

  if opt.load:

    try:

      ci = ecflow.Client("bsces110072:2501");
      
      try:

        print("Cleaning all previously loaded suites with the same name")
        ci.delete(suitename)

      except RuntimeError as e:
        print("Failed:",    e)
      
      print("Loading definition in '%s.def' into the server" % (suitename))   
      ci.load(suitedef)
      
      # print("Restarting the server. This starts job scheduling")
      # ci.restart_server()     
        
    except RuntimeError as e:
      print("Failed:",    e)

  if opt.launch:

    try:
      ci = ecflow.Client("bsces110072:2501");
      
      print("Begin the suite named '%s'" % (suitename))    
      ci.begin_suite(suitename)
      #ci.requeue(suitename)
    
    except RuntimeError as e:
      print("Failed:",    e)