#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct  2 12:29:32 2023

@author: eemili

Convert MONARCH data to regular lat-lon netcdf in MOCAGE
style, for further use in AQ

"""

import os
import time
import numpy as np
from netCDF4 import Dataset
import matplotlib.pyplot as plt
import matplotlib.colors as mc
import cartopy.crs as ccrs
from datetime import datetime as dt
from datetime import timedelta as td
from scipy.interpolate import griddata
import copy
import subprocess

def plot_2Dmap(field,lats,lons,**kwargs):
    
    filename = 'fig.png'
    # colormap = 'viridis'
    colormap = 'YlOrBr'
    if 'filename' in kwargs: filename = kwargs['filename']
    if 'colormap' in kwargs: colormap = kwargs['colormap']
    # if not os.path.exists(os.path.split(filename)[0]):
    #     os.makedirs(os.path.split(filename)[0])
    
    fig = plt.figure(dpi=300)

    # Setup a global EckertIII map with faint coastlines.
    ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
    # ax.set_global()
    if 'plotregion' in kwargs:
        plotregion = kwargs['plotregion']
        ax.set_extent([plotregion['LonMin'], plotregion['LonMax'], 
                       plotregion['LatMin'], plotregion['LatMax']], crs=ccrs.PlateCarree())
        
    ax.coastlines('110m', alpha=0.1)
    
    vmin, vmax = field.min(), field.max()
    if 'range' in kwargs and kwargs['range'] is not None:
        if "min" in kwargs['range']: vmin = kwargs['range']["min"]
        if "max" in kwargs['range']: vmax = kwargs['range']["max"]
        
    levels=np.linspace(vmin,vmax,11)
    norm = mc.BoundaryNorm(levels, 256)
    filled_c = ax.contourf(lons, lats, field, 
                           transform=ccrs.PlateCarree(),
                           cmap=colormap,
                           levels=levels,
                           norm=norm,
                           extend='both')
    
    # Add a colorbar for the filled contour.
    cbar = fig.colorbar(filled_c, orientation='horizontal')
    
    if 'title' in kwargs: ax.set_title(kwargs['title'])

    plt.show()
    # plt.savefig(filename)

def unitconv(spc):
    # MONARCH outputs are in ppm for gases and kg/m3 for aerosols 
    convdict = {'default':1.0e-6}
    if spc in convdict:
        return convdict[spc]
    else:
        return 1.0e-6

date = '2023100212'
dastep = td(hours=1)
dawindow = td(hours=36)
# exppath = '/esarchive/scratch/eemili/monarch/forecast'
# outdir = '/esarchive/scratch/eemili/monarch/forecast/converted'
exppath = '/esarchive/scratch/eemili/monarch/ensemble'
outdir = '/esarchive/scratch/eemili/monarch/ensemble/converted'
members = range(1)
# mvars = ['O3','NO2','CO','NO','SO2','NH3']
mvars = ['O3','NO2','CO','SO2']
mlevs = range(19,24)
destgrid = {'LatMin':28.05,'LatMax':72,'LonMin':-25.95,'LonMax':46,'LonRes':0.1,'LatRes':0.1,
            'nlon':720,'nlat':440}
maprange = {'O3':{'max':0.05,'min':0.01}}

if not os.path.exists(outdir): os.makedirs(outdir)

inidate = dt.strptime(date, '%Y%m%d%H')
dasteps = [inidate+i*dastep for i in range(int(dawindow/dastep)+1)]

for member in members:
    
    # Open input model file
    mfile = os.path.join(exppath,'MONARCH_d01_%s_E%04i.nc' % (inidate.strftime('%Y%m%d%H'),0))
    # mfile = os.path.join(exppath,'MONARCH_d01_%s.nc' % (inidate.strftime('%Y%m%d%H')))
    
    f = Dataset(mfile, 'r', format="NETCDF4")
    lat_grid = f['lat'][:]
    lon_grid = f['lon'][:]
    
    dest_lat = np.arange(destgrid['LatMin'],destgrid['LatMax'],destgrid['LatRes'])
    dest_lon = np.arange(destgrid['LonMin'],destgrid['LonMax'],destgrid['LonRes'])
    lon_target_mesh, lat_target_mesh = np.meshgrid(dest_lon, dest_lat)
    
    for step in dasteps[12:]:
        
        # ofile = os.path.join(outdir,'MONARCH_%s_E%04i.nc' % (step.strftime('%Y%m%d%H'),member))
        # ofile = os.path.join(outdir,'MONARCH_%s.nc' % (step.strftime('%Y%m%d%H')))
        ofile = os.path.join(outdir,'MONARCH_%s_av_an.nc' % (step.strftime('%Y%m%d%H')))
        outf = Dataset(ofile, 'w', format="NETCDF4")
    
        outf.CreationDate = np.string_(time.strftime("%c"))
    
        # Dimensions
        nclat = outf.createDimension("lat", dest_lat.size)
        nclon = outf.createDimension("lon", dest_lon.size)
        nclev = outf.createDimension("lev", len(mlevs))
        nctime = outf.createDimension("time", 1)
    
        times = outf.createVariable("time","f4",("time",))
        times.units = np.string_("seconds since %s" % (inidate.strftime('%Y-%m-%d %H:00:00')))
        times.axis = np.string_("t")
        times.standard_name = np.string_("Time axis")
        times.time_origin = np.string_(inidate.strftime('%Y-%m-%d %H:00:00'))
        times.calendar = np.string_("standard") 
        times[:] = (step-inidate).total_seconds()
    
        levels = outf.createVariable("lev","f4",("lev",))
        levels.units = np.string_("dimensionless")
        levels.axis = np.string_("Z")
        levels.standard_name = np.string_("atmosphere_hybrid_sigma_pressure_coordinate")
        levels.positive = np.string_("down")
        levels.realtopology = np.string_("linear")
        levels[:] = mlevs
    
        lats = outf.createVariable("lat","f4",("lat",))
        lats.units = np.string_("degrees_north")
        levels.axis = np.string_("Y")
        lats.standard_name = np.string_("latitude")
        lats.realtopology = np.string_("linear")
        lats[:] = dest_lat
    
        lons = outf.createVariable("lon","f4",("lon",))
        lons.units = np.string_("degrees_east")
        levels.axis = np.string_("X")
        lons.standard_name = np.string_("longitude")
        lons.realtopology = np.string_("linear")
        lons[:] = dest_lon
        
        istep = int((step-inidate)/dastep)
        
        for mvar in mvars:
            
            ncvar = np.empty((1,len(mlevs),dest_lat.size,dest_lon.size),dtype=np.float32)

            mfile = os.path.join(exppath,'MONARCH_d01_%s_av_an_%s.nc' % (inidate.strftime('%Y%m%d%H'),mvar))
            f2 = Dataset(mfile, 'r', format="NETCDF4")

            for ilev,mlev in enumerate(mlevs):
                x = f2[mvar][istep,mlev,:,:]
            
                # plot_2Dmap(x,lat_grid,lon_grid,
                #            range=maprange[mvar],
                #            title=step.strftime('%Y-%m-%d-%H')+' FG\n Avg:%.2f Min:%.2f Max:%.2f' % (x.mean(),x.min(),x.max()),
                #            plotregion=destgrid,
                #            filename=os.path.join(outdir,'monarch-%s-%s' % (step.strftime('%Y%m%d%H'),mvar)))
                
                # Interpolate on destination grid and plot (this can be greatly sped up by precomputing interpolation weights)
                regx = griddata(
                (lon_grid.flatten(), lat_grid.flatten()),  # Coordinates of curvilinear grid
                x.flatten(),                   # Values of the variable to regrid
                (lon_target_mesh, lat_target_mesh),        # Target regular grid mesh
                method='linear'  # Interpolation method, you can use 'linear', 'cubic', etc.
                )
                
                regx = regx.reshape(dest_lat.shape[0],dest_lon.shape[0])
                
                ncvar[0,ilev,:,:] = regx*unitconv(mvar)
                
                # plot_2Dmap(regx,lat_target_mesh,lon_target_mesh,
                #            range=maprange[mvar],
                #            title=step.strftime('%Y-%m-%d-%H')+' FG\n Avg:%.2f Min:%.2f Max:%.2f' % (regx.mean(),regx.min(),regx.max()),
                #            plotregion=destgrid,
                #            filename=os.path.join(outdir,'monarch-%s-%s-regrid' % (step.strftime('%Y%m%d%H'),mvar)))
            
            var = outf.createVariable(mvar,"f4",("time","lev","lat","lon",))
            var.units = np.string_("ppv")
            var.standard_name = np.string_("mole_fraction_of_%s" % (mvar))
            var[:,:,:,:] = ncvar
        
            f2.close()

        outf.close()
        
    f.close()