# -*- coding: utf-8 -*-
"""

LETKF plot suite

Plots hourly output data from MONARCH LETKF, compute misfits,
innovation statistics, model fields, analysis increments
and ensemble standard deviation

TODO:
    - [ ] use experiment namelist to set script parameters
    - [ ] generalize instrument type (only VIIRS DOD currently)
"""

import os
import numpy as np
from netCDF4 import Dataset
import matplotlib.pyplot as plt
import obsdb
from plotters import plot2Dmap
from datetime import datetime as dt
from datetime import timedelta as td
from scipy.interpolate import griddata
import copy
import subprocess

date = '2024011500'
# date = '2020060100'
mvar = 'od550du'
#mvar = 'sconcdu'
stype = 'av'
# dastep = td(hours=1)
dastep = td(hours=3)
dawindow = td(hours=24)
# isregional = False
isregional = True
# MONARCH grid information (used to rotate back observations lats,lons)
cenlon, cenlat = 20., 35.
# cenlon is tph0d, cenlat is tlm0d
exppath = '/esarchive/exp/monarch/a6vz'
# exppath = '/esarchive/exp/monarch/a6tk'
mpath = os.path.join(exppath,'regional/3hourly/ensemble-stats')
# mpath = os.path.join(exppath,'global/hourly/ensemble-stats')
# dpath = os.path.join(exppath,'original_files/011/%s' % (date))
dpath = os.path.join(exppath,'original_files/011/%s' % (date))
mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,date,stype))
pltbounds = {'LatMin':-11,'LatMax':70,'LonMin':-62,'LonMax':102}
# pltbounds = {'LatMin':15,'LatMax':55,'LonMin':-10,'LonMax':50}
# pltbounds = {'LatMin':-90,'LatMax':90,'LonMin':-180,'LonMax':180}
maprange = {'od550du':{'max':1.0,'min':0.},'sconcdu':{'max':1.0e-6,'min':0.}}
mapdiffrange = {'od550du':{'max':0.5,'min':-0.5},'sconcdu':{'max':0.2e-6,'min':-0.2e-6}}
mapstdrange = {'od550du':{'max':0.5,'min':0.},'sconcdu':{'max':0.4e-6,'min':0.}}
maplowdiffrange =  {'od550du':{'max':0.3,'min':-0.3},'sconcdu':{'max':0.1e-6,'min':-0.1e-6}}
histrange = {'od550du':{'max':1.5,'min':0.}}
histdiffrange = {'od550du':{'max':1.5,'min':-1.5}}
histlowdiffrange = {'od550du':{'max':1.,'min':-1.}}
markersize = 1.0
outdir = os.path.join(exppath,'plots',date[:-2])
framedir = os.path.join(outdir,'frames')
gifdir = os.path.join(outdir,'gifs')

if not os.path.exists(outdir): os.makedirs(outdir)
if not os.path.exists(framedir): os.makedirs(framedir)
if not os.path.exists(gifdir): os.makedirs(gifdir)

inidate = dt.strptime(date, '%Y%m%d%H')
dasteps = [inidate+i*dastep for i in range(int(dawindow/dastep)+1)]

obsfile = os.path.join(dpath,'dep.txt')
rejobsfile = os.path.join(dpath,'dep_rej.txt')

# yall = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
nobs, nused, nrej, y_ts, hofxbg_ts, hofxan_ts, fgdep_ts, andep_ts, absinno_ts = [[] for i in range(9)]
dtslot=int(dastep.total_seconds()/3600)
nslots=int(dawindow.total_seconds()/dastep.total_seconds())
# Now loop on output frequency and do some plots
for step in dasteps[1:]:
    istep = int((step-inidate)/dastep)
    print(step)
    print(istep)
    
    if mvar == 'od550du':
        # Read and plot observations, departures and hofxbg. 
        # Object are reinitialized to avoid cumulating the obs for the moment
        y = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
        yall = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
        fgdep = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
        # Read the obs here from the LETFK output (consider appending the obs, it should be automatically done)
        # but take care of the correct time with this slot shit
        y.ReadLETKFAscii(obsfile,inidate,islot=istep,dtslot=dtslot,nslots=nslots,rotated=isregional,cenlon=cenlon,cenlat=cenlat)
        fgdep.ReadLETKFAscii(obsfile,inidate,islot=istep,dtslot=dtslot,nslots=nslots,ReadDA={'ObsType':'dep'},rotated=isregional,cenlon=cenlon,cenlat=cenlat)
        naccepted = yall.ReadLETKFAscii(obsfile,inidate,islot=istep,dtslot=dtslot,nslots=nslots,rotated=isregional,cenlon=cenlon,cenlat=cenlat)
        nrejected = yall.ReadLETKFAscii(rejobsfile,inidate,islot=istep,dtslot=dtslot,nslots=nslots,rotated=isregional,cenlon=cenlon,cenlat=cenlat)
        # Flag rejected obs
        if nrejected > 0: yall.assim_mask[-nrejected:] = 0
    
        # I can interpolate both hofx_fg and hofx_an. Alternatively the first could be read from file.
        hofxbg = y - fgdep
        # hofxbg = copy.deepcopy(y)
        hofxan = copy.deepcopy(y)
        
        y.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
                                    # 'colormap':'viridis'},
                                    'colormap':'YlOrBr'},
                        PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                        outNameSuffix='y',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        y.PlotObsHisto(vmin=histrange[mvar]['min'],vmax=histrange[mvar]['max'],
                            outNameSuffix='y',outdir=framedir,subtitle='y '+step.strftime('%Y-%m-%d-%H'))
    
    # FG ensemble average
    stype = 'av'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xfg = f[mvar][istep,0,:,:]
    
    if isregional:
        lat_grid = f['lat'][:]
        lon_grid = f['lon'][:]
    else:
        grid = np.meshgrid(f['lon'][:],f['lat'][:])
        lat_grid = grid[1]
        lon_grid = grid[0]
        
    f.close()
    
    # AN ensemble average
    stype = 'av_an'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xan = f[mvar][istep,0,:,:]
    f.close()
    
    stype = 'av_inc'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xinc = f[mvar][istep,0,:,:]
    f.close()
    
    # FG ensemble standard deviation
    stype = 'std'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xfg_std = f[mvar][istep,0,:,:]
    f.close()
    
    # AN ensemble standard deviation
    stype = 'std_an'
    mfile = os.path.join(mpath,'%s_%s' % (mvar,stype),'%s_%s_%s.nc' % (mvar,inidate.strftime('%Y%m%d%H'),stype))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xan_std = f[mvar][istep,0,:,:]
    f.close()
    
    # if not xfg.mask.any():
    plot2Dmap(xfg,lat_grid,lon_grid,range=maprange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' FG\n Avg:%.2e Min:%.2e Max:%.2e' % (xfg.mean(),xfg.min(),xfg.max()),
               plotregion=pltbounds,
               filename=os.path.join(framedir,'monarch-%s-%s-fg' % (mvar,step.strftime('%Y%m%d%H'))))
    
    plot2Dmap(xan,lat_grid,lon_grid,range=maprange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' AN\n Avg:%.2e Min:%.2e Max:%.2e' % (xan.mean(),xan.min(),xan.max()),
               plotregion=pltbounds,
               filename=os.path.join(framedir,'monarch-%s-%s-an' % (mvar,step.strftime('%Y%m%d%H'))))
    
    plot2Dmap(xinc,lat_grid,lon_grid,range=mapdiffrange[mvar],
               title=step.isoformat()+' INC \n Avg:%.2e Min:%.2e Max:%.2e' % (xinc.mean(),xinc.min(),xinc.max()),
               plotregion=pltbounds,
               filename=os.path.join(framedir,'monarch-%s-%s-inc' % (mvar,step.strftime('%Y%m%d%H'))),colormap='Spectral_r')
    
    plot2Dmap(xfg_std,lat_grid,lon_grid,range=mapstdrange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' FG std\n Avg:%.2e Min:%.2e Max:%.2e' % (xfg_std.mean(),xfg_std.min(),xfg_std.max()),
               plotregion=pltbounds,
               filename=os.path.join(framedir,'monarch-%s-%s-fg-std' % (mvar,step.strftime('%Y%m%d%H'))))
    
    plot2Dmap(xan_std,lat_grid,lon_grid,range=mapstdrange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' AN std\n Avg:%.2e Min:%.2e Max:%.2e' % (xan_std.mean(),xan_std.min(),xan_std.max()),
               plotregion=pltbounds,
               filename=os.path.join(framedir,'monarch-%s-%s-an-std' % (mvar,step.strftime('%Y%m%d%H'))))

    # -------------------------------------------------------------------------
    # Interpolate analysis data at the observation locations to get the innovation values
    # This is temporary, it should disappear when the LETKF will produce all diags files
    
    if mvar == 'od550du':
        
        interpolated_data = griddata(
            (lon_grid.ravel(), lat_grid.ravel()),  # Known points
            xan.ravel(),  # Known data
            np.array((y.lons[~y.mask], y.lats[~y.mask])).T,  # Points to interpolate to
            method='linear'  # You can use 'linear', 'nearest', or other methods
        )
    
        hofxan.aod['550'][~y.mask] = np.ma.masked_invalid(interpolated_data)
        andep = y - hofxan

        absinno = andep.abs_sub(fgdep)
        # -------------------------------------------------------------------------
        
        hofxbg.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
                                'colormap':'YlOrBr'},
                        PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                        outNameSuffix='hofxbg',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        hofxbg.PlotObsHisto(vmin=histrange[mvar]['min'],vmax=histrange[mvar]['max'],
                            outNameSuffix='hofxbg',outdir=framedir,subtitle='hofxbg '+step.strftime('%Y-%m-%d-%H'))
        
        hofxan.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
                                'colormap':'YlOrBr'},
                        PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                        outNameSuffix='hofxan',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        hofxan.PlotObsHisto(vmin=histrange[mvar]['min'],vmax=histrange[mvar]['max'],
                            outNameSuffix='hofxan',outdir=framedir,subtitle='hofxan '+step.strftime('%Y-%m-%d-%H'))
        
        fgdep.PlotObsValues(ObsOnMapOpt={'range':mapdiffrange[mvar],'markersize':markersize,
                                'colormap':'RdBu_r'},
                        PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                        outNameSuffix='fgdep',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        fgdep.PlotObsHisto(vmin=histdiffrange[mvar]['min'],vmax=histdiffrange[mvar]['max'],
                           outNameSuffix='fgdep',outdir=framedir,subtitle='fgdep '+step.strftime('%Y-%m-%d-%H'))
        
        andep.PlotObsValues(ObsOnMapOpt={'range':mapdiffrange[mvar],'markersize':markersize,
                            'colormap':'RdBu_r'},
                    PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                    outNameSuffix='andep',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        andep.PlotObsHisto(vmin=histdiffrange[mvar]['min'],vmax=histdiffrange[mvar]['max'],
                           outNameSuffix='andep',outdir=framedir,subtitle='andep '+step.strftime('%Y-%m-%d-%H'))
        
        absinno.PlotObsValues(ObsOnMapOpt={'range':maplowdiffrange[mvar],'markersize':markersize,
                        'colormap':'RdBu_r'},
                PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                outNameSuffix='andep-fgdep',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        absinno.PlotObsHisto(vmin=histlowdiffrange[mvar]['min'],vmax=histlowdiffrange[mvar]['max'],
                       outNameSuffix='andep-fgdep',outdir=framedir,subtitle='|andep|-|fgdep| '+step.strftime('%Y-%m-%d-%H'))
        
        yall.PlotMaskValues(MaskOnMapOpt={'range':maprange[mvar],'markersize':markersize},
                                   PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                                   outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        
        # Compute stats and save them for a time series plot (not perfect solution since the obs
        # plots should be done by the class methods, but I haven't yet developed strategy to handle
        # multiple obs objects within the obs class)
        nobs.append(yall.nobs)
        nused.append(naccepted)
        nrej.append(nrejected)
        y_ts.append(y.Stats())
        hofxbg_ts.append(hofxbg.Stats())
        hofxan_ts.append(hofxan.Stats())
        fgdep_ts.append(fgdep.Stats())
        andep_ts.append(andep.Stats())
        absinno_ts.append(absinno.Stats())
    # This can undo selections if needed
    # y.ResetSel()

if mvar == 'od550du':
    # Plot the time series of observations here
    fig = plt.figure(figsize=(12,20),dpi=250)
    
    fig.autofmt_xdate()
    
    ax = fig.add_subplot(411)
    ax.set_title("Number of %s observations" % ('VIIRS'))
    
    ax.plot(dasteps[1:],nused,color='royalblue',\
            ls='-',lw=2,marker=None,label='Assimilated',markersize=4.0)
    ax.plot(dasteps[1:],nrej,color='black',\
        ls='--',lw=2,marker=None,label='Rejected',markersize=3.0)
    # ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%d %b, %H'))
    
    ax.grid(True)   
    ax.axis(xmin=dasteps[1],xmax=dasteps[-1])
    ax.set_xlabel('Time (UTC)')
    ax.legend(loc=1)
    
    ax = fig.add_subplot(412)
    ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
    ax.set_ylabel("%s" % ('unitless'))
    
    ax.errorbar(dasteps[1:],np.ma.array([v['550']['avg'] for v in hofxbg_ts]),
                yerr=np.ma.array([v['550']['std'] for v in hofxbg_ts]),color='royalblue',
            ls=None,marker='o',label='hofx_fg',markersize=4.0,
            capsize=6, elinewidth=1.7, markeredgewidth=1.7)
    ax.errorbar(dasteps[1:],np.ma.array([v['550']['avg'] for v in hofxan_ts]),
                yerr=np.ma.array([v['550']['std'] for v in hofxan_ts]),color='salmon',
            ls=None,marker='o',label='hofx_an',markersize=4.0,
            capsize=4, elinewidth=1.3, markeredgewidth=1.3)    
    ax.errorbar(dasteps[1:],np.ma.array([v['550']['avg'] for v in y_ts]),
                yerr=np.ma.array([v['550']['std'] for v in y_ts]),color='black',
            ls=None,marker='o',label='y',markersize=4.0,
            capsize=2, elinewidth=1., markeredgewidth=1.,alpha=0.7)
            
        
    ax.grid(True)   
    ax.axis(xmin=dasteps[1],xmax=dasteps[-1])
    ax.set_xlabel('Time (UTC)')
    ax.legend(loc=1)
    
    ax = fig.add_subplot(413)
    ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
    ax.set_ylabel("%s" % ('unitless'))
    
    ax.errorbar(dasteps[1:],np.ma.array([v['550']['avg'] for v in fgdep_ts]),
                yerr=np.ma.array([v['550']['std'] for v in fgdep_ts]),color='royalblue',
            ls=None,marker='o',label='fgdep',markersize=4.0,
            capsize=6, elinewidth=1.7, markeredgewidth=1.7)
    ax.errorbar(dasteps[1:],np.ma.array([v['550']['avg'] for v in andep_ts]),
                yerr=np.ma.array([v['550']['std'] for v in andep_ts]),color='salmon',
            ls=None,marker='o',label='andep',markersize=4.0,
            capsize=4, elinewidth=1.3, markeredgewidth=1.3)
        
    ax.grid(True)
    ax.axis(xmin=dasteps[1],xmax=dasteps[-1])
    ax.set_xlabel('Time (UTC)')
    ax.legend(loc=1)
    
    ax = fig.add_subplot(414)
    ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
    ax.set_ylabel("%s" % ('unitless'))
    
    ax.errorbar(dasteps[1:],np.ma.array([v['550']['avg'] for v in absinno_ts]),
                yerr=np.ma.array([v['550']['std'] for v in absinno_ts]),color='royalblue',
            ls=None,marker='o',label='|andep|-|fgdep|',markersize=4.0,
            capsize=6, elinewidth=1.7, markeredgewidth=1.7)
        
    ax.grid(True)
    ax.axis(xmin=dasteps[1],xmax=dasteps[-1])
    ax.set_xlabel('Time (UTC)')
    ax.legend(loc=1)
    
    plt.savefig(os.path.join(outdir,'VIIRS_%s_time_series.png' % (date[:-2])))
    plt.close(fig)
    
    np.savez(os.path.join(outdir,'VIIRS_%s_time_series.npz' % (date[:-2])),
         time=dasteps[1:],
         ntotobs=np.array(nobs),
         nusedobs=np.array(nused),
         nrejobs=np.array(nrej),
         y_avg=np.ma.array([v['550']['avg'] for v in y_ts]).filled(float(-1.0e38)),
         y_std=np.ma.array([v['550']['std'] for v in y_ts]).filled(float(-1.0e38)),
         hofxfg_avg=np.ma.array([v['550']['avg'] for v in hofxbg_ts]).filled(float(-1.0e38)),
         hofxfg_std=np.ma.array([v['550']['std'] for v in hofxbg_ts]).filled(float(-1.0e38)),
         hofxan_avg=np.ma.array([v['550']['avg'] for v in hofxan_ts]).filled(float(-1.0e38)),
         hofxan_std=np.ma.array([v['550']['std'] for v in hofxan_ts]).filled(float(-1.0e38)),
         fgdep_avg=np.ma.array([v['550']['avg'] for v in fgdep_ts]).filled(float(-1.0e38)),
         fgdep_std=np.ma.array([v['550']['std'] for v in fgdep_ts]).filled(float(-1.0e38)),
         andep_avg=np.ma.array([v['550']['avg'] for v in andep_ts]).filled(float(-1.0e38)),
         andep_std=np.ma.array([v['550']['std'] for v in andep_ts]).filled(float(-1.0e38)),
         missing_value=float(-1.0e38))

#Make some gifs at the end
for plttype in ['monarch-%s-*-fg.png' % (mvar),'monarch-%s-*-an.png' % (mvar),'monarch-%s-*-inc.png' % (mvar),
                'monarch-%s-*-fg-std.png' % (mvar),'monarch-%s-*-an-std.png' % (mvar),
                'VIIRS_550_AODMap_*_y.png','VIIRS_550_AODMap_*_fgdep.png','VIIRS_550_AODMap_*_andep.png',
                'VIIRS_550_AODMap_*_andep-fgdep.png',
                'VIIRS_550_AODMap_*_hofxbg.png','VIIRS_550_AODMap_*_hofxan.png',
                'VIIRS_AODHisto_*_andep.png','VIIRS_AODHisto_*_fgdep.png','VIIRS_AODHisto_*_andep-fgdep.png',
                'VIIRS_MaskMap_*.png']:
    subprocess.run(['convert','-delay','80','-loop','0',os.path.join(framedir,plttype),
                    os.path.join(gifdir,plttype.replace('*','anim').replace('.png','.gif'))])
        
# Here we can plot obs for all day if needed
# yall.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
#                             # 'colormap':'viridis'},
#                             'colormap':'YlOrBr'},
#                            PlotRegion=pltbounds,MapOpt={'AddGrid':False},
#                            outNameSuffix='yall',outdir=outdir)
# yall.PlotMaskValues(MaskOnMapOpt={'range':maprange[mvar],'markersize':markersize},
#                            PlotRegion=pltbounds,MapOpt={'AddGrid':False},
#                            outNameSuffix='assim_mask',outdir=outdir)

