# -*- coding: utf-8 -*-
"""

LETKF plot suite

Plots hourly output data from MONARCH LETKF, compute misfits,
innovation statistics, model fields, analysis increments
and ensemble standard deviation

TODO:
    - [ ] use experiment namelist to set script parameters
    - [ ] generalize instrument type (only VIIRS DOD currently)
"""

import os
import numpy as np
from netCDF4 import Dataset
import matplotlib.pyplot as plt
import obsdb
from plotters import plot2Dmap
from datetime import datetime as dt
from datetime import timedelta as td
from scipy.interpolate import griddata
import copy
import subprocess

date = '2024011400'
# date = '2020060100'
fcstep = td(hours=3)
fcperiod = td(hours=48)
# First step to start making the plots
first_step = 1
mvar = 'od550du'
#mvar = 'sconcdu'
# isregional = False
isregional = True
# MONARCH grid information (used to rotate back observations lats,lons)
cenlon, cenlat = 20., 35.
# cenlon is tph0d, cenlat is tlm0d
exppath = '/esarchive/exp/monarch/a6tl'
# exppath = '/esarchive/exp/monarch/a6tk'
mpath = os.path.join(exppath,'regional/3hourly')
useobs = True
opath = '/esarchive/scratch/eemili/obs/viirs/noaa20/namee/plots'
mfile = os.path.join(mpath,'%s' % (mvar),'%s_%03i_%s.nc' % (mvar,0,date))
pltbounds = {'LatMin':-11,'LatMax':70,'LonMin':-62,'LonMax':102}
# pltbounds = {'LatMin':15,'LatMax':55,'LonMin':-10,'LonMax':50}
# pltbounds = {'LatMin':-90,'LatMax':90,'LonMin':-180,'LonMax':180}
maprange = {'od550du':{'max':1.0,'min':0.},'sconcdu':{'max':1.0e-6,'min':0.}}
mapdiffrange = {'od550du':{'max':0.5,'min':-0.5},'sconcdu':{'max':0.2e-6,'min':-0.2e-6}}
mapstdrange = {'od550du':{'max':0.5,'min':0.},'sconcdu':{'max':0.4e-6,'min':0.}}
maplowdiffrange =  {'od550du':{'max':0.3,'min':-0.3},'sconcdu':{'max':0.1e-6,'min':-0.1e-6}}
histrange = {'od550du':{'max':1.5,'min':0.}}
histdiffrange = {'od550du':{'max':1.5,'min':-1.5}}
histlowdiffrange = {'od550du':{'max':1.,'min':-1.}}
markersize = 1.0
outdir = os.path.join(exppath,'plots',date[:-2])
framedir = os.path.join(outdir,'frames')
gifdir = os.path.join(outdir,'gifs')

if not os.path.exists(outdir): os.makedirs(outdir)
if not os.path.exists(framedir): os.makedirs(framedir)
if not os.path.exists(gifdir): os.makedirs(gifdir)

inidate = dt.strptime(date, '%Y%m%d%H')
fcsteps = [inidate+i*fcstep for i in range(int(fcperiod/fcstep))]

# yall = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
nobs, y_ts, hofx_ts, fcdep_ts = [[] for i in range(4)]

# Now loop on output frequency and do some plots
for step in fcsteps[first_step:]:
    istep = int((step-inidate)/fcstep)
    print(step)
    print(istep)
    
    if useobs and mvar == 'od550du':
        obsfile = os.path.join(opath,'HDAT+%s.h5' % (step.strftime('%Y%m%d')))
        # Read and plot observations, departures and hofxbg. 
        # Object are reinitialized to avoid cumulating the obs for the moment
        y = obsdb.DefInstrument('VIIRS','VIIRS_AOD',wavelength=[550])
        # Read the obs here from the LETFK output (consider appending the obs, it should be automatically done)
        # but take care of the correct time with this slot shit
        y.Read(obsfile)
        y.Select(TimeBnd=(step-fcstep*0.5,step+fcstep*0.5))
        hofx = copy.deepcopy(y)
        y.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
                                    # 'colormap':'viridis'},
                                    'colormap':'YlOrBr'},
                        PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                        outNameSuffix='y',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        y.PlotObsHisto(vmin=histrange[mvar]['min'],vmax=histrange[mvar]['max'],
                            outNameSuffix='y',outdir=framedir,subtitle='y '+step.strftime('%Y-%m-%d-%H'))
    
    # Forecast field
    mfile = os.path.join(mpath,'%s' % (mvar),'%s-%03i_%s.nc' % (mvar,0,inidate.strftime('%Y%m%d%H')))
    f = Dataset(mfile, 'r', format="NETCDF4")
    xfc = f[mvar][istep,0,:,:]
    
    if isregional:
        lat_grid = f['lat'][:]
        lon_grid = f['lon'][:]
    else:
        grid = np.meshgrid(f['lon'][:],f['lat'][:])
        lat_grid = grid[1]
        lon_grid = grid[0]
        
    f.close()
    
    # if not xfg.mask.any():
    plot2Dmap(xfc,lat_grid,lon_grid,range=maprange[mvar],
               title=step.strftime('%Y-%m-%d-%H')+' FG\n Avg:%.2e Min:%.2e Max:%.2e' % (xfc.mean(),xfc.min(),xfc.max()),
               plotregion=pltbounds,
               filename=os.path.join(framedir,'monarch-%s-%s-fc' % (mvar,step.strftime('%Y%m%d%H'))))
    
    # -------------------------------------------------------------------------
    # Interpolate forecast data at the observation locations to get the innovation values
    # This is temporary, it should disappear when the LETKF will produce all diags files
    
    if useobs and mvar == 'od550du':
        
        interpolated_data = griddata(
            (lon_grid.ravel(), lat_grid.ravel()),  # Known points
            xfc.ravel(),  # Known data
            np.array((y.lons[~y.mask], y.lats[~y.mask])).T,  # Points to interpolate to
            method='linear'  # You can use 'linear', 'nearest', or other methods
        )
    
        hofx.aod['550'][~y.mask] = np.ma.masked_invalid(interpolated_data)
        fcdep = y - hofx

        # -------------------------------------------------------------------------
        
        hofx.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
                                'colormap':'YlOrBr'},
                        PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                        outNameSuffix='hofx',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        hofx.PlotObsHisto(vmin=histrange[mvar]['min'],vmax=histrange[mvar]['max'],
                            outNameSuffix='hofx',outdir=framedir,subtitle='hofx '+step.strftime('%Y-%m-%d-%H'))
        
        
        fcdep.PlotObsValues(ObsOnMapOpt={'range':mapdiffrange[mvar],'markersize':markersize,
                                'colormap':'RdBu_r'},
                        PlotRegion=pltbounds,MapOpt={'AddGrid':False},
                        outNameSuffix='fcdep',outdir=framedir,subtitle=step.strftime('%Y-%m-%d-%H'))
        fcdep.PlotObsHisto(vmin=histdiffrange[mvar]['min'],vmax=histdiffrange[mvar]['max'],
                           outNameSuffix='fcdep',outdir=framedir,subtitle='fcdep '+step.strftime('%Y-%m-%d-%H'))
        
        # Compute stats and save them for a time series plot (not perfect solution since the obs
        # plots should be done by the class methods, but I haven't yet developed strategy to handle
        # multiple obs objects within the obs class)
        nobs.append(y.nobs)
        y_ts.append(y.Stats())
        hofx_ts.append(hofx.Stats())
        fcdep_ts.append(fcdep.Stats())
    # This can undo selections if needed
    # y.ResetSel()

if useobs and mvar == 'od550du':
    # Plot the time series of observations here
    fig = plt.figure(figsize=(12,20),dpi=250)
    
    fig.autofmt_xdate()
    
    ax = fig.add_subplot(311)
    ax.set_title("Number of %s observations" % ('VIIRS'))
    
    ax.plot(fcsteps[first_step:],nobs,color='royalblue',\
            ls='-',lw=2,marker=None,label='Assimilated',markersize=4.0)
    # ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%d %b, %H'))
    
    ax.grid(True)   
    ax.axis(xmin=fcsteps[first_step],xmax=fcsteps[-1])
    ax.set_xlabel('Time (UTC)')
    ax.legend(loc=1)
    
    ax = fig.add_subplot(312)
    ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
    ax.set_ylabel("%s" % ('unitless'))
    
    ax.errorbar(fcsteps[first_step:],np.ma.array([v['550']['avg'] for v in hofx_ts]),
                yerr=np.ma.array([v['550']['std'] for v in hofx_ts]),color='royalblue',
            ls=None,marker='o',label='hofx_fc',markersize=4.0,
            capsize=6, elinewidth=1.7, markeredgewidth=1.7) 
    ax.errorbar(fcsteps[first_step:],np.ma.array([v['550']['avg'] for v in y_ts]),
                yerr=np.ma.array([v['550']['std'] for v in y_ts]),color='black',
            ls=None,marker='o',label='y',markersize=4.0,
            capsize=2, elinewidth=1., markeredgewidth=1.,alpha=0.7)
                 
    ax.grid(True)   
    ax.axis(xmin=fcsteps[first_step],xmax=fcsteps[-1])
    ax.set_xlabel('Time (UTC)')
    ax.legend(loc=1)
    
    ax = fig.add_subplot(313)
    ax.set_title("%s statistics (%s)" % ('VIIRS DOD','unitless'))
    ax.set_ylabel("%s" % ('unitless'))
    
    ax.errorbar(fcsteps[first_step:],np.ma.array([v['550']['avg'] for v in fcdep_ts]),
                yerr=np.ma.array([v['550']['std'] for v in fcdep_ts]),color='royalblue',
            ls=None,marker='o',label='fcdep',markersize=4.0,
            capsize=6, elinewidth=1.7, markeredgewidth=1.7)
        
    ax.grid(True)
    ax.axis(xmin=fcsteps[first_step],xmax=fcsteps[-1])
    ax.set_xlabel('Time (UTC)')
    ax.legend(loc=1)
    
    plt.savefig(os.path.join(outdir,'VIIRS_%s_time_series.png' % (date[:-2])))
    plt.close(fig)

    np.savez(os.path.join(outdir,'VIIRS_%s_time_series.npz' % (date[:-2])),
             time=fcsteps[first_step:],
             nobs=np.array(nobs),
             y_avg=np.ma.array([v['550']['avg'] for v in y_ts]).filled(float(-1.0e38)),
             y_std=np.ma.array([v['550']['std'] for v in y_ts]).filled(float(-1.0e38)),
             hofxfc_avg=np.ma.array([v['550']['avg'] for v in hofx_ts]).filled(float(-1.0e38)),
             hofxfc_std=np.ma.array([v['550']['std'] for v in hofx_ts]).filled(float(-1.0e38)),
             fcdep_avg=np.ma.array([v['550']['avg'] for v in fcdep_ts]).filled(float(-1.0e38)),
             fcdep_std=np.ma.array([v['550']['std'] for v in fcdep_ts]).filled(float(-1.0e38)),
             missing_value=float(-1.0e38))

#Make some gifs at the end
for plttype in ['monarch-%s-*-fc.png' % (mvar),
                'VIIRS_550_AODMap_*_y.png','VIIRS_550_AODMap_*_fcdep.png',
                'VIIRS_550_AODMap_*_hofx.png',
                'VIIRS_AODHisto_*_fcdep.png']:
    subprocess.run(['convert','-delay','80','-loop','0',os.path.join(framedir,plttype),
                    os.path.join(gifdir,plttype.replace('*','anim').replace('.png','.gif'))])
        
# Here we can plot obs for all day if needed
# yall.PlotObsValues(ObsOnMapOpt={'range':maprange[mvar],'markersize':markersize,
#                             # 'colormap':'viridis'},
#                             'colormap':'YlOrBr'},
#                            PlotRegion=pltbounds,MapOpt={'AddGrid':False},
#                            outNameSuffix='yall',outdir=outdir)
# yall.PlotMaskValues(MaskOnMapOpt={'range':maprange[mvar],'markersize':markersize},
#                            PlotRegion=pltbounds,MapOpt={'AddGrid':False},
#                            outNameSuffix='assim_mask',outdir=outdir)

