# -*- coding: utf-8 -*-
"""
AOD instrument class
Can be satellite or groundbased sun photometer

Created on Wed Oct 17 11:28:56 2018

@author: plu
"""
from __future__ import print_function, division
import os
import numpy as np
import h5py
from baseinstr import BaseInstrument
import matplotlib.pyplot as plt
from datetime import datetime, timedelta
import utils_Hobs_mini as UtilsASCII
import copy
from rotation_routines import tll, rtll


class AODInstrument(BaseInstrument):
    """ """
    def __init__(self, name, **kwargs):
        BaseInstrument.__init__(self,name,**kwargs)
        if "wavelength" in kwargs:
            self.wavelength = kwargs["wavelength"]   # The measured wavelengths in nm (float, could be used for plots)
            self.nwv = len(kwargs["wavelength"]) # The number of measured wavelengths
            self.wllist = ['%i' % (wl) for wl in self.wavelength] # List of wavelengths as string (used for hdf5 I/O)
            self.aod = {wv : np.ma.empty((0),dtype=np.float32)
                        for wv in self.wllist}
            self.cov = {wv : np.ma.empty((0),dtype=np.float32)
                        for wv in self.wllist}
            self.units = {wv : "dimless" for wv in self.wllist}
        else:
            self.aod=[]
            self.cov=[]
        # The assimilation mask
        self.assim_mask = np.ma.empty((0), dtype=np.int32)
        self.daimon_type="AOD"

    # Some operator overloading to add/substract/divide batch of measurements
    # taken at the same time and place (including possibly averages)
    def __add__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for wv in self.wllist:
                result.aod[wv] = self.aod[wv] + other.aod[wv]
                if 'avgaod' in self.__dict__:
                    result.avgaod[wv] = self.avgaod[wv] + other.avgaod[wv]
        else:
            print('Cannot sum obs: mismatch in geolocalizations')
        
        return result
        
    def __sub__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for wv in self.wllist:
                result.aod[wv] = self.aod[wv] - other.aod[wv]
                if 'avgaod' in self.__dict__:
                    result.avgaod[wv] = self.avgaod[wv] - other.avgaod[wv]
        else:
            print('Cannot diff obs: mismatch in geolocalizations')
        
        return result
    
    def __truediv__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for wv in self.wllist:
                result.aod[wv] = self.aod[wv] / other.aod[wv]
                if 'avgaod' in self.__dict__:
                    result.avgaod[wv] = self.avgaod[wv] / other.avgaod[wv]
                result.units[wv] = "%"
        else:
            print('Cannot divide obs: mismatch in geolocalizations')
        
        return result
    
    def abs_sub(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for wv in self.wllist:
                result.aod[wv] = np.ma.abs(self.aod[wv]) - np.ma.abs(other.aod[wv])
                if 'avgaod' in self.__dict__:
                    result.avgaod[wv] = np.ma.abs(self.avgaod[wv]) - np.ma.abs(other.avgaod[wv])
        else:
            print('Cannot absolute diff obs: mismatch in geolocalizations')
        
        return result

    def Read(self, filename, **kwargs):
        """
        Reads multi-wavelengths observations from MOCAGE HDF5
        
        If assim is True reads assimilation output as well
        """                                   

        self.OpenFile(filename,mode="r")
        nobs = self.ReadGeoValues(**kwargs) 
        
        if nobs > 0:
            subgrp = ""
            obstype = "Y"
            if 'ReadAssim' in kwargs.keys():
                if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']
                if 'ObsType' in kwargs['ReadAssim']: obstype = kwargs['ReadAssim']['ObsType']  
            
            obsgrp = self.f[self.name+subgrp+"/OBSERVATIONS"]
            
            # Init the wavelength with those found in the HDF5 if not done before
            if not hasattr(self,"nwv"):
                self.wavelength = []
                for el in obsgrp.keys():
                    self.wavelength.append(int(el))
                self.nwv = len(self.wavelength) # The number of measured wavelengths
                self.wllist = ['%i' % (wl) for wl in self.wavelength]
                self.aod = {wv : np.ma.empty((0),dtype=np.float32) \
                    for wv in self.wllist}
                self.units = {wv : "dimless" for wv in self.wllist}
                        
            # Read the observations
            for wv in self.wllist:
                print(">>> Reading %s, type %s, found %i obs" % (wv,obstype,len(obsgrp[wv+"/"+obstype])))
                self.aod[wv] = np.ma.concatenate((self.aod[wv],np.ma.array(np.squeeze(obsgrp[wv+"/"+obstype][:]),dtype=self.aod[wv].dtype)))
                #self.nobs = self.nobs + len(self.aod[wv])

        self.CloseFile()
        
    def ReadLETKFAscii(self,filename,initime,**kwargs):
        
        # A very specific readers for MONARCH-DA departure files
        # The obs file time has to be provided by the caller
        dtslot = kwargs['dtslot']
        nslots = kwargs['nslots']
        obstype = "Y"
        itype = 3
        rotated = False
        if 'rotated' in kwargs: 
            rotated = kwargs['rotated']
        if rotated:    
            cenlon = kwargs['cenlon']
            cenlat = kwargs['cenlat']
        islot = kwargs['islot']
        if 'ReadDA' in kwargs:
            if 'ObsType' in kwargs['ReadDA']: obstype = kwargs['ReadDA']['ObsType']

        if obstype == 'dep':  itype = 5

        data = np.loadtxt(filename)
        
        if data.shape[0] > 0:
            # Reduce data to only wanted timeslots
            if islot is not None:
                idx = (data[:,0] == (-nslots+islot)).nonzero()[0]
                data = data[idx,:]

        nobs = data.shape[0]
        if nobs > 0:
            # Implement the selection in time for the reader
            self.nobs += nobs # Increase the instrument counter
            self.time = np.ma.concatenate(
                (self.time,
                 np.ma.array(
                    # self.time_orig+grp["Timestamp"],
                    np.array([initime+timedelta(hours=islot*dtslot) for i in range(nobs)]),
                    dtype=self.np_typetime)))
            
            if rotated:
                newlons, newlats = rtll(np.radians(data[:,1]),np.radians(data[:,2]),cenlon,
                                        np.cos(np.radians(cenlat)),np.sin(np.radians(cenlat)))
            else:
                newlons, newlats = data[:,1], data[:,2]
            
            self.lats = np.ma.concatenate(
                (self.lats,
                 # np.ma.array(grp["Longitude"],
                 np.ma.array(newlats,
                             dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate(
                (self.lons,
                 # np.ma.array(grp["Longitude"],
                 np.ma.array(newlons,
                             dtype=self.lons.dtype)))
            self.mask = np.concatenate(
            (self.mask,np.zeros((nobs),dtype=bool))) # By default all read obs are accepted
            
            self.assim_mask = np.ma.concatenate(
                (self.assim_mask,np.ma.ones((nobs),dtype=np.int32))) # By default all read obs are considered assimilated
            
            # Init the wavelength with those found in the HDF5 if not done before
            if not hasattr(self,"nwv"): 
                print('The wavelength must be initialized before when reading ASCII files')
                return
            else:
                # Read the observations
                for wv in self.wllist:
                    print(">>> Reading %s, type %s, found %i obs" % (wv,obstype,data.shape[0]))
                    self.aod[wv] = np.ma.concatenate((self.aod[wv],np.ma.array(data[:,itype],dtype=self.aod[wv].dtype)))
        return nobs

    def WriteAODHeader(self):
        
        subgrp = self.f[self.name].create_group("OBSERVATIONS")
        for wv in self.wllist:
            subgrp.create_group(wv)

    def ConvertMocageASCII(self, date, inputdir, **kwargs):
        """
        Reads monowavelength AOD from old MOCAGE ASCII
        """

        wv = self.wllist[0] # Old ASCII are monowavelength, so take only one if multiple species are present (the first)!  

        # Use the old MOCAGE HDAT reader for now                      
        obstype, obsdate, obsname, obslat, obslon, obsval, dummy = \
            UtilsASCII.LecHOBS(os.path.join(inputdir, \
            'HDAT+'+date.isoformat()[:10].replace("-","")),obstype='totcol')
            
        obsname = [n[0] for n in obsname] # Keep only instrument name
        obsdate = [datetime.strptime(t,"%Y%m%d%H%M") for t in obsdate] # Datetime conv       
        obsval = [o[0] for o in obsval] # Remove singleton dimension for surface observations

        self.nobs += len(obsdate) # Increase the instrument counter
        
        if len(obsdate) > 0:
            self.aod[wv] = np.ma.concatenate((self.aod[wv],np.ma.array(obsval,dtype=np.float32)))
            self.lats = np.ma.concatenate((self.lats,np.ma.array(obslat,dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate((self.lons,np.ma.array(obslon,dtype=self.lons.dtype)))
            self.time = np.ma.concatenate((self.time,np.ma.array(obsdate,dtype=self.np_typetime)))
            self.mask = np.concatenate((self.mask,np.zeros((len(obsdate)),dtype=bool))) # By default all obs are accepted
        
    def Write(self, filename, **kwargs):
        """ Writes multispecies observations in MOCAGE HDF5. """                                   
        # Prepare file
        self.OpenFile(filename)
        self.WriteHeader()
        self.WriteAODHeader()
        
        # Surface concentration specific code
        time_dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset("Timestamp",(self.nobs,),dtype=self.h5_typetime)
        h5arrtype = np.dtype('(1,)f')
        for wv in self.wllist:
            dset =  self.f[self.name+"/OBSERVATIONS/"+wv].create_dataset('Y',(self.nobs,),dtype=h5arrtype,maxshape=(None,))
            self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
            
        # Write actual data in the file
        self.WriteGeoValues()

        if self.nobs > 0:
            for wv in self.wllist:
                dset = self.f[self.name+"/OBSERVATIONS/"+wv].require_dataset("Y",(self.nobs,),dtype=h5arrtype)
                dset[...] = np.array([val for val in self.aod[wv][~self.mask][:,np.newaxis]],dtype=self.aod[wv].dtype)[self.sorted_idx]
        
        self.CloseFile()
        print(">>> HDF5 written")

    # A comprehensive printer for AOD stats
    def Stats(self,**kwargs):
        
        if self.nobs > 1:
            self.PrintGeoStats()
            stats = {wv:{'avg':self.aod[wv].mean(),'std':self.aod[wv].std()} for wv in self.wllist}
        else:
            stats = {wv:{'avg':np.ma.masked,'std':np.ma.masked} for wv in self.wllist}
        return stats
        
    # A comprehensive plotter for aod data
    def Plot(self,outdir="./",**kwargs):
    
        if self.nobs > 0:
            if 'makeplots' in kwargs:
                if 'ObsLoc' in kwargs['makeplots']: self.PlotObsLocation(outdir=outdir,**kwargs)
                if 'ObsHisto' in kwargs['makeplots']: self.PlotObsHisto(outdir=outdir)
                if 'ObsOnMap' in kwargs['makeplots']: self.PlotObsValues(outdir=outdir,**kwargs)
                if 'TimeSeries' in kwargs['makeplots']: self.PlotTimeSeries(outdir=outdir,**kwargs)
            else:
                self.PlotObsLocation(outdir=outdir,**kwargs)
                self.PlotObsHisto(outdir=outdir)
                self.PlotObsValues(outdir=outdir,**kwargs)
                self.PlotTimeSeries(outdir=outdir,**kwargs)
        else:
            print('>> No obs to be plotted')

    def Select(self,**kwargs):
        
        self.SelectGeo(**kwargs)
        for wv in self.wllist: self.aod[wv].mask = self.mask
        
    def ResetSel(self,**kwargs):
        
        self.ResetSelGeo(**kwargs)
        for wv in self.wllist: self.aod[wv].mask = self.mask

    def Average(self,**kwargs):
        
        avgmsk = self.TimeAveragerMask(**kwargs)
        
        self.avgaod = {wv : np.ma.empty((self.avgtime.size),dtype=self.aod[wv].dtype) \
            for wv in self.wllist}
        self.stdaod = {wv : np.ma.empty((self.avgtime.size),dtype=self.aod[wv].dtype) \
                       for wv in self.wllist}
        for wv in self.wllist:
            self.avgaod[wv][:] = np.ma.masked
            for it in range(self.avgtime.size):
                if avgmsk[it].any(): 
                    self.avgaod[wv][it] = np.ma.mean(self.aod[wv][avgmsk[it]])
                    self.stdaod[wv][it] = np.ma.std(self.aod[wv][avgmsk[it]])

    def PlotObsHisto(self,outdir="./", outNameSuffix='',**kwargs):
        
        # add underscore in suffix
        if outNameSuffix != '':
            outNameSuffix = '_' + outNameSuffix
            
        subtitle = None
        if 'subtitle' in kwargs: subtitle = kwargs['subtitle']
        
        if self.nobs > 0:
            fig = plt.figure()
            
            filename = os.path.join(outdir,"%s_AODHisto_%s_%s%s" % \
                    (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                     self.time.compressed().max().astype(str)[:13].replace("-",""),outNameSuffix))
    
            print('>>> Printing plot:', filename)
            
            for wv in self.wllist:
                
                ax = fig.add_subplot(111)
                titlestr = "%s AOD %s (%s)" % (self.name,wv,self.units[wv])
                if subtitle is not None: titlestr = titlestr + '\n ' + subtitle
                ax.set_title(titlestr)
                ax.set_xlabel("Avg:%4.2e Std:%4.2e N: %i" % \
                              (self.aod[wv].mean(),np.ma.std(self.aod[wv]),self.nobs))
                ax.hist(self.aod[wv].compressed(),bins=35)
    
                vmin, vmax = self.aod[wv].min(), self.aod[wv].max()
                if 'vmin' in kwargs: vmin = kwargs['vmin']
                if 'vmax' in kwargs: vmax = kwargs['vmax']
                ax.set_xlim((vmin,vmax))
                
            fig.savefig(filename)
            plt.close(fig)

    def PlotObsValues(self, outdir="./", figsize=(10, 5), outNameSuffix='',
                      **kwargs):
        """ Plot AOD values on a map.

        Parameters
        ----------
        outdir : str
            Output directory.
        figsize : couple of scalars
            Size of the matplotlib figure.
        outNameSuffix : str
            Suffix to add to the output file name.
        ObsOnMapOpt : dict
            Dictionary defining plot settings. Its keys can be,
            markersize : int
                Size of the markers.
            colormap : str
                Name of the matplotlib colormap.
            overlapping_avg : bool
                If True, AOD values at a same lat/lon are averaged.

        """
        fig = plt.figure(figsize=figsize, dpi=200)

        # add underscore in suffix
        if outNameSuffix != '':
            outNameSuffix = '_' + outNameSuffix
            
        subtitle = None
        if 'subtitle' in kwargs: subtitle = kwargs['subtitle']

        markersize = 2.5
        colormap = "afmhot_r"
        overlapping_avg = False
        if 'ObsOnMapOpt' in kwargs:
            if 'markersize' in kwargs["ObsOnMapOpt"]:
                markersize = kwargs["ObsOnMapOpt"]['markersize']
            if 'colormap' in kwargs["ObsOnMapOpt"]:
                colormap = kwargs["ObsOnMapOpt"]['colormap']
            if 'overlapping_avg' in kwargs["ObsOnMapOpt"]:
                overlapping_avg = kwargs["ObsOnMapOpt"]['overlapping_avg']

        if self.nobs > 0:
            for wv in self.wllist:
                ax = fig.add_subplot(111)
                filename = os.path.join(
                    outdir,
                    "{0}_{1}_AODMap_{2}_{3}{4}".format(
                        self.name,
                        wv,
                        self.time.compressed().min().astype(str)[:13].replace("-", ""),
                        self.time.compressed().max().astype(str)[:13].replace("-", ""),
                        outNameSuffix))
    
                # Variable related keywords have priority over global keywords
                lvarkwargs = {}
                if "ObsOnMapOpt" in kwargs:
                    if wv in kwargs["ObsOnMapOpt"]: 
                        lvarkwargs = kwargs["ObsOnMapOpt"][wv]
                    else:
                        lvarkwargs = kwargs["ObsOnMapOpt"]
    
                # set range
                vmin = self.aod[wv].compressed().min()
                vmax = self.aod[wv].compressed().max()
                if "range" in lvarkwargs:
                    if 'min' in lvarkwargs['range']: vmin = lvarkwargs["range"]["min"]
                    if 'max' in lvarkwargs['range']: vmax = lvarkwargs["range"]["max"]
    
                print('>>> Printing plot:', filename)
    
                m = self.SetBaseMap(ax, **kwargs)
    
                if overlapping_avg:
                    import pandas as pd
                    df = pd.DataFrame({'lat': self.lats.compressed(),
                                       'lon': self.lons.compressed(),
                                       'aod': self.aod[wv].compressed()})
                    df = df.groupby(['lat', 'lon']).mean().reset_index()
                    (x, y) = m(df['lon'], df['lat'])
                    values = df['aod']
                else:
                    (x, y) = m(self.lons.compressed(), self.lats.compressed())
                    values = self.aod[wv].compressed()
    
                sc = m.scatter(x, y, markersize, marker='o',
                               c=values,
                               edgecolors='none',
                               # edgecolors='black',lw=0.2,
                               cmap=colormap, zorder=10,
                               vmin=vmin, vmax=vmax
                               )
    
                plt.colorbar(sc, orientation='horizontal', pad=0.05, extend='both')
    
                title = ('{} {} N={}' +
                              '\nAvg: {:.2f} Min: {:.2f} Max: {:.2f} ({})').format(
                    self.name, wv, self.nobs,
                    self.aod[wv].mean(), self.aod[wv].min(), self.aod[wv].max(),
                    self.units[wv])
                if subtitle is not None: title = title + '\n ' + subtitle
                
                ax.set_title(title)
    
            plt.tight_layout()
            fig.savefig(filename)
            plt.close(fig)

    def PlotTimeSeries(self,outdir="./",**kwargs):
        
        figsize=(10,5)
        lkwargs = {}
        
        if 'TimeSeriesOpt' in kwargs:
            lkwargs = kwargs["TimeSeriesOpt"]
            if "figsize" in lkwargs: figsize = lkwargs["figsize"]
            
        if 'avgaod' not in self.__dict__: self.Average(**kwargs)
        
        fig = plt.figure(figsize=figsize,dpi=200)
        
        filename = os.path.join(outdir,"%s_AODTimeSeries_%s_%s" % \
                (self.name,self.avgtime.min().astype(str)[:13].replace("-",""), \
                 self.avgtime.max().astype(str)[:13].replace("-","")))
        
        print('>>> Printing plot:', filename)
         
        for wv in self.wllist:
            
            ax = fig.add_subplot(111)
            
            ax = self.SetTimePlot(fig,**kwargs)
            
            ax.set_title("%s (%s)" % (wv,self.units[wv]))
            ax.set_ylabel("%s" % (self.units[wv]))

            ax.plot(self.avgtime.astype(datetime),self.avgaod[wv],color='royalblue',\
                    ls=None,lw=None,marker='o',label=None,markersize=4.0)
            
            if 'yrange' in lkwargs: 
                ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
    
        fig.savefig(filename)
        plt.close(fig)

    def make_super_obs(
            self, resolution, grid_min_lat=-90., grid_min_lon=-180.,
            averaging='median', location='centroid'):
        """ Compute super-observations.

        Super-observations are computed as the a median of all the
        observations contained in a same mesh of the grid. It is
        located at the mean latitudes and longitudes of these observations.
        The grid used starts from -90 in latitude and -180 in longitude,
        and is a regular lat/lon grid with resolution defined in argument.

        Parameters
        ----------
        resolution : float
            Resolution of the regular lat/lon grid used for computation.
        grid_min_lat, grid_min_lon : float
            Minimum latitude and longitude of the regular lat/lon grid
            used for computation.
        averaging : str
            Type of averaging used to compute super obs values. It can
            be 'median' or 'mean'.
        location : str
            Way of choosing lat/lon coordinates for super obs. It can
            be 'centroid' (arithmetic mean position of all the points
            in the cell) or 'center' (center of the cell).

        """
        import pandas as pd

        # save data
        df = pd.DataFrame({'time': self.time.compressed().tolist(),
                           'lats': self.lats.compressed(),
                           'lons': self.lons.compressed()})
        aod = {wv: self.aod[wv].compressed() for wv in self.wllist}

        # reinitialize self
        BaseInstrument.__init__(self, self.name)
        self.aod = {wv: np.ma.empty((0), dtype=np.float32)
                     for wv in self.wllist}

        # find lat/lon index of corresponding grid cell
        df['lat_idx'] = ((df.lats - grid_min_lat)/resolution).astype(int)
        df['lon_idx'] = ((df.lons - grid_min_lon)/resolution).astype(int)

        time_unique = np.unique(df.time)
        for t in time_unique:
            time_idx = (df.time == t).values
            sub_df = df.loc[time_idx]

            # find lat/lon of super obs
            if location == 'centroid':
                supobs_latlon = sub_df[
                    ['lat_idx', 'lon_idx', 'lats', 'lons']].groupby(
                        ['lat_idx', 'lon_idx'], as_index=False).mean()
            elif location == 'center':
                supobs_latlon = sub_df[
                    ['lat_idx', 'lon_idx', 'lats', 'lons']].groupby(
                        ['lat_idx', 'lon_idx'], as_index=False).count()
                supobs_latlon.lats = (grid_min_lat +
                    supobs_latlon.lat_idx*resolution + resolution/2.)
                supobs_latlon.lons = (grid_min_lon +
                    supobs_latlon.lon_idx*resolution + resolution/2.)
            else:
                raise Exception("location argument must be either" +
                                "'centroid' or 'center'.")

            # compute super obs values
            for wv in self.wllist:
                sub_df = sub_df.assign(aod=aod[wv][time_idx])
                if averaging == 'median':
                    avg_val = sub_df[['lat_idx', 'lon_idx', 'aod']].groupby(
                        ['lat_idx', 'lon_idx'], as_index=False).median()
                elif averaging == 'mean':
                    avg_val = sub_df[['lat_idx', 'lon_idx', 'aod']].groupby(
                        ['lat_idx', 'lon_idx'], as_index=False).mean()
                else:
                    raise Exception("averaging argument must be either" +
                                    "'median or 'mean'.")
                self.aod[wv] = np.ma.concatenate(
                    (self.aod[wv],
                     np.ma.array(avg_val['aod'], dtype=np.float32)))

            nobs = supobs_latlon.shape[0]
            self.nobs += nobs
            self.lats = np.ma.concatenate(
                (self.lats, np.ma.array(supobs_latlon.lats,
                                        dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate(
                (self.lons, np.ma.array(supobs_latlon.lons,
                                        dtype=self.lons.dtype)))
            self.time = np.ma.concatenate(
                (self.time, np.ma.array(nobs*[t],
                                        dtype=self.np_typetime)))
            self.mask = np.concatenate((self.mask,
                                        np.zeros((nobs), dtype=bool)))

    def PlotMaskValues(self,outdir="./",**kwargs):
        
        if self.assim_mask.size == 0: return
        print(">>> Plotting assimilation mask on a map")
        
        figsize = (8,8)
        markersize = 2
        colormap = "Set1"
        lkwargs = {}
        subtitle = None
        
        if 'subtitle' in kwargs: subtitle = kwargs['subtitle']
        if 'MaskOnMapOpt' in kwargs: lkwargs = kwargs["MaskOnMapOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        
        fig = plt.figure(figsize=figsize,dpi=200)
            
        ax = fig.add_subplot(1,1,1)
            
        m = self.SetBaseMap(ax,**kwargs)
        
        fld = self.assim_mask.compressed()

        title = 'Assim Mask \n Nmiss: %i,Nrej: %i, Nok: %i' % \
                             ((fld == -2).sum(),(fld == 0).sum(),(fld == 1).sum())
        if subtitle is not None: title = title + '\n ' + subtitle
        ax.set_title(title)
                
        sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o',\
                               c=fld,cmap=colormap,latlon=True,zorder=10,vmin=-2,vmax=1)

        plt.colorbar(sc,orientation='horizontal')   
            
        fig.savefig(os.path.join(outdir,"%s_MaskMap_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)