# coding: utf8

from __future__ import print_function
import os
import sys
from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np

from baseinstr import BaseInstrument
from concinstr import ConcInstrument

import utils_Hobs_mini as UtilsASCII

try:
    import seaborn as sns
except ImportError:
    pass


class IntQtyInstrument(ConcInstrument):
    """ """
    def __init__(self, name, species=None, **kwargs):
        # Init with user defined species if given
        # The behaviour is as follows:
        # - no species init by users -> all species found in h5 are read (units and convfctr reader to be done)
        # - species init by the user -> only specified ones are read
        # - for ascii conversion error is raised if the user does not specify the species
        ConcInstrument.__init__(self, name, species=species)
        if species is not None:
            if "nmaxlevs" in kwargs:
                self.nmaxlevs = kwargs["nmaxlevs"]
            else:
                raise Exception("nmaxlevs (max number of vertical levels) " +
                                "argument is required.")

            self.pcol = {spc : np.ma.empty((0), dtype=np.float32)
                         for spc in self.species}
            self.avk = {spc : np.ma.empty((0, self.nmaxlevs), dtype=np.float32)
                        for spc in self.species}
            self.vert = np.ma.empty((0, self.nmaxlevs), dtype=np.float32)
            self.vertunit = 'abc'  # Pa or meters
            self.vertorient = 'abc'  # up or down
            if "VertGridCnst" in kwargs:
                self.vertgridcnst = kwargs["VertGridCnst"]
            else:
                self.vertgridcnst = 0
            self.VertDsetName = 'abc'  # Pressure or Altitude
            self.cov = {spc : np.ma.empty((0, self.nmaxlevs, self.nmaxlevs),
                                          dtype=np.float32)
                        for spc in self.species}

            self.vert.fill_value = self.missing_value
            for spc in self.species:
                self.cov[spc].fill_value = self.missing_value
                self.avk[spc].fill_value = self.missing_value

        # moved from concinstr (diverging in profile and intqty)
        self.units = {spc : "DU" for spc in self.species}
        self.cnvfctr = {spc : 1.0 for spc in self.species}

        self.hascov = False
        self.daimon_type = "Intqty"

    def Select(self, **kwargs):
        """ """
        if 'ValBnd' in kwargs.keys():  # Selection of values
            valbnd = kwargs['ValBnd']
            for spc in self.species:
                valmsk = ~((self.pcol[spc] > valbnd[0]) &
                           (self.pcol[spc] < valbnd[1]))
                if valmsk.any(): self.mask = self.mask | valmsk

        self.SelectGeo(**kwargs)
        # multidim array
        self.vert[self.mask, :] = np.ma.masked
        # multispecies arrays
        for spc in self.species:
            self.pcol[spc].mask = self.mask
            #multidim arrays
            self.avk[spc][self.mask,:] = np.ma.masked
            if self.hascov:
                self.cov[spc][self.mask,:,:] = np.ma.masked

    def Read(self, filename, **kwargs):
        """ Reads multi-species observations from MOCAGE HDF5.
        
        If assim is True reads assimilation output as well

        If reads several input files, we assume that they have the
        same format (same vertical coordinate for instance).

        """

        self.OpenFile(filename,mode="r")
        self.ReadGeoValues(**kwargs) 

        subgrp = ""
        obstype = "Y"
        if 'ReadAssim' in kwargs.keys():
            if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']
            if 'ObsType' in kwargs['ReadAssim']: obstype = kwargs['ReadAssim']['ObsType']  

        #Read Vertical values
        geogrp = self.f[self.name+subgrp+"/GEOLOCALIZATION"]
        npvert=[]
        nmaxlevs=1
        for el in geogrp.keys():
            if el=='Altitude' or el=='Pressure':
                self.VertDsetName = el
                self.vertgridcnst=geogrp[el].attrs["IsGridConstant"]
                self.vertunit = geogrp[el].attrs["Units"]
                self.vertorient = geogrp[el].attrs["Orientation"]
                if self.vertgridcnst==0:
                    npvert = geogrp[el][:]
                    self.nmaxlevs=npvert.shape[1]

        #Read variables
        obsgrp = self.f[self.name+subgrp+"/OBSERVATIONS"]

        # Init the species with those found in the HDF5 if not done before
        if not hasattr(self,"species"):
            self.species = []
            for el in obsgrp.keys():
                self.species.append(el)
            self.pcol = {spc : np.ma.empty((0),dtype=np.float32)
                         for spc in self.species}
            self.avk = {spc : np.ma.empty((0,self.nmaxlevs),dtype=np.float32)
                        for spc in self.species}
            self.vert = np.ma.empty((0,self.nmaxlevs),dtype=np.float32)

        for spc in self.species:

            print(">>> Reading %s, type %s, found %i obs" % (spc,obstype,len(obsgrp[spc+"/"+obstype])))
            self.pcol[spc] = np.ma.concatenate((self.pcol[spc],np.ma.array(np.squeeze(obsgrp[spc+"/"+obstype][:]),dtype=self.pcol[spc].dtype)))
            self.avk[spc] = np.ma.concatenate((self.avk[spc],np.ma.array(np.squeeze(obsgrp[spc+"/AveragingKernels"][:]),dtype=self.avk[spc].dtype)))
            if "ErrorCovariance" in obsgrp[spc].keys():
                self.hascov = True
                self.cov[spc] = np.ma.concatenate((self.cov[spc],np.ma.array(np.squeeze(obsgrp[spc+"/ErrorCovariance"][:]),dtype=self.cov[spc].dtype)))
            self.units[spc] = obsgrp[spc+"/"+obstype].attrs["Units"]
            self.cnvfctr[spc] = 1.0
            self.missing_value = obsgrp[spc+"/"+obstype].attrs["MissingValue"]
        self.vert=np.ma.concatenate((self.vert,np.ma.array((npvert),dtype=self.vert.dtype)))

    def ConvertMocageASCII(self,date,inputdir):
        """ Reads monospecies observations from old MOCAGE ASCII. """                          
        spc = self.species[0] # Old ASCII are monospecies, so take only one if multiple species are present (the first)!
        self.vertunit = 'Pa' #Pa or meters
        self.vertorient = 'up' #up or down
        self.vertgridcnst = 1
        self.VertDsetName = 'Pressure' #pressure or altitude
            
        # Use the old MOCAGE HDAT reader for now
        obstype, obsdate, obsname, obslat, obslon, obsval, obspress = \
            UtilsASCII.LecHOBS(os.path.join(inputdir, \
            'HDAT+'+date.isoformat()[:10].replace("-","")),obstype='partcol')
        npval = np.ma.array(obsval,dtype=self.cov[spc].dtype).squeeze()
        
        # Avk is mandatory until now
        ostypetmp, datetmp, nametmp, lattmp, lontmp, obsavk  =  \
            UtilsASCII.LecHAVK(os.path.join(inputdir,'HAVK+'+date.isoformat()[:10].replace("-","")))
            
        nobs = len(obsdate)
            
        if  os.path.isfile(os.path.join(inputdir,'HCOV+'+date.isoformat()[:10].replace("-",""))): 
            self.hascov = True
            ostypetmp, datetmp, nametmp, lattmp, lontmp, obscov, presstmp  =  \
                UtilsASCII.LecHCOV(os.path.join(inputdir, \
                'HCOV+'+date.isoformat()[:10].replace("-","")))
            npcov = np.ma.array(obscov,dtype=self.cov[spc].dtype).squeeze()
            
        obsname = [n[0]+n[1] for n in obsname] # Reduce old strings found in HDATs to a single site code
        obsdate = [datetime.strptime(t,"%Y%m%d%H%M") for t in obsdate] # Datetime conversion
        
        # Convert observation lists to numpy arrays for vertical values
        nppress = np.full((nobs,self.nmaxlevs),self.missing_value,dtype=self.vert.dtype)
        npavk = np.full((nobs,self.nmaxlevs),self.missing_value,dtype=self.avk[spc].dtype)
        
        for ip in range(nobs):
            nppress[ip,:len(obspress[ip])] = obspress[ip]
            npavk[ip,:len(obspress[ip])] = obsavk[ip]
        
        self.nobs += nobs # Increase the instrument counter
        
        if self.vertgridcnst == 1 and self.vert.size == 0 : 
            self.vert = np.ma.concatenate((self.vert,np.ma.array(nppress[0,:])[np.newaxis,:]))
        
        if nobs > 0:
            if self.vertgridcnst == 0: self.vert = np.ma.concatenate((self.vert,np.ma.array(nppress)))
            self.pcol[spc] = np.ma.concatenate((self.pcol[spc],npval))
            self.avk[spc] = np.ma.concatenate((self.avk[spc],npavk))
            self.lats = np.ma.concatenate((self.lats,np.ma.array(obslat,dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate((self.lons,np.ma.array(obslon,dtype=self.lons.dtype)))
            self.time = np.ma.concatenate((self.time,np.ma.array(obsdate,dtype=self.time.dtype)))
            if self.hascov: self.cov[spc] = np.ma.concatenate((self.cov[spc],npcov))
            self.mask = np.concatenate((self.mask,np.zeros((len(obsdate)),dtype=bool))) # By default all obs are accepted

    def Write(self, filename, **kwargs):
        """ Writes multispecies observations in MOCAGE HDF5. """

        if self.nobs == 0:
            print("Nothing to write to the HDF5 file")
        else:
            # Prepare file
            self.OpenFile(filename)
            self.WriteHeader()
            self.WriteConcHeader()

            #define datasets

            #time dataset
            time_dset = self.f[self.name + "/GEOLOCALIZATION"].require_dataset(
                "Timestamp", (self.nobs, ), dtype=self.h5_typetime)
            #define datatypes
            h5pcoltype = np.dtype('(1,)f')
            h5arrtype = np.dtype('(%i,)f' % (self.nmaxlevs))
            h5matrixtype = np.dtype('(%i,%i,)f' % (self.nmaxlevs, self.nmaxlevs))

            # Vertical dataset
            self.CheckVertCoo()
            
            if self.vertgridcnst == 1:
                dset = self.f[self.name+"/GEOLOCALIZATION"].create_dataset(
                    self.VertDsetName,
                    (1, ),
                    dtype=h5arrtype,
                    maxshape=(None, ))
            else:
                dset = self.f[self.name+"/GEOLOCALIZATION"].create_dataset(
                    self.VertDsetName,
                    (self.nobs, ),
                    dtype=h5arrtype,
                    maxshape=(None, ))

            # Vert coo attributes
            dset.attrs['Units'] = np.string_(self.vertunit)
            dset.attrs['Orientation'] = np.string_(self.vertorient)
            dset.attrs['MissingValue'] = self.missing_value
            dset.attrs['IsGridConstant'] = self.vertgridcnst
            self.attach_scale_2_dataset(dset, time_dset, 't', 'Timestamp')

            # IntQty dataset
            for spc in self.species:
                #partcol
                dset = self.f[self.name + "/OBSERVATIONS/" + spc].create_dataset('Y', (self.nobs,), dtype=h5pcoltype,
                                                                                 maxshape=(None,))
                dset.attrs['MaxNumLevels'] = self.nmaxlevs
                dset.attrs['Units'] = np.string_(self.units[spc])
                dset.attrs['MissingValue'] = self.missing_value
                self.attach_scale_2_dataset(dset, time_dset, 't', "Timestamp")

                #avk
                dset = self.f[self.name + "/OBSERVATIONS/" + spc].create_dataset('AveragingKernels', (self.nobs,), dtype=h5arrtype,
                                                                                 maxshape=(None,))
                dset.attrs['MaxNumLevels'] = self.nmaxlevs
                dset.attrs['Units'] = np.string_('unitless')
                dset.attrs['MissingValue'] = self.missing_value
                self.attach_scale_2_dataset(dset, time_dset, 't', "Timestamp")

                # Covariance dataset (optional)
                if self.hascov:
                    dset = self.f[self.name + "/OBSERVATIONS/" + spc].create_dataset(
                        'ErrorCovariance', (self.nobs,), dtype=h5matrixtype, maxshape=(None,))
                    dset.attrs['MaxNumLevels'] = self.nmaxlevs
                    dset.attrs['Units'] = np.string_(self.units[spc]+'-2')
                    dset.attrs['MissingValue'] = self.missing_value
                    self.attach_scale_2_dataset(dset, time_dset, 't', "Timestamp")

            # Write actual data in the file
            self.WriteGeoValues()
            
            if self.vertgridcnst == 1:
                dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset(self.VertDsetName,(1,),dtype=h5arrtype)
                dset[...] = self.vert.filled(fill_value=self.missing_value)
            else:
                dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset(self.VertDsetName,(self.nobs,),dtype=h5arrtype)
                dset[...] = np.array([val for val in self.vert[~self.mask,...].filled(fill_value=self.missing_value)],dtype=self.vert.dtype)[self.sorted_idx]

            for spc in self.species:
                #partcol
                dset = self.f[self.name + "/OBSERVATIONS/" + spc].require_dataset("Y", (self.nobs,), dtype=h5pcoltype)
                dset[...] = np.array([val for val in self.pcol[spc][~self.mask][:,np.newaxis]], dtype=self.pcol[spc].dtype)[
                    self.sorted_idx]
                #avk
                dset = self.f[self.name + "/OBSERVATIONS/" + spc].require_dataset("AveragingKernels", (self.nobs,), dtype=h5arrtype)
                dset[...] = np.array([val for val in self.avk[spc][~self.mask, ...].filled(fill_value=self.missing_value)], dtype=self.avk[spc].dtype)[
                    self.sorted_idx]
                # Covariance dataset (optional)
                if self.hascov:
                    dset = self.f[self.name + "/OBSERVATIONS/" + spc].require_dataset("ErrorCovariance", (self.nobs,),
                                                                                      dtype=h5matrixtype)
                    dset[...] = np.array([val for val in self.cov[spc][~self.mask, ...].filled(fill_value=self.missing_value)], dtype=self.cov[spc].dtype)[
                        self.sorted_idx]

            self.CloseFile()
            print(">>> HDF5 written ")

    def PlotObsHisto(self, outdir="./"):
        """ """
        fig = plt.figure()

        filename = os.path.join(outdir.decode('utf-8'), "%s_PColHisto_%s_%s" % \
                                (self.name, self.time.compressed().min().astype(str)[:13].replace("-", ""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-", "")))

        print('>>> Printing plot:', filename)

        for spc in self.species:
            ax = fig.add_subplot(111)
            ax.set_title("%s (%s)" % (spc, self.units[spc]))
            ax.set_xlabel("Avg:%4.2e Std:%4.2e N: %i" % \
                          (self.pcol[spc].mean() * self.cnvfctr[spc],
                           np.ma.std(self.pcol[spc]) * self.cnvfctr[spc], self.nobs))
            ax.hist(self.pcol[spc].compressed() * self.cnvfctr[spc], bins=35, normed=True)

        fig.savefig(filename)
        plt.close(fig)

    def PlotObsValues(self, outdir="./", figsize=(10, 5), outNameSuffix='', **kwargs):
        """ """
        fig = plt.figure(figsize=figsize, dpi=72)

        #add underscore if suffix to write
        if outNameSuffix != '': outNameSuffix = '_' + outNameSuffix

        markersize = 2
        colormap = "afmhot"
        if 'ObsOnMapOpt' in kwargs:
            if 'markersize' in kwargs["ObsOnMapOpt"]: markersize = kwargs["ObsOnMapOpt"]['markersize']
            if 'colormap' in kwargs["ObsOnMapOpt"]: colormap = kwargs["ObsOnMapOpt"]['colormap']

        for spc in self.species:

            ax = fig.add_subplot(111)

            if 'VertPlot' in kwargs: #Plot the altitude of columns
                filename = os.path.join(outdir.decode('utf-8'), "%s_%s_AltColMap_%s_%s%s" % \
                                    (self.name, spc, self.time.compressed().min().astype(str)[:13].replace("-", ""), \
                                     self.time.compressed().max().astype(str)[:13].replace("-", ""), \
                                     outNameSuffix))
            else:
                filename = os.path.join(outdir, "%s_%s_ConcMap_%s_%s%s" % \
                                    (self.name, spc, self.time.compressed().min().astype(str)[:13].replace("-", ""), \
                                     self.time.compressed().max().astype(str)[:13].replace("-", ""), \
                                     outNameSuffix))

            if 'VertPlot' in kwargs: #Plot the altitude of columns
                vmin = self.vert[:,1].compressed().min()
                vmax = self.vert[:,1].compressed().max()
            else:
                vmin = self.pcol[spc].compressed().min() * self.cnvfctr[spc]
                vmax = self.pcol[spc].compressed().max() * self.cnvfctr[spc]

            # Select the range based on kwargs
            if "range" in kwargs["ObsOnMapOpt"]:
                if kwargs["ObsOnMapOpt"]["range"]["type"] == "simple":
                    vmin = kwargs["ObsOnMapOpt"]["range"]["min"]
                    vmax = kwargs["ObsOnMapOpt"]["range"]["max"]

            print('>>> Printing plot:', filename)

            m = self.SetBaseMap(ax, **kwargs)

            if 'VertPlot' in kwargs: #Plot the altitude of columns
                sc = m.scatter(self.lons.compressed(), self.lats.compressed(), markersize, marker='o', \
                        c=self.vert[:,1].compressed(), cmap=colormap, \
                           latlon=True, zorder=10, vmin=vmin, vmax=vmax)
            else:
                sc = m.scatter(self.lons.compressed(), self.lats.compressed(), markersize, marker='o', \
                           c=self.pcol[spc].compressed() * self.cnvfctr[spc], cmap=colormap, \
                           latlon=True, zorder=10, vmin=vmin, vmax=vmax)

            plt.colorbar(sc, orientation='horizontal')

            if 'VertPlot' in kwargs: #Plot the altitude of columns
                ax.set_title('%s %s N=%i \nAvg: %.1f Min: %.1f Max: %.1f (%s)' % ("Altitude", spc, self.nobs, \
                                                                              self.vert[:,1].mean(),self.vert[:,1].min(),\
                                                                              self.vert[:,1].max(),self.vertunit))

            else:
                ax.set_title('%s %s N=%i \nAvg: %.1f Min: %.1f Max: %.1f (%s)' % (self.name, spc, self.nobs, \
                                                                              self.pcol[spc].mean() * self.cnvfctr[spc], \
                                                                              self.pcol[spc].min() * self.cnvfctr[spc],
                                                                              self.pcol[spc].max() * self.cnvfctr[spc], \
                                                                              self.units[spc]))

            fig.savefig(filename)
            plt.close(fig)

    def PlotTimeSeries(self, outdir="./", **kwargs):
        """ """
        self.Average()

        fig = plt.figure(figsize=(10, 5), dpi=200)

        filename = os.path.join(outdir.decode('utf-8'), "%s_ConcTimeSeries_%s_%s" % \
                                (self.name, self.avgtime.min().astype(str)[:13].replace("-", ""),
                                 self.avgtime.max().astype(str)[:13].replace("-", "")))

        print('>>> Printing plot:', filename)

        for spc in self.species:
            ax = fig.add_subplot(111)

            ax = self.SetTimePlot(fig, **kwargs)

            ax.set_title("%s (%s)" % (spc, self.units[spc]))
            ax.set_ylabel('%s' % (self.units[spc]))

            ax.plot(self.avgtime.astype(datetime), self.avgpcol[spc] * self.cnvfctr[spc], color='royalblue',
                    ls='-', lw=2, marker=None, label=None, markersize=4.0)

        fig.savefig(filename)
        plt.close(fig)

    def Average(self, **kwargs):
        """ """
        avgmsk = self.TimeAveragerMask()

        self.avgpcol = {spc: np.ma.empty((self.avgtime.size), dtype=self.pcol[spc].dtype) \
                        for spc in self.species}
        for spc in self.species:
            self.avgpcol[spc][:] = np.ma.masked
            for it in range(self.avgtime.size):
                if avgmsk[it].any(): self.avgpcol[spc][it] = np.ma.mean(self.pcol[spc][avgmsk[it]])

    def CheckVertCoo(self):
        """ """
        if self.vertunit in ("m", "meter", "meters", "metre", "metres"):
            self.vertunit = "m"
            self.VertDsetName = "Altitude"
        elif self.vertunit in ("km", "kilometer", "kilometers", "kilometre", "kilometres"):
            self.vert *= 1000 #km to m
            self.vertunit = "m"
            self.VertDsetName = "Altitude"
        elif self.vertunit in ("Pa", "Pascal", "Pascals"):
            self.vertunit = "Pa"
            self.VertDsetName = "Pressure"
        elif self.vertunit in ("hPa", "hectoPascal", "hectoPascals", "mbar", "milibar", "milibars"):
            self.vert *= 100 #hPa to Pa
            self.vertunit = "Pa"
            self.VertDsetName = "Pressure"
        else:
            raise ValueError("The vertical unit should be in meters or Pascals.!")

        # check orientation
        #take only first row
        vrow = np.trim_zeros(self.vert[0,:], 'b') #remove trailing zeros if any
        if np.all(vrow[1:] > vrow[:-1]):
            if self.vertunit == "m":
                self.vertorient = "up"
            elif self.vertunit == "Pa":
                self.vertorient = "down"
        elif np.all(vrow[1:] < vrow[:-1]):
            if self.vertunit == "m":
                self.vertorient = "down"
            elif self.vertunit == "Pa":
                self.vertorient = "up"
        else:
            raise ValueError("Vertical coordinate is not monotonically increasing or decreasing.")
        #if np.all(vrow[1:,:] >= vrow[:-1,:], axis=2) #whole array
        print("CHECK OK")

    def Plot(self, outdir="./", **kwargs):
        """ A comprehensive plotter for pcol data. """
        sns.set() # Apply seaborn beautificator
        if self.nobs > 0:
            if 'makeplots' in kwargs:
                if 'ObsLoc' in kwargs['makeplots']: self.PlotObsLocation(outdir=outdir, **kwargs)
                if 'ObsHisto' in kwargs['makeplots']: self.PlotObsHisto(outdir=outdir)
                if 'ObsOnMap' in kwargs['makeplots']: self.PlotObsValues(outdir=outdir, **kwargs)
                if 'TimeSeries' in kwargs['makeplots']: self.PlotTimeSeries(outdir=outdir,**kwargs)
            else:
                self.PlotObsLocation(outdir=outdir, **kwargs)
