# coding: utf8
"""
Generic instrument class

Created on Mon Feb  6 10:38:56 2017

@author: emili
"""
from __future__ import print_function
import os
import numpy as np
import h5py
import time
from datetime import datetime, timedelta
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, FormatStrFormatter
from mpl_toolkits.basemap import Basemap
try:
    import seaborn as sns
except ImportError:
    pass
try:
    from netCDF4 import Dataset # needed for netcdf storage
except ImportError:
    pass
try:
	from skyfield import api, almanac # For day-night computation
except ImportError:
    pass
# from geopy.distance import great_circle


class BaseInstrument(object):
    """ Generic instrument class.

    Parameters
    ----------
    name : str
        Name of the instrument.

    """
    np_typetime = "datetime64[s]"
    h5_typetime = 'int32'
    time_orig = np.datetime64("1970-01-01T00:00:00")
    missing_value = float(-1.0e38)
    ra = 6.3781*1e6

    def __init__(self, name, **kwargs):
        self.name = name
        self.nobs = 0  # The number of obs for the current instrument
        # (must always correspond to (~self.mask).sum())
        self.lats = np.ma.empty((0), dtype=np.float32)
        self.lons = np.ma.empty((0), dtype=np.float32)
        self.time = np.ma.empty((0), dtype=self.np_typetime)
        self.mask = np.empty((0), dtype=bool)  # The base observation mask,
        # useful as standalone array because np masks can be one-sized
        # or instrument dependent
        self.f = None  # HDF5 or NETCDF file pointer

    def WriteHeader(self):
        """ Prepare the HDF5 instrument and geoloc group. """
        grp = self.f.create_group(self.name)
        # Instrument header
        grp.attrs['InstrumentLongName'] = np.string_("some_long_name")
        # Measurement type
        grp.attrs['MeasurementType'] = np.string_(self.daimon_type) #np.string_(type(self).__name__)
        # Measurement product and version
        grp.attrs['ProductName'] = np.string_("some_product_info")
        # Date of creation
        grp.attrs['CreationDate'] = np.string_(time.strftime("%c"))
        # Version of the code used to create the hdf5 file
        grp.attrs['CreationCode'] = np.string_('mocageobs ver. 0.1')
        
        # Geoloc axes
        # Time dataset (it is a scale dimension, unlimited)
        subgrp = grp.create_group("GEOLOCALIZATION")
        time_dset = subgrp.create_dataset("Timestamp",(self.nobs,),dtype=self.h5_typetime,maxshape=(None))
        time_dset.attrs['Units'] = np.string_('seconds UTC since 1970 01 01')    
        # Latitudes dataset
        dset = subgrp.create_dataset("Latitude",(self.nobs,),dtype=self.lats.dtype,maxshape=(None))
        self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        dset.attrs['Units'] = np.string_('Degrees North')
        # Longitudes dataset
        dset = subgrp.create_dataset("Longitude",(self.nobs,),dtype=self.lons.dtype,maxshape=(None))
        self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        dset.attrs['Units'] = np.string_('Degrees East')
        # Time human readable
        dset = subgrp.create_dataset("TimeHuman",(self.nobs,),dtype='S19',maxshape=(None))
        self.attach_scale_2_dataset(dset,time_dset,'t',"TimeHuman")
        dset.attrs['Units'] = np.string_('date UTC in ISO 8601 format, YYYY-MM-DDTHH:MM:SS')

    def WriteGeoValues(self):
        """ Compute and store an ordering index.

        Compute and store an ordering index to write obs in strictly
        increasing time order, followed by lats and lons ordering.

        """
        self.sorted_idx = np.lexsort(
            (self.lons.compressed(),
             self.lats.compressed(),
             (self.time-self.time_orig).compressed()))
        grp = self.f[self.name+"/GEOLOCALIZATION"]
        dset = grp.require_dataset("Timestamp", (self.nobs, ),
                                   dtype=self.h5_typetime)
        dset[...] = (self.time-self.time_orig).compressed().astype(
            self.h5_typetime)[self.sorted_idx]
        dset = grp.require_dataset("TimeHuman", (self.nobs, ), dtype='S19')
        dset[...] = self.time.compressed().astype(
            np.dtype("S"))[self.sorted_idx]
            # np.dtype("S19"))[self.sorted_idx]
        dset = grp.require_dataset("Latitude", (self.nobs, ),
                                   dtype=self.lats.dtype)
        dset[...] = self.lats.compressed()[self.sorted_idx]
        dset = grp.require_dataset("Longitude", (self.nobs, ),
                                   dtype=self.lons.dtype)
        dset[...] = self.lons.compressed()[self.sorted_idx]

    def ReadGeoValues(self, **kwargs):
        """ """
        subgrp = ""

        if 'ReadAssim' in kwargs.keys():
            if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']

        print('Opening ',self.name+subgrp)
        grp = self.f[self.name+subgrp+"/GEOLOCALIZATION"]
        if "Timestamp" in grp:
            nobs = grp["Timestamp"].size
            # Implement the selection in time for the reader
            self.nobs += nobs # Increase the instrument counter
            self.time = np.ma.concatenate(
                (self.time,
                 np.ma.array(
                    # self.time_orig+grp["Timestamp"],
                    self.time_orig+grp["Timestamp"][:],
                    dtype=self.np_typetime)))
            self.lats = np.ma.concatenate(
                (self.lats,
                 # np.ma.array(grp["Longitude"],
                 np.ma.array(grp["Latitude"][:],
                             dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate(
                (self.lons,
                 # np.ma.array(grp["Longitude"],
                 np.ma.array(grp["Longitude"][:],
                             dtype=self.lons.dtype)))
            self.mask = np.concatenate(
            (self.mask,np.zeros((nobs),dtype=bool))) # By default all read obs are accepted
            return nobs
        else:
            return 0

    def ReadNCGeoValues(self, **kwargs):
        """ """

        grp = self.f[self.name+"/GEOLOCALIZATION"]

        if "Timestamp" in grp.variables:
            nobs = grp["Timestamp"].size
            # Implement the selection in time for the reader
            self.nobs += nobs # Increase the instrument counter
            self.time = np.ma.concatenate(
                (self.time,
                 np.ma.array(
                    # self.time_orig+grp["Timestamp"],
                    self.time_orig+grp["Timestamp"][:],
                    dtype=self.np_typetime)))
            self.lats = np.ma.concatenate(
                (self.lats,
                 # np.ma.array(grp["Longitude"],
                 np.ma.array(grp["Latitude"][:],
                             dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate(
                (self.lons,
                 # np.ma.array(grp["Longitude"],
                 np.ma.array(grp["Longitude"][:],
                             dtype=self.lons.dtype)))
            self.mask = np.concatenate(
            (self.mask,np.zeros((nobs),dtype=bool))) # By default all read obs are accepted
            return nobs
        else:
            return 0

    def OpenFile(self, filename, mode='a', **kwargs):
        """ Open a HDF5/NETCDF4

        The opened file in stored in the attribute self.f.

        Parameters
        ----------
        filename : str
            Path to the input h5 file.
        mode : str
            Opening mode passed to h5py.File function.

        """
        ftype = 'h5'
        if 'ftype' in kwargs: ftype = kwargs['ftype']
        if filename[-3:] != '.'+ftype:
            filename = filename+'.'+ftype
        
        if ftype == 'h5':    
            self.f = h5py.File(filename, mode)
        elif ftype == 'nc': 
            self.f = Dataset(filename, mode, format="NETCDF4")
        else:
            print('File type not supported')    

    def CloseFile(self):
        """ Close the HDF5 file stored in self.f. """
        self.f.close()
        self.f = None  # Avoids messing up when performing deepcopy of the
        # obs instances

    def isColocatedWith(self, other, consider_mask=False):
        """ Checks wether two instances of the obs are colocated. """
        if consider_mask:
            res = (
                self.nobs > 0 and
                self.nobs == other.nobs and
                (self.lats[~self.mask] == other.lats[~other.mask]).all() and
                (self.lons[~self.mask] == other.lons[~other.mask]).all() and
                (self.time[~self.mask] == other.time[~other.mask]).all()
            )
        else:
            res = (
                self.nobs > 0 and
                self.nobs == other.nobs and
                (self.lats == other.lats).all() and
                (self.lons == other.lons).all() and
                (self.time == other.time).all()
            )
        return res

    def ColocateWith(self, other, **kwargs):
        """ Colocate in space and time two observational datasets. """
        # The first dataset (self) is the base one. It can only loose data based on "other".
        # Default colocation is very strict. Can be tuned depending on the application
        dmaxtime = 5 # 90                    # Maximum distance in seconds (commented values were used for IASI paper!)
        dmaxdist = 1e3 # 10e3                  # Maximum distance in meters (commented values were used for IASI paper!)
        # If exactdistcalc=True uses great circle computations instead of simple euclidean distance
        # on the sphere (veeery slow, use only with few pixels)
        exactdistcalc = False
        # If bigmem = True it avoids the slow python loop on pixels using a distance matrix. Can saturate 
        # system memory for large observational datasets (>10000 obs). Use with caution.
        # Example: usage on 200000 and 100000 nobs datasets consumes up to 250 Gb RAM. Speedup wrt standard method o(10) 
        bigmem, storematrix, usematrix = False, False, False
        if "Accelerate" in kwargs: 
            bigmem = kwargs["Accelerate"]
            if "StoreMatrix" in kwargs: storematrix = kwargs["StoreMatrix"]
            if "UseMatrix" in kwargs: usematrix = kwargs["UseMatrix"]

        print(">>> Computing colocation...")
        #np.seterr(all='raise')
        if bigmem:
            if not usematrix:
                mask = np.zeros((self.time.size,other.time.size),dtype=bool)
                print("Computing lat mask...")
                tmpmask = np.abs(np.ma.subtract.outer(self.lats,other.lats))*(np.pi/180*self.ra) > dmaxdist
                if tmpmask.any(): mask = mask | tmpmask
                print("Computing lon mask...")
                tmpmask = np.abs(np.ma.subtract.outer(self.lons,other.lons))*(np.pi/180*self.ra) > dmaxdist
                if tmpmask.any(): mask = mask | tmpmask
                print("Computing time mask...")
                tmpmask = np.abs(np.ma.subtract.outer(self.time,other.time).astype(np.float32)) > dmaxtime
                if tmpmask.any(): mask = mask | tmpmask
            else:
                # The other dataset provides the mask
                mask = other.colocmask.T
                
            if storematrix: self.colocmask = mask
            
            print("Masking...")
            # Update the previous mask
            self.mask = self.mask | ((~mask).sum(axis=1) == 0)
        else:
            # Loop only on unmasked elements of self and search possible neighbors in other
            for i in (~self.mask).nonzero()[0]:
                mask = np.zeros((other.time.size),dtype=bool)
                tmpmask = np.abs((self.time[i]-other.time).astype(np.float64)) > dmaxtime
                if tmpmask.any(): mask = mask | tmpmask
#                if exactdistcalc:
#                    # Extremely slow great circle calculation
#                    dist = np.array([great_circle((self.lats[i],other.lats[j]),(self.lons[i],other.lons[j])).meters \
#                                     for j in (~other.mask).nonzero()[0]])
#                else:
                # Approximation considering euclidean space on the sphere (should work since I'm interested in small distances)
                tmpmask = np.abs(self.lats[i]-other.lats)*np.pi/180.*self.ra > dmaxdist
                if tmpmask.any(): mask = mask | tmpmask
                tmpmask = np.abs(self.lons[i]-other.lons)*np.pi/180.*self.ra > dmaxdist
                if tmpmask.any(): mask = mask | tmpmask
                # Update the previous mask
                self.mask[i] = self.mask[i] | ((~mask).sum() == 0)
                
        # Print info and update the geographical mask   
        nobs = (~self.mask).sum()
        if nobs < self.nobs:
            print(">> Obs colocation with %s " % (other.name))
            print(">> Retained %i obs over %i" % (nobs,self.nobs))

    def SelectGeo(self, **kwargs):
        """ Computes mask for space-time selection. """
        # Default options
        timebnd, latbnd, lonbnd, custombnd, daynight = None, None, None, None, None
        if 'TimeBnd' in kwargs.keys(): timebnd = kwargs['TimeBnd']
        if 'LatBnd' in kwargs.keys(): latbnd = kwargs['LatBnd']
        if 'LonBnd' in kwargs.keys(): lonbnd = kwargs['LonBnd']
        if 'DayNight' in kwargs.keys(): daynight = kwargs['DayNight']
        if 'CustomBnd' in kwargs.keys(): custombnd = kwargs['CustomBnd']
        
        if timebnd is not None:
            print("Selecting obs in time between ",timebnd[0]," and ",timebnd[1])
            timemsk = ~((self.time > timebnd[0]) & (self.time < timebnd[1]))
            if timemsk.any(): self.mask = self.mask | timemsk
        if latbnd is not None: 
            print("Selecting obs latitudes between ",latbnd[0]," and ",latbnd[1])
            latmsk = ~((self.lats > latbnd[0]) & (self.lats < latbnd[1]))
            if latmsk.any(): self.mask = self.mask | latmsk
        if lonbnd is not None:
            print("Selecting obs longitudes between ",lonbnd[0]," and ",lonbnd[1])
            lonmsk = ~((self.lons > lonbnd[0]) & (self.lons < lonbnd[1]))
            if lonmsk.any(): self.mask = self.mask | lonmsk
        if daynight is not None:
            print("Selecting obs only during ",daynight)
            daynightmsk = ComputeDayNight(self.lats,self.lons,self.time,daynight)
            if daynightmsk.any(): self.mask = self.mask | daynightmsk
        if custombnd is not None:
            print("Selecting obs with a geographical custom bounds")
            custommsk = ComputeCustomMask(self.lats,self.lons,custombnd)
            if custommsk.any(): self.mask = self.mask | custommsk
        
        if 'DataThinning' in kwargs: self.DataThinning(**kwargs)
        
        if 'Colocation' in kwargs:
            clkwargs = kwargs['Colocation']
            for other in clkwargs["DataList"]:
                self.ColocateWith(other,**clkwargs)

        if self.mask.any():
            print(">>> %i observations selected over %i" % ((~self.mask).sum(),self.nobs))
            self.time.mask = self.mask
            self.lats.mask = self.mask
            self.lons.mask = self.mask
            self.nobs = (~self.mask).sum()
            
    def ResetSelGeo(self):
        """ Reset the selection masks. """
        self.mask = np.zeros((self.time.size),dtype=bool)
        self.time.mask = self.mask
        self.lats.mask = self.mask
        self.lons.mask = self.mask
        # self.assim_mask.mask = self.mask
        self.nobs = (~self.mask).sum()

    def DataThinning(self, **kwargs):
        """ Reduce the number of obs. """
        # Define a (coarse) grid for data thinning
        # Default grid option
        nx, ny = 10, 10
        latmin, latmax = self.lats.min(), self.lats.max()
        lonmin, lonmax = self.lons.min(), self.lons.max()
        regfactor = 2 # Put 2 here instead of 1 if you want to avoid neighbors. But works well only with very coarse thinning and many obs
        
        if "Bounds" in kwargs["DataThinning"]:
           latmin = kwargs["DataThinning"]["Bounds"]["LatMin"]
           latmax = kwargs["DataThinning"]["Bounds"]["LatMax"]
           lonmin = kwargs["DataThinning"]["Bounds"]["LonMin"]
           lonmax = kwargs["DataThinning"]["Bounds"]["LonMax"]
            
        latrange = latmax - latmin
        lonrange = lonmax - lonmin
        
        if "Resolution" in kwargs["DataThinning"]:
            nx = int(lonrange/kwargs["DataThinning"]["Resolution"]["dlon"])
            ny = int(latrange/kwargs["DataThinning"]["Resolution"]["dlat"])
            
        if nx < 2 or ny < 2: 
            print("Warning: Skipping data thinning. The grid is degenerated. Use another grid setting")
            return
            
        clats, dlats = np.linspace(latmin,latmax,num=ny,endpoint=False,retstep=True)
        clons, dlons = np.linspace(lonmin,lonmax,num=nx,endpoint=False,retstep=True)
        cgrid = np.meshgrid(clons,clats)
        clatgrid = cgrid[1].flatten()
        clongrid = cgrid[0].flatten()
        
        timemsk = self.TimeAveragerMask()
        thinmask = np.ones((self.time.size),dtype=bool)
        
        # Perform data thinning hourly
        for it,t in enumerate(self.avgtime):
            if self.avgnobs[it] > 0:
                print("Data thinning loop at %s" % (t.astype(datetime).isoformat()))
                # Loop on the grid and keep only first obs located in the pixel
                for lat,lon in zip(clatgrid,clongrid): 
                    latmsk = np.ma.masked_outside(self.lats,lat,lat+dlats/regfactor,copy=True).mask
                    lonmsk = np.ma.masked_outside(self.lons,lon,lon+dlons/regfactor,copy=True).mask
                    latlonmsk = np.ma.mask_or(latmsk,lonmsk)
                    latlonmsk = np.ma.mask_or(latlonmsk,timemsk[it])
                    if (~latlonmsk).any(): thinmask[(~latlonmsk).nonzero()[0][0]] = False
                
        print(">>> Performed data thinning:")
        print(">>> %i observations selected over %i" % ((~thinmask).sum(),self.nobs)  )
            
        self.mask = np.ma.mask_or(self.mask,thinmask)
        
    def TimeAveragerMask(self,**kwargs):
        """
        An averager to compute timeseries (spatial averages)
        It returns the selector to the caller
        """
        freq = 1
        if 'AvgTimeSerFreq' in kwargs: freq = kwargs['AvgTimeSerFreq']
        tstart = self.time.compressed().min().astype(datetime)
        tstart = datetime(tstart.year,tstart.month,tstart.day,tstart.hour)
        tend = self.time.compressed().max().astype(datetime)
        tend = datetime(tend.year,tend.month,tend.day,tend.hour,59,59)
        # Fixed to hourly frequency for now, make it changeable afterwards
        avgtimestep = timedelta(hours=freq)
        self.avgtime = np.ma.array([tstart+avgtimestep/2+avgtimestep*i for i in \
                                      range(int((tend-tstart).total_seconds()/ \
                                                avgtimestep.total_seconds()+1))],dtype=self.np_typetime)
        self.avgnobs = np.ma.zeros(self.avgtime.size)
        timemsk = []
        for it in range(self.avgtime.size):
            timemsk.append((self.time > self.avgtime[it]-np.timedelta64(avgtimestep)/2) & \
                           (self.time <= self.avgtime[it]+np.timedelta64(avgtimestep)/2))
                           # (self.time <= self.avgtime[it]+np.timedelta64(avgtimestep)/2) & (self.mask == 0)) # Should consider mask?
            self.avgnobs[it] = timemsk[it].sum()

        return timemsk
    
    def LatLonAveragerMask(self,**kwargs):
        """
        An averager to compute spatial (lat,lon) statistics
        It returns the an obs selector to the caller
        """
        # Define a grid for the averager
        # Default grid option
        nx, ny = 20, 20
        latmin, latmax = self.lats.min(), self.lats.max()
        lonmin, lonmax = self.lons.min(), self.lons.max()
        
        if "Bounds" in kwargs:
           latmin = kwargs["Bounds"]["LatMin"]
           latmax = kwargs["Bounds"]["LatMax"]
           lonmin = kwargs["Bounds"]["LonMin"]
           lonmax = kwargs["Bounds"]["LonMax"]
            
        latrange = latmax - latmin
        lonrange = lonmax - lonmin
        
        if "Resolution" in kwargs:
            nx = int(lonrange/kwargs["Resolution"]["dlon"])
            ny = int(latrange/kwargs["Resolution"]["dlat"])
            
        if nx < 2 or ny < 2: 
            print("Warning: Skipping data density plot. The grid is degenerated. Use another grid setting")
            return
            
        # The latlongrid contains the borders of each grid point (needed later for basemap plots)
        clats, dlats = np.linspace(latmin,latmax,num=ny+1,endpoint=True,retstep=True)
        clons, dlons = np.linspace(lonmin,lonmax,num=nx+1,endpoint=True,retstep=True)
        self.latlongrid = np.meshgrid(clons,clats)
        
        # The computations are done for each grid cell, upper and leftmost borders are not computed twice
        clatgrid = self.latlongrid[1][:-1,:].flatten()
        clongrid = self.latlongrid[0][:,:-1].flatten()
        self.obsdensity = np.ma.zeros((ny,nx))
        latlonmsk = []
        
        print(">>> Computing obs density...")
        
        for ig,(lat,lon) in enumerate(zip(clatgrid,clongrid)):
            ilat, ilon = int((lat - latmin)/dlats), int((lon - lonmin)/dlons)
            latmsk = np.ma.masked_outside(self.lats,lat,lat+dlats,copy=True).mask
            lonmsk = np.ma.masked_outside(self.lons,lon,lon+dlons,copy=True).mask
            latlonmsk.append((~np.ma.mask_or(latmsk,lonmsk)).nonzero()[0])
            # latlonmsk.append(~(latmsk | lonmsk | self.mask).nonzero()[0]) # Should consider mask?
            self.obsdensity[ilat,ilon] = len(latlonmsk[ig])

        return latlonmsk

    def SpaceAveragerMask(self,**kwargs):
        """
        An averager to compute statistics (temporal averages) at original obs locations
        It returns the selector to the caller
        """

        latlonlist = []
        latlonidx = []
        for iobs in range(self.time.size):
        # for iobs in list((~self.mask).nonzero()[0]): # Should consider mask?
            if (self.lats[iobs],self.lons[iobs]) in latlonlist:
                idx = latlonlist.index((self.lats[iobs],self.lons[iobs]))
                latlonidx[idx].append(iobs)
            else:
                latlonlist.append((self.lats[iobs],self.lons[iobs]))
                latlonidx.append([iobs])

        return latlonlist, latlonidx

    def PrintGeoStats(self):
        
        print(">>> GEOLOCALIZATION infos:")
        print("Total obs = %i ,\nn = %i are not masked " % (len(self.time),self.nobs))
        print("TimeMin:%s TimeMax: %s" % (self.time.compressed().min(),self.time.compressed().max()))
        print("LatMin:%5.2f LatMax:%5.2f" % (self.lats.min(),self.lats.max()))
        print("LonMin:%5.2f LonMax:%5.2f" % (self.lons.min(),self.lons.max()))

    def PlotObsLocation(self,outdir="./",**kwargs):

        figsize = (15,10)
        marker = 'o'
        markersize = 2.5
        markercolor = 'black'

        if 'ObsLocationOpt' in kwargs:
            if "figsize" in kwargs["ObsLocationOpt"]: figsize = kwargs["ObsLocationOpt"]["figsize"]
            if 'marker' in kwargs['ObsLocationOpt']: marker = kwargs['ObsLocationOpt']['marker'] 
            if 'markersize' in kwargs['ObsLocationOpt']: markersize = kwargs['ObsLocationOpt']['markersize']
            if 'markercolor' in kwargs['ObsLocationOpt']: markercolor = kwargs['ObsLocationOpt']['markercolor']
            
        fig = plt.figure(figsize=figsize,dpi=200)
        
        print(">>> Plotting observation location")
        
        ax = fig.gca()
        
        m = self.SetBaseMap(ax,**kwargs)

        (x, y) = m(self.lons.compressed(), self.lats.compressed())

        m.scatter(x,y,markersize,\
                  c=markercolor,marker=marker,edgecolors='black',alpha=0.7, \
                  zorder=10)
        
        ax.set_title('%s obs location (N=%i)' % (self.name,self.nobs))
        fig.savefig(os.path.join(outdir,"%s_ObsLoc_%s_%s" % \
                            (self.name.replace('.',''),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
    def PlotNobsTSer(self,outdir="./",**kwargs):
        
        fig = plt.figure(figsize=(10,5),dpi=200)
        
        self.TimeAveragerMask(**kwargs)
        
        filename = os.path.join(outdir,"%s_NobsTimeSeries_%s_%s" % \
                (self.name.replace('.',''),self.avgtime.min().astype(str)[:13].replace("-",""), \
                 self.avgtime.max().astype(str)[:13].replace("-","")))
        
        print('>>> Printing plot:', filename)
        
        ax = fig.add_subplot(111)
        
        ax = self.SetTimePlot(fig,**kwargs)
        
        ax.set_title("%s" % ('Number of observed locations'))
        barwidth = 1./24. # More or less one hour width
        # barwidth = 1. # More or less one hour width
        ax.bar(self.avgtime.astype(datetime),self.avgnobs,barwidth,label=None)
    
        fig.savefig(filename)
        plt.close(fig)

    def PlotObsDensity(self,outdir="./",**kwargs):

        figsize = (10,6)
        colormap = "jet"
        lkwargs = {}

        if 'ObsDensityOpt' in kwargs: lkwargs = kwargs["ObsDensityOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"] 
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
            
        self.LatLonAveragerMask(**lkwargs)
            
        fig = plt.figure(figsize=figsize,dpi=200)
        
        print(">>> Plotting observation density")
        
        ax = fig.gca()
                            
        vmin = self.obsdensity.min()
        vmax = self.obsdensity.max()
        
        if "range" in lkwargs:
            vmin = lkwargs["range"]["min"]
            vmax = lkwargs["range"]["max"]
        
        m = self.SetBaseMap(ax,**kwargs)

        ax.set_title('%s obs density (N=%i)' % (self.name,self.nobs))

        fld = np.ma.masked_equal(self.obsdensity,0)
        
        mesh = m.pcolormesh(self.latlongrid[0],self.latlongrid[1],fld,\
                  cmap=colormap,vmin=vmin,vmax=vmax,
                  zorder=10,latlon=True)
        
        plt.colorbar(mesh,orientation='horizontal')
                
        fig.savefig(os.path.join(outdir,"%s_ObsDensity_%s_%s" % \
                            (self.name.replace('.',''),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
    def SetBaseMap(self,ax,**kwargs):
        """ """
        addgrid = True
        resol = 'l'
        dlat = 10.
        dlon = 20.

        if 'MapOpt' in kwargs:
            if "AddGrid" in kwargs["MapOpt"]: addgrid = kwargs["MapOpt"]["AddGrid"]
            if "Resolution" in kwargs["MapOpt"]: resol = kwargs["MapOpt"]["Resolution"]
            if "GridSpacing" in kwargs["MapOpt"]:
                dlat = kwargs["MapOpt"]["GridSpacing"]["dlat"]
                dlon = kwargs["MapOpt"]["GridSpacing"]["dlon"]

        if 'MapProjection' in kwargs:
            mapproj = kwargs["MapProjection"]["type"]
            latc, lonc, satdist = 0.0, 0.0, 1e6
            if "LatCenter" in kwargs["MapProjection"]: latc = kwargs["MapProjection"]["LatCenter"]
            if "LonCenter" in kwargs["MapProjection"]: lonc = kwargs["MapProjection"]["LonCenter"]
            if "SatDist" in kwargs["MapProjection"]: satdist = kwargs["MapProjection"]["SatDist"]*1e3
            m = Basemap(projection=mapproj,lat_0=latc,lon_0=lonc,satellite_height=satdist,resolution=resol)
        else:
            m = Basemap(projection='cyl',resolution=resol)

        m.drawmapboundary()
        m.fillcontinents()

        if 'MapOpt' in kwargs:
            if 'coastlines' in kwargs["MapOpt"]:
                m.drawcoastlines(**kwargs["MapOpt"]['coastlines'])

        if addgrid: 
            m.drawparallels(np.arange(-90.,90.,dlat))
            m.drawmeridians(np.arange(-180.,180.,dlon))

        if 'PlotRegion' in kwargs and (m.projection not in ('nsper','ortho','geos','robin')):
            # your extent in lat/lon (dec degrees)
            ulx = kwargs["PlotRegion"]["LonMin"]
            uly = kwargs["PlotRegion"]["LatMax"]
            lrx = kwargs["PlotRegion"]["LonMax"]
            lry = kwargs["PlotRegion"]["LatMin"]
            # transform coordinates to map projection
            xmin, ymin = m(ulx, lry)
            xmax, ymax = m(lrx, uly)
            ax.set_xlim(xmin, xmax)
            ax.set_ylim(ymin, ymax)

        return m

    def SetTimePlot(self,fig,**kwargs):
        
        fig.autofmt_xdate() # Need fig to set this
        
        ax = fig.gca()
        
        if 'TimeSeries' in kwargs:
            if 'yrange' in kwargs['TimeSeries']:
                majorLocator   = MultipleLocator(int((kwargs['TimeSeries']['yrange'][1]-kwargs['TimeSeries']['yrange'][0])/5))
                minorLocator   = MultipleLocator(int((kwargs['TimeSeries']['yrange'][1]-kwargs['TimeSeries']['yrange'][0])/10))
                ax.yaxis.set_major_locator(majorLocator)
                ax.yaxis.set_minor_locator(minorLocator)
    
            if 'tickformat' in kwargs['TimeSeries']:
                majorFormatter = FormatStrFormatter(kwargs['TimeSeries']['tickformat'])
                ax.yaxis.set_major_formatter(majorFormatter)
                
            if 'yrange' in kwargs['TimeSeries']: ax.set_ylim(kwargs['TimeSeries']['yrange'][0],kwargs['TimeSeries']['yrange'][1])   
          
            if 'cycle' in kwargs['TimeSeries']:
                if 'xrange' in kwargs['TimeSeries']: ax.set_xlim(kwargs['TimeSeries']['xrange'][0],kwargs['TimeSeries']['xrange'][1])
            else:
                ax.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%d %b, %H'))

        ax.grid(True)   
        ax.set_xlabel('Time (UTC)')
        ax.legend(loc=2)

        return ax

    def Regrid(self):
        pass

    def attach_scale_2_dataset(self, dataset, scale_dimension_dset,
                               scale_label, scale_name, dim=0):
        """ Attach a scale dimension dataset to another dataset. """
        dataset.dims[dim].label = scale_label
        scale_dimension_dset.make_scale(scale_name)
        dataset.dims[dim].attach_scale(scale_dimension_dset)
        
def ComputeDayNight(lats,lons,times,daynight='day'):
    
    ts = api.load.timescale()
    e = api.load('de421.bsp')
    
    daynightmsk = np.zeros((len(times)),dtype=bool)
    libout = {'Day':True,'Night':False}
    
    for i,(lat,lon,t) in enumerate(zip(lats,lons,times)):
        pos = api.Topos(latitude_degrees=lat,longitude_degrees=lon)
        f = almanac.sunrise_sunset(e, pos)
        daynightmsk[i] = \
            f(ts.utc(t.astype(datetime).replace(tzinfo=api.utc))) != libout[daynight]
        
    return daynightmsk

def ComputeCustomMask(lats,lons,custombnd):
    mask = False
    return mask
    
    