# -*- coding: utf-8 -*-
"""
Surface observation class

Created on Thu Feb  2 17:37:58 2017

@author: emili
"""
from __future__ import print_function, division
import os
import sys
import numpy as np
from baseinstr import BaseInstrument
from concinstr import ConcInstrument
import seaborn as sns
import matplotlib.pyplot as plt
import copy

class ColConcInstrument(ConcInstrument):
    
    def __init__(self,name,**kwargs):
        
        BaseInstrument.__init__(self,name,**kwargs)
        # Init with user defined species if given
        # The behavior is as follows:
        # - no species init by users -> all species found in h5 are read (units and convfctr reader to be done)
        # - species init by the user -> only specified ones are read
        if "species" in kwargs:
            ConcInstrument.__init__(self,name,species=kwargs["species"]) 
            if "nmaxlevs" in kwargs: 
                self.nmaxlevs = kwargs["nmaxlevs"]
                self.InitArrays()
            else:
                print("Error: you need to init the max number of vertical levels (nmaxlevs keyword)!")
                sys.exit()

        self.press_isconst = False # Wether the pressure axe is constant or variable for each measurement
        self.press_coorddir = 'downward' # The orientation of the vertical axis 

        if "ConstPressureGrid" in kwargs: self.press_isconst = bool(kwargs["ConstPressureGrid"])

        self.hascov = False # Switch to treat or not error covariances
        self.hasavk = False # Switch to treat or not averaging kernels
        
        self.daimon_type = "Column"
    
    # Some operator overloading to add/substract/divide batch of measurements
    # taken at the same time and place (including possibly averages)
    def __add__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for spc in self.species:
                result.col[spc] = self.col[spc] + other.col[spc]
                if 'avgcol' in self.__dict__:
                    # Need to recompute standard deviation after the sum.
                    # Consider making std optional and avoid calling the averager
                    result.Average()
                    #result.avgcol[spc] = self.avgcol[spc] + other.avgcol[spc]
        else:
            print('Cannot sum obs: mismatch in geolocalizations')
        
        return result
        
    def __sub__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for spc in self.species:
                result.col[spc] = self.col[spc] - other.col[spc]
                if 'avgcol' in self.__dict__:
                    # Need to recompute standard deviation for the difference.
                    # Consider making std optional and avoid calling the averager
                    result.Average()
                    #result.avgcol[spc] = self.avgcol[spc] - other.avgcol[spc]
        else:
            print('Cannot diff obs: mismatch in geolocalizations')
        
        return result
    
    def __truediv__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for spc in self.species:
                result.col[spc] = self.col[spc] / other.col[spc]
                if 'avgcol' in self.__dict__:
                    result.avgcol[spc] = self.avgcol[spc] / other.avgcol[spc]
                    result.stdcol[spc] = self.stdcol[spc] / other.avgcol[spc]
                result.units[spc] = "%"
                result.cnvfctr[spc] = 100
        else:
            print('Cannot divide obs: mismatch in geolocalizations')
        
        return result
    
    # Init the data arrays for the profile instrument
    def InitArrays(self):
        
        self.col = {spc : np.ma.empty((0,self.nmaxlevs),dtype=np.float32) \
                for spc in self.species}
        self.press = np.ma.empty((0,self.nmaxlevs+1),dtype=np.float32)
        self.cov = {spc : np.ma.empty((0,self.nmaxlevs,self.nmaxlevs),dtype=np.float32) \
                for spc in self.species}
        self.prior = {spc : np.ma.empty((0,self.nmaxlevs),dtype=np.float32) \
                for spc in self.species}
        # Consider distinct levels in and levels out for averaging kernels matrices
        self.avk = {spc : np.ma.empty((0,self.nmaxlevs,self.nmaxlevs),dtype=np.float32) \
                for spc in self.species}
        # The assimilation mask
        self.outmask = {spc : np.ma.empty((0,self.nmaxlevs),dtype=np.int32) \
                for spc in self.species}
        
        self.press.fill_value = self.missing_value
        for spc in self.species:
            self.col[spc].fill_value = self.missing_value
            self.prior[spc].fill_value = self.missing_value
            self.cov[spc].fill_value = self.missing_value
            self.avk[spc].fill_value = self.missing_value
            
        # Units and conversion factor for plots
        self.units = {spc: "DU" for spc in self.species}
        #if spc == 'Totash': # kept for memory, was it really working?
        #    self.units[spc]='kg/m2'
        self.cnvfctr = {spc: 1 for spc in self.species}
#        
#    def __mul__(self,other):

    def Read(self,filename,closefile=True,**kwargs):
        """
        Reads multispecies observations from MOCAGE HDF5
        
        If assim is True reads assimilation output as well
        """                                   
        self.OpenFile(filename,mode="r")
        nobs = self.ReadGeoValues(**kwargs)      
        
        if nobs > 0:
            
            subgrp = ""
            obstype = "Y"
            if 'ReadAssim' in kwargs:
                if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']
                if 'ObsType' in kwargs['ReadAssim']: obstype = kwargs['ReadAssim']['ObsType']
                self.hascov, self.hasavk = False, False # Set back to false if I'm reading assimilation output, avoids all unintentional avk and cov usage afterwards
            
            obsgrp = self.f[self.name+subgrp+"/OBSERVATIONS"]
            geogrp = self.f[self.name+subgrp+"/GEOLOCALIZATION"]

            # Init the species with those found in the HDF5 if not done before
            # Init the number of levels with hdf5 attribute
            if not hasattr(self,"nspc"):
                ConcInstrument.__init__(self,[el for el in obsgrp.keys()])
                self.nmaxlevs = geogrp["Pressure"].attrs['MaxNumLevels']-1
                self.InitArrays()
            if 'IsGridConstant' in geogrp["Pressure"].attrs:
                self.press_isconst = bool(geogrp["Pressure"].attrs['IsGridConstant']) # To do, attribute not yet in HSTAT. This works only for HDAT reader
            # Read the observations
            if self.press_isconst and self.press.size > 0 : 
                # Don't read constant pressure more than once
                pass
            else:
                # I don't know why but fill value needed to be forced back after concatenation (numpy bug?)
                press = np.ma.masked_values(np.ma.array(geogrp["Pressure"][:],dtype=self.press.dtype),self.missing_value,rtol=1e-03)
                self.press = np.ma.array(np.ma.concatenate((self.press,press)),fill_value=self.missing_value)
            
            for spc in self.species:
                print(">>> Reading %s, found %i obs" % (spc,len(obsgrp[spc+"/"+obstype]))) 
                prof = np.ma.masked_values(np.ma.array(obsgrp[spc+"/"+obstype][:],dtype=self.col[spc].dtype),self.missing_value,rtol=1e-03)
                # I don't know why but fill value needed to be forced back after concatenation (numpy bug)
                self.col[spc] = np.ma.array(np.ma.concatenate((self.col[spc],prof)),fill_value=self.missing_value)
                if 'ReadAssim' in kwargs: 
                    # Assimilation mask from daimon output
                    if obstype in ["Hx_ana","Hx_bkg"]:
                        self.outmask[spc] = np.ma.concatenate((self.outmask[spc],np.ma.array(obsgrp[spc+"/Mask_"+obstype[-3:]][:],dtype=self.outmask[spc].dtype)))                     
                if self.hascov:
                    cov =  np.ma.masked_values(np.ma.array(obsgrp[spc+"/ErrorCovariance"][:],\
                                                            dtype=self.cov[spc].dtype),self.missing_value,rtol=1e-03)
                    self.cov[spc] = np.ma.array(np.ma.concatenate((self.cov[spc],cov)),fill_value=self.missing_value)
                if self.hasavk: 
                    avk = np.ma.masked_values(np.ma.array(obsgrp[spc+"/AveragingKernels"][:],\
                                                            dtype=self.cov[spc].dtype),self.missing_value,rtol=1e-03)
                    self.avk[spc] = np.ma.array(np.ma.concatenate((self.avk[spc],avk)),fill_value=self.missing_value)
                    prof = np.ma.masked_values(np.ma.array(obsgrp[spc+"/Apriori"][:],dtype=self.prior[spc].dtype),self.missing_value,rtol=1e-03)
                    self.prior[spc] = np.ma.array(np.ma.concatenate((self.prior[spc],prof)),fill_value=self.missing_value)
                
        if closefile: self.CloseFile()
        
        # Consistency check for AVK array dimensions in hdf5 storage
#        for i in range(self.nobs):
#            self.col['O3'][i,:] = np.ma.dot(self.avk[spc][i,:,:].T,self.prior['O3'][i,:])
            
    def Write(self,filename,closefile=True,**kwargs):
        """
        Writes multispecies observations in MOCAGE HDF5
        """                                   
        # Prepare file
        self.OpenFile(filename)
        self.WriteHeader()
        self.WriteConcHeader()
        
        time_dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset("Timestamp",(self.nobs,),dtype=self.h5_typetime)
        
        h5presstype = np.dtype('(%i,)f' % (self.nmaxlevs+1))
        h5arrtype = np.dtype('(%i,)f' % (self.nmaxlevs))
        h5matrixtype = np.dtype('(%i,%i,)f' % (self.nmaxlevs,self.nmaxlevs))
        
        # Pressure dataset
        if self.press_isconst:
            dset = self.f[self.name+"/GEOLOCALIZATION"].create_dataset('Pressure',(1,),dtype=h5presstype,maxshape=(None,))
        else:
            dset = self.f[self.name+"/GEOLOCALIZATION"].create_dataset('Pressure',(self.nobs,),dtype=h5presstype,maxshape=(None,))
            
        # Measurement attributs
        dset.attrs['MaxNumLevels'] = self.nmaxlevs+1
        dset.attrs['Units'] = np.string_('Pa')
        dset.attrs['Orientation'] = np.string_(self.press_coorddir)
        dset.attrs['IsGridConstant'] = np.int32(self.press_isconst)
        dset.attrs['MissingValue'] = self.missing_value
        self.attach_scale_2_dataset(dset,time_dset,'t','Timestamp')
        
        # Profiles dataset
        for spc in self.species:
            dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset('Y',(self.nobs,),dtype=h5arrtype,maxshape=(None,))
            dset.attrs['MaxNumLevels'] = self.nmaxlevs
            dset.attrs['Units'] = np.string_(self.units[spc])
            dset.attrs['MissingValue'] = self.missing_value
            self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
            
            # Covariance dataset (optional)
            if self.hascov:
                dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset( \
                              'ErrorCovariance',(self.nobs,),dtype=h5matrixtype,maxshape=(None,))
                dset.attrs['MaxNumLevels'] = self.nmaxlevs
                dset.attrs['Units'] = np.string_('DU-2') # Verify
                dset.attrs['MissingValue'] = self.missing_value
                self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
            # Avk dataset (optional)
            if self.hasavk:
                dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset( \
                              'AveragingKernels',(self.nobs,),dtype=h5matrixtype,maxshape=(None,))
                dset.attrs['MaxNumLevels'] = self.nmaxlevs
                dset.attrs['Units'] = np.string_('Unitless')
                dset.attrs['MissingValue'] = self.missing_value
                self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
                
                dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset('Apriori',(self.nobs,),dtype=h5arrtype,maxshape=(None,))
                dset.attrs['MaxNumLevels'] = self.nmaxlevs
                dset.attrs['Units'] = np.string_(self.units[spc])
                dset.attrs['MissingValue'] = self.missing_value
                self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        
        if self.nobs > 0:
            # Write actual data in the file
            self.WriteGeoValues()
            
            # To debug
#            self.press = self.press[:,::-1]
#            self.col['O3'] = self.col['O3'][:,::-1]
            
            if self.press_isconst:
                dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset("Pressure",(1,),dtype=h5presstype)
                dset[...] = self.press.filled()
            else:
                dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset("Pressure",(self.nobs,),dtype=h5presstype)
                dset[...] = np.array([val for val in self.press[~self.mask,...].filled()],dtype=self.press.dtype)[self.sorted_idx]
            
            for spc in self.species:
                dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset("Y",(self.nobs,),dtype=h5arrtype)
                dset[...] = np.array([val for val in self.col[spc][~self.mask,...].filled()],dtype=self.col[spc].dtype)[self.sorted_idx]
                 # Covariance dataset (optional)
                if self.hascov:
                    dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset("ErrorCovariance",(self.nobs,),dtype=h5matrixtype)
                    dset[...] = np.array([val for val in self.cov[spc][~self.mask,...].filled()],dtype=self.cov[spc].dtype)[self.sorted_idx]
                # Avk dataset (optional)
                if self.hasavk:
                    dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset("AveragingKernels",(self.nobs,),dtype=h5matrixtype)
                    dset[...] = np.array([val for val in self.avk[spc][~self.mask,...].filled()],dtype=self.avk[spc].dtype)[self.sorted_idx]
                    dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset("Apriori",(self.nobs,),dtype=h5arrtype)
                    dset[...] = np.array([val for val in self.prior[spc][~self.mask,...].filled()],dtype=self.prior[spc].dtype)[self.sorted_idx]

        if closefile:
            self.CloseFile()
            print(">>> HDF5 written ")
        
    def Select(self,**kwargs):
        
        if self.nobs > 0:
            
            self.SelectGeo(**kwargs)
            
            if not self.press_isconst: self.press[self.mask,:] = np.ma.masked
            
            # Mask the entire profile for the geographically masked points
            for spc in self.species: 
                self.col[spc][self.mask,:] = np.ma.masked
                if self.hascov: self.cov[spc][self.mask,:,:] = np.ma.masked
                if self.hasavk: 
                    self.prior[spc][self.mask,:] = np.ma.masked
                    self.avk[spc][self.mask,:,:] = np.ma.masked
                
                # Selection of vertical levels based on user requested bounds
                if 'PressBnd' in kwargs:
                    pmax = kwargs['PressBnd']['max']
                    pmin = kwargs['PressBnd']['min']
                    if self.press_isconst:
                        print("!!! Warning: never been tested yet")
                        pmask = ((self.press[0,:] < pmin) | (self.press[0,:] > pmax))
                        colmsk = (pmask[1:] | pmask[:-1]) # Transpose interface mask to col mask
                        if pmask.any():
                            #self.press[0,pmask] = np.ma.masked #Keep the original pressure array to avoid some nasty behavior
                            self.col[spc][:,colmsk] = np.ma.masked
                            if self.hasavk: 
                                self.avk[spc][:,:,colmsk] = np.ma.masked
                                self.prior[spc][:,colmsk] = np.ma.masked
                            if self.hascov: 
                                self.cov[spc][:,colmsk,:] = np.ma.masked
                                self.cov[spc][:,:,colmsk] = np.ma.masked
                    else:
                        pmask = np.ma.masked_outside(self.press,pmin,pmax).mask
                        colmsk = (pmask[:,1:] | pmask[:,:-1])
                        self.col[spc].mask = colmsk
                        if self.hascov: 
                            covmsk = np.repeat(colmsk[:,:,None],self.nmaxlevs,axis=2)
                            self.cov[spc].mask = covmsk | np.transpose(covmsk,axes=(0,2,1)) # Trick for building square matrix mask
                        if self.hasavk:
                            self.prior[spc].mask = colmsk
                            for iz in range(self.nmaxlevs): self.avk[spc][:,iz,:].mask = colmsk 
                            # The following reduces the avk input levels and pressure grids as well
                            if 'AvkPressBnd' in kwargs:
                                pmax = kwargs['AvkPressBnd']['max']
                                pmin = kwargs['AvkPressBnd']['min']
                                self.press = np.ma.masked_outside(self.press,pmin,pmax)                                
                                col2msk = (self.press.mask[:,1:] | self.press.mask[:,:-1])                                
                                for iz in range(self.nmaxlevs): self.avk[spc][:,:,iz].mask = self.avk[spc][:,:,iz].mask | col2msk
                        
                self.nobs = (~self.mask).sum()
                
            # Vertical aggregation of columns from the original vertical grid to a coarser one
            if 'SumColumns' in kwargs:
                sfct = 4
                if 'NumOfLayers' in kwargs['SumColumns']: sfct = kwargs['SumColumns']['NumOfLayers']
                oldobs = copy.deepcopy(self) # Copy the original object and init a new one
                self.nmaxlevs = self.nmaxlevs / sfct 
                self.InitArrays()
                self.press = np.ma.concatenate((self.press,oldobs.press[:,::sfct]))
                for spc in self.species:
                    vsum = np.ma.cumsum(oldobs.col[spc],axis=1)
                    vsum[:,sfct:] = vsum[:,sfct:] - vsum[:,:-sfct]
                    self.col[spc] = np.ma.concatenate((self.col[spc],vsum[:,(sfct-1)::sfct]))
                    if self.hasavk:
                        vsum = np.ma.cumsum(oldobs.prior[spc],axis=1)
                        vsum[:,sfct:] = vsum[:,sfct:] - vsum[:,:-sfct]
                        self.prior[spc] = np.ma.concatenate((self.prior[spc],vsum[:,(sfct-1)::sfct]))
                        vsum = oldobs.avk[spc].cumsum(axis=2)
                        vsum[:,:,sfct:] = vsum[:,:,sfct:] - vsum[:,:,:-sfct]
                        self.avk[spc] = np.ma.concatenate((self.avk[spc],vsum[:,::sfct,(sfct-1)::sfct])) # Sampling should be replaced by some averaging???
                    if self.hascov:
                        vsum = np.ma.zeros((self.nobs,self.nmaxlevs,self.nmaxlevs),dtype=np.float32,fill_value=self.missing_value)
                        for h in range(self.nmaxlevs):
                            for k in range(self.nmaxlevs):
                                vsum[:,h,k] = oldobs.cov[spc][:,h*sfct:(h+1)*sfct,k*sfct:(k+1)*sfct].sum(axis=(1,2))
                                # Totally not sure about this!!!
                        self.cov[spc] = np.ma.concatenate((self.cov[spc],vsum))
                        
    def Average(self,AvgType='global',**kwargs):
        # Vertical obs averager
        # avgpress and avgprof are added to the data structure
        # to be used for further processing or plotting
        # (previously existent averages are erased on this call)
        
        # Default average pressure grid
        if self.press_isconst: 
            self.avgpress = np.ma.copy(self.press[0,:])
        else: 
            # Just average all pressures for the average grid
            self.avgpress = np.ma.mean(self.press,axis=0)
            
        # Compute time series of average profiles
        if AvgType == 'timeseries':
            
            avgmsk = self.TimeAveragerMask()
            
            self.avgcol = {spc : np.ma.empty((self.avgtime.size,self.avgpress.size),dtype=self.col[spc].dtype,\
                fill_value=self.missing_value) for spc in self.species}
            self.stdcol = {spc : np.ma.empty((self.avgtime.size,self.avgpress.size),dtype=self.col[spc].dtype,\
                fill_value=self.missing_value) for spc in self.species}
            self.avgnobs = {spc : np.zeros((self.avgtime.size,self.avgpress.size),dtype=np.int32) \
                for spc in self.species}
            if self.hascov: 
                self.avgcov = {spc : np.ma.empty((self.avgtime.size,self.avgpress.size),dtype=self.cov[spc].dtype,\
                fill_value=self.missing_value) for spc in self.species}
            if self.hasavk: 
                self.avgavk = {spc : np.ma.empty((self.avgtime.size,self.avgpress.size),dtype=self.avk[spc].dtype,\
                fill_value=self.missing_value) for spc in self.species}
            
            for spc in self.species:
                # Init
                self.avgcol[spc][:] = np.ma.masked
                self.stdcol[spc][:] = np.ma.masked
                if self.hascov: self.avgcov[spc][:] = np.ma.masked
                if self.hasavk: self.avgavk[spc][:] = np.ma.masked
                
                for it in range(self.avgtime.size):
                    if avgmsk[it].any(): 
                        self.avgnobs[spc][it,:] = self.col[spc][avgmsk[it],:].count(axis=0)
                        self.avgcol[spc][it,:] = np.ma.mean(self.col[spc][avgmsk[it],:],axis=0)
                        self.stdcol[spc][it,:] = np.ma.std(self.col[spc][avgmsk[it],:],axis=0)
                        if self.hascov: self.avgcov[spc][it,:] = np.ma.mean(self.cov[spc][avgmsk[it],:,:],axis=0)
                        if self.hasavk: self.avgavk[spc][it,:] = np.ma.mean(self.avk[spc][avgmsk[it],:,:],axis=0)
                            
        # Compute one global average profile
        elif AvgType == 'global':
            
            self.avgcol = {}
            self.stdcol = {}
            self.avgnobs = {}
            if self.hascov: self.avgcov = {}
            if self.hasavk: self.avgavk = {}
            
            for spc in self.species:
                self.avgnobs[spc] = self.col[spc].count(axis=0)
                self.avgcol[spc] = np.ma.mean(self.col[spc],axis=0)
                self.stdcol[spc] = np.ma.std(self.col[spc],axis=0)
                if self.hascov: self.avgcov[spc] = np.ma.mean(self.cov[spc],axis=0)
                if self.hasavk: self.avgavk[spc] = np.ma.mean(self.avk[spc],axis=0)

        else:
            print("Error: I do not understand this type of averaging")
            sys.exit()
                        
    def PrintObsStats(self,**kwargs):
        
        doseplevs = False
        if 'LevelsSelection' in kwargs: 
            levs = kwargs["LevelsSelection"]
            doseplevs = True
            
        print(">>> Profile obs infos:")
        print("Total obs = %i ,\nn = %i are not masked " % (len(self.time),self.nobs))

        for spc in self.species:
            if not doseplevs:
                print(spc+" Stats - ConcMin:%5.3e %s ConcMax:%5.3e %s" % \
                (self.col[spc].min()*self.cnvfctr[spc],self.units[spc],self.col[spc].max()*self.cnvfctr[spc],self.units[spc]))
            else:
                if self.press_isconst: 
                    press = np.ma.copy(self.press[0,:])
                else: 
                    press = np.ma.mean(self.press,axis=0)
                for ilev in levs:
                    print(spc+" Stats between %.1f-%.1f hPa - ConcMin:%5.3e %s ConcMax:%5.3e %s" % \
                    (press[ilev]/100,press[ilev+1]/100,self.col[spc][:,ilev].min()*self.cnvfctr[spc],self.units[spc],\
                     self.col[spc][:,ilev].max()*self.cnvfctr[spc],self.units[spc]))
                    
    def PlotObsHisto(self,outdir="./",**kwargs):
        
        figsize = (15,15)
        doseplevs = False
        lkwargs = {}
        
        if 'LevelsSelection' in kwargs: 
            levs = kwargs["LevelsSelection"]
            doseplevs = True
        if 'ObsHistoOpt' in kwargs: lkwargs = kwargs["ObsHistoOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        
        print(">>> Plotting profile level values histograms")
        
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_ConcHisto_%s_%s" % \
                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            fig = plt.figure(figsize=figsize,dpi=200)
            
            # Variable related keywords have priority over global keywords
            lvarkwargs = {}

            if spc in lkwargs: 
                lvarkwargs = lkwargs[spc]
            else:
                lvarkwargs = lkwargs
                    
            if "range" in lvarkwargs:
                vmin = lvarkwargs["range"]["min"]
                vmax = lvarkwargs["range"]["max"]
                historange =(vmin,vmax)
            else:
                historange = None
            
            if not doseplevs:
                ax = fig.add_subplot(1,1,1)
                ax.set_title("%s (%s) all levels" % (spc,self.units[spc]))
                ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                          (self.col[spc].mean()*self.cnvfctr[spc],np.ma.std(self.col[spc])*self.cnvfctr[spc]))
                ax.hist(self.col[spc].compressed()*self.cnvfctr[spc],bins=35,density=True,range=historange)
            else:
                nsubplt = int(np.ceil(np.sqrt(len(levs))))
                if self.press_isconst: 
                    press = np.ma.copy(self.press[0,:])
                else: 
                    press = np.ma.mean(self.press,axis=0)
                for iplt,ilev in enumerate(levs):
                    ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                    ax.set_title("%s (%s) at %.1f-%.1f hPa" % (spc,self.units[spc],press[ilev]/100,press[ilev+1]/100))
                    ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                          (self.col[spc][:,ilev].mean()*self.cnvfctr[spc],np.ma.std(self.col[spc][:,ilev])*self.cnvfctr[spc]))
                    ax.hist(self.col[spc][:,ilev].compressed()*self.cnvfctr[spc],bins=35,density=True,range=historange)
                    
            plt.savefig(filename)
            plt.close(fig)
            
    def PlotLevelsHisto(self,outdir="./",**kwargs):
        
        if self.press_isconst: return
        
        figsize = (25,25)
        doseplevs = False
        lkwargs = {}
        
        if 'LevelsSelection' in kwargs: 
            levs = kwargs["LevelsSelection"]
            doseplevs = True
        if 'LevelsHistoOpt' in kwargs: lkwargs = kwargs["LevelsHistoOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        
        print(">>> Plotting profile level values histograms")
            
        filename = os.path.join(outdir,"%s_LevelsHisto_%s_%s" % \
            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
             self.time.compressed().max().astype(str)[:13].replace("-","")))
        
        fig = plt.figure(figsize=figsize,dpi=200)
                
        if "range" in lkwargs:
            vmin = lkwargs["range"]["min"]
            vmax = lkwargs["range"]["max"]
            historange =(vmin,vmax)
        else:
            historange = None
        
        if self.press_coorddir == 'upward': dp = (self.press[:,:-1]-self.press[:,1:])/100
        if self.press_coorddir == 'downward': dp = (self.press[:,1:]-self.press[:,:-1])/100
        
        if not doseplevs:
            ax = fig.add_subplot(1,1,1)
            ax.set_title("dp (hPa) for all levels")
            ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                      (dp.mean(),np.ma.std(dp)))
            ax.hist(dp.compressed(),bins=35,density=True,range=historange)
        else:
            nsubplt = int(np.ceil(np.sqrt(len(levs))))
            
            for iplt,ilev in enumerate(levs):
                ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                ax.set_title("dp (hPa) for level %i" % (ilev))
                ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                      (dp[:,ilev].mean(),np.ma.std(dp[:,ilev])))
                ax.hist(dp[:,ilev].compressed(),bins=35,density=True,range=historange)
                
        plt.savefig(filename)
        plt.close(fig)
        
    def PlotObsValues(self,outdir="./",**kwargs):
        
        figsize = (15,15)
        markersize = 2
        colormap = "jet"
        lkwargs = {}
        
        levs = range(self.nmaxlevs)
        if 'LevelsSelection' in kwargs: levs = kwargs["LevelsSelection"]
            
        if 'ObsOnMapOpt' in kwargs: lkwargs = kwargs["ObsOnMapOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
        
        print(">>> Plotting profile level values on a map")
        
        if self.press_isconst: 
            press = np.ma.copy(self.press[0,:])
        else: 
            press = np.ma.mean(self.press,axis=0)
        
        for isp,spc in enumerate(self.species):

            fig = plt.figure(figsize=figsize,dpi=200)
            
            filename = os.path.join(outdir,"%s_%s_ConcMap_%s_%s" % \
                                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            nsubplt = int(np.ceil(np.sqrt(len(levs))))
            
            for iplt,ilev in enumerate(levs):
                
                ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                
                m = self.SetBaseMap(ax,**kwargs)

                fld = self.col[spc][:,ilev]*self.cnvfctr[spc]
                
                if fld.count() > 1:
                    
                    vmin = fld.min()
                    vmax = fld.max()
                    
                    # Variable related keywords have priority over global keywords
                    lvarkwargs = {}

                    if spc in lkwargs: 
                        lvarkwargs = lkwargs[spc]
                    else:
                        lvarkwargs = lkwargs
                    
                    if "range" in lvarkwargs:
                        vmin = lvarkwargs["range"]["min"]
                        vmax = lvarkwargs["range"]["max"]
                        
                    # Shrink to the common mask that considers missing vertical points    
                    lmask = self.mask | fld.mask
                    ax.set_title('%s N=%i %.1f-%.1f hPa\n Avg: %.2f Min: %.2f Max: %.2f (%s)' % (spc,(~lmask).sum(),\
                             press[ilev]/100,press[ilev+1]/100,\
                             fld.mean(),fld.min(),fld.max(),self.units[spc]))
                    
                    sc = m.scatter(self.lons[~lmask],self.lats[~lmask],markersize,marker='o',\
                                   c=fld[~lmask],cmap=colormap,latlon=True,zorder=10,vmin=vmin,vmax=vmax)
    
                    plt.colorbar(sc,orientation='horizontal')
                
            fig.savefig(filename)
            plt.close(fig)
            
    def PlotVertProfiles(self,outdir="./",**kwargs):
        
        figsize = (15,15)
        nmaxplts = 16
        lkwargs = {}
        if 'VertProfilesOpt' in kwargs:
            lkwargs = kwargs["VertProfilesOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        
        print(">>> Plotting vertical profiles")
        
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_ConcProfiles_%s_%s" % \
                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            fig = plt.figure(figsize=figsize,dpi=200)
            
            npltsxpage = min(self.nobs,nmaxplts)
            nsubplt = int(np.ceil(np.sqrt(npltsxpage)))
            
            for iplot,ip in enumerate((~self.mask).nonzero()[0]):
                
                if iplot > npltsxpage-1: break
            
                ax = fig.add_subplot(nsubplt,nsubplt,iplot+1)
                ax.set_title("%s profile \n (%.1fN,%.1fE,%s)" % (spc,self.lats[ip],self.lons[ip],self.time[ip].astype(str)[11:16]))
                ax.set_xlabel("%s" % (self.units[spc]))
                ax.set_ylabel("%s" % ("hPa"))
                c = 'black'
                lw = 2
                if 'lw' in lkwargs: lw = lkwargs['lw']
                
                if self.press_isconst: 
                    press = (self.press[0,:-1]+self.press[0,1:])/2
                else:
                    press = (self.press[ip,:-1]+self.press[ip,1:])/2
                    
                ax.plot(self.col[spc][ip,:]*self.cnvfctr[spc],press/100,c,lw=lw)
                ax.invert_yaxis()
                if press.max()/press.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
                if 'xlog' in lkwargs: ax.set_xscale('symlog',linthreshx=1e-3)
                if 'yrange' in lkwargs: 
                    ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
                if 'xrange' in lkwargs:
                    ax.set_xlim(lkwargs['xrange']["min"],lkwargs['xrange']["max"])
                #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
                #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
                #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
                ax.grid(True,which='minor',linestyle='--')
                ax.tick_params(which='minor',labelsize=7)
                
            plt.subplots_adjust(wspace=0.2,hspace=0.3)
            plt.savefig(filename)
            plt.close(fig)
            
    def PlotAvgVertProfiles(self,outdir="./",**kwargs):
        
        figsize = (16,8)
        lkwargs = {}
        basefilename = "ConcAvgProfile"
        savedata = False
        printdata = False
        if 'VertProfilesOpt' in kwargs:
            lkwargs = kwargs["VertProfilesOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if "filename" in lkwargs: basefilename = lkwargs["filename"]
        if "savedata" in lkwargs: savedata = lkwargs["savedata"]
        if "printdata" in lkwargs: printdata = lkwargs["printdata"]

        print(">>> Plotting average profile")
        
        if 'avgpress' not in self.__dict__: self.Average(**kwargs) 
        # Done to avoid recomputing averages. Impacts the behavior when diffing obs. 
        # Now diff(avg(obs)) prevails over avg(diff(obs)) unless averaging was not done before diffs
        
        # Use half layer pressure for plots
        press = (self.avgpress[:-1]+self.avgpress[1:])/2
        
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_%s_%s_%s" % \
                (self.name,spc,basefilename,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            if savedata:
                np.savez(filename,\
                avgprof=self.avgcol[spc].filled(),avgpress=press.filled(),\
                stdprof=self.stdcol[spc].filled(),avgnobs=self.avgnobs[spc],\
                species=spc,missing_value=self.missing_value)

            if printdata:
                np.savetxt(filename+".txt",np.array([press.filled()/100,self.avgcol[spc].filled()]).T,\
                           delimiter=' , ',fmt=('%5.2f','%.3e'))
                
            # Avoid plotting very unrepresentative data
            vmsk = self.avgnobs[spc] >= max(self.avgnobs[spc].max()/3,2)
            
            fig = plt.figure(figsize=figsize,dpi=200)

            ax = fig.add_subplot(131)
            ax.set_title("%s average profile \n" % (spc))
            ax.set_xlabel("%s" % (self.units[spc]))
            ax.set_ylabel("%s" % ("hPa"))
            c = 'black'
            lw = 2
            if 'lw' in lkwargs: lw = lkwargs['lw']
            ax.plot(self.avgcol[spc][vmsk]*self.cnvfctr[spc],press[vmsk]/100,c,lw=lw)
            ax.invert_yaxis()
            if press.max()/press.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
            if 'xlog' in lkwargs: 
                if lkwargs['xlog'] == 'symlog':
                    ax.set_xscale('symlog',linthreshx=1e-3) 
                else:
                    ax.set_xscale('log')
            if 'yrange' in lkwargs: 
                ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
            if 'xrange' in lkwargs:
                ax.set_xlim(lkwargs['xrange']["min"],lkwargs['xrange']["max"])
            #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
            #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
            #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
            ax.grid(True,which='minor',linestyle='--')
            ax.tick_params(which='minor',labelsize=7)
            
            ax = fig.add_subplot(132)
            ax.set_title("%s profile standard deviation \n" % (spc))
            ax.set_xlabel("%s" % (self.units[spc]))
            ax.set_ylabel("%s" % ("hPa"))
            c = 'black'
            lw = 2
            if 'lw' in lkwargs: lw = lkwargs['lw']
            ax.plot(self.stdcol[spc][vmsk]*self.cnvfctr[spc],press[vmsk]/100,c,lw=lw)
            ax.invert_yaxis()
            if press.max()/press.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
            if 'xlog' in lkwargs: ax.set_xscale('log')
            if 'yrange' in lkwargs: 
                ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
            if 'xrange' in lkwargs:
                ax.set_xlim(max(0,lkwargs['xrange']["min"]),lkwargs['xrange']["max"])
            #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
            #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
            #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
            ax.grid(True,which='minor',linestyle='--')
            ax.tick_params(which='minor',labelsize=7)
            
            ax = fig.add_subplot(133)
            ax.set_title("Number of %s observations \n" % (spc))
            ax.set_xlabel("N")
            ax.set_ylabel("%s" % ("hPa"))
            c = 'black'
            lw = 2
            ax.plot(self.avgnobs[spc],press/100,c,lw=lw)
            ax.invert_yaxis()
            if press.max()/press.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
            if 'yrange' in lkwargs: 
                ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
            ax.grid(True,which='minor',linestyle='--')
            ax.tick_params(which='minor',labelsize=7)
                
            plt.savefig(filename)
            plt.close(fig)            
            
    def PlotAvgAuxProfiles(self,outdir="./",**kwargs):
        
        figsize = (16,8)
        lkwargs = {}
        invert = True
        if 'AuxProfilesOpt' in kwargs:
            lkwargs = kwargs["AuxProfilesOpt"]
            if "figsize" in lkwargs: figsize = lkwargs["figsize"]
            
        print(">>> Plotting avk and error covariances average profiles")
            
        if 'avgpress' not in self.__dict__: self.Average(**kwargs)
        # Done to avoid recomputing averages. Impacts the behavior when diffing obs. 
        # Now diff(avg(obs)) prevails over avg(diff(obs)) unless averaging was not done before diffs
        
        # Use half layer pressure for plots
        press = (self.avgpress[:-1]+self.avgpress[1:])/2
        
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_AuxAvgProfile_%s_%s" % \
                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            fig = plt.figure(figsize=figsize,dpi=200)
            iplt = 1
            if self.hascov:
                if invert: 
                    basetitle = 'R'
                    # Here I reinvert the cov matrices, then average the result (the operation is not linear)
                    
                    # Simmetry check for invR
                    symmetric = np.array([np.ma.allclose(self.cov[spc][i,:,:],self.cov[spc][i,:,:].T,rtol=1e-3) for i in range(self.nobs)],dtype=bool)
                    
                    # Remove masked columns in case there are some (since linalg does not like masked arrays)
                    # Attention: if the col mask is not the same for all pixels this might not work
                    # So I mask all profiles with less levels than the average profile for the moment (need to do an expensive loop otherwise)
                    nlevs = self.avgcol[spc].count()
                    levmsk = self.avgcol[spc].mask
                    # This mask remains local in this plotting routine
                    pixmask = self.mask | np.ma.make_mask(self.col[spc].count(axis=1) < nlevs)
                    
                    print("%i over %i pixels have more masked levels than the average" % (pixmask.sum(),self.nobs))
                    print("> Neglecting them for plotting the average R matrix")
                    lcov = np.zeros(((~pixmask).sum(),nlevs,nlevs))
                    tmpcov = self.cov[spc][~pixmask,:,:]
                    # Copy values to a reduced rank matrix
                    for iz,izorig in zip(range(nlevs),(~levmsk).nonzero()[0]):
                        lcov[:,iz,:] = tmpcov[:,izorig,~levmsk]
                        lcov[:,:,iz] = tmpcov[:,~levmsk,izorig]
                        
                    cond = np.linalg.cond(lcov)
                    det = np.linalg.det(lcov)
                    print("R diagnostics:")
                    print("%i covariances over %i pixels are symmetric" % (symmetric.sum(),self.nobs))
                    print("Determinant (min,max,mean):" , det.min(), det.max(), det.mean())
                    print("Condition number (min,max,mean):" , cond.min(), cond.max(), cond.mean())

                    # Invert the reduced matrix
                    tmpcov = np.ma.mean(np.linalg.pinv(lcov,rcond=1e-5),axis=0)
                    
                    # Fill again the full masked matrix
                    cov2plt = np.ma.zeros((self.nmaxlevs,self.nmaxlevs))
                    cov2plt[levmsk,:] = np.ma.masked
                    cov2plt[:,levmsk] = np.ma.masked
                    for iz,izorig in zip(range(nlevs),(~levmsk).nonzero()[0]):
                        cov2plt[izorig,~levmsk] = tmpcov[iz,:]
                        cov2plt[~levmsk,izorig] = tmpcov[:,iz]
                else:
                    basetitle = 'R-1'
                    cov2plt = self.avgcov[spc]
                
                print('Some R diags',np.ma.diag(cov2plt).max(),np.ma.diag(cov2plt).min(),np.ma.diag(cov2plt).mean())
                
                ax = fig.add_subplot(1,3,iplt)
                ax.set_title("%s average error covariance (1/diag(%s)) \n" % (spc,basetitle))
                ax.set_xlabel("%s" % (self.units[spc]))
                ax.set_ylabel("%s" % ("hPa"))
                c = 'black'
                lw = 2
                if 'lw' in lkwargs: lw = lkwargs['lw']
                ax.plot(np.ma.diag(cov2plt)*self.cnvfctr[spc],\
                        press/100,c,lw=lw)
#                ax.plot(np.ma.sqrt(np.ma.diag(self.avgcov[spc]))*self.cnvfctr[spc],\
#                        press/100,c,lw=lw)
                ax.invert_yaxis()
                if press.max()/press.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
                if 'yrange' in lkwargs: 
                    ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
                if 'covrange' in lkwargs:
                    if 'xlog' in lkwargs['covrange']: ax.set_xscale('log')
                    ax.set_xlim(lkwargs['covrange']["min"],lkwargs['covrange']["max"])
                #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
                #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
                #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
                ax.grid(True,which='minor',linestyle='--')
                ax.tick_params(which='minor',labelsize=7)
                iplt += 1
                
                ax = fig.add_subplot(1,3,iplt)
                ax.set_title("%s average error covariance (full %s) \n" % (spc,basetitle))
                ax.set_xlabel("%s" % (self.units[spc]))
                ax.set_ylabel("%s" % ("hPa"))
                c = 'black'
                lw = 2
                if 'lw' in lkwargs: lw = lkwargs['lw']
                colors = plt.cm.Spectral(np.linspace(0,1,self.nmaxlevs))
                for iz in range(self.nmaxlevs):
                    ax.plot(cov2plt[iz,:]*self.cnvfctr[spc],press/100,color=colors[iz],lw=lw)
                ax.invert_yaxis()
                if press.max()/press.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
                if 'xlog' in lkwargs: ax.set_xscale('symlog',linthreshx=1e-4)
                if 'yrange' in lkwargs: 
                    ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
                if 'invRrange' in lkwargs:
                    if 'xlog' in lkwargs['invRrange']: ax.set_xscale('log')
                    ax.set_xlim(-lkwargs['invRrange']["max"],lkwargs['invRrange']["max"])
                ax.grid(True,which='minor',linestyle='--')
                ax.tick_params(which='minor',labelsize=7)
                iplt += 1
            
            if self.hasavk:
                ax = fig.add_subplot(1,3,iplt)
                ax.set_title("%s average averaging kernels \n" % (spc))
                ax.set_ylabel("%s" % ("hPa"))
                c = 'black'
                lw = 2
                if 'lw' in lkwargs: lw = lkwargs['lw']
                colors = plt.cm.Spectral(np.linspace(0,1,self.nmaxlevs))
                for iz in range(self.nmaxlevs):
                    ax.plot(self.avgavk[spc][:,iz],press/100,color=colors[iz],lw=lw)
                ax.invert_yaxis()
                if press.max()/press.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
                #if 'xlog' in lkwargs: ax.set_xscale('symlog',linthreshx=1e-1)
                if 'yrange' in lkwargs: 
                    ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
                ax.set_xlim(-0.2,0.5)
                if 'avkrange' in lkwargs:
                    if 'xlog' in lkwargs['avkrange']: ax.set_xscale('log')
                    ax.set_xlim(-lkwargs['avkrange']["max"],lkwargs['avkrange']["max"])
                ax.grid(True,which='minor',linestyle='--')
                ax.tick_params(which='minor',labelsize=7)
                iplt += 1
                
            plt.savefig(filename)
            plt.close(fig)
    
    def PlotAuxValues(self,outdir="./",**kwargs):
        
        figsize = (12,8)
        markersize = 2
        colormap = "jet"
        lkwargs = {}
       
        print(">>> Plotting avk and error covariances on a map")
        
        if 'ObsOnMapOpt' in kwargs: lkwargs = kwargs["ObsOnMapOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
            
        for isp,spc in enumerate(self.species):

            fig = plt.figure(figsize=figsize,dpi=200)
            
            filename = os.path.join(outdir,"%s_%s_AuxMap_%s_%s" % \
                                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))

            if self.hasavk:
                
                ax = fig.add_subplot(1,2,1)
                    
                m = self.SetBaseMap(ax,**kwargs)
    
                fld = np.ma.diagonal(np.moveaxis(self.avk[spc][~self.mask,:,:], 0, -1)).sum(axis=1)
                vmin = fld.min()
                vmax = fld.max()
                
                ax.set_title("%s Avk Trace \n Avg: %.2f,Min: %.2f,Max: %.2f" % (spc,fld.mean(),vmin,vmax))
                
                # Variable related keywords have priority over global keywords
                lvarkwargs = {}
                if "AvkTrace" in lkwargs: 
                    lvarkwargs = lkwargs["AvkTrace"]
                else:
                    lvarkwargs = lkwargs
                
                if "range" in lvarkwargs:
                    vmin = lvarkwargs["range"]["min"]
                    vmax = lvarkwargs["range"]["max"]
                
                sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o', \
                           c=fld.compressed(),\
                           cmap=colormap,vmin=vmin,vmax=vmax,\
                           latlon=True,zorder=10)
    
                plt.colorbar(sc,orientation='horizontal')
                
            ax = fig.add_subplot(1,2,2)
                
            ax.set_title("Retrieved levels")
                
            m = self.SetBaseMap(ax,**kwargs)

            fld = np.ma.count(self.col[spc][~self.mask,:],axis=1)
            vmin = fld.min()
            vmax = fld.max()
            
            ax.set_title("%s Retrieved levels \n Avg: %.1f,Min: %2i,Max: %2i" % (spc,fld.mean(),vmin,vmax))
            
            # Variable related keywords have priority over global keywords
            lvarkwargs = {}

            if "RetrievedLevels" in lkwargs: 
                lvarkwargs = lkwargs["RetrievedLevels"]
            else:
                lvarkwargs = lkwargs
            
            if "range" in lvarkwargs:
                vmin = lvarkwargs["range"]["min"]
                vmax = lvarkwargs["range"]["max"]
            
            sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o', \
                       c=fld,\
                       cmap=colormap,vmin=vmin,vmax=vmax,\
                       latlon=True,zorder=10)

            plt.colorbar(sc,orientation='horizontal')
                
            fig.savefig(filename)
            plt.close(fig)
            
    def PlotMaskValues(self,outdir="./",**kwargs):
                
        print(">>> Plotting assimilation mask on a map")
        
        figsize = (15,15)
        markersize = 2
        colormap = "Set1"
        lkwargs = {}
        
        levs = range(self.nmaxlevs)
        if 'LevelsSelection' in kwargs: levs = kwargs["LevelsSelection"]
            
        if 'MaskOnMapOpt' in kwargs: lkwargs = kwargs["MaskOnMapOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
            
        if self.press_isconst: 
            press = np.ma.copy(self.press[0,:])
        else: 
            press = np.ma.mean(self.press,axis=0)
            
        for isp,spc in enumerate(self.species):

            if self.outmask[spc].size == 0: return
            
            fig = plt.figure(figsize=figsize,dpi=200)
            
            filename = os.path.join(outdir,"%s_%s_MaskMap_%s_%s" % \
                                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            nsubplt = int(np.ceil(np.sqrt(len(levs))))
            
            for iplt,ilev in enumerate(levs):
                
                ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                
                m = self.SetBaseMap(ax,**kwargs)

                fld = self.outmask[spc][~self.mask,ilev]
                
                if fld.count() > 1:

                    ax.set_title('%s Assim Mask %.1f-%.1f hPa\n Nmiss: %i,Nrej: %i, Nok: %i' % \
                                 (spc,press[ilev]/100,press[ilev+1]/100,\
                                 (fld == -2).sum(),(fld == 0).sum(),(fld == 1).sum()))
                    
                    sc = m.scatter(self.lons[~self.mask],self.lats[~self.mask],markersize,marker='o',\
                                   c=fld,cmap=colormap,latlon=True,zorder=10,vmin=-2,vmax=1)
    
                    plt.colorbar(sc,orientation='horizontal')
                
            fig.savefig(filename)
            plt.close(fig)
        
    # A comprehensive printer for profile concentration stats
    def Stats(self,**kwargs):
        
        if self.nobs > 0:
            self.PrintGeoStats()
            self.PrintObsStats(**kwargs)
        
    # A comprehensive plotter for profile concentration data
    def Plot(self,outdir="./",**kwargs):
    
        sns.set() # Apply seaborn beautificator
        if self.nobs > 0:
            if 'makeplots' in kwargs:
                if 'ObsLoc' in kwargs['makeplots']: self.PlotObsLocation(outdir=outdir,**kwargs)
                if 'NObsTimeSer' in kwargs['makeplots']: self.PlotNobsTSer(outdir=outdir,**kwargs)
                if 'ObsDensity' in kwargs['makeplots']: self.PlotObsDensity(outdir=outdir,**kwargs)
                if 'ObsHisto' in kwargs['makeplots']: self.PlotObsHisto(outdir=outdir,**kwargs)
                if 'LevelsHisto' in kwargs['makeplots']: self.PlotLevelsHisto(outdir=outdir,**kwargs)
                if 'ObsOnMap' in kwargs['makeplots']: self.PlotObsValues(outdir=outdir,**kwargs)
                if 'VertProfiles' in kwargs['makeplots']: self.PlotVertProfiles(outdir=outdir,**kwargs)
                if 'AvgVertProfiles' in kwargs['makeplots']: self.PlotAvgVertProfiles(outdir=outdir,**kwargs)
                if 'AvgAuxProfiles' in kwargs['makeplots']: self.PlotAvgAuxProfiles(outdir=outdir,**kwargs)
                if 'AuxOnMap' in kwargs['makeplots']: self.PlotAuxValues(outdir=outdir,**kwargs)
                if 'MaskOnMap' in kwargs['makeplots']: self.PlotMaskValues(outdir=outdir,**kwargs)
            else:
                self.PlotObsLocation(outdir=outdir,**kwargs)
                self.PlotObsHisto(outdir=outdir,**kwargs)
                self.PlotObsValues(outdir=outdir,**kwargs)
        else:
            print('>> No obs to be plotted')

    def MakeSuperObs(self):
        pass
