# coding: utf8
""" Lidar observation class. """

from __future__ import print_function
import os
import sys
import numpy as np
import datetime
from baseinstr import BaseInstrument
import matplotlib.pyplot as plt
import copy
from functools import reduce


class LidarInstrument(BaseInstrument):
    """ Lidar observation class.

    Parameters
    ----------
    name : str
        Name of the instrument
    wllist : list of str, optional
        List of studied wavelengths.
    nmaxlevs : int, optional
        Maximal number of vertical levels.
    hascov : bool, optional
        If True, covariance errors are computed.
    upward : bool, optional
        If True, direction of the vertical axis is set to 'UP', otherwise
        set to 'DOWN'.
    vert_isconst : bool, optional
        Tells wether the vertical axis is constant or variable for each
        measurement.
    kwargs : dict, optional
        Used to allow unexpected keyword arguments.

    """
    # missing_value = np.nan

    def __init__(self, name, wllist=None, nmaxlevs=0, hascov=False,
                 vert_name="", upward=True, vert_isconst=False, **kwargs):
        BaseInstrument.__init__(self, name)
        self.hascov = hascov
        self.daimon_type = "Lidar"
        if upward is True:
            self.vertical_direction = 'UP'
        else:
            self.vertical_direction = 'DOWN'
        self.vertName = vert_name
        self.vert_isconst = vert_isconst
        self.vert_unit = ""
        if wllist is not None:
            self.wllist = wllist
            self.nwv = len(wllist)
            self.nmaxlevs = nmaxlevs
            self.units = {wv: "" for wv in wllist}
            self._initArrays()

    def _initArrays(self):
        ''' Initialize arrays of the instrument. '''
        self.prof = {wv: np.ma.empty((0, self.nmaxlevs), dtype=np.float32)
                     for wv in self.wllist}
        self.vert = np.ma.empty((0, self.nmaxlevs), dtype=np.float32)
        self.vert.fill_value = self.missing_value

        if self.hascov is True:
            self.cov = {wv: np.ma.empty((0, self.nmaxlevs, self.nmaxlevs),
                                        dtype=np.float32)
                        for wv in self.wllist}

        for wv in self.wllist:
            self.prof[wv].fill_value = self.missing_value
            if self.hascov is True:
                self.cov[wv].fill_value = self.missing_value

    def Read(self, filename, ReadAssim=None):
        """ Reads multispecies observations from MOCAGE HDF5.

        Parameters
        ----------
        filename : str
            Path to the input h5 file.
        ReadAssim : None or dict, optional
            Parameters to read hdf5 files produced during the assimilation.
            It is a dictionary with keys 'Region' and 'ObsType'
            (e.g. {'Region': 'GLOB11', 'ObsType': 'Y'}).

        """
        self.OpenFile(filename, mode="r")
        subgrp = ""
        self.obstype = "Y"
        if ReadAssim is not None:
            nobs = self.ReadGeoValues(ReadAssim=ReadAssim)
            if 'Region' in ReadAssim: subgrp = "/"+ReadAssim['Region']
            if 'ObsType' in ReadAssim:
                self.obstype = ReadAssim['ObsType']
        else:
            nobs = self.ReadGeoValues()
        
        obsgrp = self.f[self.name+subgrp+"/OBSERVATIONS"]
        geogrp = self.f[self.name+subgrp+"/GEOLOCALIZATION"]
        if 'Pressure' in geogrp.keys():
            self.vertName = 'Pressure'
        elif 'Altitude' in geogrp.keys():
            self.vertName = 'Altitude'
        else:
            raise Exception(
                "Vertical axis not found in GEOLOCALIZATION metadata.")
        # Init the species with those found in the HDF5 if not done before
        # Init the number of levels with hdf5 attribute
        if not hasattr(self, "nwv"):
            self.wllist = [wv for wv in obsgrp.keys()]
            self.nwv = len(self.wllist)
            self.nmaxlevs = geogrp[self.vertName][:].shape[1]
            self.units = {wv: obsgrp[wv][self.obstype].attrs['Units']
                          if 'Units' in obsgrp[wv][self.obstype].attrs else ''
                          for wv in obsgrp.keys()}
            for wv in obsgrp.keys():
                if isinstance(self.units[wv], bytes):
                    self.units[wv]=self.units[wv].decode()
            self._initArrays()
        # get geolocalization attributes
        if 'Units' in geogrp[self.vertName].attrs:
            self.vert_unit = geogrp[self.vertName].attrs['Units'].decode()
        else:
            self.vert_unit = ""

        if 'Orientation' in geogrp[self.vertName].attrs:
            self.vertical_direction = geogrp[self.vertName].attrs[
                'Orientation'].decode() 
        else:
            self.vertical_direction = 'DOWN'
        if 'IsGridConstant' in geogrp[self.vertName].attrs:
            vert_isconst = bool(geogrp[self.vertName].attrs[
                                     'IsGridConstant'])
        else:
            vert_isconst = False
        # check if read file IsGridConstant attr corresponds the obj data
        if self.vert.size == 0:
            self.vert_isconst = vert_isconst
        else:
            assert(self.vert_isconst == vert_isconst)
        if 'MissingValue' in geogrp[self.vertName].attrs:
            self.missing_value = geogrp[self.vertName].attrs[
                'MissingValue']
        vert = np.ma.masked_values(
            np.ma.array(geogrp[self.vertName][:], dtype=self.vert.dtype),
            self.missing_value,
            rtol=1e-03)
        if self.vert_isconst == 0 or self.vert.size == 0:
            self.vert = np.ma.array(np.ma.concatenate((self.vert, vert)),
                                    fill_value=self.missing_value)
        else:
            if (self.vert != vert).any():
                raise Exception(
                    "Trying to read an HDAT file defined on a "
                    "different vertical grid than data already contained" +
                    "in the object.")

        if nobs > 0:
            for wv in self.wllist:
                print(">>> Reading {0}, found {1} obs".format(
                    wv, len(obsgrp[os.path.join(wv, self.obstype)])))
                prof = np.ma.masked_values(
                    np.ma.array(obsgrp[os.path.join(wv, self.obstype)][:],
                                dtype=self.prof[wv].dtype),
                    self.missing_value,
                    rtol=1e-03)

                self.prof[wv] = np.ma.array(
                    np.ma.concatenate((self.prof[wv], prof)),
                    fill_value=self.missing_value)

                if self.hascov:
                    cov = np.ma.masked_values(
                        np.ma.array(obsgrp[wv+"/ErrorCovariance"][:],
                                    dtype=self.cov[wv].dtype),
                        self.missing_value,
                        rtol=1e-03)
                    self.cov[wv] = np.ma.array(
                        np.ma.concatenate((self.cov[wv], cov)),
                        fill_value=self.missing_value)
        self.CloseFile()

    def Write(self, filename, **kwargs):
        """ Writes multispecies observations in MOCAGE HDF5.

        Parameters
        ----------
        filename : str
            Path to the output h5 file.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.

        """
        self.OpenFile(filename)
        self.WriteHeader()
        subgrp = self.f[self.name].create_group("OBSERVATIONS")
        for wv in self.wllist:
            subgrp.create_group(wv)

        # paths inside the h5 file
        geo_path = os.path.join(self.name, "GEOLOCALIZATION")
        obs_path = os.path.join(self.name, "OBSERVATIONS")

        time_dset = self.f[geo_path].require_dataset(
            "Timestamp", (self.nobs, ), dtype=self.h5_typetime)

        h5arrtype = np.dtype('({}, )f'.format(self.nmaxlevs))
        h5matrixtype = np.dtype('({0}, {0}, )f'.format(self.nmaxlevs))

        # Vertical dataset (altitude or pressure)
        if self.vert_isconst is True:
            dset = self.f[geo_path].create_dataset(
                self.vertName, (1, ), dtype=h5arrtype, maxshape=(None, ))
        else:
            dset = self.f[geo_path].create_dataset(
                self.vertName, (self.nobs, ), dtype=h5arrtype,
                maxshape=(None, ))

        # Measurement attributs
        dset.attrs['MaxNumLevels'] = self.nmaxlevs
        dset.attrs['Units'] = np.string_(self.vert_unit)
        dset.attrs['Orientation'] = np.string_(self.vertical_direction)
        dset.attrs['MissingValue'] = self.missing_value
        dset.attrs['IsGridConstant'] = int(self.vert_isconst)
        self.attach_scale_2_dataset(dset, time_dset, 't', 'Timestamp')

        # Profiles dataset
        for wv in self.wllist:
            dset = self.f[os.path.join(obs_path, wv)].create_dataset(
                'Y', (self.nobs, ), dtype=h5arrtype, maxshape=(None, ))
            dset.attrs['MaxNumLevels'] = self.nmaxlevs
            dset.attrs['Units'] = self.units[wv]
            dset.attrs['MissingValue'] = self.missing_value
            self.attach_scale_2_dataset(dset, time_dset, 't', "Timestamp")

        if self.hascov:
            dset = self.f[os.path.join(obs_path, wv)].create_dataset(
                'ErrorCovariance', (self.nobs, ), dtype=h5matrixtype,
                maxshape=(None, ))
            dset.attrs['MaxNumLevels'] = self.nmaxlevs
            dset.attrs['Units'] = np.string_('')
            dset.attrs['MissingValue'] = self.missing_value
            self.attach_scale_2_dataset(dset, time_dset, 't',
                                        "Timestamp")

        # Write actual data in the file
        self.WriteGeoValues()
        if self.nobs == 0:
            print("Nothing to write to the HDF5 file")
        else:
            if self.vert_isconst is True:
                dset = self.f[geo_path].require_dataset(
                    self.vertName, (1, ), dtype=h5arrtype)
                dset[...] = self.vert.filled(
                    fill_value=self.missing_value)
            else:
                dset = self.f[geo_path].require_dataset(
                    self.vertName, (self.nobs, ), dtype=h5arrtype)
                dset[...] = np.array(
                    [val for val in self.vert[~self.mask, ...].filled(
                        fill_value=self.missing_value)],
                    dtype=self.vert.dtype)[self.sorted_idx]
            for wv in self.wllist:
                dset = self.f[os.path.join(obs_path, wv)].require_dataset(
                    "Y", (self.nobs, ), dtype=h5arrtype)
                dset[...] = np.array(
                    [val for val in self.prof[wv][~self.mask, ...].filled(
                        fill_value=self.missing_value)],
                    dtype=self.prof[wv].dtype)[self.sorted_idx]
                if self.hascov:
                    dset = self.f[
                        os.path.join(obs_path, wv)].require_dataset(
                            "ErrorCovariance",
                            (self.nobs, ),
                            dtype=h5matrixtype)
                    dset[...] = np.array(
                        [val for val in
                         self.cov[wv][~self.mask, ...].filled(
                            fill_value=self.missing_value)],
                        dtype=self.cov[wv].dtype)[self.sorted_idx]

        self.CloseFile()
        print(">>> HDF5 written ")

    def Select(self, TimeBnd=None, LatBnd=None, LonBnd=None, minVert=None,
               **kwargs):
        """ Make a space-time selection by setting a mask on the data.

        Parameters
        ----------
        TimeBnd : couple of datetime.datetime objects, optional
            Time boundaries (min, max).
        LatBnd : couple of float, optional
            Latitudes boundaries (min, max).
        LonBnd : couple of float, optional
            Longitudes boundaries (min, max).
        minVert : int, optional
            Minimal number of values on the vertical axis to consider an
            observation significant.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.

        """
        self.SelectGeo(TimeBnd=TimeBnd, LatBnd=LatBnd, LonBnd=LonBnd,
                       **kwargs)
        if self.vert_isconst is False:
            self.vert[self.mask, :] = np.ma.masked
        for wv in self.wllist:
            self.prof[wv][self.mask, :] = np.ma.masked
            if self.hascov:
                self.cov[wv][self.mask, :] = np.ma.masked

        if minVert is not None:
            self.select_significant(minVert)

    def select_significant(self, minVert):
        """ Mask obs with less than minVert values along the vertical axis.

        Parameters
        ----------
        minVert : int
            Minimal number of values on the vertical axis to consider an
            observation significant.

        """
        mask_dict = {wv: [np.sum(~self.prof[wv][i, :].mask) < minVert
                          for i in range(self.nobs)]
                     for wv in self.wllist}
        mask = reduce(np.logical_or, list(mask_dict.values()) + [self.mask])
        self.mask[mask] = True
        self.time[mask] = np.ma.masked
        self.lats[mask] = np.ma.masked
        self.lons[mask] = np.ma.masked
        self.nobs = (~self.mask).sum()
        if self.vert_isconst is False:
            self.vert[self.mask, :] = np.ma.masked
        for wv in self.wllist:
            self.prof[wv][self.mask, :] = np.ma.masked
            if self.hascov:
                self.cov[wv][self.mask, :] = np.ma.masked

    def verticalRegrid(self, z, interp):
        """ Interpolate data on a new vertical scale.

        Interpolation type is linear.

        Parameters
        ----------
        z : 1D array-like
            If interp is 'linear' or 'loglinear', z is the list of vertical
            levels for interpolation.
            If interp is 'layerAverage', z is the list of layers boundaries
            (e.g. if z = [1, 2, 4, 8], resulting vertical levels will be
            [1.5, 3.0, 6.0] and interpolated values will correspond to the mean
            of values contained in layers [(1, 2), (2, 4), (4, 8)]).
        interp : str
            Type of interpolation among:
              * 'layerAverage'
              * 'linear'
              * 'loglinear'

        """
        # make sure z is a list
        z = list(z)
        z.sort()

        # define layers in for layer-averaging interpolation
        if interp is 'layerAverage':
            # write layers as tuple
            layers = [(l, h) for l, h in zip(z[:-1], z[1:])]
            # find center of layers
            z = [np.mean((l, h)) for l, h in zip(z[:-1], z[1:])]

        # save old levels and profiles
        levels = copy.deepcopy(self.vert)
        levels.fill_value = np.nan
        profiles = copy.deepcopy(self.prof)
        for wv in self.wllist:
            profiles[wv].fill_value = np.nan

        # reinitialize self
        self.nmaxlevs = len(z)
        self._initArrays()
        self.vert = np.ma.concatenate(
            (self.vert,
             np.ma.array(np.array(z)[np.newaxis, :],
                         dtype=self.vert.dtype)))
        self.vert.fill_value = self.missing_value

        # interpolation
        for i in range(self.nobs):
            if self.vert_isconst is True:
                levs = levels[0, :].filled(fill_value=self.missing_value)
            else:
                levs = levels[i, :].filled(fill_value=self.missing_value)
            for wv in self.wllist:
                if interp is 'loglinear':
                    new_prof = np.interp(
                        np.log(z),
                        np.log(levs),
                        profiles[wv][i, :].filled(
                            fill_value=self.missing_value),
                        left=self.missing_value,
                        right=self.missing_value)
                elif interp is 'linear':
                    new_prof = np.interp(
                        z,
                        levs,
                        profiles[wv][i, :].filled(
                            fill_value=self.missing_value),
                        left=self.missing_value,
                        right=self.missing_value)
                elif interp is 'layerAverage':
                    new_prof = []
                    for lay in layers:
                        with np.warnings.catch_warnings():
                            np.warnings.filterwarnings(
                                'ignore')
                            idx = np.logical_and(levs >= lay[0],
                                                 levs < lay[1])
                            new_prof.append(np.nanmean(
                                profiles[wv][i, :].filled(
                                    fill_value=np.nan)[idx]))
                else:
                    raise Exception('Unknown interpolation type.')
                self.prof[wv] = np.ma.concatenate(
                    (self.prof[wv],
                     np.ma.array(np.array(new_prof)[np.newaxis, :],
                                 dtype=self.prof[wv].dtype)))
                self.prof[wv].fill_value = self.missing_value

        for wv in self.wllist:
            self.prof[wv][np.isnan(self.prof[wv])] = np.ma.masked

        # after interpolation vertical levels are constant between obs
        self.vert_isconst = True

    def Plot(self, outdir="./", makeplots=[], **kwargs):
        """ Computes different types of plots.

        Parameters
        ----------
        outdir : str, optional
            Output directory where to save the figures.
        makeplots : list of str, optional
            List of the different types of charts to drawn. Available
            possibilities are:
              * 'ObsLoc' (baseinstr.BaseInstrument.PlotObsLocation)
              * 'ObsScatter' (lidar_instr.LidarInstrument.scatterPlot)
        kwargs : dict, optional
            Keyword arguments passed to PlotObsLocation.

        """
        if self.nobs > 0:
            if 'ObsLoc' in makeplots:
                self.PlotObsLocation(outdir=outdir, **kwargs)
            if 'ObsScatter' in makeplots:
                self.scatterPlot(outdir=outdir, **kwargs)
            if 'VertProfile' in makeplots:
                self.PlotVertProfiles(outdir=outdir, **kwargs)
        else:
            print('>> No obs to be plotted')

    def scatterPlot(self,  others=[], outdir="./", ylim=None,
                     invert_yaxis=False, scatter_kw={'cmap': 'CMRmap_r'},
                     **kwargs):
        """ Scatter plot of observations.

        For ground-based lidars, for each station, a chart is made with
        time as abscissa. Whereas, for sattelite lidar, a unique chart
        is produced.

        Parameters
        ----------
        others :
            Others instruments objects to represent.
        outdir : str, optional
            Output directory where to save the figures.
        ylim : list of float
            Optional argument. Delimiters for y axis.
        invert_yaxis : bool
            If True, y axis is inverted.
        scatter_kw : dict
            Additional keyword arguments passed to matplotlib.pyplot.scatter.

        """
        print(">>> observation scatter plot")
        plotname = self.name
        # Shape of the subplot
        if len(others) == 0:
            columns = 1
            rows = 1
        elif len(others) in [1, 2]:
            columns = 1
            rows = len(others)+1
        else:
            columns = round((len(others)+1)/2)
            rows = 2

        plt.close('all')
        for wv in self.wllist:
            # Scatter plot min,max - self as reference for all plots
            if 'vmin' not in scatter_kw:
                scatter_kw['vmin'] = np.min(self.prof[wv])
            if 'vmax' not in scatter_kw:
                scatter_kw['vmax'] = np.max(self.prof[wv])

            if self.vertical_direction == 'DOWN':
                # Spatial LIDAR profiles
                # Set subplot
                fig, axs = plt.subplots(rows, columns, figsize=(15, 10),
                                        sharey=True, squeeze=False)
                if ylim is not None:
                    axs[0, 0].set_ylim(bottom=min(ylim[:]), top=max(ylim[:]))
                if invert_yaxis:
                    plt.gca().invert_yaxis()

                plotname += "_scatter_plot_{wv}.png"
                file_path = os.path.join(outdir, plotname)
                # Loop on instrument
                for ax, instr in zip(axs.ravel(), [self] + others):
                    ax, im = instr._scatter_plot(
                        ax,
                        idx=np.logical_not(instr.mask),
                        wv=wv,
                        title=(instr.name),
                        scatter_kw=scatter_kw,
                        # title=("Wavelength = {wv}",
                    )

                fig.tight_layout()
                fig.subplots_adjust(bottom=0.2)
                cbar_ax = fig.add_axes([0.15, 0.07, 0.7, 0.03])
                cbar = fig.colorbar(im, cax=cbar_ax, format='%.0e',
                                    orientation='horizontal')  # shrink=0.95
                cbar.ax.text(1.05, 0.1, self.units[wv], va='center', ha='left',
                             fontdict=dict(size='20'))
                cbar.ax.tick_params(labelsize=20)
                plt.savefig(file_path.format(wv=wv))
                plt.close()

            else:
                # Ground LIDAR profiles and others...
                plotname += "_scatter_plot_{lat},{lon}_{wv}.png"
                file_path = os.path.join(outdir, plotname)
                # find the different station locations
                coord = np.stack((self.lats.compressed(),
                                  self.lons.compressed()), axis=1)
                stations = np.unique(coord, axis=0)
                print("{} stations found:\n".format(stations.shape[0]),
                      stations)
                for lat, lon in stations:
                    # Set subplot
                    fig, axs = plt.subplots(rows, columns, figsize=(15, 10),
                                            sharey=True, squeeze=False)
                    if ylim is not None:
                        axs[0, 0].set_ylim(bottom=min(ylim[:]),
                                           top=max(ylim[:]))
                    if invert_yaxis:
                        plt.gca().invert_yaxis()

                    idx = reduce(np.logical_and,
                                 [self.lats == lat, self.lons == lon,
                                  ~self.mask])
                    # Loop on instruments
                    for ax, instr in zip(axs.ravel(), [self] + others):
                        ax, im = instr._scatter_plot(
                            ax,
                            idx=idx,
                            wv=wv,
                            title=(instr.name),
                            scatter_kw=scatter_kw,
                        )
                    fig.tight_layout()
                    fig.subplots_adjust(bottom=0.2)
                    cbar_ax = fig.add_axes([0.15, 0.07, 0.7, 0.03])
                    cbar = fig.colorbar(im, cax=cbar_ax, format='%.0e',
                                        orientation='horizontal')
                    cbar.ax.text(1.05, 0.1, self.units[wv],
                                 va='center', ha='left',
                                 fontdict=dict(size='20'))
                    cbar.ax.tick_params(labelsize=20)
                    plt.savefig(file_path.format(wv=wv, lat=lat, lon=lon))
                    plt.close()

    def _scatter_plot(self, ax, idx, wv, title, scatter_kw={}):
        """ Scatter plot of choosen observations.

        Parameters
        ----------
        ax : axes.Axes object
            Figure subplot.
        idx : 1D array
            Boolean array of length the number of observations in the object.
        wv : str
            Wavelength.
        title : str
            Title of the figure. It must contain {wv} instead of the
            wavelength value ({wv} will be automatically replaced).
        scatter_kw : dict
            Additional keyword arguments passed to matplotlib.pyplot.scatter.

        """
        times = np.array(self.time[idx].tolist())

        for i, x in enumerate(times):
            if self.vert_unit == 'Pa':
                convert_coef = 100  # Pressure in hPa
                ylabel = 'hPa'
            else:  # Keep unit in that case
                convert_coef = 1
                ylabel = self.vert_unit

            if self.vert_isconst is False:
                y = self.vert[idx, :][i]/convert_coef
            else:
                y = self.vert[0, :]/convert_coef

            y[y < 0] = np.nan
            c = self.prof[wv][idx, :][i]
            # c[c < 0] = np.nan
            im = ax.scatter(
                np.full(len(y), x, dtype='object'),
                y,
                c=c,
                zorder=10,
                **scatter_kw,
            )
        if 'Pa' in self.vert_unit:
            ax.set_yscale('log')
        ax.set_ylabel(ylabel, fontsize=20)
        ax.set_xlim(left=np.min(self.time).tolist(),
                    right=np.max(self.time).tolist())
        ax.tick_params(axis='both', which='major', labelsize=15)
        plt.gcf().autofmt_xdate(bottom=0.3, rotation=45, ha='right')
        ax.set_title(title.format(wv=wv), fontweight="bold", size=20)
        ax.grid(b=True, which='major', linewidth=2)
        ax.grid(b=True, which='minor', linewidth=1)
        ax.label_outer()
        return(ax, im)

    def make_super_obs(self, resolution):
        """ Compute super-observations. """
        print("Super-observations computing is not implemented " +
              "for LidarInstrument.")

    def PlotVertProfiles(self, others=[], outdir="./",**kwargs):
        
        #figsize = (15,15)
        nmaxplts = 500
        lkwargs = {}
        if 'VertProfilesOpt' in kwargs:
            lkwargs = kwargs["VertProfilesOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        
        print(">>> Plotting vertical profiles")
        
        for iwv,wv in enumerate(self.wllist):
            
            filename = os.path.join(outdir,"%s_%s_Profiles_%s_%s" % \
                (self.name,wv,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            fig = plt.figure(dpi=200)
            #figsize=figsize
            npltsxpage = min(self.nobs,nmaxplts)
            nsubplt = int(np.ceil(np.sqrt(npltsxpage)))
            
            for iplot,ip in enumerate((~self.mask).nonzero()[0]):
                if iplot > npltsxpage-1: break
                
                fig = plt.figure(dpi=200)
                #ax = fig.add_subplot(nsubplt,nsubplt,iplot+1)
                ax = fig.add_subplot(111)
                ax.set_title("%s nm profile \n (%.1fN,%.1fE,%s)" % (\
                    wv,self.lats[ip],self.lons[ip],\
                    self.time[ip].astype(str)[11:16]))
                if self.units[wv] == 'm^-1' :#'$"%s"$' % (self.units[wv])
                    ax.set_xlabel('$m^{-1}$')
                else :
                    ax.set_xlabel("%s" % (self.units[wv]))
                ax.set_ylabel("%s" % ("hPa"))
                #c = 'black'
                lw = 2
                if 'lw' in lkwargs: lw = lkwargs['lw']
                
                if self.vert_isconst: 
                    press = self.vert[0, :]
                else:
                    press = self.vert[ip, :]
                    
                ax.plot(self.prof[wv][ip, :], press/100, c='black', lw=lw,
                        label=self.obstype)
                for obj in others:
                    ax.plot(obj.prof[wv][ip,:], press/100, lw=lw,
                            label=obj.obstype)
                ax.invert_yaxis()
                if press.max()/press.min() >= 20 or \
                    'ylog' in lkwargs: ax.set_yscale('log')
                if 'xlog' in lkwargs: ax.set_xscale('symlog',linthreshy=1e-1)
                if 'yrange' in lkwargs: 
                    ax.set_ylim(lkwargs['yrange']["min"],
                            lkwargs['yrange']["max"])
                if 'xrange' in lkwargs:
                    ax.set_xlim(lkwargs['xrange']["min"],
                            lkwargs['xrange']["max"])
                #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
                #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
                #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
                ax.grid(True,which='minor',linestyle='--')
                ax.tick_params(which='minor',labelsize=7)
                ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
                ax.legend(loc='upper right')
            #plt.subplots_adjust(wspace=0.2,hspace=0.3)
                plt.savefig(filename+'_'+str(iplot))
                plt.close(fig)

    def to_altitude(self):
        """ """
        if self.vertName == 'Altitude':
            pass
        elif self.vertName == 'Pressure':
            if self.vert_unit not in ['Pa']:
                raise Exception(("Vertical unit is {} and must be " +
                                 "Pa.").format(self.vert_unit))
            P0 = 101325.
            self.vert = np.log(P0/self.vert)/0.00012
            self.vert_unit = 'meters'
            self.vertName = 'Altitude'
        else:
            raise Exception(
                "Vertical coordinate must be either 'Pressure' or 'Altitude'")

    def to_pressure(self):
        """ """
        if self.vertName == 'Pressure':
            pass
        elif self.vertName == 'Altitude':
            if self.vert_unit not in ['meter', 'meters', 'm']:
                raise Exception(("Vertical unit is {} and must be " +
                                 "meters.").format(self.vert_unit))
            P0 = 101325.
            self.vert = P0*np.exp(-0.00012*self.vert)
            self.vert_unit = 'Pa'
            self.vertName = 'Pressure'
        else:
            raise Exception(
                "Vertical coordinate must be either 'Pressure' or 'Altitude'")
