# coding: utf8
""" MOCAGE observations tools. """

from __future__ import print_function
import copy
import glob
import os
import sys
from datetime import datetime, timedelta, time
from itertools import groupby
import h5py
from dateutil.parser import parse
from six import string_types  # for python 2/3 compatibility
import importlib
import gc
import numpy as np
try:
    import netCDF4
    # import xarray # needed for netcdf manipulation
except ImportError:
    pass

def Convert(instrname, instrtype, period, inputdir="./", outputdir="./",
            make_super_obs=None, **kwargs):
    """ Convert obs from native format to MOCAGE HDF5.

    It is the principal function to be used in the user script.

    Parameters
    ----------
    instrname : str
        Name of the instrument.
    instrType : str
        Name of the class corresponding to the instrument. It must be
        one of
            * SACS_SO2
            * MODIS_AOD
            * EarlinetLidar
            * CaliopLidar
            * PMAP_AOD
            * MSGashEUMETSAT
    period : couple of datetime.datetime
        Time period defined as (fisrt day, last day).
    outputdir : str, optional
        Output directory.
    make_super_obs : dict or None, optional
        Keyword arguments passed to instrType.make_super_obs. If not None,
        super-observations are computed.
    kwargs : kwargs
        Keyword arguments passed to instrType.__init__, instrType.Convert,
        instrType.Select, instrType.Stats, instrType.Plot, instrType.Write.

    """
    outputfreq = 'daily'
    if "options" in kwargs:
        if "OutputFrequency" in kwargs["options"]:
            outputfreq = kwargs["options"]["OutputFrequency"]

    if not os.path.exists(outputdir):
        os.makedirs(outputdir)

    for subperiod in SplitPeriod(period, outputfreq):
        obs = DefInstrument(instrname, instrtype, **kwargs)
        obs.Convert(subperiod["period"], inputdir, **kwargs)

        obs.Select(**kwargs)

        if make_super_obs is not None:
            obs.make_super_obs(**make_super_obs)

        if 'makestats' in kwargs:
            obs.Stats(**kwargs)

        if 'makeplots' in kwargs:
            obs.Plot(outputdir, **kwargs)

        obs.Write(os.path.join(outputdir, 'HDAT+'+subperiod["filename"]),
                  **kwargs)

        del obs
        gc.collect()


def DefInstrument(instrName, instrType, **kwargs):
    """ Build an instrument object of the willing class.

    Parameters
    ----------
    instrName : str
        Name of the instrument.
    instrType : str
        Name of the class corresponding to the instrument. It must be
        one of 
            * SurfConcInstrument
            * ProfConcInstrument
            * IntQtyInstrument
            * IASI_L1C
            * GOME2_L1B
            * RadioSondeInstrument
            * SOFRIDProfInstrument
            * SACS_SO2
            * AODInstrument
            * MODIS_AOD
            * LidarInstrument
            * EarlinetLidar
            * PMAP_AOD
            * MSGashEUMETSAT
            * WoudcInstrument
    kwargs : kwargs
        Keyword arguments passed to the class constructor.

    """
    # This is a complete list of all available obs classes
    # with corresponding python filenames.
    # It is used to import only the instruments selected by the user
    instr_modules = {"SurfConcInstrument": "surfconc_instr",
                     "ProfConcInstrument": "profconc_instr",
                     "IntQtyInstrument": "IntQty_instr",
                     "IASI_L1C": "user_instruments.IASI_L1C_instr",
                     "GOME2_L1B": "user_instruments.GOME2_L1B_instr",
                     "RadioSondeInstrument": "user_instruments.radiosonde_instr",
                     "SOFRIDProfInstrument": "user_instruments.SOFRID_prof_instr",
                     "SACS_SO2": "user_instruments.SACS_SO2_instr",
                     "AODInstrument": "AOD_instr",
                     "ColConcInstrument": "colconc_instr",
                     "ConcInstrument": "concinstr",
                     "MODIS_AOD": "user_instruments.MODIS_AOD_instr",
                     "AERUSGEO_AOD": "user_instruments.AERUSGEO_AOD_instr",
                     "CALIOP_AOD": "user_instruments.CALIOP_AOD_instr",
                     "EarlinetLidar": "user_instruments.earlinet_lidar_instr",
                     "EprofileLidar": "user_instruments.eprofile_lidar_instr",
                     "LidarInstrument": "lidar_instr",
                     "CaliopLidar": "user_instruments.caliop_lidar_instr",
                     "PMAP_AOD": "user_instruments.PMAP_AOD_instr",
                     "MSGashEUMETSAT": "user_instruments.MSGashEUMETSAT_IntQty_instr",
                     "EeaSurfaceObservations": "user_instruments.eea_surf_instr",
                     "TropomiInstr": "user_instruments.tropomi_intqty_instr",
                     "WoudcInstrument": "user_instruments.woudc_instr",
                     "ATOM_flights": "user_instruments.atom_instr",
                     "IagosFlight": "user_instruments.iagos_instr",
                     "CAP_ASH": "user_instruments.cap_ash_instr",
                     "VIIRS_AOD": "user_instruments.VIIRS_AOD_instr",
                     }

    # Load the correct instrument module here
    module = importlib.import_module(instr_modules[instrType])
    # Consider making the module import list globally available
    # It will allow using 'isinstance' to allow for class specific
    # coding in all utilities. For the moment I use type().__name__
    # but I cannot check for inherited types

    # This defines the class to be used
    myinstr_cls = getattr(module, instrType)

    return myinstr_cls(instrName, **kwargs)


def SplitPeriod(period, outputfreq):
    """ Split a period according to a time step.

    Split a period into daily, monthly ... time frames and return also
    appropriate filenaming for MOCAGE.

    Parameters
    ----------
    period : couple of datetime.datetime
        Time period defined as (fisrt day, last day).
    outputfreq : str
        Currently, it can only be equal to 'daily'.

    Returns
    -------
    List of dictionaries
        Each dictionary {"period": (start, end), "filename": name} corresponds
        to a subperiod.

    """
    if outputfreq == 'daily':
        firstday = datetime(period[0].year, period[0].month, period[0].day)
        dates = [(firstday + timedelta(days=i),
                  firstday + timedelta(days=i+1))
                 for i in range((period[1]-period[0]).days)]
        names = [d[0].isoformat().replace('-', '')[:8] for d in dates]
        subperiods = [{"period": date, "filename": name}
                      for date, name in zip(dates, names)]
    else:
        raise Exception("ERROR: {} not yet implemented".format(outputfreq))

    return subperiods


def DefPeriod(start, end):
    """ Convert start/end argument to datetime objects.

    Parameters
    ----------
    start, end: str
        Start/end date of the period. They can be given in 'yyyymmdd' format
        or any format that can be parsed with dateutil.parser

    Returns
    -------
    couple of datetime.datetime
        Time period define as (fisrt day, last day).

    """
    try:
        starttime = parse(start)  # works well with many possible date inputs,
        # including 'only integer strings' except %Y%m%d%H
    except:
        starttime = datetime.strptime(start, "%Y%m%d%H")

    try:
        endtime = parse(end)
    except:
        endtime = datetime.strptime(end, "%Y%m%d%H")

    if endtime < starttime:
        raise ValueError('In DefPeriod, endtime is before starttime!')

    return (starttime, endtime)


def GenerateObs(instrname, instrtype, outputdir="./", **kwargs):
    """ An observation generator. """
    obs = DefInstrument(instrname,instrtype,**kwargs)
    # Reads the Hx variable in in input HSTAT, depending on the ReadAssim dictionary
    if "FromFile" in kwargs:
        obs.Read(kwargs["fromfile"]["filename"],**kwargs)
        date = os.path.basename(kwargs["fromfile"]["filename"]).split('+')[1][:8]
    elif "random" in kwargs:
        obs.Generate(**kwargs)
        date = kwargs["random"]["date"]
    # Sometime necessary when auxiliary obs inputs are not found in HSTAT files,
    # in this case they are taken from filename1 (HDAT), whereas Y is taken from filename2 (HSTAT)
    elif "FromTwoFiles" in kwargs:
        obs.Read(kwargs["FromTwoFiles"]["filename1"])
        date = os.path.basename(kwargs["FromTwoFiles"]["filename1"])[5:13]
        obs2 = DefInstrument(instrname,instrtype,**kwargs)
        obs2.Read(kwargs["FromTwoFiles"]["filename2"],**kwargs)
        obs.ReplaceWith(obs2)
        if "Perturb" in kwargs and kwargs["Perturb"] == True: obs.Perturb(**kwargs)
    else:
        print("Decide the type of generation (random or fromfile)")
        sys.exit()

    if 'makestats' in kwargs:
        obs.Stats(**kwargs)

    if not os.path.exists(outputdir) :
        os.makedirs(outputdir)

    if 'makeplots' in kwargs:
        obs.Plot(outdir=outputdir,**kwargs)

    obs.Write(os.path.join(outputdir, 'HDAT+'+date), **kwargs)


def MocageAscii2Hdf5(instrname, instrtype, period, inputdir="./",
                     outputdir="./", **kwargs):
    """ Convert obs from old ASCII format to MOCAGE HDF5. """         
    if not os.path.exists(outputdir) : os.makedirs(outputdir)
    
    for day in SplitPeriod(period,'daily'): 
        obs = DefInstrument(instrname,instrtype,**kwargs)
        obs.ConvertMocageASCII(day["period"][0],inputdir,**kwargs)
        if 'makestats' in kwargs: obs.Stats(**kwargs)
        if 'makeplots' in kwargs: obs.Plot(outdir=outputdir,**kwargs)
        obs.Write(os.path.join(outputdir,'HDAT+'+day["filename"]))
        del obs
        gc.collect()


def AggregateInstr(indir=None, outdir=None, filefreq="daily", filetype="h5"):
    """ Combine HDF5/NETCDF files with different instruments.

    Files can be searched in more directories
    If outdir is not given, used indir[0] for saving files

    """
    period = None

    if indir == None:
        raise ValueError('indir argument cannot be empty')

    if isinstance(indir, string_types):
        indir = [indir] #to make sure that inputdir is iterable

    if outdir == None:
        outdir = indir[0]
        
    if not os.path.exists(outdir) : os.makedirs(outdir)

    if filefreq == "yearly":
        sl = 9
    elif filefreq == "monthly":
        sl = 11
    elif filefreq == "daily":
        sl = 13
    else:
        raise ValueError('filefreq argument is not recognized')

    if period is not None:
        gooddates = [subperiod["filename"] for subperiod in\
         SplitPeriod(DefPeriod(period[0],period[1]),'daily')]    

    files = []
    for dir in indir:
        files += [f for f in glob.glob(dir+"/HDAT*.%s*" % (filetype))]
    if period is not None: 
        files = [f for f in files if os.path.basename(f)[5:-3] in gooddates]

    bnames = map(os.path.basename, files)
    hdates = [s[s.find('HDAT'):s.find('HDAT')+sl] for s in bnames]
    tags = zip(hdates, files) #list of tuples
    tags = sorted(tags, key=lambda x:x[0]) #sort by the first value, necessary for groupby

    for k, v in groupby(tags, key=lambda x: x[0]):
        infiles = [x[1] for x in v] #second values in returned tuples
        outfile = (outdir + '/' + k + '.%s' % (filetype)).replace('//','/')
        print('For date:', k[5:], '- ', len(infiles), 'files to aggregate')
        if outfile in infiles:
            raise ValueError('Stop! File ' + outfile +' is used as input and output.')
        if filetype == 'h5': AggregateHdf5(infiles,outfile)
        if filetype == 'nc': AggregateNcdf(infiles,outfile)

def AggregateHdf5(filelist, outputfile, mergetype="copy"):
    """ Aggregate two or more HDF5 files. """

    fout = h5py.File(outputfile,'w')

    for filein in filelist:
        fin = h5py.File(filein,'r')
        for instr in fin:
            if instr not in fout:
                if mergetype is 'copy':
                    h5py.h5o.copy(fin.id, instr.encode(), fout.id, instr.encode())
            #fout[instr] = h5py.ExternalLink(filein,instr)
        fin.close()

    fout.close()

def AggregateNcdf(filelist, outputfile):

    mode = 'w'
    for fn, filein in enumerate(filelist):
        if fn > 0: mode = 'a'
        with netCDF4.Dataset(filein) as src, netCDF4.Dataset(outputfile, mode) as dst:
            recursivecopy(src,dst)

def recursivecopy(root,dst):

    for grp in root.groups:
        if grp not in dst.groups: dst.createGroup(grp)
        copy_netcdf_group(root[grp], dst[grp])
        recursivecopy(root[grp],dst[grp])

def PlotObs(instrname,instrtype,inputfile,outputdir="./",**kwargs):
    """ Snapshots visualization/stats of one hdf5 file

    Plot snapshots of observations from MOCAGE HDF5

    """
    obs = DefInstrument(instrname,instrtype,**kwargs)
    
    obs.Read(inputfile,**kwargs)

    if not os.path.exists(outputdir) : os.makedirs(outputdir)
    
    obs.Select(**kwargs)
    if 'makestats' in kwargs: obs.Stats(**kwargs)
    obs.Plot(outdir=outputdir,**kwargs)


def PlotObsTimeSeries(instrname, instrtype, inputdir, outputdir="./",
                      **kwargs):
    """ Time series visualization/stats of multiple hdf5 files

    Plot timeseries of observations from MOCAGE HDF5

    """
    obs = DefInstrument(instrname,instrtype,**kwargs)
    
    hdatlist = sorted([f for f in os.listdir(inputdir) if f[:4] == 'HDAT'])

    if 'TimeBnd' in kwargs: 
        hdatlist = [f for f in hdatlist if 
            (datetime.strptime(f.split('.')[0][-8:], "%Y%m%d") >= kwargs['TimeBnd'][0] and 
             datetime.strptime(f.split('.')[0][-8:], "%Y%m%d") <= kwargs['TimeBnd'][1])]
    
    for f in hdatlist: obs.Read(os.path.join(inputdir,f),**kwargs)

    if not os.path.exists(outputdir) : os.makedirs(outputdir)
    
    obs.Select(**kwargs)
    if 'makestats' in kwargs: obs.Stats(**kwargs)
    obs.PlotNobsTSer(outdir=outputdir,**kwargs)
    obs.Plot(outdir=outputdir,**kwargs)
    

def PreProcessObs(instrname, instrtype, period, inputdir="./", outputdir="./",
                  **kwargs):
    """ Preproccessing of one DAIMON hdf5 file. """
    inputfreq = 'daily'
    if "options" in kwargs:
        if "InputFrequency" in kwargs["options"]:
            inputfreq = kwargs["options"]["InputFrequency"]
    
    if not os.path.exists(outputdir) : os.makedirs(outputdir)
    
    for subperiod in SplitPeriod(period,inputfreq):
        
        obs = DefInstrument(instrname,instrtype,**kwargs)
        
        try:
            inputfile = glob.glob(os.path.join(inputdir,'HDAT+'+subperiod["filename"]+'.*'))[0]
            print('Found ', inputfile)
        except:
            print('No valid obs files found for this period ',subperiod["filename"])

        obs.Read(inputfile,**kwargs)
    
        obs.Select(**kwargs)
        if 'makestats' in kwargs: obs.Stats(**kwargs)
        if 'makeplots' in kwargs: obs.Plot(outdir=outputdir,**kwargs)
        
        obs.Write(os.path.join(outputdir,os.path.basename(inputfile).split('.')[0]),**kwargs)

        del obs
        gc.collect()

def PlotObsAssim(instrname, instrtype, inputfile, outputdir="./", **kwargs):
    """ Plot snapshots of observations and model

    Snapshots visualization/stats of DAIMON assimilation output for one
    hdf5 file

    """
    # Observations first 
    print(">>> Y :")
    y = DefInstrument(instrname,instrtype,**kwargs)
    y.Read(inputfile,**kwargs)
    
    if not os.path.exists(outputdir) : os.makedirs(outputdir)
    
    outdir = os.path.join(outputdir,"Y")
    if not os.path.exists(outdir) : os.makedirs(outdir)
    
    y.Select(**kwargs)
    if 'makestats' in kwargs: y.Stats(**kwargs)
    y.Plot(outdir=outdir,**kwargs)

    # Better implement an obs cleaner maybe
    if 'ObsType' in kwargs:
        otypes = kwargs['ObsType']
    else:
        otypes = ('Hx_bkg','Hx_ana')
        
    if 'DiffType' in kwargs:
        difftypes = kwargs['DiffType']
    else:
        difftypes = ('Absolute',)
    
    hx, misfit = {}, {}    

    for otype in otypes:
        
        print(">>> "+otype)
        outdir = os.path.join(outputdir,otype)
        if not os.path.exists(outdir) : os.makedirs(outdir)

        hx[otype] = DefInstrument(instrname,instrtype,**kwargs)
        kwargs['ReadAssim']['ObsType'] = otype
        hx[otype].Read(inputfile,**kwargs)
        hx[otype].Select(**kwargs)
        if 'makestats' in kwargs: hx[otype].Stats(**kwargs)
        hx[otype].Plot(outdir=outdir,**kwargs)
         
        for dtype in difftypes:
            print(">>> %s misfit: " % (dtype))
            outdir = os.path.join(outputdir,"Misfit_"+otype+"_"+dtype)
            if not os.path.exists(outdir) : os.makedirs(outdir)
        
            # Replace the kwargs with specific keywords for diff plotting
            misfitpltkw = copy.deepcopy(kwargs)
            if 'MisfitPlotKwargs' in kwargs:
                if dtype in kwargs['MisfitPlotKwargs']: 
                    for key in kwargs['MisfitPlotKwargs'][dtype].keys():
                        misfitpltkw[key] = kwargs['MisfitPlotKwargs'][dtype][key]
                else:
                    for key in kwargs['MisfitPlotKwargs'].keys():
                        misfitpltkw[key] = kwargs['MisfitPlotKwargs'][key]
                    
            misfit[otype] = hx[otype]-y
            if dtype == 'PercentOfObs': misfit[otype] = misfit[otype]/y
            if dtype == 'PercentOfMod': misfit[otype] = misfit[otype]/hx[otype]
            if 'makestats' in kwargs: misfit[otype].Stats(**misfitpltkw)
            misfit[otype].Plot(outdir=outdir,**misfitpltkw)


def daterange(d1, d2):
    """ """
    d1 = datetime.combine(d1, time.min)  # truncate to the midnight
    d2 = enddtrange(d2)
    # in daterange, as in range, the last date is not inclusive
    return (d1 + timedelta(days=i) for i in range((d2 - d1).days))

def enddtrange(dt):
    """ """
    if dt == datetime.combine(dt, time.min):
        return dt # just return it - it's already midnight
    else:
        # find a forthcoming midnight
        return datetime.combine(dt + timedelta(days=1), time.min)

def copy_netcdf_group(src, dst):

        dst.setncatts(src.__dict__)

        for name, dimension in src.dimensions.items():
            if name not in dst.dimensions.keys():
                dst.createDimension(name, (len(dimension) if not dimension.isunlimited() else None))

        for name, variable in src.variables.items():
            if name not in dst.variables.keys():
                fill_value = None
                if hasattr(src[name], "_FillValue"): fill_value = src[name]._FillValue
                x = dst.createVariable(name, variable.datatype, 
                    variable.dimensions, fill_value=fill_value)
                # copy variable attributes
                dst[name].setncatts({k: src[name].getncattr(k) 
                    for k in src[name].ncattrs() if k not in ["_FillValue"]})
                dst[name][:] = src[name][:]
                