# coding: utf8
""" Profile concentration module

@author: emili

"""
from __future__ import print_function, division
import os
import sys
import numpy as np
from datetime import datetime
from baseinstr import BaseInstrument
from concinstr import ConcInstrument
try:
    import utils_Hobs_mini as UtilsASCII # needed for conversion of old MOCAGE obs files
except ImportError:
    pass
import matplotlib as mpl
import matplotlib.pyplot as plt
import copy
import time
try:
    import seaborn as sns # needed for nicer plots
except ImportError:
    pass
try:
    import MV2 as MV  # only for vertical regridding
    import cdutil.vertical  # only for vertical regridding
except ImportError:
    pass
try:
    from netCDF4 import Dataset # needed for netcdf storage
except ImportError:
    pass
from itertools import cycle

class ProfConcInstrument(ConcInstrument):
    """ Profile concentration intrument class.

    Init with user defined species if given
    The behavior is as follows:
      - no species init by users -> all species found in h5 are read (units
        andconvfctr reader to be done)
      - species init by the user -> only specified ones are read
      - for ascii conversion error is raised if the user does not specify the
        species

    Parameters
    ----------
    name : str
        Name of the instrument.
    species : list of str, optional
        List of species stored in the object.
    nmaxlevs : int
        Maximal number of vertical levels.

    """
    def __init__(self, name, species=None, nmaxlevs=None, **kwargs):
        ConcInstrument.__init__(self, name, species=species, **kwargs)
        if nmaxlevs is not None:
            if species is not None:
                self.nmaxlevs = nmaxlevs
                self.InitArrays()
            else:
                raise Exception("if you set the max number of vertical" +
                                "levels (nmaxlevs keyword), you also need " +
                                "to specify species argument !")

        # set the pressure axis constant or variable for each measurement
        self.press_isconst = False
        if "ConstPressureGrid" in kwargs:
            self.press_isconst = bool(kwargs["ConstPressureGrid"])

        # orientation of the vertical axis
        self.press_coorddir = 'upward'

        self.hascov = False  # switch to treat or not error covariances
        self.hasavk = False  # switch to treat or not averaging kernels

        self.daimon_type = "Profile"
    
    # Some operator overloading to add/substract/divide batch of measurements
    # taken at the same time and place (including possibly averages)
    def __add__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for spc in self.species:
                result.prof[spc] = self.prof[spc] + other.prof[spc]
                if 'avgprof' in self.__dict__:
                    # Need to recompute standard deviation after the sum.
                    # Consider making std optional and avoid calling the averager
                    result.Average(AverageOpt={"PressureGrid":other.avgpress})
                    #result.avgprof[spc] = self.avgprof[spc] + other.avgprof[spc]
                if 'columns' in self.__dict__: 
                    for col in self.columns["ColsBounds"]:
                        result.columns[spc][col] = self.columns[spc][col] + other.columns[spc][col]
        else:
            print('Cannot sum obs: mismatch in geolocalizations')
        
        return result
        
    def __sub__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for spc in self.species:
                result.prof[spc] = self.prof[spc] - other.prof[spc]
                if 'avgprof' in self.__dict__:
                    # Need to recompute standard deviation for the difference.
                    # Consider making std optional and avoid calling the averager
                    result.Average(AverageOpt={"PressureGrid":other.avgpress})
                    #result.avgprof[spc] = self.avgprof[spc] - other.avgprof[spc]
                if 'columns' in self.__dict__: 
                    for col in self.columns["ColsBounds"]:
                        result.columns[spc][col] = self.columns[spc][col] - other.columns[spc][col]
        else:
            print('Cannot diff obs: mismatch in geolocalizations')
        
        return result
    
    def __truediv__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            for spc in self.species:
                result.prof[spc] = self.prof[spc] / other.prof[spc]
                if 'avgprof' in self.__dict__:
                    result.avgprof[spc] = self.avgprof[spc] / other.avgprof[spc]
                    result.stdprof[spc] = self.stdprof[spc] / other.avgprof[spc]
                result.units[spc] = "%"
                result.cnvfctr[spc] = 100
                if 'columns' in self.__dict__: 
                    for col in self.columns["ColsBounds"]:
                        result.columns[spc][col] = self.columns[spc][col]/other.columns[spc][col]*100
                    result.columns[spc]["units"] = "%"
        else:
            print('Cannot divide obs: mismatch in geolocalizations')
        
        return result
    
    def ScatterWith(self,other,outdir="./",**kwargs):
        
        figsize = (8,8)
        lkwargs = {}
        printdata = True
        
        if 'ColsHistoOpt' in kwargs: lkwargs = kwargs["ColsHistoOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        # Compute scatter plots between two datasets
        # Done only if partial columns are already available
        if self.isColocatedWith(other) and 'columns' in self.__dict__:
            for spc in self.species:
                
                filename = os.path.join(outdir,"%s_%s_%s_%s_%s" % \
                    (self.name,spc,"ColumnsScatter",self.time.compressed().min().astype(str)[:13].replace("-",""), \
                     self.time.compressed().max().astype(str)[:13].replace("-","")))
                
                if printdata: 
                    f=open(filename+".txt","w")
                    f.write("%10s %5s %9s %9s %9s %9s %9s %9s %5s\n" %
                            (spc+" col","n","mean(Y)","std(Y)","mean(Hx)","std(Hx)","bias","std","corr"))
                            
                fig = plt.figure(figsize=figsize,dpi=200)
                # Variable related keywords have priority over global keywords
                lvarkwargs = {}
    
                if spc in lkwargs: 
                    lvarkwargs = lkwargs[spc]
                else:
                    lvarkwargs = lkwargs
                                        
                nsubplt = int(np.ceil(np.sqrt(len(self.columns["ColsBounds"]))))
                
                for iplt,col in enumerate(self.columns["ColsBounds"].keys()):
                    ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                    ax.scatter(self.columns[spc][col],other.columns[spc][col],s=2) 
                    if "range" in lvarkwargs:
                        if col in lvarkwargs["range"]:
                            vmin = lvarkwargs["range"][col]["min"]
                            vmax = lvarkwargs["range"][col]["max"]
                        else:
                            vmin = lvarkwargs["range"]["min"]
                            vmax = lvarkwargs["range"]["max"]
                        pltrange =(vmin,vmax)
                    else:
                        pltrange = (min(self.columns[spc][col].min(),other.columns[spc][col].min())*0.9,\
                                    max(self.columns[spc][col].max(),other.columns[spc][col].max())*1.1)
                    ax.set_xlim(pltrange)
                    ax.set_ylim(pltrange)
                    bias, std, corr = \
                       (other.columns[spc][col]-self.columns[spc][col]).mean(), \
                        np.ma.std(self.columns[spc][col]-other.columns[spc][col]),\
                        np.ma.corrcoef(np.ma.vstack((self.columns[spc][col],other.columns[spc][col])))[0,1]
                    ax.set_title("%s %s column \n Bias:%4.2e Std:%4.2e r:%4.2f" % \
                       (spc,col,bias,std,corr))
                    ax.set_xlabel("Y (%s)" % (self.columns[spc]["units"]))
                    ax.set_ylabel("Hx (%s)" % (self.columns[spc]["units"]))
                    if printdata: f.write("%10s %5i %5.3e %5.3e %5.3e %5.3e %5.3e %5.3e %5.3f\n" \
                                          % (col,self.columns[spc][col].count(),\
                                             self.columns[spc][col].mean(),np.ma.std(self.columns[spc][col]), \
                                             other.columns[spc][col].mean(),np.ma.std(other.columns[spc][col]), \
                                             bias,std,corr))
                    
                plt.subplots_adjust(wspace=0.2,hspace=0.4)    
                plt.savefig(filename)
                plt.close(fig)
                if printdata: f.close()
                
        return
    
    # Init the data arrays for the profile instrument
    def InitArrays(self):
        
        self.prof = {spc : np.ma.empty((0,self.nmaxlevs),dtype=np.float32) \
                for spc in self.species}
        self.press = np.ma.empty((0,self.nmaxlevs),dtype=np.float32)
        # Consider storage optimization for both covariances and avk
        self.cov = {spc : np.ma.empty((0,self.nmaxlevs,self.nmaxlevs),dtype=np.float32) \
                for spc in self.species}
        # Consider distinct levels in and levels out for averaging kernels matrices
        self.avk = {spc : np.ma.empty((0,self.nmaxlevs,self.nmaxlevs),dtype=np.float32) \
                for spc in self.species}
        # The assimilation mask
        self.outmask = {spc : np.ma.empty((0,self.nmaxlevs),dtype=np.int32) \
                for spc in self.species}
        
        self.press.fill_value = self.missing_value
        for spc in self.species:
            self.prof[spc].fill_value = self.missing_value
            self.cov[spc].fill_value = self.missing_value
            self.avk[spc].fill_value = self.missing_value
            
        # Units and conversion factor for plots
        self.units = {spc: "ppb" \
                if spc[:2] != "PM" else "mug/m3" for spc in self.species}
        self.cnvfctr = {spc: 1e9 for spc in self.species}
#        
#    def __mul__(self,other):

    def Read(self, filename, **kwargs):

        if filename.split('.')[-1] == 'h5':
            self.Readh5(filename, **kwargs)
        elif filename.split('.')[-1] == 'nc':
            self.Readnc(filename, **kwargs)
        else:
            print('Format not recognized')

    def Readh5(self, filename, closefile=True, **kwargs):
        """ Reads multispecies observations from MOCAGE HDF5

        If ReadAssim is True, reads assimilation output as well.

        Parameters
        ----------
        filename : str
            Name of the file to read.
        closefile : bool
            Define wether to close the h5 file after reading.
        ReadAssim : dict
            If ReadAssim argument is provided, the file is considered to be a
            HSTAT (a h5 file resulting from MOCAGE assimilation or validation
            run). In this case, 'ReadAssim' must be a dictionary with two
            keys:

                - 'Region' : name of the domain.
                - 'Obstype' : name of the variable to read
                  ("Hx_ana", "Hx_bkg", ...).

        """
        self.OpenFile(filename, mode="r",**kwargs)
        nobs = self.ReadGeoValues(**kwargs)

        if nobs > 0:
            subgrp = ""
            obstype = "Y"
            if 'ReadAssim' in kwargs:
                if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']
                if 'ObsType' in kwargs['ReadAssim']:
                    obstype = kwargs['ReadAssim']['ObsType']
                # Set back to false if I'm reading assimilation output, avoids
                # all unintentional avk and cov usage afterwards
                self.hascov, self.hasavk = False, False

            obsgrp = self.f[self.name+subgrp+"/OBSERVATIONS"]
            geogrp = self.f[self.name+subgrp+"/GEOLOCALIZATION"]
            # Init the species with those found in the HDF5 if not done before
            if not hasattr(self, "nspc"):
                species=[el for el in obsgrp.keys()]
                self.nspc = len(species)
                self.species = species
            # Init the number of levels with hdf5 attribute
            if 'ReadAssim' in kwargs:
                if not hasattr(self, "press"):
                    self.nmaxlevs = geogrp["Pressure"][:].shape[1]
                    self.press_isconst = (geogrp["Pressure"][:].shape[0] == 1)
                    self.InitArrays()
                    press0 = geogrp["Pressure"][1].tolist()
                    press_is_sorted = (sorted(press0) == press0)
                    self.press_coorddir = ('downward' if press_is_sorted
                                       else 'upward')
            else:
                if not hasattr(self, "press"):
                    self.nmaxlevs = geogrp["Pressure"].attrs['MaxNumLevels']
                    self.press_isconst = bool(
                        geogrp["Pressure"].attrs['IsGridConstant']
                    )
                    self.InitArrays()
                    self.press_coorddir = geogrp["Pressure"].attrs['Orientation']
                    self.press_coorddir = self.press_coorddir.decode()

            # Read the observations
            if self.press_isconst and self.press.size > 0:
                # Don't read constant pressure more than once
                pass
            else:
                # I don't know why but fill value needed to be forced back
                # after concatenation (numpy bug ?)
                press = np.ma.masked_values(
                    np.ma.array(geogrp["Pressure"][:], dtype=self.press.dtype),
                    self.missing_value,
                    rtol=1e-03,
                )
                self.press = np.ma.array(
                    np.ma.concatenate((self.press, press)),
                    fill_value=self.missing_value,
                )

            for spc in self.species:
                print(">>> Reading {}, found {} obs".format(
                    spc, len(obsgrp[spc+"/"+obstype])))
                prof = np.ma.masked_values(
                    np.ma.array(obsgrp[spc+"/"+obstype][:],
                                dtype=self.prof[spc].dtype),
                    self.missing_value,
                    rtol=1e-03,
                )
                # I don't know why but fill value needed to be forced back
                # after concatenation (numpy bug ?)
                self.prof[spc] = np.ma.array(
                    np.ma.concatenate((self.prof[spc], prof)),
                    fill_value=self.missing_value,
                )
                if 'ReadAssim' in kwargs:
                    # Assimilation mask from daimon output
                    if obstype in ["Hx_ana", "Hx_bkg", "Hx_dry"]:
                        self.outmask[spc] = np.ma.concatenate(
                            (self.outmask[spc],
                             np.ma.array(obsgrp[spc+"/Mask_"+obstype[-3:]][:],
                                         dtype=self.outmask[spc].dtype))
                        )
                if self.hascov:
                    cov = np.ma.masked_values(
                        np.ma.array(obsgrp[spc+"/ErrorCovariance"][:],
                                    dtype=self.cov[spc].dtype),
                        self.missing_value,
                        rtol=1e-03,
                    )
                    self.cov[spc] = np.ma.array(
                        np.ma.concatenate((self.cov[spc], cov)),
                        fill_value=self.missing_value,
                    )
                if self.hasavk:
                    avk = np.ma.masked_values(
                        np.ma.array(obsgrp[spc+"/AveragingKernels"][:],
                                    dtype=self.cov[spc].dtype),
                        self.missing_value,
                        rtol=1e-03,
                    )
                    self.avk[spc] = np.ma.array(
                        np.ma.concatenate((self.avk[spc], avk)),
                        fill_value=self.missing_value
                    )

        if closefile:
            self.CloseFile()

    def Readnc(self, filename, closefile=True, **kwargs):

        self.OpenFile(filename,mode="r",ftype='nc')
        nobs = self.ReadNCGeoValues(**kwargs)

        if nobs > 0:
            subgrp = ""
            obstype = "Y"
            if 'ReadAssim' in kwargs.keys():
                if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']
                if 'ObsType' in kwargs['ReadAssim']: obstype = kwargs['ReadAssim']['ObsType'] 
                # Set back to false if I'm reading assimilation output, avoids
                # all unintentional avk and cov usage afterwards
                self.hascov, self.hasavk = False, False 
            
            instrgrp = self.f[self.name]
            obsgrp = self.f[self.name+subgrp+"/OBSERVATIONS"]

            # Init the species with those found in the HDF5 if not done before
            if not hasattr(self,"nspc"):
                self.species = [el for el in obsgrp.variables]
                self.nspc = len(self.species)

            # Init the number of levels with hdf5 attribute
            if 'ReadAssim' in kwargs:
                if not hasattr(self, "press"):
                    self.nmaxlevs = instrgrp['level'].shape[0]
                    self.press_isconst = instrgrp['level'].IsGridConstant
                    self.InitArrays()
                    self.press_coorddir = instrgrp['level'].Orientation
            else:
                if not hasattr(self, "press"):
                    self.nmaxlevs = instrgrp['level'].shape[0]
                    self.press_isconst = instrgrp['level'].IsGridConstant
                    self.InitArrays()
                    self.press_coorddir = instrgrp['level'].Orientation

            # Read the observations
            if self.press_isconst and self.press.size > 0:
                # Don't read constant pressure more than once
                pass
            else:
                # I don't know why but fill value needed to be forced back
                # after concatenation (numpy bug ?)
                press = np.ma.masked_values(
                    np.ma.array(instrgrp["level"][:], dtype=self.press.dtype),
                    self.missing_value,
                    rtol=1e-03,
                )
                self.press = np.ma.array(
                    np.ma.concatenate((self.press, np.ma.reshape(press,(1,self.nmaxlevs)))),
                    fill_value=self.missing_value,
                )

            # Read the observations
            for spc in self.species:
                print(">>> Reading %s, type %s, found %i obs" % (spc,obstype,len(obsgrp[spc])))
                prof = np.ma.masked_values(np.ma.array(obsgrp[spc][:],dtype=self.prof[spc].dtype),
                    self.missing_value,
                    rtol=1e-03)
                print(prof.shape)
                self.prof[spc] = np.ma.array(np.ma.concatenate((self.prof[spc],prof)),fill_value=self.missing_value)
                print(self.prof[spc].shape)

            if 'ReadAssim' in kwargs: 
                # Assimilation mask from daimon output
                if obstype in ["Hx_ana","Hx_bkg"]:
                    self.assim_mask[spc] = np.ma.concatenate((self.assim_mask[spc],np.ma.array(obsgrp[spc+"/Mask_"+obstype[-3:]][:][:,0],
                        dtype=self.assim_mask[spc].dtype)))

        if closefile:
            self.CloseFile()

    def ConvertMocageASCII(self,date,inputdir):
        """
        Reads monospecies observations from old MOCAGE ASCII
        """                          
        
        spc = self.species[0] # Old ASCII are monospecies, so take only one if multiple species are present (the first)!
        
        # Use the old MOCAGE HDAT reader for now
        obstype, obsdate, obsname, obslat, obslon, obsval, obspress = \
            UtilsASCII.LecHOBS(os.path.join(inputdir, \
            'HDAT+'+date.isoformat()[:10].replace("-","")),obstype='profile')
            
        nobs = len(obsdate)
            
        if  os.path.isfile(os.path.join(inputdir,'HCOV+'+date.isoformat()[:10].replace("-",""))): 
            self.hascov = True
            ostypetmp, datetmp, nametmp, lattmp, lontmp, obscov, presstmp  =  \
                UtilsASCII.LecHCOV(os.path.join(inputdir, \
                'HCOV+'+date.isoformat()[:10].replace("-","")))
            npcov = np.full((nobs,self.nmaxlevs,self.nmaxlevs),self.missing_value,dtype=self.cov[spc].dtype)
            
        # Conversion for AVK not yet done since I do not have any old profile-avk instrument
            
        obsname = [n[0]+n[1] for n in obsname] # Reduce old strings found in HDATs to a single site code
        obsdate = [datetime.strptime(t,"%Y%m%d%H%M") for t in obsdate] # Datetime conversion
        
        # Convert observation lists to numpy arrays
        
        nppress = np.full((nobs,self.nmaxlevs),self.missing_value,dtype=self.press.dtype)
        npval = np.full((nobs,self.nmaxlevs),self.missing_value,dtype=self.prof[spc].dtype)
        
        for ip in range(nobs):
            nppress[ip,:len(obspress[ip])] = obspress[ip]
            npval[ip,:len(obspress[ip])] = obsval[ip]
            if self.hascov: npcov[ip,:len(obspress[ip]),:len(obspress[ip])] = obscov[ip]
        
        self.nobs += nobs # Increase the instrument counter
        
        if nobs > 0:
            # I don't know why but fill value needed to be forced back after concatenation (numpy bug)
            self.press = np.ma.array(np.ma.concatenate((self.press,np.ma.masked_values(np.ma.array(nppress),self.missing_value))),fill_value=self.missing_value)
            self.prof[spc] = np.ma.array(np.ma.concatenate((self.prof[spc],np.ma.masked_values(np.ma.array(npval),self.missing_value))),fill_value=self.missing_value)
            self.lats = np.ma.concatenate((self.lats,np.ma.array(obslat,dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate((self.lons,np.ma.array(obslon,dtype=self.lons.dtype)))
            self.time = np.ma.concatenate((self.time,np.ma.array(obsdate,dtype=self.time.dtype)))
            if self.hascov: self.cov[spc] = np.ma.array(np.ma.concatenate((self.cov[spc],np.ma.masked_values(np.ma.array(npcov),self.missing_value))),fill_value=self.missing_value)
            self.mask = np.concatenate((self.mask,np.zeros((len(obsdate)),dtype=bool))) # By default all obs are accepted
        
    def Write(self,filename,closefile=True,**kwargs):
        """
        Writes multispecies observations in MOCAGE HDF5
        """                                   
        # Prepare file
        self.OpenFile(filename)
        self.WriteHeader()
        self.WriteConcHeader()
        
        time_dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset("Timestamp",(self.nobs,),dtype=self.h5_typetime)
        
        h5arrtype = np.dtype('(%i,)f' % (self.nmaxlevs))
        h5matrixtype = np.dtype('(%i,%i,)f' % (self.nmaxlevs,self.nmaxlevs))
        
        # Pressure dataset
        if self.press_isconst:
            dset = self.f[self.name+"/GEOLOCALIZATION"].create_dataset('Pressure',(1,),dtype=h5arrtype,maxshape=(None,))
        else:
            dset = self.f[self.name+"/GEOLOCALIZATION"].create_dataset('Pressure',(self.nobs,),dtype=h5arrtype,maxshape=(None,))
            
        # Measurement attributs
        dset.attrs['MaxNumLevels'] = self.nmaxlevs
        dset.attrs['Units'] = np.string_('Pa')
        dset.attrs['Orientation'] = np.string_(self.press_coorddir)
        dset.attrs['IsGridConstant'] = np.int32(self.press_isconst)
        dset.attrs['MissingValue'] = self.missing_value
        self.attach_scale_2_dataset(dset,time_dset,'t','Timestamp')
        
        # Profiles dataset
        for spc in self.species:
            dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset('Y',(self.nobs,),dtype=h5arrtype,maxshape=(None,))
            dset.attrs['MaxNumLevels'] = self.nmaxlevs
            dset.attrs['Units'] = np.string_('vmr')
            dset.attrs['MissingValue'] = self.missing_value
            self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
            
            # Covariance dataset (optional)
            if self.hascov:
                dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset( \
                              'ErrorCovariance',(self.nobs,),dtype=h5matrixtype,maxshape=(None,))
                dset.attrs['MaxNumLevels'] = self.nmaxlevs
                dset.attrs['Units'] = np.string_('vmr-2')
                dset.attrs['MissingValue'] = self.missing_value
                self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
            # Avk dataset (optional)
            if self.hasavk:
                dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset( \
                              'AveragingKernels',(self.nobs,),dtype=h5matrixtype,maxshape=(None,))
                dset.attrs['MaxNumLevels'] = self.nmaxlevs
                dset.attrs['Units'] = np.string_('Unitless')
                dset.attrs['MissingValue'] = self.missing_value
                self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        
        if self.nobs > 0:
            # Write actual data in the file
            self.WriteGeoValues()
            
            if self.press_isconst:
                dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset("Pressure",(1,),dtype=h5arrtype)
                dset[...] = self.press.filled()
            else:
                dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset("Pressure",(self.nobs,),dtype=h5arrtype)
                dset[...] = np.array([val for val in self.press[~self.mask,...].filled()],dtype=self.press.dtype)[self.sorted_idx]
            
            for spc in self.species:
                dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset("Y",(self.nobs,),dtype=h5arrtype)
                dset[...] = np.array([val for val in self.prof[spc][~self.mask,...].filled()],dtype=self.prof[spc].dtype)[self.sorted_idx]
                 # Covariance dataset (optional)
                if self.hascov:
                    dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset("ErrorCovariance",(self.nobs,),dtype=h5matrixtype)
                    dset[...] = np.array([val for val in self.cov[spc][~self.mask,...].filled()],dtype=self.cov[spc].dtype)[self.sorted_idx]
                # Avk dataset (optional)
                if self.hasavk:
                    dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset("AveragingKernels",(self.nobs,),dtype=h5matrixtype)
                    dset[...] = np.array([val for val in self.avk[spc][~self.mask,...].filled()],dtype=self.avk[spc].dtype)[self.sorted_idx]

        if closefile:
            self.CloseFile()
            print(">>> HDF5 written ")

    def Select(self,**kwargs):
        
        if self.nobs > 0:
            
            self.SelectGeo(**kwargs)
            
            if not self.press_isconst: self.press[self.mask,:] = np.ma.masked
            
            # Mask the entire profile for the geographically masked points
            for spc in self.species: 
                self.prof[spc][self.mask,:] = np.ma.masked
                if self.hascov: self.cov[spc][self.mask,:,:] = np.ma.masked
                if self.hasavk: self.avk[spc][self.mask,:,:] = np.ma.masked
                
                # Selection of vertical levels based on user requested bounds
                if 'PressBnd' in kwargs:
                    pmax = kwargs['PressBnd']['max']
                    pmin = kwargs['PressBnd']['min']
                    if self.press_isconst:
                        pmask = ((self.press[0,:] < pmin) | (self.press[0,:] > pmax))
                        if pmask.any():
                            #self.press[0,pmask] = np.ma.masked #Keep the original pressure array to avoid some nasty behavior
                            self.prof[spc][:,pmask] = np.ma.masked
                            if self.hasavk: 
                                self.avk[spc][:,:,pmask] = np.ma.masked
                            if self.hascov: 
                                self.cov[spc][:,pmask,:] = np.ma.masked
                                self.cov[spc][:,:,pmask] = np.ma.masked
                    else:
                        self.press = np.ma.masked_outside(self.press,pmin,pmax)
                        self.prof[spc].mask = self.press.mask
                        print("Not coded yet")
                        sys.exit()
                        #if self.hasavk: self.avk[spc].mask = 
                        #if self.hascov: self.cov[spc].mask =                    
            
    def Average(self,AvgType='global',**kwargs):
        # Vertical obs averager
        # avgpress and avgprof are added to the data structure
        # to be used for further processing or plotting
        # (previously existent averages are erased on this call)
        
        # Default average pressure grid
        if self.press_isconst: 
            self.avgpress = np.ma.copy(self.press[0,:])
        else: 
            # Default array for regridding
            self.avgpress = np.ma.array([ 101325., 100543., 98588., 95744., 92246., 88280., 83995., 79509., 74912., 70273., \
                          65643., 61060., 56554., 52146., 47854., 43695., \
                          39681., 35828., 32150., 28660., 25371., 22294., \
                          19436., 16795., 14384., 12204., 10205.,  8518., \
                          6997.,  5673.,  4529.,  3551.,  2726.,  2040.,  \
                          1481.,  1037., 695., 440.7, 261.1, 142., 69., 29., 10.],dtype=self.press.dtype,fill_value=self.missing_value)
        
        if 'AverageOpt' in kwargs:
            if 'PressureGrid' in kwargs['AverageOpt']: self.avgpress = np.ma.array(kwargs['AverageOpt']['PressureGrid'],\
                dtype=self.press.dtype,fill_value=self.missing_value)

        # Vertical averages have pressure axes oriented downward
        self.avgpress.sort()
        
        # Compute time series of average profiles
        if AvgType == 'timeseries':
            
            avgmsk = self.TimeAveragerMask()
            
            self.avgprof = {spc : np.ma.empty((self.avgtime.size,self.avgpress.size),dtype=self.prof[spc].dtype,\
                fill_value=self.missing_value) for spc in self.species}
            self.stdprof = {spc : np.ma.empty((self.avgtime.size,self.avgpress.size),dtype=self.prof[spc].dtype,\
                fill_value=self.missing_value) for spc in self.species}
            self.avgnobs = {spc : np.zeros((self.avgtime.size,self.avgpress.size),dtype=np.int32) \
                for spc in self.species}
            if self.hascov: 
                self.avgcov = {spc : np.ma.empty((self.avgtime.size,self.avgpress.size),dtype=self.cov[spc].dtype,\
                fill_value=self.missing_value) for spc in self.species}
            if self.hasavk: 
                self.avgavk = {spc : np.ma.empty((self.avgtime.size,self.avgpress.size),dtype=self.avk[spc].dtype,\
                fill_value=self.missing_value) for spc in self.species}
            
            for spc in self.species:
                # Init
                self.avgprof[spc][:] = np.ma.masked
                self.stdprof[spc][:] = np.ma.masked
                if self.hascov: self.avgcov[spc][:] = np.ma.masked
                if self.hasavk: self.avgavk[spc][:] = np.ma.masked
                
                for it in range(self.avgtime.size):
                    if avgmsk[it].any(): 
                        # The following is ok only when pressure levels are constant (the pressure grid remain the same)
                        if self.press_isconst: 
                            self.avgnobs[spc][it,:] = self.prof[spc][avgmsk[it],:].count(axis=0)
                            self.avgprof[spc][it,:] = np.ma.mean(self.prof[spc][avgmsk[it],:],axis=0)
                            self.stdprof[spc][it,:] = np.ma.std(self.prof[spc][avgmsk[it],:],axis=0)
                            if self.hascov: self.avgcov[spc][it,:] = np.ma.mean(self.cov[spc][avgmsk[it],:,:],axis=0)
                            if self.hasavk: self.avgavk[spc][it,:] = np.ma.mean(self.avk[spc][avgmsk[it],:,:],axis=0)
                        # The following is for varying pressure levels (average is computed on a common pressure grid)
                        else:
                            regridprofs = self.VertRegrid(self.prof[spc][avgmsk[it],:],self.press[avgmsk[it],:])
                            self.avgnobs[spc] = regridprofs.count(axis=0)
                            self.avgprof[spc][it,:] = np.ma.mean(regridprofs,axis=0)
                            self.stdprof[spc][it,:] = np.ma.std(regridprofs,axis=0)
                            if self.hascov: print("WARNING: I do not know yet how to vertically regrid error covariances")
                            if self.hasavk: print("WARNING: I do not know yet how to vertically regrid averaging kernel")
                            
        # Compute one global average profile
        elif AvgType == 'global':
            
            self.avgprof = {}
            self.stdprof = {}
            self.avgnobs = {}
            if self.hascov: self.avgcov = {}
            if self.hasavk: self.avgavk = {}
            
            for spc in self.species:
                # The following is ok only when pressure levels are constant (the pressure grid remain the same)
                if self.press_isconst: 
                    self.avgnobs[spc] = self.prof[spc].count(axis=0)
                    self.avgprof[spc] = np.ma.mean(self.prof[spc],axis=0)
                    self.stdprof[spc] = np.ma.std(self.prof[spc],axis=0)
                    if self.hascov: self.avgcov[spc] = np.ma.mean(self.cov[spc],axis=0)
                    if self.hasavk: self.avgavk[spc] = np.ma.mean(self.avk[spc],axis=0)
                # The following is for varying pressure levels (average is computed on a common pressure grid)
                else:
                    regridprofs = self.VertRegrid(self.prof[spc],self.press,self.avgpress)
                    self.avgnobs[spc] = regridprofs.count(axis=0)
                    self.avgprof[spc] = np.ma.mean(regridprofs,axis=0)
                    self.stdprof[spc] = np.ma.std(regridprofs,axis=0)
                    if self.hascov: print("WARNING: I do not know yet how to vertically regrid error covariances")
                    if self.hasavk: print("WARNING: I do not know yet how to vertically regrid averaging kernel")
        else:
            print("Error: I do not understand this type of averaging")
            sys.exit()
            
    def VertRegrid(self,profilesin,pressuresin,pressureout):
        """
        Regridder for vertical profiles. Input data is numpy.ma array
        """   
        
        # Use MV arrays locally. Annoying!!!
        profilesin = MV.array(profilesin,copy=1)
        pressuresin = MV.array(pressuresin,copy=1)
        
        if self.press_coorddir == 'upward': 
            profilesin = profilesin[:,::-1]
            pressuresin = pressuresin[:,::-1]
        
        print(">>> Vertical regridding...")
        profilesout = cdutil.vertical.logLinearInterpolation(profilesin,pressuresin,pressureout,axis='1')
        
        # Return a masked array to the caller
        return profilesout(raw=1).transpose()
    
    def ComputeCols(self,**kwargs):
        """
        Computation of total or partial columns for further stats
        """
        columns = {"TROPO":(34000,101500)}
        if "ColsBounds" in kwargs: columns = kwargs["ColsBounds"]
        
        self.columns = {spc : { col : np.ma.empty((0),dtype=np.float32) \
                             for col in columns.keys()} for spc in self.species}
        self.columns["ColsBounds"] = columns
        for spc in self.species: 
            self.columns[spc]["units"] = "DU"
            
        if self.press_isconst:
            if self.press_coorddir == "downward":
                dp = (self.press[0,2:]-self.press[0,:-2])/2
                dp = np.append(dp,(self.press[0,-1]-self.press[0,-2])/2)
                dp = np.insert(dp,0,self.press[0,0]+(self.press[0,1]-self.press[0,0])/2)
            elif self.press_coorddir == "upward":
                dp = -(self.press[0,2:]-self.press[0,:-2])/2
                dp = np.insert(dp,0,(self.press[0,0]-self.press[0,1])/2)
                dp = np.append(dp,self.press[0,-1]+(self.press[0,-2]-self.press[0,-1])/2)
            else:
                print("Pressure array order not defined")
                return 
            for col in columns:
                colidx = (self.press[0,:] > columns[col][0]) & (self.press[0,:] < columns[col][1])
                if colidx.any():
                    for spc in self.species:
                        self.columns[spc][col] = np.ma.sum(self.prof[spc][:,colidx]*dp[colidx],axis=1)*1e6*0.0078914
                        # Mask columns that contained too many missing values at some levels (more than 2/3)
                        self.columns[spc][col][self.prof[spc][:,colidx].count(axis=1) <= colidx.sum()*0.6] = np.ma.masked                  
                else:
                    print("Warning: no pressure interval available for: %s" % (col))
                    return
        else:
           # Could be dangerous to do if the vertical grids differ significantly
           print("Not yet available")  
            
        return
                        
    def PrintObsStats(self,**kwargs):
        
        doseplevs = False
        if 'LevelsSelection' in kwargs and self.press_isconst: 
            levs = kwargs["LevelsSelection"]
            doseplevs = True
            
        print(">>> Profile obs infos:")
        print("Total obs = %i ,\nn = %i are not masked " % (len(self.time),self.nobs))

        for spc in self.species:
            if not doseplevs:
                print(spc+" Stats - ConcMin:%5.3e %s ConcMax:%5.3e %s" % \
                (self.prof[spc].min()*self.cnvfctr[spc],self.units[spc],self.prof[spc].max()*self.cnvfctr[spc],self.units[spc]))
            else:
                for ilev in levs:
                    print(spc+" Stats at level %.1f hPa - ConcMin:%5.3e %s ConcMax:%5.3e %s" % \
                    (self.press[0,ilev]/100,self.prof[spc][:,ilev].min()*self.cnvfctr[spc],self.units[spc],\
                     self.prof[spc][:,ilev].max()*self.cnvfctr[spc],self.units[spc]))
                    
    def PlotObsHisto(self,outdir="./",**kwargs):
        
        figsize = (15,15)
        doseplevs = False
        lkwargs = {}
        
        if 'LevelsSelection' in kwargs and self.press_isconst: 
            levs = kwargs["LevelsSelection"]
            doseplevs = True
        if 'ObsHistoOpt' in kwargs: lkwargs = kwargs["ObsHistoOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        
        print(">>> Plotting profile level values histograms")
        
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_ConcHisto_%s_%s" % \
                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            fig = plt.figure(figsize=figsize,dpi=200)
            
            # Variable related keywords have priority over global keywords
            lvarkwargs = {}

            if spc in lkwargs: 
                lvarkwargs = lkwargs[spc]
            else:
                lvarkwargs = lkwargs
                    
            if "range" in lvarkwargs:
                vmin = lvarkwargs["range"]["min"]
                vmax = lvarkwargs["range"]["max"]
                historange =(vmin,vmax)
            else:
                historange = None
            
            if not doseplevs:
                ax = fig.add_subplot(1,1,1)
                ax.set_title("%s (%s) all levels" % (spc,self.units[spc]))
                ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                          (self.prof[spc].mean()*self.cnvfctr[spc],np.ma.std(self.prof[spc])*self.cnvfctr[spc]))
                ax.hist(self.prof[spc].compressed()*self.cnvfctr[spc],bins=35,density=True,range=historange)
            else:
                nsubplt = int(np.ceil(np.sqrt(len(levs))))
                
                for iplt,ilev in enumerate(levs):
                    ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                    ax.set_title("%s (%s) at level %.1f hPa" % (spc,self.units[spc],self.press[0,ilev]/100))
                    ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                          (self.prof[spc][:,ilev].mean()*self.cnvfctr[spc],np.ma.std(self.prof[spc][:,ilev])*self.cnvfctr[spc]))
                    ax.hist(self.prof[spc][:,ilev].compressed()*self.cnvfctr[spc],bins=35,density=True,range=historange)
                    
            plt.savefig(filename)
            plt.close(fig)

    def PlotColsHisto(self,outdir="./",**kwargs):
        
        figsize = (15,15)
        lkwargs = {}
        
        if 'ColsHistoOpt' in kwargs: lkwargs = kwargs["ColsHistoOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        
        print(">>> Plotting vertical columns histograms")
        
        if 'columns' not in self.__dict__: self.ComputeCols(**lkwargs)
        
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_ColumnsHisto_%s_%s" % \
                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            fig = plt.figure(figsize=figsize,dpi=200)
            
            # Variable related keywords have priority over global keywords
            lvarkwargs = {}

            if spc in lkwargs: 
                lvarkwargs = lkwargs[spc]
            else:
                lvarkwargs = lkwargs
                                
            nsubplt = int(np.ceil(np.sqrt(len(self.columns["ColsBounds"]))))
            
            for iplt,col in enumerate(self.columns["ColsBounds"].keys()):
                if "range" in lvarkwargs:
                    if col in lvarkwargs["range"]:
                        vmin = lvarkwargs["range"][col]["min"]
                        vmax = lvarkwargs["range"][col]["max"]
                    else:
                        vmin = lvarkwargs["range"]["min"]
                        vmax = lvarkwargs["range"]["max"]
                    historange =(vmin,vmax)
                else:
                    historange = None
                ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                ax.set_title("%s (%s) %s column" % (spc,self.columns[spc]["units"],col))
                ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                      (self.columns[spc][col].mean(),np.ma.std(self.columns[spc][col])))
                ax.hist(self.columns[spc][col].compressed(),bins=35,density=True,range=historange)
            
            plt.subplots_adjust(wspace=0.2,hspace=0.4)        
            plt.savefig(filename)
            plt.close(fig)
        
    def PlotObsValues(self,outdir="./",**kwargs):
        
        if not self.press_isconst: return 
        
        figsize = (15,15)
        markersize = 2
        colormap = "jet"
        lkwargs = {}
        
        levs = range(self.nmaxlevs)
        if 'LevelsSelection' in kwargs: levs = kwargs["LevelsSelection"]
            
        if 'ObsOnMapOpt' in kwargs: lkwargs = kwargs["ObsOnMapOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
        
        print(">>> Plotting profile level values on a map")
        
        for isp,spc in enumerate(self.species):

            fig = plt.figure(figsize=figsize,dpi=200)
            
            filename = os.path.join(outdir,"%s_%s_ConcMap_%s_%s" % \
                                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            nsubplt = int(np.ceil(np.sqrt(len(levs))))
            
            for iplt,ilev in enumerate(levs):
                
                ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                
                m = self.SetBaseMap(ax,**kwargs)

                fld = self.prof[spc][:,ilev]*self.cnvfctr[spc]
                
                if fld.count() > 1:
                    
                    vmin = fld.min()
                    vmax = fld.max()
                    
                    # Variable related keywords have priority over global keywords
                    lvarkwargs = {}

                    if spc in lkwargs: 
                        lvarkwargs = lkwargs[spc]
                    else:
                        lvarkwargs = lkwargs
                    
                    if "range" in lvarkwargs:
                        vmin = lvarkwargs["range"]["min"]
                        vmax = lvarkwargs["range"]["max"]
                        
                    # Shrink to the common mask that considers missing vertical points    
                    lmask = self.mask | fld.mask
                    ax.set_title('%s N=%i %.1f hPa\n Avg: %.2f Min: %.2f Max: %.2f (%s)' % (spc,(~lmask).sum(),self.press[0,ilev]/100,\
                             fld.mean(),fld.min(),fld.max(),self.units[spc]))
                    
                    sc = m.scatter(self.lons[~lmask],self.lats[~lmask],markersize,marker='o',\
                                   c=fld[~lmask],cmap=colormap,latlon=True,zorder=10,vmin=vmin,vmax=vmax)
    
                    plt.colorbar(sc,orientation='horizontal')
                
            fig.savefig(filename)
            plt.close(fig)
            
    def PlotVertProfiles(self,outdir="./",**kwargs):
        
        figsize = (20,20)
        nmaxplts = 24
        lkwargs = {}
        if 'VertProfilesOpt' in kwargs:
            lkwargs = kwargs["VertProfilesOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        
        print(">>> Plotting vertical profiles")
        
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_ConcProfiles_%s_%s" % \
                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            fig = plt.figure(figsize=figsize,dpi=200)
            
            npltsxpage = min(self.nobs,nmaxplts)
            nsubplt = int(np.ceil(np.sqrt(npltsxpage)))
            
            for iplot,ip in enumerate((~self.mask).nonzero()[0]):
                
                if iplot > npltsxpage-1: break
            
                ax = fig.add_subplot(nsubplt,nsubplt,iplot+1)
                ax.set_title("%s profile \n (%.1fN,%.1fE,%s)" % (spc,self.lats[ip],self.lons[ip],self.time[ip].astype(str)[11:16]))
                # ax.set_xlabel("%s" % (self.units[spc]))
                # ax.set_ylabel("%s" % ("hPa"))
                c = 'black'
                lw = 2
                if 'lw' in lkwargs: lw = lkwargs['lw']
                
                if self.press_isconst: 
                    press = self.press[0,:]
                else:
                    press = self.press[ip,:]
                    
                ax.plot(self.prof[spc][ip,:]*self.cnvfctr[spc],press/100,c,lw=lw)
                ax.invert_yaxis()
                if press.max()/press.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
                if 'xlog' in lkwargs: ax.set_xscale('symlog',linthresh=1e-1)
                if 'yrange' in lkwargs: 
                    ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
                if 'xrange' in lkwargs:
                    ax.set_xlim(lkwargs['xrange']["min"],lkwargs['xrange']["max"])
                #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
                #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
                #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
                ax.grid(True,which='minor',linestyle='--')
                ax.tick_params(which='minor',labelsize=7)
                
            plt.subplots_adjust(wspace=0.2,hspace=0.3)
            plt.savefig(filename)
            plt.close(fig)

    def plot_vertical_profiles(
            self, others=[], outdir="./", figsize=(8, 8), nmaxplts=np.inf,
            colors=['k'], labels=None, linestyles=['-'], title=None,
            xlabel=None, xlog=False, ylog=False,
            xlim={'min': None, 'max': None}, ylim={'min': None, 'max': None},
            filename=None, **kwargs):
        """ Plot vertical profiles.

        Draw vertical profiles for one or several colocated instruments.

        Parameters
        ----------
        others : list of ProfConcInstrument objects
            Other instruments to display on the same figure.
        outdir : str, optional
            Output directory where to save the figures.
        figsize : couple of scalars
            Figure size (as define in matplotlib.pyplot.figure).
        nmaxplts : int
            maximum number of plots to produce.
        colors : list
            List of colors corresponding to each curves (self + others).
        labels : list
            List of labels corresponding to each curves (self + others).
        xlog : bool
            If True, x axis scale type is set to log.
        ylog : bool
            If True, y axis scale type is set to log.
        xlim, ylim : dict
            Axis limits given as {'min': ..., 'max': ...}.
        xlabel : None or str
            Labels for x axis.
        filename : None or str
            Filename that can contain '{species}', '{name}' and '{time}'.

        """
        # test if additionnal objects are synchronised with self
        for obj in others:
            if not self.isColocatedWith(obj, consider_mask=True):
                print("Warning: objects not colocated. Profiles are not " +
                      "produced.")
                return

        print(">>> Plotting vertical profiles")
        for spc in self.species:

            for iobs in range(self.nobs):
                if iobs > nmaxplts-1:
                    break
                fig = plt.figure(figsize=figsize)
                ax = plt.gca()

                # set title
                if title is None:
                    plt.title(
                        "{} profile \n ({:.1f}N, {:.1f}E, {})".format(
                            spc,
                            self.lats[~self.mask][iobs],
                            self.lons[~self.mask][iobs],
                            self.time[~self.mask][iobs].astype(str)),
                        fontsize='xx-large',
                    )
                else:
                    if hasattr(self, 'id'):
                        plt.title(
                            title.format(
                                species=spc,
                                id=self.id[~self.mask][iobs],
                                lat=self.lats[~self.mask][iobs],
                                lon=self.lons[~self.mask][iobs],
                                time=self.time[~self.mask][iobs].astype(str)),
                            fontsize='xx-large',
                        )
                    else:
                        plt.title(
                            title.format(
                                species=spc,
                                lat=self.lats[~self.mask][iobs],
                                lon=self.lons[~self.mask][iobs],
                                time=self.time[~self.mask][iobs].astype(str)),
                            fontsize='xx-large',
                        )

                # set axes labels
                if xlabel is None:
                    plt.xlabel(self.units[spc], fontsize='xx-large')
                else:
                    plt.xlabel(xlabel, fontsize='xx-large')
                plt.ylabel("hPa", fontsize='xx-large')

                # plot
                if labels is None:
                    labels = [obj.name for obj in [self] + others]
                for obj, c, lab, ls in zip([self] + others, cycle(colors), labels,
                                           cycle(linestyles)):
                    if obj.press_isconst:
                        press = obj.press[0, :]
                    else:
                        press = obj.press[~self.mask][iobs, :]
                    plt.plot(
                        obj.prof[spc][~obj.mask][iobs, :].filled(
                            np.nan)*obj.cnvfctr[spc],
                        press/100,
                        c=c,
                        label=lab,
                        linestyle=ls,
                    )
                plt.legend(fontsize='xx-large')

                ax.invert_yaxis()
                if ylog:
                    plt.yscale('log')
                if xlog:
                    plt.xscale('symlog', linthresh=1e-1)
                plt.xlim(xlim["min"], xlim["max"])
                plt.ylim(ylim["min"], ylim["max"])
                plt.grid(True, which='minor', linestyle='--')
                plt.tick_params(which='both', labelsize='large')
                plt.tight_layout()

                # output figure
                if outdir is None:
                    plt.show()
                else:
                    if filename is None:
                        output_path = os.path.join(
                            outdir,
                            "{name}_{species}_ConcProfiles_{time}",
                        )
                    else:
                        output_path = os.path.join(
                            outdir,
                            filename,
                        )
                    if hasattr(self, 'id'):
                        output_path += "_{}".format(self.id[~self.mask][iobs])
                    plt.savefig(
                        output_path.format(
                            name=self.name,
                            species=spc,
                            time=self.time[~self.mask][iobs],
                        )
                    )
                    plt.close(fig)
            
    def PlotAvgVertProfiles(self,outdir="./",**kwargs):
        
        figsize = (16,8)
        lkwargs = {}
        basefilename = "ConcAvgProfile"
        savedata = False
        printdata = False
        if 'VertProfilesOpt' in kwargs:
            lkwargs = kwargs["VertProfilesOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if "filename" in lkwargs: basefilename = lkwargs["filename"]
        if "savedata" in lkwargs: savedata = lkwargs["savedata"]
        if "printdata" in lkwargs: printdata = lkwargs["printdata"]

        print(">>> Plotting average profile")
        
        if 'avgpress' not in self.__dict__: self.Average(**kwargs) 
        # Done to avoid recomputing averages. Impacts the behavior when diffing obs. 
        # Now diff(avg(obs)) prevails over avg(diff(obs)) unless averaging was not done before diffs
            
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_%s_%s_%s" % \
                (self.name,spc,basefilename,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            if savedata:
                np.savez(filename,\
                avgprof=self.avgprof[spc].filled(),avgpress=self.avgpress.filled(),\
                stdprof=self.stdprof[spc].filled(),avgnobs=self.avgnobs[spc],\
                species=spc,missing_value=self.missing_value)

            if printdata:
                np.savetxt(filename+".txt",np.array([self.avgpress.filled()/100,self.avgprof[spc].filled()]).T,\
                           delimiter=' , ',fmt=('%5.2f','%.3e'))
                
            # Avoid plotting very unrepresentative data
            vmsk = self.avgnobs[spc] >= max(self.avgnobs[spc].max()/3,2)
            
            fig = plt.figure(figsize=figsize,dpi=200)

            ax = fig.add_subplot(131)
            ax.set_title("%s average profile \n" % (spc))
            ax.set_xlabel("%s" % (self.units[spc]))
            ax.set_ylabel("%s" % ("hPa"))
            c = 'black'
            lw = 2
            if 'lw' in lkwargs: lw = lkwargs['lw']
            ax.plot(self.avgprof[spc][vmsk]*self.cnvfctr[spc],self.avgpress[vmsk]/100,c,lw=lw)
            ax.invert_yaxis()
            if self.avgpress.max()/self.avgpress.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
            if 'xlog' in lkwargs: ax.set_xscale('symlog',linthresh=1e-1)
            if 'yrange' in lkwargs: 
                ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
            if 'xrange' in lkwargs:
                ax.set_xlim(lkwargs['xrange']["min"],lkwargs['xrange']["max"])
            #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
            #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
            #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
            ax.grid(True,which='minor',linestyle='--')
            ax.tick_params(which='minor',labelsize=7)
            
            ax = fig.add_subplot(132)
            ax.set_title("%s profile standard deviation \n" % (spc))
            ax.set_xlabel("%s" % (self.units[spc]))
            ax.set_ylabel("%s" % ("hPa"))
            c = 'black'
            lw = 2
            if 'lw' in lkwargs: lw = lkwargs['lw']
            ax.plot(self.stdprof[spc][vmsk]*self.cnvfctr[spc],self.avgpress[vmsk]/100,c,lw=lw)
            ax.invert_yaxis()
            if self.avgpress.max()/self.avgpress.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
            if 'xlog' in lkwargs: ax.set_xscale('symlog',linthresh=1e-1)
            if 'yrange' in lkwargs: 
                ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
            if 'xrange' in lkwargs:
                ax.set_xlim(max(0,lkwargs['xrange']["min"]),lkwargs['xrange']["max"])
            #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
            #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
            #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
            ax.grid(True,which='minor',linestyle='--')
            ax.tick_params(which='minor',labelsize=7)
            
            ax = fig.add_subplot(133)
            ax.set_title("Number of %s observations \n" % (spc))
            ax.set_xlabel("N")
            ax.set_ylabel("%s" % ("hPa"))
            c = 'black'
            lw = 2
            ax.plot(self.avgnobs[spc],self.avgpress/100,c,lw=lw)
            ax.invert_yaxis()
            if self.avgpress.max()/self.avgpress.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
            if 'yrange' in lkwargs: 
                ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
            ax.grid(True,which='minor',linestyle='--')
            ax.tick_params(which='minor',labelsize=7)
                
            plt.savefig(filename)
            plt.close(fig)            
            
    def PlotAvgAuxProfiles(self,outdir="./",**kwargs):
        
        figsize = (10,10)
        lkwargs = {}
        if 'VertProfilesOpt' in kwargs:
            lkwargs = kwargs["VertProfilesOpt"]
            if "figsize" in lkwargs: figsize = lkwargs["figsize"]
            
        print(">>> Plotting avk and error covariances average profiles")
            
        if 'avgpress' not in self.__dict__: self.Average(**kwargs)
        # Done to avoid recomputing averages. Impacts the behavior when diffing obs. 
        # Now diff(avg(obs)) prevails over avg(diff(obs)) unless averaging was not done before diffs
            
        for isp,spc in enumerate(self.species):
            
            filename = os.path.join(outdir,"%s_%s_AuxAvgProfile_%s_%s" % \
                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            fig = plt.figure(figsize=figsize,dpi=200)

            if self.hascov:
                ax = plt.subplot2grid((8, 2), (0, 0), rowspan=6)
                ax.set_title("%s average error covariance \n" % (spc))
                ax.set_xlabel("%s" % (self.units[spc]))
                ax.set_ylabel("%s" % ("hPa"))
                c = 'black'
                lw = 2
                if 'lw' in lkwargs: lw = lkwargs['lw']
                ax.plot(1/np.ma.sqrt(np.ma.diag(self.avgcov[spc]))*self.cnvfctr[spc],self.avgpress/100,c,lw=lw)
                ax.invert_yaxis()
                if self.avgpress.max()/self.avgpress.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
                if 'xlog' in lkwargs: ax.set_xscale('symlog',linthresh=1e-1)
                if 'yrange' in lkwargs: 
                    ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
                if 'xrange' in lkwargs:
                    ax.set_xlim(0,lkwargs['xrange']["max"])
                #ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,50,70,150,250,500,750]))
                #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
                #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
                ax.grid(True,which='minor',linestyle='--')
                ax.tick_params(which='minor',labelsize=7)
            
            if self.hasavk:
                ax = plt.subplot2grid((8, 2), (0, 1), rowspan=6)
                ax.set_title("%s average averaging kernels \n" % (spc))
                ax.set_ylabel("%s" % ("hPa"))
                c = 'black'
                lw = 2
                if 'lw' in lkwargs: lw = lkwargs['lw']
                cmap = mpl.cm.gist_ncar_r
                colors = cmap(np.linspace(0,1,self.nmaxlevs))
                for iz in range(self.nmaxlevs):
                    ax.plot(self.avgavk[spc][:,iz],self.avgpress/100,color=colors[iz],lw=lw)
                ax.invert_yaxis()
                if self.avgpress.max()/self.avgpress.min() >= 20 or 'ylog' in lkwargs: ax.set_yscale('log')
                #if 'xlog' in lkwargs: ax.set_xscale('symlog',linthresh=1e-1)
                if 'yrange' in lkwargs: 
                    ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
                #if 'xrange' in lkwargs:
                ax.set_xlim(-0.25,0.3)
                ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%d'))
                ax.yaxis.set_minor_locator(plt.FixedLocator([2,3,5,7,20,30,50,70,150,200,300,400,500,750]))
                #ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.1f'))
                #ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.1f'))
                ax.grid(True,which='minor',linestyle='--')
                ax.tick_params(which='minor',labelsize=7)

                ax = plt.subplot2grid((8, 2), (7, 1))
                norm = mpl.colors.Normalize(vmin=0, vmax=self.nmaxlevs-1)
                cbar = mpl.colorbar.ColorbarBase(ax, cmap=cmap,
                                               norm=norm, ticks=[9,13,16,21,26,31,35],
                                               orientation='horizontal')
                cbar.ax.set_xticklabels(['20','55', '100','220','400','600','800']) 
                
            plt.savefig(filename)
            plt.close(fig)
    
    def PlotAuxValues(self,outdir="./",**kwargs):
        
        figsize = (12,8)
        markersize = 2
        colormap = "jet"
        lkwargs = {}
       
        print(">>> Plotting avk and error covariances on a map")
        
        if 'ObsOnMapOpt' in kwargs: lkwargs = kwargs["ObsOnMapOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
            
        for isp,spc in enumerate(self.species):

            fig = plt.figure(figsize=figsize,dpi=200)
            
            filename = os.path.join(outdir,"%s_%s_AuxMap_%s_%s" % \
                                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))

            if self.hasavk:
                
                ax = fig.add_subplot(1,2,1)
                    
                m = self.SetBaseMap(ax,**kwargs)
    
                fld = np.ma.diagonal(np.moveaxis(self.avk[spc][~self.mask,:,:], 0, -1)).sum(axis=1)
                vmin = fld.min()
                vmax = fld.max()
                
                ax.set_title("%s Avk Trace \n Avg: %.2f,Min: %.2f,Max: %.2f" % (spc,fld.mean(),vmin,vmax))
                
                # Variable related keywords have priority over global keywords
                lvarkwargs = {}
                if "AvkTrace" in lkwargs: 
                    lvarkwargs = lkwargs["AvkTrace"]
                else:
                    lvarkwargs = lkwargs
                
                if "range" in lvarkwargs:
                    vmin = lvarkwargs["range"]["min"]
                    vmax = lvarkwargs["range"]["max"]
                
                sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o', \
                           c=fld.compressed(),\
                           cmap=colormap,vmin=vmin,vmax=vmax,\
                           latlon=True,zorder=10)
    
                plt.colorbar(sc,orientation='horizontal')
                
            ax = fig.add_subplot(1,2,2)
                
            ax.set_title("Retrieved levels")
                
            m = self.SetBaseMap(ax,**kwargs)

            fld = np.ma.count(self.prof[spc][~self.mask,:],axis=1)
            vmin = fld.min()
            vmax = fld.max()
            
            ax.set_title("%s Retrieved levels \n Avg: %.1f,Min: %2i,Max: %2i" % (spc,fld.mean(),vmin,vmax))
            
            # Variable related keywords have priority over global keywords
            lvarkwargs = {}

            if "RetrievedLevels" in lkwargs: 
                lvarkwargs = lkwargs["RetrievedLevels"]
            else:
                lvarkwargs = lkwargs
            
            if "range" in lvarkwargs:
                vmin = lvarkwargs["range"]["min"]
                vmax = lvarkwargs["range"]["max"]
            
            sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o', \
                       c=fld,\
                       cmap=colormap,vmin=vmin,vmax=vmax,\
                       latlon=True,zorder=10)

            plt.colorbar(sc,orientation='horizontal')
                
            fig.savefig(filename)
            plt.close(fig)
            
    def PlotMaskValues(self,outdir="./",**kwargs):
        
        if not self.press_isconst: return
        
        print(">>> Plotting assimilation mask on a map")
        
        figsize = (15,15)
        markersize = 2
        colormap = "Set1"
        lkwargs = {}
        
        levs = range(self.nmaxlevs)
        if 'LevelsSelection' in kwargs: levs = kwargs["LevelsSelection"]
            
        if 'MaskOnMapOpt' in kwargs: lkwargs = kwargs["MaskOnMapOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
            
        for isp,spc in enumerate(self.species):

            if self.outmask[spc].size == 0: return
            
            fig = plt.figure(figsize=figsize,dpi=200)
            
            filename = os.path.join(outdir,"%s_%s_MaskMap_%s_%s" % \
                                (self.name,spc,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            nsubplt = int(np.ceil(np.sqrt(len(levs))))
            
            for iplt,ilev in enumerate(levs):
                
                ax = fig.add_subplot(nsubplt,nsubplt,iplt+1)
                
                m = self.SetBaseMap(ax,**kwargs)

                fld = self.outmask[spc][~self.mask,ilev]
                
                if fld.count() > 1:

                    ax.set_title('%s Assim Mask %.1f hPa\n Nmiss: %i,Nrej: %i, Nok: %i' % \
                                 (spc,self.press[0,ilev]/100,\
                                 (fld == -2).sum(),(fld == 0).sum(),(fld == 1).sum()))
                    
                    sc = m.scatter(self.lons[~self.mask],self.lats[~self.mask],markersize,marker='o',\
                                   c=fld,cmap=colormap,latlon=True,zorder=10,vmin=-2,vmax=1)
    
                    plt.colorbar(sc,orientation='horizontal')
                
            fig.savefig(filename)
            plt.close(fig)
        
    # A comprehensive printer for profile concentration stats
    def Stats(self,**kwargs):
        
        if self.nobs > 0:
            self.PrintGeoStats()
            self.PrintObsStats(**kwargs)
        
    # A comprehensive plotter for profile concentration data
    def Plot(self,outdir="./",**kwargs):
    
        sns.set() # Apply seaborn beautificator
        if self.nobs > 0:
            if 'makeplots' in kwargs:
                if 'ObsLoc' in kwargs['makeplots']: self.PlotObsLocation(outdir=outdir,**kwargs)
                if 'NObsTimeSer' in kwargs['makeplots']: self.PlotNobsTSer(outdir=outdir,**kwargs)
                if 'ObsDensity' in kwargs['makeplots']: self.PlotObsDensity(outdir=outdir,**kwargs)
                if 'ObsHisto' in kwargs['makeplots']: self.PlotObsHisto(outdir=outdir,**kwargs)
                if 'ColsHisto' in kwargs['makeplots']: self.PlotColsHisto(outdir=outdir,**kwargs)
                if 'ObsOnMap' in kwargs['makeplots']: self.PlotObsValues(outdir=outdir,**kwargs)
                if 'VertProfiles' in kwargs['makeplots']: self.PlotVertProfiles(outdir=outdir,**kwargs)
                if 'AvgVertProfiles' in kwargs['makeplots']: self.PlotAvgVertProfiles(outdir=outdir,**kwargs)
                if 'AvgAuxProfiles' in kwargs['makeplots']: self.PlotAvgAuxProfiles(outdir=outdir,**kwargs)
                if 'AuxOnMap' in kwargs['makeplots']: self.PlotAuxValues(outdir=outdir,**kwargs)
                if 'MaskOnMap' in kwargs['makeplots']: self.PlotMaskValues(outdir=outdir,**kwargs)
            else:
                self.PlotObsLocation(outdir=outdir,**kwargs)
                self.PlotObsHisto(outdir=outdir,**kwargs)
                self.PlotObsValues(outdir=outdir,**kwargs)
        else:
            print('>> No obs to be plotted')

    def MakeSuperObs(self):
        pass

    def plot_time_series(self, others, outdir=None, filename=None, colors=[],
                         labels=[], markers=[]):
        """ Plot the timeseries of concentration values.

        Two figures are displayed. The one above shows the pressure function
        of the time, and the one below shows the concentration values.

        Parameters
        ----------
        others : list of ProfConcInstrument objects
            Other instruments to display on the same figure.
        outdir : str
            Path where to save the figure. If None, the figure is displayed
            in a popping window.
        filename : str
            Name of the ouput file.
        colors, labels, markers : list
            List of colors/labels/markers corresponding to each
            curves (self + others).

        """
        # check that objects has only one vertical level
        if self.nmaxlevs > 1:
            raise Exception("plot_time_series must be used with object with " +
                            "only one vertical level.")
        # test if additionnal objects are synchronised with self
        for obj in others:
            assert(self.isColocatedWith(obj))
        # define default parameters
        if not colors:
            colors = [None for i in range(len(others) + 1)]
        if not labels:
            labels = [obj.name for obj in [self] + others]
        if not markers:
            markers = [None for i in range(len(others) + 1)]

        # plot
        fig, ax = plt.subplots(
            self.nspc + 1,
            1,
            sharex=True,
            figsize=(7, 3*(self.nspc + 1)),
        )
        # Pression
        ax[0].invert_yaxis()
        ax[0].plot(self.time, self.press/100.0, linewidth=1)
        ax[0].set(xlabel='time', ylabel='pressure (hPa)')
        ax[0].grid()
        for i, spe in enumerate(self.species):
            # Variable
            for obj, c, l, m in zip([self] + others, colors, labels, markers):
                ax[i+1].plot(
                    obj.time,
                    obj.prof[spe].squeeze().filled(np.nan)*obj.cnvfctr[spe],
                    c=c, label=l, marker=m,
                )
            ax[i+1].set(
                xlabel='time',
                ylabel="{} ({})".format(spe, obj.units[spe]),
            )
            ax[i+1].grid()
            ax[i+1].legend()
        fig.autofmt_xdate(rotation=45, ha='right')

        # save figure
        plt.tight_layout()
        if filename is None:
            filename = "{}_{}_timeseries.png".format(
                spe, self.time[0].astype(datetime).strftime("%Y-%m-%d"))
        plt.savefig(os.path.join(outdir, filename))
        plt.close()

    def vertical_regrid(self, z, interp):
        """ Interpolate data on a new vertical scale.

        Interpolation type is linear.

        Parameters
        ----------
        z : 1D array-like
            If interp is 'linear' or 'loglinear', z is the list of vertical
            levels for interpolation.
            If interp is 'layerAverage', z is the list of layers boundaries
            (e.g. if z = [1, 2, 4, 8], resulting vertical levels will be
            [1.5, 3.0, 6.0] and interpolated values will correspond to the mean
            of values contained in layers [(1, 2), (2, 4), (4, 8)]).
        interp : str
            Type of interpolation among:

              * 'layerAverage'
              * 'linear'
              * 'loglinear'

        """
        # make sure z is a list
        z = list(z)
        z.sort()

        # define layers in for layer-averaging interpolation
        layers = None
        if interp == 'layerAverage':
            # write layers as tuple
            layers = [(l, h) for l, h in zip(z[:-1], z[1:])]
            # find center of layers
            z = [np.mean((l, h)) for l, h in zip(z[:-1], z[1:])]

        # save old levels and profiles
        levels = copy.deepcopy(self.press)
        levels.fill_value = np.nan
        profiles = copy.deepcopy(self.prof)
        for spe in self.species:
            profiles[spe].fill_value = np.nan

        # reinitialize self
        self.nmaxlevs = len(z)
        self.InitArrays()
        self.press_coorddir = 'downward'
        self.press = np.ma.concatenate(
            (self.press,
             np.ma.array(np.array(z)[np.newaxis, :],
                         dtype=self.press.dtype)))
        self.press.fill_value = self.missing_value

        if self.press_isconst is True:
            levs = np.repeat(levels, self.nobs, axis=0)
        else:
            levs = levels

        # interpolation
        for spe in self.species:
            self.prof[spe] = self._vert_regrid(
                profiles[spe], levs, z,
                interp=interp, layers=layers,
            )

        # after interpolation vertical levels are constant between obs
        self.press_isconst = True

    def _vert_regrid(self, profilesin, pressuresin, pressureout,
                     interp='loglinear', layers=None):
        """ Regridder for vertical profiles.

        Input data is numpy.ma array.

        Parameters
        ----------
        profilesin : numpy.ma array
            Input profiles array with as many lines as there are observations.
        pressuresin : numpy.ma array
            Input vertical levels array with as many lines as there are
            observations.
        pressureout : list
            If interp is 'linear' or 'loglinear', z is the list of vertical
            levels for interpolation.
            If interp is 'layerAverage',
        interp : str
            Type of interpolation among:

              * 'layerAverage'
              * 'linear'
              * 'loglinear'
        layers : list of scalar couples
            Needed only if interp is 'layerAverage'. It corresponds to layers
            where to average concentrations.

        Returns
        -------
        numpy.ma array
            Interpolated profiles.

        """
        # make sure z is a list
        z = list(pressureout)

        if self.press_coorddir == 'upward':
            profilesin = profilesin[:, ::-1]
            pressuresin = pressuresin[:, ::-1]

        res = []
        for i in range(profilesin.shape[0]):
            assert((pressuresin[i, :].mask == profilesin[i, :].mask).all())
            prof = profilesin[i, :].compressed()
            press = pressuresin[i, :].compressed()
            if interp == 'loglinear':
                new_prof = np.interp(
                    np.log(z),
                    np.log(press),
                    prof,
                    left=self.missing_value,
                    right=self.missing_value,
                )
            elif interp == 'linear':
                new_prof = np.interp(
                    z,
                    press,
                    prof,
                    left=self.missing_value,
                    right=self.missing_value,
                )
            elif interp == 'layerAverage':
                new_prof = []
                for lay in layers:
                    with np.warnings.catch_warnings():
                        np.warnings.filterwarnings('ignore')
                        idx = np.logical_and(press >= lay[0], press < lay[1])
                        mean = np.nanmean(prof[idx])
                        if np.isnan(mean):
                            new_prof.append(self.missing_value)
                        else:
                            new_prof.append(mean)
            else:
                raise Exception('Unknown interpolation type.')
            res.append(new_prof)

        res = np.ma.array(res, fill_value=self.missing_value)
        res.mask = (res == self.missing_value)
        return res

