# -*- coding: utf-8 -*-
"""
Surface observation class

Created on Thu Feb  2 17:37:58 2017

@author: emili
"""
from __future__ import print_function, division
import os
import sys
import numpy as np
from baseinstr import BaseInstrument
try:
    import seaborn as sns
except ImportError:
    pass
import matplotlib.pyplot as plt
try:
    import cdms2 as cdms
except ImportError:
    pass
try:
    import MV2 as MV
except ImportError:
    pass
try:
    import vacumm.misc.grid.regridding as regrid
except ImportError:
    pass
import datetime
import calendar
import copy
import json
from json import encoder
import h5py
    
class SpectralSounderInstrument(BaseInstrument):
    """
    Class for hyperspectral imagers (eg. IASI,GOME etc...)
    """
    wltol = 1e-2

    def __init__(self,**kwargs):
        
        # Spectral sounder initialization
        self.wltype = "Wavelengths"
        wldtype = np.float32
        
        if "WlType" in kwargs: self.wltype = kwargs["WlType"]
        if self.wltype == "Channels": wldtype = np.int32
        
        self.nwl = 0 # The dimension of the spectrum (init via the raw files converter or via the hdf5 reader)
        self.wl = np.empty((0),dtype=wldtype) # As above
        
        # Observations
        self.spectra = np.ma.empty((0,0),dtype=np.float32) # The spectrum
        self.cov = np.ma.empty((0,0),dtype=np.float32) # The spectrum error covariance
        
        # The assimilation mask
        self.outmask = np.ma.empty((0,0),dtype=np.int32)
        
        # Observation geometry
        self.sunzen = np.ma.empty((0),dtype=np.float32)    # Sun zenith angle
        self.sunazim = np.ma.empty((0),dtype=np.float32)   # Sun azimuth angle
        self.satzen = np.ma.empty((0),dtype=np.float32)    # Satellite zenith angle
        self.satazim = np.ma.empty((0),dtype=np.float32)   # Satellite azimuth angle
        
        self.hascov = False # Switch to treat or not error covariances
        
        self.auxctrlvars = [] # A list for auxiliary control variables (optional, only used for assimilation output)
            
        self.cnvfctr = 1 # The scale factor for plotting
        
    # Some operator overloading to add/substract batch of measurements
    # taken at the same time and place
    def __add__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            result.spectra = self.spectra + other.spectra
            if 'avgspectra' in self.__dict__: result.Average()
            if len(self.auxctrlvars) > 0:
                for var in self.auxctrlvars:
                    result.__dict__[var] = self.__dict__[var] + other.__dict__[var]
        else:
            print('Cannot sum obs: mismatch in geolocalizations')
        
        return result
        
    def __sub__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            result.spectra = self.spectra - other.spectra
            if 'avgspectra' in self.__dict__: result.Average()
            if len(self.auxctrlvars) > 0:
                for var in self.auxctrlvars:
                    result.__dict__[var] = self.__dict__[var] - other.__dict__[var]
        else:
            print('Cannot diff obs: mismatch in geolocalizations')
        
        return result
    
    def __truediv__(self,other):
        
        result = copy.deepcopy(self) # Copy the original object
        
        if self.isColocatedWith(other):
            result.spectra = self.spectra / other.spectra
            if 'avgspectra' in self.__dict__: 
                result.avgspectra = self.avgspectra / other.avgspectra
                result.stdspectra = self.stdspectra / other.avgspectra
            if len(self.auxctrlvars) > 0:
                for var in self.auxctrlvars:
                    result.__dict__[var] = self.__dict__[var] / other.__dict__[var]
            result.rad_units = "%"        
            result.cnvfctr = 100
        else:
            print('Cannot divide obs: mismatch in geolocalizations')
        
        return result
    
    def ReplaceWith(self,other):
        
        if self.isColocatedWith(other):
            self.spectra = other.spectra
        else:
            print('Cannot set obs equal: mismatch in geolocalizations')
        
    def Read(self,filename,closefile=True,**kwargs):
        """
        Reads radiance observations from MOCAGE HDF5
        
        """
        init = False
        if self.nwl == 0: init = True
                                   
        self.OpenFile(filename,mode="r")
        nobs = self.ReadGeoValues(**kwargs)
        
        if nobs > 0:
            # Use the following to retrieve data with the matrix shape
            #data = np.array([d for d in g["Y"]],dtype=np.float32)
            # Implement the time selection in the reader
            subgrp = ""
            obstype = "Y"
            if 'ReadAssim' in kwargs:
                if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']
                if 'ObsType' in kwargs['ReadAssim']: obstype = kwargs['ReadAssim']['ObsType']
                self.hascov = False
                
            grp = self.f[self.name+subgrp+"/GEOMETRY"]
            self.sunzen = np.ma.concatenate((self.sunzen,np.ma.array(grp["SunZenithAngle"],dtype=self.sunzen.dtype)))
            self.sunazim = np.ma.concatenate((self.sunazim,np.ma.array(grp["SunAzimuthAngle"],dtype=self.sunazim.dtype)))
            self.satzen = np.ma.concatenate((self.satzen,np.ma.array(grp["SatZenithAngle"][:],dtype=self.satzen.dtype)))
            self.satazim = np.ma.concatenate((self.satazim,np.ma.array(grp["SatAzimuthAngle"][:],dtype=self.satazim.dtype)))
            
            grp = self.f[self.name+subgrp+"/OBSERVATIONS"]
            # Build the wavelengths array or check if wavelengths are different then previous reads
            foundwl = np.array(np.squeeze(grp[self.wltype][:]),dtype=self.wl.dtype)
            if init:
                # The wavelength array is filled with the first granule
                self.nwl = foundwl.size
                self.spectra = np.ma.append(self.spectra,np.ma.empty((0,self.nwl),dtype=self.spectra.dtype),axis=1)
                if self.hascov: self.cov = np.ma.append(self.cov,np.ma.empty((0,self.nwl),dtype=self.cov.dtype),axis=1)
                self.outmask = np.ma.append(self.outmask,np.ma.empty((0,self.nwl),dtype=self.outmask.dtype),axis=1)
                self.wl = np.append(self.wl,foundwl)
            else:
                if (np.abs(self.wl - foundwl) > 1e-3).any():
                    print('Wavelength arrays not matching among different granules')
                    sys.exit()
                    
            spectra = np.ma.masked_values(np.ma.array(grp[obstype][:],dtype=self.spectra.dtype),self.missing_value,rtol=1e-03)
            # I don't know why but fill value needed to be forced back after concatenation (numpy bug)
            self.spectra = np.ma.array(np.ma.concatenate((self.spectra,spectra)),fill_value=self.missing_value)
            
            if self.hascov:
                cov = np.ma.masked_values(np.ma.array(grp["ErrorCovariance"][:],dtype=self.cov.dtype),self.missing_value,rtol=1e-03)
                # I don't know why but fill value needed to be forced back after concatenation (numpy bug)
                self.cov = np.ma.array(np.ma.concatenate((self.cov,cov)),fill_value=self.missing_value)
    
            if 'ReadAssim' in kwargs: 
                # Assimilation mask from daimon output
                if obstype in ["Hx_ana","Hx_bkg"]:
                    self.outmask = np.ma.concatenate((self.outmask,np.ma.array(grp["Mask_"+obstype[-3:]][:],dtype=self.outmask.dtype)))

            # The following section exploits python automatic HDF5 visiting. I need to take care of not reading twice the
            # same data when, for example, a dataset is both in the diagnostic and control folders of the HDF5 file.
            self.dontreadlist = []
                
            # Read H controlled params here (available only in DAIMON output HDF5)
            if 'ReadAssim' in kwargs and 'AuxControl' in kwargs['ReadAssim'] and obstype in ['Hx_ana','Hx_bkg']:
                self.auxctrlvars = kwargs['ReadAssim']['AuxControl']
                self.auxctrltype = obstype[-3:]
                grp = self.f[self.name+subgrp+"/PARAMETERS_CTRL"]
                grp.visititems(self.ExtractCtrlVar)
                self.dontreadlist.extend(self.auxctrlvars) # Avoid double reading the same variables later
            
            # Read surface parameters
            if "SURFACE" in self.f[self.name+subgrp]:
                grp = self.f[self.name+subgrp+"/SURFACE"]
                grp.visititems(self.ExtractVar)
                    
        if closefile: self.CloseFile()
        
        return nobs
        
    def Write(self,filename,closefile=True,**kwargs):
        """
        Writes multispecies observations in MOCAGE HDF5
        """                                   
        # Open file
        self.OpenFile(filename)
        
        # Prepare header and datasets for satellite radiances
        self.WriteHeader()
        
        time_dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset("Timestamp",(self.nobs,),dtype=self.h5_typetime)
        
        grp = self.f[self.name]
        
        # Add satellite specific fields to geolocation group
        subgrp = grp.create_group("GEOMETRY")
        dset = subgrp.create_dataset("SunZenithAngle",(self.nobs,),dtype=self.sunzen.dtype,maxshape=(None))
        self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        dset.attrs['units'] = np.string_('Degrees')
        dset = subgrp.create_dataset("SunAzimuthAngle",(self.nobs,),dtype=self.sunazim.dtype,maxshape=(None))
        self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        dset.attrs['units'] = np.string_('Degrees')
        dset = subgrp.create_dataset("SatZenithAngle",(self.nobs,),dtype=self.satzen.dtype,maxshape=(None))
        self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        dset.attrs['units'] = np.string_('Degrees')
        dset = subgrp.create_dataset("SatAzimuthAngle",(self.nobs,),dtype=self.satazim.dtype,maxshape=(None))
        self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        dset.attrs['units'] = np.string_('Degrees')
        
        subgrp = grp.create_group("SURFACE")
        for var in self.surfvarlist:
            dset = subgrp.create_dataset(var,(self.nobs,),dtype=self.__dict__[var].dtype,maxshape=(None))
            self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
            #dset.attrs['units'] = np.string_('Degrees') # Include units soon
        
        subgrp = grp.create_group("OBSERVATIONS")
        h5arrtype = np.dtype('(%i,)f' % (self.nwl))
        dset =  subgrp.create_dataset('Y',(self.nobs,),dtype=h5arrtype,maxshape=(None,))
        self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
        dset.attrs['units'] = np.string_(self.rad_units)
        
        if self.hascov:
            dset =  subgrp.create_dataset('ErrorCovariance',(self.nobs,),dtype=h5arrtype,maxshape=(None,))
            self.attach_scale_2_dataset(dset,time_dset,'t',"Timestamp")
            dset.attrs['units'] = np.string_(self.rad_units)
        
        if self.wltype == "Channels":
            h5wlarrtype = np.dtype('(%i,)i' % (self.nwl))
        else:
            h5wlarrtype = h5arrtype
        dset =  subgrp.create_dataset(self.wltype,(1,),dtype=h5wlarrtype)
        dset.attrs['units'] = np.string_('nm')
        
        # Write actual data in the file
        if self.nobs > 0:
            
            self.WriteGeoValues()
            
            dset = self.f[self.name+"/GEOMETRY"].require_dataset("SunZenithAngle",(self.nobs,),dtype=self.sunzen.dtype)
            dset[...] = self.sunzen.compressed()[self.sorted_idx]
            dset = self.f[self.name+"/GEOMETRY"].require_dataset("SunAzimuthAngle",(self.nobs,),dtype=self.sunazim.dtype)
            dset[...] = self.sunazim.compressed()[self.sorted_idx]
            dset = self.f[self.name+"/GEOMETRY"].require_dataset("SatZenithAngle",(self.nobs,),dtype=self.satzen.dtype)
            dset[...] = self.satzen.compressed()[self.sorted_idx]
            dset = self.f[self.name+"/GEOMETRY"].require_dataset("SatAzimuthAngle",(self.nobs,),dtype=self.satazim.dtype)
            dset[...] = self.satazim.compressed()[self.sorted_idx]
            
            for var in self.surfvarlist:
                dset = self.f[self.name+"/SURFACE"].require_dataset(var,(self.nobs,),dtype=self.__dict__[var].dtype)
                dset[...] = self.__dict__[var].compressed()[self.sorted_idx]
            
            dset = self.f[self.name+"/OBSERVATIONS"].require_dataset("Y",(self.nobs,),dtype=h5arrtype)
            dset[...] = np.array([val for val in self.spectra[~self.mask,...].filled()],dtype=self.spectra.dtype)[self.sorted_idx]
            if self.hascov:
                dset = self.f[self.name+"/OBSERVATIONS"].require_dataset("ErrorCovariance",(self.nobs,),dtype=h5arrtype)
                dset[...] = np.array([val for val in self.cov[~self.mask,...].filled()],dtype=self.cov.dtype)[self.sorted_idx]
            dset = self.f[self.name+"/OBSERVATIONS"].require_dataset(self.wltype,(1,),dtype=h5wlarrtype)
            dset[...] = self.wl[np.newaxis,:]
        
        if closefile:
            self.CloseFile()
            print(">>> HDF5 written ")
        
        if 'SaveWavelengthsIn' in kwargs:
            with open(kwargs['SaveWavelengthsIn'],'w') as outfile:
                if self.wltype == "Channels":
                    json.dump({self.wltype:[int(wl) for wl in self.wl]},outfile)
                else:
                    encoder.FLOAT_REPR = lambda o: format(o, '.4f')
                    json.dump({self.wltype:[float(wl) for wl in self.wl]},outfile)
                    
    def Perturb(self,**kwargs):
        
        if self.hascov:
            # Normal perturbation
            epsilon = np.random.randn(self.nobs,self.nwl) * (1/np.ma.sqrt(self.cov))
            self.spectra = self.spectra + epsilon
        else:
            print("Perturbation not yet available without given covariance")
                    
    def Average(self,**kwargs):
        # Average spectrum
        self.avgspectra = np.ma.mean(self.spectra,axis=0)
        self.avgspectra.fill_value = self.missing_value # Numpy bug? 
        self.stdspectra = np.ma.std(self.spectra,axis=0)
        self.stdspectra.fill_value = self.missing_value # Numpy bug? 
        if self.hascov:
            # I average sqrt(R) instead of R-1, then inverse again. 
            # Since I usually plot sqrt(R) this looks more reasonable
            self.avgcov = 1.0/np.ma.mean(1.0/np.ma.sqrt(self.cov),axis=0)**2
            self.avgcov.fill_value = self.missing_value # Numpy bug? 
            self.stdcov = 1.0/np.ma.std(1.0/np.ma.sqrt(self.cov),axis=0)**2
            self.stdcov.fill_value = self.missing_value # Numpy bug? 
        
    def PlotAvgSpectra(self,outdir="./",**kwargs):
        
        figsize=(10,10)
        lkwargs, llkwargs = {}, {}
        savedata = False
        varlist = ['spectra']
        
        if 'AvgSpectraOpt' in kwargs: lkwargs = kwargs["AvgSpectraOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if "savedata" in lkwargs: savedata = lkwargs["savedata"]
        if "varlist" in lkwargs:  varlist = lkwargs["varlist"]
        
        for var in varlist:
            
            fig = plt.figure(figsize=figsize,dpi=300)
            filename = os.path.join(outdir,"%s_AvgRad%s_%s_%s" % \
                            (self.name,var.title(),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-","")))
        
            print('>>> Plotting %s' % (var.title()))
            
            ax = fig.add_subplot(211)
            
            ax.set_title(r' Average %s' % (var.title()))
                    
            if 'avgspectra' not in self.__dict__: self.Average(**kwargs)
            
            avgname = "avg%s" % (var)
            
            if var == 'cov':
                y = 1.0/(np.ma.sqrt(self.__dict__[avgname])*self.cnvfctr)
            elif var == 'relcov':
                y = 1.0/np.ma.sqrt(self.avgcov)/self.avgspectra*100
            else:
                y = self.__dict__[avgname]*self.cnvfctr
            
            ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                          (y.mean(),np.ma.std(y)))
            ax.set_ylabel("%s" % (self.rad_units))
    
            if 'LogScale' in lkwargs:
                if lkwargs['LogScale'] == 'symlog' and 'linthreshy' in lkwargs: 
                    llkwargs['linthreshy']=lkwargs['linthreshy']
                ax.set_yscale(lkwargs['LogScale'],**llkwargs)
            if 'range' in lkwargs:
                if var in lkwargs['range']:
                    vmin, vmax = lkwargs["range"][var]["min"], lkwargs["range"][var]["max"]
                else:
                    vmin, vmax = lkwargs['range']['ymin'],lkwargs['range']['ymax']
                ax.set_ylim(vmin,vmax)
            ax.yaxis.grid(True)
            ax.grid(True,which='minor',linestyle='--')
                
            ax.plot(self.wl,y)
            
            ax = fig.add_subplot(212)
            
            ax.set_title(r' %s Standard Deviation ' % var.title())
            
            stdname = "std%s" % (var)
            
            if var == 'cov':
                y = 1.0/(np.ma.sqrt(self.__dict__[stdname])*self.cnvfctr)
            elif var == 'relcov':
                y = 1.0/np.ma.sqrt(self.stdcov)/self.avgspectra*100
            else:
                y = self.__dict__[stdname]*self.cnvfctr
            
            ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                          (y.mean(),np.ma.std(y)))
            ax.set_ylabel("%s" % (self.rad_units))
    
            if 'LogScale' in lkwargs:
                ax.set_yscale(lkwargs['LogScale'],**llkwargs)
            if 'range' in lkwargs:
                if var in lkwargs['range']:
                    vmin, vmax = lkwargs["range"][var]["min"], lkwargs["range"][var]["max"]
                else:
                    vmin, vmax = lkwargs['range']['ymin'],lkwargs['range']['ymax']
                ax.set_ylim(0,vmax)
                
            ax.yaxis.grid(True)
            ax.grid(True,which='minor',linestyle='--')
                
            ax.plot(self.wl,y)
            
            fig.savefig(filename)
            plt.close(fig)
            
            if savedata:
                np.savez(filename,\
                wl=self.wl,wltype=self.wltype,\
                avgspct=self.__dict__[avgname].filled(),\
                stdspct=self.__dict__[stdname].filled(),\
                missing_value=self.missing_value)
    
    def PrintObsStats(self,**kwargs):
        
        print(">>> Radiance obs infos:")
        print("Total obs = %i ,\nn = %i are not masked " % (len(self.time),self.nobs))
        print("RadMin:%5.3e RadMax:%5.3e" % (self.spectra.min()*self.cnvfctr,self.spectra.max()*self.cnvfctr))
        if self.hascov: print("CovMin:%5.3e CovMax:%5.3e" % (self.cov.min()*self.cnvfctr,self.cov.max()*self.cnvfctr))
            
    def PlotObsHisto(self,outdir="./",**kwargs):
        
        print('>>> Plotting radiances histograms')
        
        figsize = (15,15)
        nmaxplts = 16
        aggregatewl = True
        historange = None
        varlist = ['spectra']
        lkwargs = {}
        
        if 'ObsHistoOpt' in kwargs: lkwargs = kwargs["ObsHistoOpt"]
        if "varlist" in lkwargs:  varlist = lkwargs["varlist"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'npltsperpage' in lkwargs: nmaxplts = lkwargs["npltsperpage"]
        if 'spectral' in lkwargs: aggregatewl = False
        
        for var in varlist:
            
            fig = plt.figure(figsize=figsize)
            
            if aggregatewl:
                
                ax = fig.add_subplot(1,1,1)
                ax.set_title('All wavelengths')
                # Select the range based on kwargs
                if "range" in lkwargs:
                    if var in lkwargs['range']:
                        vmin, vmax = lkwargs["range"][var]["min"], lkwargs["range"][var]["max"]
                    else:
                        vmin, vmax = lkwargs['range']['min'],lkwargs['range']['max']
                    historange = (vmin,vmax)
                
                if var == 'cov':
                    values = 1.0/(np.ma.sqrt(self.__dict__[var])*self.cnvfctr)
                elif var == 'relcov':
                    values = 1.0/np.ma.sqrt(self.cov)/self.spectra*100
                else:
                    values = self.__dict__[var]*self.cnvfctr
                    
                ax.set_xlabel("Avg:%4.2e Std:%4.2e (%s)" % \
                                      (values.mean(),np.ma.std(values),self.rad_units))
                ax.hist(values.compressed(),bins=35,density=True,range=historange)
    
            else:
                
                wls = range(0,self.nwl,self.nwl/nmaxplts)
                if "wavelengths" in lkwargs["spectral"]:
                    wls = [iw for iw in range(self.nwl) if  \
                           np.abs(self.wl[iw]-np.array(lkwargs["spectral"]["wavelengths"])).min() < self.wltol*2] 
                npltsxpage = min(len(wls),nmaxplts)
                nsubplt = int(np.ceil(np.sqrt(npltsxpage)))
                    
                for ip,ich in enumerate(wls[:npltsxpage]):
                    
                    ax = fig.add_subplot(nsubplt,nsubplt,ip+1)
                    ax.set_title(r'$\lambda$: %4.1f' % (self.wl[ich]))
                    
                    # Select the range based on kwargs
                    if "range" in lkwargs:
                        if var in lkwargs['range']: 
                            llkwargs = lkwargs['range'][var]
                        else:
                            llkwargs = lkwargs['range']
                        if "type" in llkwargs and llkwargs["type"] == "FuncOfWl":
                            for ir, wlrange in enumerate(llkwargs["WlRegions"]):
                                if wlrange[0] <= self.wl[ich] < wlrange[1]: 
                                    vmin = llkwargs["limits"][ir][0]
                                    vmax = llkwargs["limits"][ir][1]
                        else:
                            vmin, vmax = llkwargs["min"], llkwargs["max"]
                        historange = (vmin,vmax)
                        
                    if var == 'cov':
                        values = 1.0/(np.ma.sqrt(self.__dict__[var][:,ich])*self.cnvfctr)
                    elif var == 'relcov':
                        values = 1.0/np.ma.sqrt(self.cov[:,ich])/self.spectra[:,ich]*100
                    else:
                        values = self.__dict__[var][:,ich]*self.cnvfctr
                    
                    if values.count() > 0:
                        ax.set_xlabel("Avg:%4.2e Std:%4.2e (%s)" % \
                                      (values.mean(),np.ma.std(values),self.rad_units))
                        ax.hist(values.compressed(),bins=35,density=True,range=historange)
                
            fig.savefig(os.path.join(outdir,"%s_%sHisto_%s_%s" % \
                                (self.name,var.title(),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-",""))))
            plt.close(fig)
        
    def PlotObsValues(self,outdir="./",**kwargs):
        
        print(">>> Plotting radiances on a map")
        
        figsize = (15,15)
        nmaxplts = 16
        aggregatewl = True
        markersize = 2
        colormap = "viridis"
        varlist = ['spectra']
        lkwargs = {}
        
        if 'ObsOnMapOpt' in kwargs: lkwargs = kwargs["ObsOnMapOpt"]
        if "varlist" in lkwargs:  varlist = lkwargs["varlist"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
        if 'npltsperpage' in lkwargs: nmaxplts = lkwargs["npltsperpage"]
        if 'spectral' in lkwargs: aggregatewl = False
        
        for var in varlist:
            
            fig = plt.figure(figsize=figsize,dpi=300)
            
            if aggregatewl:
                
                ax = fig.add_subplot(1,1,1)
                    
                m = self.SetBaseMap(ax,**kwargs)
                
                if var == 'cov':
                    fld = 1.0/(np.ma.sqrt(self.__dict__[var].mean(axis=1))*self.cnvfctr)
                elif var == 'relcov':
                    fld = 1.0/np.ma.sqrt(self.cov.mean(axis=1))/self.spectra.mean(axis=1)*100
                else:
                    fld = self.__dict__[var].mean(axis=1)*self.cnvfctr
                
                if fld.count() > 0:
                    
                    vmin = fld.min()
                    vmax = fld.max()
                    
                    ax.set_title('N=%i (%s) \n Avg: %.1f Min: %.1f Max: %.1f' % \
                                 (self.nobs,self.rad_units,fld.mean(),vmin,vmax))
                    
                    # Select the range based on kwargs
                    if "range" in lkwargs:
                        if var in lkwargs['range']:
                            vmin, vmax = lkwargs["range"][var]["min"], lkwargs["range"][var]["max"]
                        else:
                            vmin, vmax = lkwargs['range']['min'],lkwargs['range']['max']
                                    
                    sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o', \
                                   c=fld.compressed(),cmap=colormap,\
                                            latlon=True,zorder=10,vmin=vmin,vmax=vmax)
        
                    plt.colorbar(sc,orientation='horizontal')
                
            else:
            
                wls = range(0,self.nwl,self.nwl/nmaxplts)
                if "wavelengths" in lkwargs["spectral"]:
                    wls = [iw for iw in range(self.nwl) if \
                           np.abs(self.wl[iw]-np.array(lkwargs["spectral"]["wavelengths"])).min() < self.wltol*2]    
                npltsxpage = min(len(wls),nmaxplts)
                nsubplt = int(np.ceil(np.sqrt(npltsxpage)))
                
                for ip,ich in enumerate(wls[:npltsxpage]):
                    
                    ax = fig.add_subplot(nsubplt,nsubplt,ip+1)
                    
                    m = self.SetBaseMap(ax,**kwargs)
        
                    if var == 'cov':
                        fld = 1.0/(np.ma.sqrt(self.__dict__[var][:,ich])*self.cnvfctr)
                    elif var == 'relcov':
                        fld = 1.0/np.ma.sqrt(self.cov[:,ich])/self.spectra[:,ich]*100
                    else:
                        fld = self.__dict__[var][:,ich]*self.cnvfctr
                    
                    if fld.count() > 1:
        
                        vmin = fld.min()
                        vmax = fld.max()
                        # Shrink to the common mask that considers missing vertical points    
                        lmask = self.mask | fld.mask
                        
                        ax.set_title('Wl: %4.1f N=%i (%s) \n Avg: %.1f Min: %.1f Max: %.1f' % \
                                     (self.wl[ich],(~lmask).sum(),self.rad_units,fld.mean(),vmin,vmax))
                        
                        # Select the range based on kwargs
                        if "range" in lkwargs:
                            if var in lkwargs['range']: 
                                llkwargs = lkwargs['range'][var]
                            else:
                                llkwargs = lkwargs['range']
                            if "type" in llkwargs and llkwargs["type"] == "FuncOfWl":
                                for ir, wlrange in enumerate(llkwargs["WlRegions"]):
                                    if wlrange[0] <= self.wl[ich] < wlrange[1]: 
                                        vmin = llkwargs["limits"][ir][0]
                                        vmax = llkwargs["limits"][ir][1]
                            else:
                                vmin, vmax = llkwargs["min"], llkwargs["max"]
                        
                        sc = m.scatter(self.lons[~lmask],self.lats[~lmask],markersize,marker='o', \
                                       c=fld[~lmask], cmap=colormap,\
                                                latlon=True,zorder=10,vmin=vmin,vmax=vmax)
            
                        plt.colorbar(sc,orientation='horizontal')
                
            fig.savefig(os.path.join(outdir,"%s_%sMap_%s_%s" % \
                                (self.name,var.title(),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-",""))))
            plt.close(fig)
        
    def PrintGeometryStats(self,**kwargs):
        
        print(">>> Observations geometry info:")
        print("SunZenMin:%5.2f SunZenMax:%5.2f" % (self.sunzen.min(),self.sunzen.max()))
        print("SunAzMin:%5.2f SunAzMax:%5.2f" % (self.sunazim.min(),self.sunazim.max()))
        print("SatZenMin:%5.2f SatZenMax:%5.2f" % (self.satzen.min(),self.satzen.max()))
        print("SatAzMin:%5.2f SatAzMax:%5.2f" % (self.satazim.min(),self.satazim.max()))
        
    def PrintAuxStats(self):
        
        varlist = self.surfvarlist
        print(">>> Auxiliary vars infos:")
        for var in varlist:
            print("%sMin:%5.2f %sMax:%5.2f" % (var,self.__dict__[var].min(),var,self.__dict__[var].max()))
    
    def PlotGeometryHisto(self,outdir="./",**kwargs):
        
        fig = plt.figure(figsize=(20,8),dpi=200)
        
        print('>>> Plotting observation geometry histograms')
        
        ax = fig.add_subplot(1,4,1)
        ax.set_title(r'Sun Zenith Angle')
        ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                      (self.sunzen.mean(),np.ma.std(self.sunzen)))
        ax.hist(self.sunzen.compressed(),bins=35,density=True)
        
        ax = fig.add_subplot(1,4,2)
        ax.set_title(r'Sun Azimuth Angle')
        ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                      (self.sunazim.mean(),np.ma.std(self.sunazim)))
        ax.hist(self.sunazim.compressed(),bins=35,density=True)
        
        ax = fig.add_subplot(1,4,3)
        ax.set_title(r'Sat Zenith Angle')
        ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                      (self.satzen.mean(),np.ma.std(self.satzen)))
        ax.hist(self.satzen.compressed(),bins=35,density=True)
        
        ax = fig.add_subplot(1,4,4)
        ax.set_title(r'Sat Azimuth Angle')
        ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                      (self.satazim.mean(),np.ma.std(self.satazim)))
        ax.hist(self.satazim.compressed(),bins=35,density=True)
            
        fig.savefig(os.path.join(outdir,"%s_ObsGeometryHisto_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
    def PlotAuxHisto(self,outdir="./",**kwargs):
        
        print(">>> Plotting histograms of auxiliary variables")
        
        figsize=(12,12)
        varlist = self.surfvarlist
        
        if 'ObsAuxOpt' in kwargs:
            if "figsize" in kwargs["ObsAuxOpt"]: figsize = kwargs["ObsAuxOpt"]["figsize"]
            if 'variables' in kwargs["ObsAuxOpt"]: varlist = kwargs["ObsAuxOpt"]['variables']
        
        fig = plt.figure(figsize=figsize,dpi=200)
        
        nsubplt = int(np.ceil(np.sqrt(len(varlist))))
        
        for ip,var in enumerate(varlist):
            
            ax = fig.add_subplot(nsubplt,nsubplt,ip+1)
            
            if var == "ScatteringAngle":
                fld = np.ma.arccos(np.ma.sin(np.pi/180.*self.satzen)*np.ma.sin(np.pi/180.*self.sunzen) \
                    * np.ma.cos(np.pi/180.*(self.sunazim-self.satazim)) \
                    - np.ma.cos(np.pi/180.*self.satzen)*np.ma.cos(np.pi/180.*self.sunzen)) / np.pi * 180.0
            else: 
                fld = self.__dict__[var]
                
            vmin = fld.min()
            vmax = fld.max()
            # Select the range based on kwargs
            if "ObsAuxOpt" in kwargs and "range" in kwargs["ObsAuxOpt"]:
                if var in kwargs["ObsAuxOpt"]["range"]:
                    vmin = kwargs["ObsAuxOpt"]["range"][var]["min"]
                    vmax = kwargs["ObsAuxOpt"]["range"][var]["max"]
            
            ax.set_title('%s N=%i \nAvg: %.2f Min: %.2f Max: %.2f' % (var,self.nobs,\
                         fld.mean(),fld.min(),fld.max()))
            
            ax.hist(fld.compressed(),bins=35,density=True,range=(vmin,vmax))
            
        fig.subplots_adjust(wspace=0.3,hspace=0.3)
            
        fig.savefig(os.path.join(outdir,"%s_AuxHisto_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
    def PlotAuxValues(self,outdir="./",**kwargs):
        
        print(">>> Plotting surface variables on a map")
        
        figsize=(12,12)
        markersize = 1.5
        colormap = 'viridis'
        varlist = self.surfvarlist
        
        if 'ObsAuxOpt' in kwargs:
            if "figsize" in kwargs["ObsAuxOpt"]: figsize = kwargs["ObsAuxOpt"]["figsize"]
            if 'markersize' in kwargs["ObsAuxOpt"]: markersize = kwargs["ObsAuxOpt"]['markersize']
            if 'colormap' in kwargs["ObsAuxOpt"]: colormap = kwargs["ObsAuxOpt"]['colormap']
            if 'variables' in kwargs["ObsAuxOpt"]: varlist = kwargs["ObsAuxOpt"]['variables']
        
        fig = plt.figure(figsize=figsize,dpi=200)
        
        nsubplt = int(np.ceil(np.sqrt(len(varlist))))
        
        for ip,var in enumerate(varlist):
            
            ax = fig.add_subplot(nsubplt,nsubplt,ip+1)
            
            if var == "ScatteringAngle":
                fld = np.ma.arccos(np.ma.sin(np.pi/180.*self.satzen)*np.ma.sin(np.pi/180.*self.sunzen) \
                    * np.ma.cos(np.pi/180.*(self.sunazim-self.satazim)) \
                    - np.ma.cos(np.pi/180.*self.satzen)*np.ma.cos(np.pi/180.*self.sunzen)) / np.pi * 180.0
            else: 
                fld = self.__dict__[var]*self.cnvfctr
            
            vmin = fld.min()
            vmax = fld.max()
            # Select the range based on kwargs
            if "ObsAuxOpt" in kwargs and "range" in kwargs["ObsAuxOpt"]:
                if var in kwargs["ObsAuxOpt"]["range"]:
                    vmin = kwargs["ObsAuxOpt"]["range"][var]["min"]
                    vmax = kwargs["ObsAuxOpt"]["range"][var]["max"]
            
            ax.set_title('%s N=%i \nAvg: %.2f Min: %.2f Max: %.2f' % (var,self.nobs,\
                         fld.mean(),fld.min(),fld.max()))
            
            m = self.SetBaseMap(ax,**kwargs)
    
            sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o', \
                           c=fld.compressed(), cmap=colormap,\
                                    latlon=True,zorder=10,vmin=vmin,vmax=vmax)
            plt.colorbar(sc,orientation='horizontal')
            
        fig.savefig(os.path.join(outdir,"%s_SurfaceMap_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
    def PlotMaskValues(self,outdir="./",**kwargs):
        
        if self.outmask.size == 0: return
        print(">>> Plotting assimilation mask on a map")
        
        figsize = (8,8)
        markersize = 2
        colormap = "Set1"
        lkwargs = {}
        
        if 'MaskOnMapOpt' in kwargs: lkwargs = kwargs["MaskOnMapOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        
        fig = plt.figure(figsize=figsize,dpi=200)
            
        ax = fig.add_subplot(1,1,1)
            
        m = self.SetBaseMap(ax,**kwargs)
        
        fld = self.outmask[~self.mask,0] # Until now entire spectra are masked.
        
        if fld.count() > 1:

            ax.set_title('Assim Mask \n Nmiss: %i,Nrej: %i, Nok: %i' % \
                                 ((fld == -2).sum(),(fld == 0).sum(),(fld == 1).sum()))
                    
            sc = m.scatter(self.lons[~self.mask],self.lats[~self.mask],markersize,marker='o',\
                                   c=fld,cmap=colormap,latlon=True,zorder=10,vmin=-2,vmax=1)
    
            plt.colorbar(sc,orientation='horizontal')   
            
        fig.savefig(os.path.join(outdir,"%s_MaskMap_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
    def DiagR(self,other,outdir="./",**kwargs):
        
        print(">>> Diagnose R matrix using Desroziers")
        
        figsize = (26,6)
        lkwargs = {}
        savedata = True
        
        if 'DesroziersStatsOpt' in kwargs: lkwargs = kwargs["DesroziersStatsOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if "savedata" in lkwargs: savedata = lkwargs["savedata"]
        
        if self.isColocatedWith(other):
        
            R  = np.zeros((self.nwl,self.nwl))   # R covariance  matrix 
            
            MeanOP = - self.spectra.astype(np.float64).mean(axis=0)
            MeanOA = - other.spectra.astype(np.float64).mean(axis=0)           
            
            for i in (~self.mask).nonzero()[0]:
                #outprod = np.ma.outer((-self.spectra[i,:]-MeanOP).T,(-self.spectra[i,:]-MeanOA)) # Originally bugged from Mohammad script
                #outprod = np.ma.outer((-self.spectra[i,:]-MeanOP).T,(-other.spectra[i,:]-MeanOA))
                outprod = np.ma.outer(self.spectra[i,:].T,other.spectra[i,:])
                outprod.fill_value = 0
                R = R + outprod.filled()

            nmat = np.outer(self.spectra[(~self.mask).nonzero()[0],:].count(axis=0),np.ones(self.nwl)).astype(np.int) # The matrix of the number of valid values
            
            R = R / nmat
            
            CorrMatrix = np.zeros((self.nwl,self.nwl))

            for i in range (self.nwl):
                for j in range (self.nwl):
                    CorrMatrix[i,j] = R[i,j]/(np.sqrt(R[i,i]*R[j,j]))
                
            fig = plt.figure(figsize=figsize,dpi=200)
                
            ax = fig.add_subplot(1,4,1)
            
            im = plt.imshow(R,origin='lower',extent=[self.wl[0],self.wl[-1],self.wl[0],self.wl[-1]] ,cmap='jet')
            plt.colorbar(im),#cax=cax)
            crange = [R.min(),R.max()]
            if 'crange' in lkwargs: crange = lkwargs['crange']
            plt.clim(crange["min"],crange["max"]);
            plt.title('Observation error covariance R ')
            plt.xlabel(self.wltype)
            plt.ylabel(self.wltype)
            
            ax = fig.add_subplot(1,4,2)
            
            im = plt.imshow(CorrMatrix,origin='lower',extent=[self.wl[0],self.wl[-1],self.wl[0],self.wl[-1]] ,cmap='jet')
            plt.colorbar(im),#cax=cax)
            plt.clim(0.3,1);
            plt.title('Interchannel error correlation ')
            plt.xlabel(self.wltype)
            plt.ylabel(self.wltype)
            
            ax = fig.add_subplot(1,4,3)
            
            ax.plot(self.wl,np.sqrt(np.diag(R)), color = 'red',label= 'error observation matrix')
            ax.set_ylim(0.2,3.0)
            plt.legend()
            plt.xlabel(self.wltype)
            plt.ylabel (self.rad_units)
            
            ax = fig.add_subplot(1,4,4)
            
            ax.plot(self.wl,np.diag(nmat), color = 'red',label= 'Nobs')
            plt.legend()
            plt.xlabel(self.wltype)
            plt.ylabel ("Nobs")
            
            filename = os.path.join(outdir,"%s_R_%s_%s" % \
                                (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            fig.savefig(filename)
            
            plt.close(fig)
            
            if savedata:
                np.savez(filename,\
                wl=self.wl,wltype=self.wltype,\
                R=R,\
                missing_value=self.missing_value)
        else:
            print("Cannot compute Desroziers stats. Misfits are not colocated.")
        
    def ExtractSurfaceHeight(self,lats,lons):
        # Extract a topography database and interpolate at obs location
        
        # Speed up possible by keeping array in memory
        databasefile = "/archive/pae/emili/radiances/ancillary/orography/GMTED2010_15n030_0125deg.nc"
        fid = cdms.open(databasefile)
        z = fid('elevation',raw=1)
        # A very annoying copy just because the TEMIS input netcdf is not well defined!!!
        zlons=cdms.createAxis(fid('longitude',raw=1))
        zlats=cdms.createAxis(fid('latitude',raw=1))
        fid.close()
        zlons.designateLongitude()
        zlons.standard_name = 'longitude'
        zlons.id = 'lon'
        zlons.units = 'degrees_east'
        zlats.designateLatitude()
        zlats.standard_name = 'latitude'
        zlats.id = 'lat'
        zlats.units = 'degrees_north'
        z = MV.array(z,axes=(zlats,zlons),id='elevation')
        zint = regrid.grid2xy(z,lons,lats,method='linear')(raw=1)
        # I used to do nearest neighbor method, but on 28th August 2019 it started producing random results 
        # without any explicable reason. Switched to linear interpolation, 
        # but this option produces masked values close to -180 and +180 lon (probably depending on max and min lon).
        
        self.SurfHeight = np.ma.concatenate((self.SurfHeight,zint.astype(self.SurfHeight.dtype))) # Why not np.ma.concatenate???        
        
        # I do a temporary dirty fix which mask all observations that get masked extractions. Get rid of this vacuum when you can
        self.mask = np.ma.mask_or(self.mask,self.SurfHeight.mask)
        
    def ExtractSurfaceVarsFromIFS(self,lats,lons,time,varlist=[],\
                                  inputdir="/archive/pae/emili/ifs_data/0125X0125",bestfct=False,monthlyfiles=True):
        
        # Extract a list of variables from IFS forecast and interpolate at obs locations

        # Use some sort of average time
        metbasetime = time.min()+np.timedelta64((time-time.min()).T.mean(),'s')
        metbasetimedt = metbasetime.astype(datetime.datetime)
        
        timethreshold = 60 # in minutes
        if np.abs((time-metbasetime).T).max() > np.timedelta64(timethreshold*60,'s'):
            print("%s" % ("!!! Warning: matching with meteorological fields possibly unprecise"))
            noutsiders = (np.abs((time-metbasetime).T) > np.timedelta64(timethreshold*60,'s')).sum()
            print("Number of pixels with time mismatch greater than %i minutes: %i" % (timethreshold,noutsiders))
        
        if bestfct and metbasetimedt.hour >= 12: 
            # use the 12 UTC forecast for the afternoon
            inputdir = os.path.join(inputdir,'surf-12+24h-fct') 
        else:
            inputdir = os.path.join(inputdir,'surf-00+24h-fct')
        
        if monthlyfiles:
            inputdir = os.path.join(inputdir,"%4i" % \
                (metbasetimedt.year))
            fid = cdms.open(os.path.join(inputdir,'IFS-surf-fct+%4i%02i%02i-%4i%02i%02i.nc' % \
                                         (metbasetimedt.year,metbasetimedt.month,1,\
                                          metbasetimedt.year,metbasetimedt.month,\
                                          calendar.monthrange(metbasetimedt.year, metbasetimedt.month)[1])))
        else:
            inputdir = os.path.join(inputdir,"%4i/%02i" % \
                (metbasetimedt.year,metbasetimedt.month))
            fid = cdms.open(os.path.join(inputdir,'IFS-surf-fct+%4i%02i%02i.nc' % \
                                         (metbasetimedt.year,metbasetimedt.month,metbasetimedt.day)))
        
        # Correspondence between local names and IFS netcdf names
        vardict = { \
                   #'SurfSkinTemperature':'sst',\
                   'SurfSkinTemperature':'skt',\
                   'Surf2mTemperature':'t2m',\
                   'Surf2mPressure':'sp', \
                   'Surf10mUWind':'u10', \
                   'Surf10mVWind':'v10',\
                   'TotalCloudCover':'tcc'}
        
        meteotime = np.array(fid['time'].asdatetime(),dtype=self.np_typetime)
        # Find where the granule stays among the meteotimes
        timeidx = np.searchsorted(meteotime,metbasetime)
        # Define time extraction times and interpolation factors (not so good. Stop it)
        if timeidx == 0: 
            # Outside of the interval
            print("!!! Warning: IASI date before first IFS available time, matching with meteorological fields possibly unprecise")
            wbfr, waft = 0.0, 1.0
        elif timeidx == meteotime.size:
            # Outside of the interval
            print("!!! Warning: IASI date after last IFS available time, matching with meteorological fields possibly unprecise")
            wbfr, waft = 1.0, 0.0
        else:
            # Inside the interval
            meteotimebfr, meteotimeaft  = meteotime[timeidx-1], meteotime[timeidx]
            wbfr = (meteotimeaft-metbasetime)/(meteotimeaft-meteotimebfr)
            waft = (metbasetime-meteotimebfr)/(meteotimeaft-meteotimebfr)
            
        # Some other time spent to adjust axis for space interpolations
        llons = copy.deepcopy(lons)
        
        # This is needed when using global IFS fields (otherwise gets messy W of greenwich)
        llons[llons < 0] = llons[ llons < 0 ] + 360.0
                
        # Read data and interpolate in time and space for all variables
        for var in varlist:
            
            # The special cases of outside border conditions are treated with logical expressions
            tmpfld = fid(vardict[var],time=(slice(timeidx-int(timeidx != 0),timeidx+int(timeidx == 0))),squeeze=1)*wbfr \
                    + fid(vardict[var],time=(slice(timeidx-int(timeidx == meteotime.size),timeidx+int(timeidx != meteotime.size))),squeeze=1)*waft
           
            self.__dict__[var] = np.ma.concatenate((self.__dict__[var],\
                                            regrid.grid2xy(tmpfld,llons,lats,method='linear')(raw=1).astype(self.__dict__[var].dtype)))
            # I used to do nearest neighbor method, but on 28th August 2019 it started producing random results 
            # without any explicable reason. Switched to linear interpolation. not working perfectly though
        
            # I do a temporary dirty fix which mask all observations that get masked extractions.
            # Get rid of this vacuum interpolator when you can
            self.mask = np.ma.mask_or(self.mask,self.__dict__[var].mask)
            
        fid.close()
        
    # This is a generic extractor that fills only variables that are already defined in the class and not
    # present in the exclusion list
    def ExtractVar(self,name,object):
        print('Found '+name)
        varname = name.split('/')[-1]
        if isinstance(object,h5py.Dataset) and varname in self.__dict__ and varname not in self.dontreadlist:
            print('Reading dataset', name)
            self.__dict__[varname] = np.concatenate((self.__dict__[varname],\
                         np.ma.array(object[:],dtype=object[:].dtype)))
        return None
    
    # This is a less generic extractor that fills only variables that are already defined in the class
    # depending on the requested type (analaysis or background)    
    def ExtractCtrlVar(self,name,object):
        print('Found '+name)
        varname = name.split('/')[-1].split('_')[0]
        vartype = name.split('/')[-1].split('_')[1]
        if isinstance(object,h5py.Dataset) and varname in self.__dict__ \
            and varname in self.auxctrlvars and vartype == self.auxctrltype:
            print('Reading dataset', name)
            self.__dict__[varname] = np.concatenate((self.__dict__[varname],\
                         np.ma.array(object[:],dtype=object[:].dtype)))
        return None
        
    def MakeSuperObs(self):
        pass
    