# coding: utf8
""" Surface observation Module.

@author: emili

"""
from __future__ import print_function, division
import os
import numpy as np
from datetime import datetime, timedelta
from baseinstr import BaseInstrument
from concinstr import ConcInstrument
try:
    import utils_Hobs_mini as UtilsASCII # needed for conversion of old MOCAGE obs files
except ImportError:
    pass
import matplotlib.pyplot as plt
import copy
try:
    import seaborn as sns
except ImportError:
    pass
try:
    from netCDF4 import Dataset # needed for netcdf storage
except ImportError:
    pass

class SurfConcInstrument(ConcInstrument):
    """ Surface observation class.

    Init with user defined species if given
    The beahvior is as follows:
      - no species init by users -> all species found in h5 are read (units
        and convfctr reader to be done)
      - species init by the user -> only specified ones are read
      - for ascii conversion error is raised if the user do not specify
        the species

    """
    def __init__(self, name, species=None, **kwargs):
        ConcInstrument.__init__(self, name, species)
        
        self.logobs = False
        if "LogValues" in kwargs and kwargs["LogValues"]:
            self.logobs = True

        if species is not None:
            self.conc = {spc: np.ma.empty((0), dtype=np.float32)
                         for spc in self.species}
            self.cov = {spc: np.ma.empty((0), dtype=np.float32)
                        for spc in self.species}

            # The assimilation mask
            self.assim_mask = {spc: np.ma.empty((0), dtype=np.int32)
                        for spc in self.species}

            # Units and conversion factor for plots
            if self.logobs:
                self.units = {spc: "log(vmr)" for spc in self.species}
                self.cnvfctr = {spc: 1 for spc in self.species}
            else:
                self.units = {spc: "ppb" if spc[:2] != "PM" else "mug/m3"
                              for spc in self.species}
                self.cnvfctr = {spc: 1e9 for spc in self.species}
            for spe in self.species:
                self.conc[spe].fill_value = self.missing_value
                self.cov[spe].fill_value = self.missing_value

        self.hascov = False  # Switch to treat or not error covariances
        self.daimon_type = "Surface"

    def __add__(self, other):
        """ Add measurements taken at the same time and place. """
        result = copy.deepcopy(self)  # Copy the original object

        if self.isColocatedWith(other):
            for spc in self.species:
                result.conc[spc] = self.conc[spc] + other.conc[spc]
                if 'avgconc' in self.__dict__: result.Average()
        else:
            print('Cannot sum obs: mismatch in geolocalizations')

        return result

    def __sub__(self, other):
        """ Substract measurements taken at the same time and place. """
        result = copy.deepcopy(self) # Copy the original object

        if self.isColocatedWith(other):
            for spc in self.species:
                result.conc[spc] = self.conc[spc] - other.conc[spc]
                if 'avgconc' in self.__dict__: result.Average()
        else:
            print('Cannot sum obs: mismatch in geolocalizations')

        return result

    def __truediv__(self, other):
        """ Divide measurements taken at the same time and place. """
        result = copy.deepcopy(self)
        if self.isColocatedWith(other):
            for spc in self.species:
                result.conc[spc] = self.conc[spc] / other.conc[spc]
                if 'avgconc' in self.__dict__: 
                    result.avgconc[spc] = self.avgconc[spc]/other.avgconc[spc]
                    result.stdconc[spc] = self.stdconc[spc]/other.avgconc[spc]
                    result.avgconcsp[spc] = self.avgconcsp[spc]/other.avgconcsp[spc]
                    result.stdconcsp[spc] = self.stdconcsp[spc]/other.avgconcsp[spc]
                result.units[spc] = "%"
                result.cnvfctr[spc] = 100
        else:
            print('Cannot sum obs: mismatch in geolocalizations')

        return result

    def Read(self, filename, **kwargs):

        if filename.split('.')[-1] == 'h5':
            self.Readh5(filename, **kwargs)
        elif filename.split('.')[-1] == 'nc':
            self.ReadNC(filename, **kwargs)
        else:
            print('Format not recognized')

    def Readh5(self, filename, **kwargs):
        """ Reads multispecies observations from MOCAGE HDF5.

        If assim is True reads assimilation output as well.

        """                                   
        self.OpenFile(filename,mode="r",**kwargs)
        nobs = self.ReadGeoValues(**kwargs) 

        if nobs > 0:
            subgrp = ""
            obstype = "Y"
            if 'ReadAssim' in kwargs.keys():
                if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']
                if 'ObsType' in kwargs['ReadAssim']: obstype = kwargs['ReadAssim']['ObsType']  
            
            obsgrp = self.f[self.name+subgrp+"/OBSERVATIONS"]

            # Init the species with those found in the HDF5 if not done before
            if not hasattr(self,"nspc"):
                self.species = [el for el in obsgrp.keys()]
                self.nspc = len(self.species)
                self.conc = {spc : np.ma.empty((0),dtype=np.float32) \
                    for spc in self.species}
                self.cov = {spc : np.ma.empty((0),dtype=np.float32) \
                    for spc in self.species}
                self.assim_mask = {spc : np.ma.empty((0),dtype=np.int32) \
                    for spc in self.species}
                #moved from concinstr (diverging in profile and intqty)
                if not self.logobs:
                    self.units = {spc : "ppb" \
                            if spc[:2] != "PM" else "mug/m3" for spc in self.species}
                    self.cnvfctr = {spc : 1e9 for spc in self.species}
                else:
                    self.units = {spc: "log(vmr)" for spc in self.species}
                    self.cnvfctr = {spc : 1 for spc in self.species}

            # Read the observations
            for spc in self.species:
                print(">>> Reading %s, type %s, found %i obs" % (spc,obstype,len(obsgrp[spc+"/"+obstype])))
                conc = np.ma.masked_values(np.ma.array(obsgrp[spc+"/"+obstype][:][:,0],dtype=self.conc[spc].dtype),
                    self.missing_value,
                    rtol=1e-03)
                self.conc[spc] = np.ma.array(np.ma.concatenate((self.conc[spc],conc)),fill_value=self.missing_value)
                print(self.conc[spc].min(),self.conc[spc].max())

            if 'ReadAssim' in kwargs: 
                # Assimilation mask from daimon output
                if obstype in ["Hx_ana","Hx_bkg"]:
                    self.assim_mask[spc] = np.ma.concatenate((self.assim_mask[spc],np.ma.array(obsgrp[spc+"/Mask_"+obstype[-3:]][:][:,0],
                        dtype=self.assim_mask[spc].dtype)))

        self.CloseFile()

    def ReadNC(self, filename, **kwargs):

        self.OpenFile(filename,mode="r",ftype='nc')
        nobs = self.ReadNCGeoValues(**kwargs)

        if nobs > 0:
            subgrp = ""
            obstype = "Y"
            if 'ReadAssim' in kwargs.keys():
                if 'Region' in kwargs['ReadAssim']: subgrp = "/"+kwargs['ReadAssim']['Region']
                if 'ObsType' in kwargs['ReadAssim']: obstype = kwargs['ReadAssim']['ObsType']  
            
            obsgrp = self.f[self.name+subgrp+"/OBSERVATIONS"]

            # Init the species with those found in the HDF5 if not done before
            if not hasattr(self,"nspc"):
                self.species = [el for el in obsgrp.variables]
                self.nspc = len(self.species)
                self.conc = {spc : np.ma.empty((0),dtype=np.float32) \
                    for spc in self.species}
                self.cov = {spc : np.ma.empty((0),dtype=np.float32) \
                    for spc in self.species}
                self.assim_mask = {spc : np.ma.empty((0),dtype=np.int32) \
                    for spc in self.species}
                #moved from concinstr (diverging in profile and intqty)
                if not self.logobs:
                    self.units = {spc : "ppb" \
                            if spc[:2] != "PM" else "mug/m3" for spc in self.species}
                    self.cnvfctr = {spc : 1 for spc in self.species}
                else:
                    self.units = {spc: "log(vmr)" for spc in self.species}
                    self.cnvfctr = {spc : 1 for spc in self.species}

            # Read the observations
            for spc in self.species:
                print(">>> Reading %s, type %s, found %i obs" % (spc,obstype,len(obsgrp[spc])))
                # Warning, netcdf values are in ppb whereas I use to have vmr in hdf files
                conc = np.ma.masked_values(np.ma.array(obsgrp[spc][:],dtype=self.conc[spc].dtype),
                    self.missing_value,
                    rtol=1e-03)
                self.conc[spc] = np.ma.array(np.ma.concatenate((self.conc[spc],conc*1.e-9)),fill_value=self.missing_value)

            if 'ReadAssim' in kwargs: 
                # Assimilation mask from daimon output
                if obstype in ["Hx_ana","Hx_bkg"]:
                    self.assim_mask[spc] = np.ma.concatenate((self.assim_mask[spc],np.ma.array(obsgrp[spc+"/Mask_"+obstype[-3:]][:][:,0],
                        dtype=self.assim_mask[spc].dtype)))

        self.CloseFile()

    def Write(self, filename, **kwargs):
        """ Writes multispecies observations in MOCAGE HDF5. """
        # Prepare file
        self.OpenFile(filename)
        self.WriteHeader()
        self.WriteConcHeader()

        # Surface concentration specific code
        time_dset = self.f[self.name+"/GEOLOCALIZATION"].require_dataset(
            "Timestamp", (self.nobs, ), dtype=self.h5_typetime)
        h5arrtype = np.dtype('(1,)f')
        for spc in self.species:
            dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset(
                'Y', (self.nobs, ), dtype=h5arrtype, maxshape=(None, ))
            self.attach_scale_2_dataset(dset, time_dset, 't', "Timestamp")
            # Covariance dataset (optional)
            if self.hascov:
                dset =  self.f[self.name+"/OBSERVATIONS/"+spc].create_dataset(
                    'ErrorCovariance', (self.nobs, ), dtype=h5arrtype,
                    maxshape=(None, ))
                self.attach_scale_2_dataset(dset, time_dset, 't', "Timestamp")

        # Write actual data in the file
        self.WriteGeoValues()

        for spc in self.species:
            dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset(
                "Y", (self.nobs, ), dtype=h5arrtype)
            if self.nobs > 0:
                dset[...] = np.array(
                    [val for val in self.conc[spc][~self.mask][:, np.newaxis]],
                    dtype=self.conc[spc].dtype)[self.sorted_idx]
                # Covariance dataset (optional)
                if self.hascov:
                    dset = self.f[self.name+"/OBSERVATIONS/"+spc].require_dataset(
                        "ErrorCovariance", (self.nobs, ), dtype=h5arrtype)
                    dset[...] = np.array(
                        [val for val in self.cov[spc][~self.mask][:, np.newaxis]],
                        dtype=self.cov[spc].dtype)[self.sorted_idx]
            else:
                print(">>> No observations to write")

        self.CloseFile()
        print(">>> HDF5 written")

    def Select(self, **kwargs):
        """ """
        if self.nobs > 0:
            self.SelectGeo(**kwargs)

            for spc in self.species: 
                # If assimilation mask is available mask rejected obs from stats 
                # (multispecies distinction lost, first masked species masks the entire location currently)
                if self.assim_mask[spc].size > 0:
                    print(">>> Masking %i pixels over %i rejected by the assimilation" % ((self.assim_mask[spc] == 0).sum(),self.nobs))
                    self.mask = np.ma.mask_or(self.mask,self.assim_mask[spc] == 0)
                self.conc[spc][self.mask] = np.ma.masked
                if self.hascov:
                    self.cov[spc][self.mask] = np.ma.masked

            # This should be activated, but it does not allow misfit computation in case of assim_mask.
            # A workaround should be found
            # self.nobs = (~self.mask).sum()   

    def ConvertMocageASCII(self, date, inputdir, **kwargs):
        """ Reads monospecies observations from old MOCAGE ASCII. """

        spc = self.species[0]  # Old ASCII are monospecies, so take only one if multiple species are present (the first)!  
                        
        # Use the old MOCAGE HDAT reader for now                      
        obstype, obsdate, obsname, obslat, obslon, obsval, dummy = \
            UtilsASCII.LecHOBS(os.path.join(inputdir, \
            'HDAT+'+date.isoformat()[:10].replace("-","")),obstype='totcol')

        nobs = len(obsdate)

        if  os.path.isfile(os.path.join(inputdir,'HCOV+'+date.isoformat()[:10].replace("-",""))): 
            self.hascov = True
            ostypetmp, datetmp, nametmp, lattmp, lontmp, obscov, presstmp  =  \
                UtilsASCII.LecHCOV(os.path.join(inputdir, \
                'HCOV+'+date.isoformat()[:10].replace("-", "")))
            
        if 'MinutesToSubtractToTime' in kwargs:   # time should be < 24h00 for datetime.strptime below
            obsdate = [np.string_(int(t)-(40+kwargs["MinutesToSubtractToTime"])) for t in obsdate] # substract the minutes in kwargs to obsdate HHMM
        obsname = [n[0]+n[1] for n in obsname] # Reduce old strings found in HDATs to a single site code
        obsdate = [datetime.strptime(t,"%Y%m%d%H%M") for t in obsdate] # Datetime conv       
        obsval = [o[0] for o in obsval] # Remove singleton dimension for surface observations
        if self.hascov: obscov = [o[0,0] for o in obscov] # Remove singleton dimension for surface observations
        
        self.nobs += nobs # Increase the instrument counter
        
        if nobs > 0:
            self.conc[spc] = np.ma.concatenate((self.conc[spc],np.ma.array(obsval,dtype=self.conc[spc].dtype)))
            self.lats = np.ma.concatenate((self.lats,np.ma.array(obslat,dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate((self.lons,np.ma.array(obslon,dtype=self.lons.dtype)))
            self.time = np.ma.concatenate((self.time,np.ma.array(obsdate,dtype=self.np_typetime)))
            if self.hascov: self.cov[spc] = np.ma.concatenate((self.cov[spc],np.ma.array(obscov,dtype=self.cov[spc].dtype)))
            self.mask = np.concatenate((self.mask,np.zeros((nobs),dtype=bool))) # By default all obs are accepted

    def Average(self,**kwargs):
        """ """
        
        avgmsk = self.TimeAveragerMask()

        self.avgnobstm = {spc : np.ma.empty((self.avgtime.size),dtype=np.int32) \
            for spc in self.species}
        self.avgconc = {spc : np.ma.empty((self.avgtime.size),dtype=self.conc[spc].dtype) \
            for spc in self.species}
        self.stdconc = {spc : np.ma.empty((self.avgtime.size),dtype=self.conc[spc].dtype) \
            for spc in self.species}
        self.avgrejected = {spc : np.ma.empty((self.avgtime.size),dtype=self.assim_mask[spc].dtype) \
            for spc in self.species}
        for spc in self.species:
            self.avgconc[spc][:] = np.ma.masked
            self.stdconc[spc][:] = np.ma.masked
            for it in range(self.avgtime.size):
                if avgmsk[it].any(): 
                    self.avgconc[spc][it] = np.ma.mean(self.conc[spc][avgmsk[it]])
                    self.stdconc[spc][it] = np.ma.std(self.conc[spc][avgmsk[it]])
                    self.avgnobstm[spc][it] = self.conc[spc][avgmsk[it]].count()
                    if self.assim_mask[spc].size > 0:
                        self.avgrejected[spc][it] = (self.assim_mask[spc][avgmsk[it]] == 0).sum()
                    
        latlons, spidx = self.SpaceAveragerMask()
        
        self.avglat = np.ma.array([val[0] for val in latlons])
        self.avglon = np.ma.array([val[1] for val in latlons])
        self.avgnobssp = {spc : np.ma.empty((len(latlons)),dtype=np.int32) \
            for spc in self.species}
        self.avgconcsp = {spc : np.ma.empty((len(latlons)),dtype=self.conc[spc].dtype) \
            for spc in self.species}
        self.stdconcsp = {spc : np.ma.empty((len(latlons)),dtype=self.conc[spc].dtype) \
            for spc in self.species}
        self.avgrejectedsp = {spc : np.ma.empty((len(latlons)),dtype=self.assim_mask[spc].dtype) \
            for spc in self.species}
            
        for spc in self.species:
            self.avgconcsp[spc][:] = np.ma.masked
            self.stdconcsp[spc][:] = np.ma.masked
            for io in range(len(latlons)): 
                self.avgconcsp[spc][io] = np.ma.mean(self.conc[spc][spidx[io]])
                self.stdconcsp[spc][io] = np.ma.std(self.conc[spc][spidx[io]])
                self.avgnobssp[spc][io] = self.conc[spc][spidx[io]].count()
                if self.assim_mask[spc].size > 0:
                    self.avgrejectedsp[spc][io] = (self.assim_mask[spc][spidx[io]] == 0).sum()

    def PrintObsStats(self):
        """ """
        print(">>> Surface obs infos:")
        print("Total obs = %i ,\nn = %i are not masked " % (len(self.time),self.nobs))
        for spc in self.species:
            print(spc+" ConcMin:%5.3e %s ConcMax:%5.3e %s" % \
                (self.conc[spc].min()*self.cnvfctr[spc],self.units[spc],self.conc[spc].max()*self.cnvfctr[spc],self.units[spc]))

    def PlotObsHisto(self,outdir="./",**kwargs):
        """ """
        figsize = (6,6)
        lkwargs = {}
        if 'ObsHistoOpt' in kwargs: lkwargs = kwargs["ObsHistoOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        
        fig = plt.figure(figsize=figsize,dpi=200)
        
        filename = os.path.join(outdir,"%s_ConcHisto_%s_%s" % \
                (self.name.replace('.',''),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                 self.time.compressed().max().astype(str)[:13].replace("-","")))

        print('>>> Printing plot:', filename)
        
        for spc in self.species:
            
            # Variable related keywords have priority over global keywords
            lvarkwargs = {}
            if spc in lkwargs: 
                lvarkwargs = lkwargs[spc]
            else:
                lvarkwargs = lkwargs
                    
            if "range" in lvarkwargs:
                vmin = lvarkwargs["range"]["min"]
                vmax = lvarkwargs["range"]["max"]
                historange =(vmin,vmax)
            else:
                historange = None
            
            ax = fig.add_subplot(111)
            ax.set_title("%s (%s)" % (spc,self.units[spc]))
            ax.set_xlabel("Avg:%4.2e Std:%4.2e N: %i" % \
                          (self.conc[spc].mean()*self.cnvfctr[spc],np.ma.std(self.conc[spc])*self.cnvfctr[spc],self.conc[spc].count()))
            ax.hist(self.conc[spc].compressed()*self.cnvfctr[spc],bins=35,density=True,range=historange)
            
        fig.savefig(filename)
        plt.close(fig)
        
    def PlotObsValues(self, outdir="./", **kwargs):
        """ """
        lkwargs = {}
        figsize=(6,8)
        markersize = 2
        colormap = "afmhot"
        if 'ObsOnMapOpt' in kwargs: lkwargs = kwargs["ObsOnMapOpt"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']

        fig = plt.figure(figsize=figsize,dpi=200)

        for spc in self.species:

            ax = fig.add_subplot(111)

            filename = os.path.join(outdir,"%s_%s_ConcMap_%s_%s" % \
                                (self.name.replace('.',''),spc.replace('.',''),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))

            vmin = self.conc[spc].compressed().min()*self.cnvfctr[spc]
            vmax = self.conc[spc].compressed().max()*self.cnvfctr[spc]
            # Select the range based on kwargs

            # Variable related keywords have priority over global keywords
            lvarkwargs = {}
            if spc in lkwargs: 
                lvarkwargs = lkwargs[spc]
            else:
                lvarkwargs = lkwargs

            if "range" in lvarkwargs:
                vmin = lvarkwargs["range"]["min"]
                vmax = lvarkwargs["range"]["max"]

            print('>>> Printing plot:', filename)

            m = self.SetBaseMap(ax,**kwargs)

            sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o', \
                           c=self.conc[spc].compressed()*self.cnvfctr[spc], cmap=colormap,\
                                    latlon=True,zorder=10,vmin=vmin,vmax=vmax)

            plt.colorbar(sc,orientation='horizontal')

            ax.set_title('%s %s N=%i \nAvg: %.1f Min: %.1f Max: %.1f (%s)' % (self.name,spc,self.conc[spc].count(),\
                         self.conc[spc].mean()*self.cnvfctr[spc],\
                         self.conc[spc].min()*self.cnvfctr[spc],self.conc[spc].max()*self.cnvfctr[spc],\
                         self.units[spc]))

        fig.savefig(filename)
        plt.close(fig)

    def PlotTimeSeries(self,outdir="./",**kwargs):
        """ """
        figsize=(10,15)
        lkwargs = {}
        savedata = True
        
        if 'TimeSeriesOpt' in kwargs: lkwargs = kwargs["TimeSeriesOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if "savedata" in lkwargs: savedata = lkwargs["savedata"]
            
        if 'avgconc' not in self.__dict__: self.Average(**kwargs)
        
        fig = plt.figure(figsize=figsize,dpi=200)
        
        filename = os.path.join(outdir,"%s_ConcTimeSeries_%s_%s" % \
                (self.name.replace('.',''),self.avgtime.min().astype(str)[:13].replace("-",""), \
                 self.avgtime.max().astype(str)[:13].replace("-","")))
        
        print('>>> Printing plot:', filename)
         
        for spc in self.species:
            
            ax = fig.add_subplot(311)
            
            ax = self.SetTimePlot(fig,**kwargs)
            
            ax.set_title("%s average (%s)" % (spc,self.units[spc]))
            ax.set_ylabel("%s" % (self.units[spc]))

            ax.plot(self.avgtime.astype(datetime),self.avgconc[spc]*self.cnvfctr[spc],color='royalblue',\
                    ls='-',lw=2,marker=None,label=None,markersize=4.0)
            
            if 'yrange' in lkwargs: 
                ax.set_ylim(lkwargs['yrange']["min"],lkwargs['yrange']["max"])
                
            ax = fig.add_subplot(312)
            
            ax = self.SetTimePlot(fig,**kwargs)
            
            ax.set_title("%s standard deviation (%s)" % (spc,self.units[spc]))
            ax.set_ylabel("%s" % (self.units[spc]))

            ax.plot(self.avgtime.astype(datetime),self.stdconc[spc]*self.cnvfctr[spc],color='royalblue',\
                    ls='-',lw=2,marker=None,label=None,markersize=4.0)
            
            if 'yrange' in lkwargs: 
                ax.set_ylim(max(0,lkwargs['yrange']["min"]),lkwargs['yrange']["max"])

            ax = fig.add_subplot(313)
            
            ax = self.SetTimePlot(fig,**kwargs)
            
            ax.set_title("%s observations" % (spc))

            ax.plot(self.avgtime.astype(datetime),self.avgnobstm[spc],color='royalblue',\
                    ls='-',lw=2,marker=None,label=None,markersize=4.0)

            if self.assim_mask[spc].size > 0:
                ax.plot(self.avgtime.astype(datetime),self.avgrejected[spc],color='black',\
                    ls='--',lw=2,marker=None,label=None,markersize=3.0)

            # Warning: save and plot only made for monospecies currently
            if savedata:
                np.savez(filename,\
                time=self.avgtime.astype(datetime),\
                avg=self.avgconc[spc].filled(),\
                std=self.stdconc[spc].filled(),\
                missing_value=self.missing_value)
    
        fig.savefig(filename)
        plt.close(fig)

    def PlotTimeAverages(self,outdir="./",**kwargs):

        lkwargs = {}
        figsize=(18,18)
        markersize = 2
        colormap = "afmhot"
        savedata = True
        if 'ObsOnMapOpt' in kwargs: lkwargs = kwargs["ObsOnMapOpt"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
        if "savedata" in lkwargs: savedata = lkwargs["savedata"]

        fig = plt.figure(figsize=figsize,dpi=200)

        if 'avgconcsp' not in self.__dict__: self.Average(**kwargs)

        for spc in self.species:

            ax = fig.add_subplot(221)

            filename = os.path.join(outdir,"%s_AvgConcMap_%s_%s" % \
                                (self.name.replace('.',''),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                                 self.time.compressed().max().astype(str)[:13].replace("-","")))
            
            vmin = self.avgconcsp[spc].min()*self.cnvfctr[spc]
            vmax = self.avgconcsp[spc].max()*self.cnvfctr[spc]
            # Select the range based on kwargs

            # Variable related keywords have priority over global keywords
            lvarkwargs = {}
            if spc in lkwargs: 
                lvarkwargs = lkwargs[spc]
            else:
                lvarkwargs = lkwargs

            if "range" in lvarkwargs:
                vmin = lvarkwargs["range"]["min"]
                vmax = lvarkwargs["range"]["max"]

            print('>>> Printing plot:', filename)

            m = self.SetBaseMap(ax,**kwargs)

            sc = m.scatter(self.avglon,self.avglat,markersize,marker='o', \
                           c=self.avgconcsp[spc]*self.cnvfctr[spc], cmap=colormap,\
                                    latlon=True,zorder=10,vmin=vmin,vmax=vmax)

            plt.colorbar(sc,orientation='horizontal')

            ax.set_title('%s %s Avg N=%i \nAvg: %.1f Min: %.1f Max: %.1f (%s)' % (self.name,spc,self.avgconcsp[spc].count(),\
                         self.avgconcsp[spc].mean()*self.cnvfctr[spc],\
                         self.avgconcsp[spc].min()*self.cnvfctr[spc],self.avgconcsp[spc].max()*self.cnvfctr[spc],\
                         self.units[spc]))
                
            ax = fig.add_subplot(222)
            
            vmin = self.stdconcsp[spc].min()*self.cnvfctr[spc]
            vmax = self.stdconcsp[spc].max()*self.cnvfctr[spc]
            # Select the range based on kwargs

            if "range" in lvarkwargs:
                if not self.logobs:
                    vmin = max(0,lvarkwargs["range"]["min"])
                    vmax = lvarkwargs["range"]["max"]
                else:
                    vmin = 0
                    vmax = 1

            m = self.SetBaseMap(ax,**kwargs)

            sc = m.scatter(self.avglon,self.avglat,markersize,marker='o', \
                           c=self.stdconcsp[spc]*self.cnvfctr[spc], cmap="YlOrRd",\
                                    latlon=True,zorder=10,vmin=vmin,vmax=vmax)

            plt.colorbar(sc,orientation='horizontal')

            ax.set_title('%s %s Std N=%i \nAvg: %.1f Min: %.1f Max: %.1f (%s)' % (self.name,spc,self.stdconcsp[spc].count(),\
                         self.stdconcsp[spc].mean()*self.cnvfctr[spc],\
                         self.stdconcsp[spc].min()*self.cnvfctr[spc],self.stdconcsp[spc].max()*self.cnvfctr[spc],\
                         self.units[spc]))

            ax = fig.add_subplot(223)
            
            vmin = self.avgnobssp[spc].min()
            vmax = self.avgnobssp[spc].max()

            m = self.SetBaseMap(ax,**kwargs)

            sc = m.scatter(self.avglon,self.avglat,markersize,marker='o', \
                           c=self.avgnobssp[spc], cmap="YlOrRd",\
                                    latlon=True,zorder=10,vmin=vmin,vmax=vmax)

            plt.colorbar(sc,orientation='horizontal')

            ax.set_title('%s %s Nobs N=%i \nAvg: %.1f Min: %i Max: %i ' % (self.name,spc,self.avgnobssp[spc].count(),\
                         self.avgnobssp[spc].mean(),\
                         self.avgnobssp[spc].min(),self.avgnobssp[spc].max()))

            if self.assim_mask[spc].size > 0:
                ax = fig.add_subplot(224)
            
                vmin = self.avgrejectedsp[spc].min()
                vmax = self.avgrejectedsp[spc].max()

                m = self.SetBaseMap(ax,**kwargs)

                sc = m.scatter(self.avglon,self.avglat,markersize,marker='o', \
                               c=self.avgrejectedsp[spc], cmap="YlOrRd",\
                                        latlon=True,zorder=10,vmin=vmin,vmax=vmax)

                plt.colorbar(sc,orientation='horizontal')

                ax.set_title('%s %s Rej N=%i \nAvg: %.1f Min: %i Max: %i ' % (self.name,spc,self.avgrejectedsp[spc].count(),\
                             self.avgrejectedsp[spc].mean(),\
                             self.avgrejectedsp[spc].min(),self.avgrejectedsp[spc].max()))

                        # Warning: save and plot only made for monospecies currently
            if savedata:
                np.savez(filename,\
                lon=self.avglon,\
                lat=self.avglat,\
                nobs=self.avgnobssp[spc],
                avg=self.avgconcsp[spc].filled(),\
                std=self.stdconcsp[spc].filled(),\
                missing_value=self.missing_value)

        fig.savefig(filename)
        plt.close(fig)
            
    def Stats(self,**kwargs):
        """ A comprehensive printer for surface concentration stats. """
        if self.nobs > 0:
            self.PrintGeoStats()
            self.PrintObsStats()

    def Plot(self,outdir="./",**kwargs):
        """ A comprehensive plotter for surface concentration data. """
        sns.set() # Apply seaborn beautificator
        if self.nobs > 0:
            if 'makeplots' in kwargs:
                if 'ObsLoc' in kwargs['makeplots']:
                    self.PlotObsLocation(outdir=outdir,**kwargs)
                if 'NObsTimeSer' in kwargs['makeplots']:
                    self.PlotNobsTSer(outdir=outdir,**kwargs)
                if 'ObsHisto' in kwargs['makeplots']:
                    self.PlotObsHisto(outdir=outdir,**kwargs)
                if 'ObsOnMap' in kwargs['makeplots']:
                    self.PlotObsValues(outdir=outdir,**kwargs)
                if 'TimeSeries' in kwargs['makeplots']:
                    self.PlotTimeSeries(outdir=outdir,**kwargs)
                if 'TimeAverages' in kwargs['makeplots']:
                    self.PlotTimeAverages(outdir=outdir,**kwargs)
            else:
                self.PlotObsLocation(outdir=outdir,**kwargs)
                self.PlotObsHisto(outdir=outdir)
                self.PlotObsValues(outdir=outdir,**kwargs)
                self.PlotTimeSeries(outdir=outdir,**kwargs)
        else:
            print('>> No obs to be plotted')

    def make_super_obs(self, resolution):
        """ Compute super-observations.

        For concentration values, a superobs is the a median of all the
        observations contained in a same mesh of the grid. It is
        located at the mean latitudes and longitudes of these observations.
        The grid used starts from -90. in latitude and -180. in longitude
        and is a regular lat/lon grid with resolution defined in argument.

        For covariance error, superobs are also computed as the squared
        inverse of the mean of values inside a same mesh.

        Parameters
        ----------
        resolution : float
            Resolution of the regular lat/lon grid used for computation.

        """
        import pandas as pd

        # save data
        df = pd.DataFrame({'time': self.time.compressed().tolist(),
                           'lats': self.lats.compressed(),
                           'lons': self.lons.compressed()})
        conc = {spe: self.conc[spe][~self.mask] for spe in self.species}
        if self.hascov:
            cov = {spe: self.cov[spe][~self.mask] for spe in self.species}
        # reinitialize self
        BaseInstrument.__init__(self, self.name)
        self.conc = {spe: np.ma.empty((0), dtype=np.float32)
                     for spe in self.species}
        if self.hascov:
            self.cov = {spe: np.ma.empty((0), dtype=np.float32)
                        for spe in self.species}

        # find lat/lon index of nearest point on global grid with
        # resolution defined in argument
        glob_lat_min = -90.
        glob_lon_min = -180.
        df['lat_idx'] = ((df.lats - glob_lat_min)/resolution).astype(int)
        df['lon_idx'] = ((df.lons - glob_lon_min)/resolution).astype(int)

        time_unique = np.unique(df.time)
        for t in time_unique:
            time_idx = (df.time == t).values
            sub_df = df.loc[time_idx]
            mean_latlon = sub_df[
                ['lat_idx', 'lon_idx', 'lats', 'lons']].groupby(
                    ['lat_idx', 'lon_idx'], as_index=False).mean()
            for spe in self.species:
                sub_df = sub_df.assign(conc=conc[spe][time_idx])
                median_conc = sub_df[['lat_idx', 'lon_idx', 'conc']].groupby(
                    ['lat_idx', 'lon_idx'], as_index=False).median()
                self.conc[spe] = np.ma.concatenate(
                    (self.conc[spe],
                     np.ma.array(median_conc['conc'], dtype=np.float32)))
                if self.hascov:
                    sub_df = sub_df.assign(cov=cov[spe][time_idx])
                    supobs_cov = 1./(sub_df[
                        ['lat_idx', 'lon_idx', 'cov']].groupby(
                            ['lat_idx', 'lon_idx'], as_index=False).mean())**2
                    self.cov[spe] = np.ma.concatenate(
                        (self.cov[spe],
                         np.ma.array(supobs_cov['cov'], dtype=np.float32)))

            nobs = mean_latlon.shape[0]
            self.nobs += nobs
            self.lats = np.ma.concatenate(
                (self.lats, np.ma.array(mean_latlon.lats,
                                        dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate(
                (self.lons, np.ma.array(mean_latlon.lons,
                                        dtype=self.lons.dtype)))
            self.time = np.ma.concatenate(
                (self.time, np.ma.array(nobs*[t],
                                        dtype=self.np_typetime)))
            self.mask = np.concatenate((self.mask,
                                        np.zeros((nobs), dtype=bool)))

        print("{spe} total number of observations after averaging: {nobs}".format(
            spe=self.species[0], nobs=self.nobs))
