# coding: utf8

import os
import numpy as np
import datetime
from pyhdf.SD import SD
import glob
from functools import reduce
import ipdb

from AOD_instr import AODInstrument

class CALIOP_AOD(AODInstrument):
    """ Class for AERUSGEO AOD data """
    
    def __init__(self, name, **kwargs):
        AODInstrument.__init__(self, name, **kwargs)

    def Convert(self, period, inputdir, prefix, dset_name, wavelength,
                filter_max=1e30, **kwargs):
        """ Read data from a directory.

        File names format must be:
        {prefix}.{year}-{month}-{day}T??-??-??ZD.hdf
        for ex : CAL_LID_L15_Exp-Beta-V3-40.2017-02-22T01-30-00ZD.hdf

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input directory.
        prefix : str
            Input filename prefix.
        dset_name : str
            Name of the hdf5 dataset where to find the coefficients.
        wavelength : str
            Wavelength.
        filter_max : float, optional
            Threshold above which values are considered missing.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.

        """
        nbdays = (period[1] - period[0]).days
        dates = (period[0] + datetime.timedelta(d) for d in range(nbdays))
        file_list = []
        for d in dates:
            file_list.extend(
                glob.glob(os.path.join(
                    inputdir,
                    "{prefix}.{date}T??-??-??Z?.hdf".format(
                        prefix=prefix, date=d.strftime('%Y-%m-%d')))))
        if len(file_list) > 0:
            self.convert_list_of_files(file_list=file_list,
                                       filter_max=filter_max,
                                       dset_name=dset_name,
                                       wavelength=wavelength)
        else:
            print("No file found for period {}".format(period))


    def convert_list_of_files(self, file_list, dset_name, wavelength,
                              filter_max=1e30):
        """ Reads necessary data from files.

        Works for a single wavelength.

        Parameters
        ----------
        file_list : list of str
            List of input files to read.
        dset_name : str
            Name of the hdf5 dataset where to find the coefficients.
        wavelength : str
            Wavelength.
        filter_max : float, optional
            Threshold above which values are considered missing.
        """

        for f in file_list:
            # open and read filename
            ds = SD(f)
            dsets = ds.datasets()
            dsNames = [k for k in dsets.keys()]  # convert dict_keys to list

            # get coords
            lats = ds.select('Latitude')[:, 1]
            lons = ds.select('Longitude')[:, 1]

            # get times
            file_times = [datetime.datetime.strptime(str(int(t)).zfill(6), '%y%m%d') +
                          datetime.timedelta(days=(t % 1))
                          for t in np.squeeze(
                            ds.select('Profile_UTC_Time')[:, 1])]
            times = file_times

            # get number of observations and number of vertical levels
            nobs = ds.select(dset_name)[:].shape[0]
            #nobs += file_nobs

            # get var data
            data = ds.select(dset_name)[:].ravel()

            # get var unit
            if 'units' in ds.select(dset_name).attributes():
                units = ds.select(dset_name).units
            else:
                units = ''

            #miss_value = AOD.getfillvalue()  # -9999 # Valeur missing values AOD
            #qlt_land = QAFLAG.get()  # flag quality  0 1 2 3
            ######

            # check data integrity
            if len(np.unique(units)) != 1:
                raise Exception('Different units found in files: {}'.format(
                    np.unique(units)))
            wv = str(wavelength[0])

            # define instrument
            #self.__init__(self.name, wavelength=wv,
            #              vert_name=self.vertName)
            self.units = {wv: units[0]}
            self.nobs = self.nobs + nobs
            self.aod[wv] = np.ma.concatenate(
                    (self.aod[wv],
                     np.ma.array(data, dtype=self.aod[wv].dtype)))
            self.aod[wv].fill_value = self.missing_value
            self.lats = np.ma.concatenate(
                (self.lats, np.ma.array(lats, dtype=self.lats.dtype)))
            self.lats.fill_value = self.missing_value
            self.lons = np.ma.concatenate(
                (self.lons, np.ma.array(lons, dtype=self.lons.dtype)))
            self.lons.fill_value = self.missing_value
            self.mask = np.concatenate(
                (self.mask, np.zeros((nobs), dtype=bool)))
            self.time = np.ma.concatenate(
                (self.time, np.ma.array(times, dtype=self.np_typetime)))


        # mask negative values and values above filter_max
        idx = reduce(np.logical_or, [self.aod[wv] < 0,
                                     self.aod[wv] > filter_max])
        self.aod[wv][idx] = np.ma.masked
