# -*- coding: utf-8 -*-
"""
IASI observation class

Created on Thu Feb  2 17:37:58 2017

@author: emili
"""
from __future__ import print_function
import os
import copy
import numpy as np
from datetime import datetime, timedelta
from baseinstr import BaseInstrument
from spectralsounderinstr import SpectralSounderInstrument
try:
    import seaborn as sns
except ImportError:
    pass
import matplotlib.pyplot as plt
import glob
try:
    import cdms2 as cdms
except ImportError:
    pass
try:
    import genutil
except ImportError:
    pass
import fnmatch

class IASI_L1C(SpectralSounderInstrument):
    """
    Class for IASI L1C data
    """
    def __init__(self,name,**kwargs):
        
        BaseInstrument.__init__(self,name,**kwargs)

        # The initialization of the spectral instrument with a subselection of surface variables
        SpectralSounderInstrument.__init__(self,WlType="Channels")
        
        # Input for RTTOV
        self.SurfLandSeaMask = np.ma.empty((0),dtype=np.int32)
        self.SurfSkinTemperature = np.ma.empty((0),dtype=np.float32) # (can be controlled by assimilation)
        self.Surf2mTemperature = np.ma.empty((0),dtype=np.float32)
        self.Surf2mPressure = np.ma.empty((0),dtype=np.float32)
        self.Surf10mUWind = np.ma.empty((0),dtype=np.float32)
        self.Surf10mVWind = np.ma.empty((0),dtype=np.float32)
        self.SurfHeight = np.ma.empty((0),dtype=np.float32)
        
        self.surfvarlist = [var for var in self.__dict__ if var[:4] == 'Surf']
        
        # Some diagnostics that can be found in assimilation output
        self.SurfEmissivity = np.ma.empty((0,0),dtype=np.float32) # (can also be controlled by assimilation)
        self.JacobianMatrix = np.ma.empty((0,0,0),dtype=np.float32)
        self.Profile = np.ma.empty((0,0),dtype=np.float32)
        self.P = np.ma.empty((0,0),dtype=np.float32)
        
        # Only used for preprocessin in Python
        self.qflag = np.ma.empty((0),dtype=np.int16) # IASI quality flag
        
        # Auxiliary info from AVHRR (only available for IASI v5 L1B) or EUMETSAT L2
        self.CloudF = np.ma.empty((0),dtype=np.int8)
        self.LandF = np.ma.empty((0),dtype=np.int8)
        self.CloudQFlag = np.ma.empty((0),dtype=np.int8)
        
        self.daimon_type = "Rttov"
        self.rad_units = "mW/m2/sr/cm-1"
        
    def Read(self,filename,**kwargs):
        
        init = False
        if self.nwl == 0: init = True
        nobs = SpectralSounderInstrument.Read(self,filename,closefile=False,**kwargs)

        if init and nobs > 0: 
            self.SurfEmissivity = np.ma.append(self.SurfEmissivity,np.ma.empty((0,self.nwl),dtype=self.SurfEmissivity.dtype),axis=1)
            if 'ReadAssim' in kwargs and 'Diagnostics' in kwargs['ReadAssim'] and kwargs['ReadAssim']['Diagnostics']:
                subgrp = "/"+kwargs['ReadAssim']['Region']
                grp = self.f[self.name+subgrp+"/DIAGNOSTICS"]
                arr = grp["P"][:]
                self.nlevs = arr.shape[1]
                self.P = np.ma.append(self.P,np.ma.empty((0,self.nlevs),dtype=self.P.dtype),axis=1)
                self.Profile = np.ma.append(self.Profile,np.ma.empty((0,self.nlevs),dtype=self.Profile.dtype),axis=1)
                self.JacobianMatrix = np.ma.append(self.JacobianMatrix,np.ma.empty((0,self.nwl,0),dtype=self.JacobianMatrix.dtype),axis=1)
                self.JacobianMatrix = np.ma.append(self.JacobianMatrix,np.ma.empty((0,self.nwl,self.nlevs),dtype=self.JacobianMatrix.dtype),axis=2)
        
        if 'ReadAssim' in kwargs and 'Diagnostics' in kwargs['ReadAssim'] and kwargs['ReadAssim']['Diagnostics'] \
            and kwargs['ReadAssim']['ObsType'] == 'Hx_bkg':
            subgrp = "/"+kwargs['ReadAssim']['Region']
            grp = self.f[self.name+subgrp+"/DIAGNOSTICS"]
            grp.visititems(self.ExtractVar)
        
        self.CloseFile()
        
    def Convert(self,period,inputdir,**kwargs):
        """
        Reads radiance observations from IASI L1C EUMETSAT Netcdf
        
        - Region, time and channel selection done here to avoid too much data ingest
        """
        # Default options
        timebnd, latbnd, lonbnd = None, None, None
        
        if 'TimeBnd' in kwargs.keys(): timebnd = kwargs['TimeBnd']
        if 'LatBnd' in kwargs.keys(): latbnd = kwargs['LatBnd']
        if 'LonBnd' in kwargs.keys(): lonbnd = kwargs['LonBnd']
        
        self.channels = [1479, 1509, 1513, 1521, 1536, 1574, 1578, 1579, 1585, 1587, 1626, 1639, \
                         1643, 1652, 1658, 1671]
        if "Channels" in kwargs: self.channels = kwargs["Channels"]
        
        if "RemoveDust" in kwargs:
            # Include channels needed for Dust spectral signature computation (if not already there)
            self.channels = sorted(self.channels + [ch for ch in [737,1311,1804,2230] if ch not in self.channels])
        
        if self.nwl == 0:
            # The wavelenght array is filled with the first granule
            self.nwl = len(self.channels)
            # Extend the spectral dimension of the spectra to receive data later
            self.spectra = np.ma.append(self.spectra,np.ma.empty((0,self.nwl),dtype=self.spectra.dtype),axis=1)
            self.wl = np.append(self.wl,np.array(self.channels,dtype=np.float32))
        
        matches = []
        for root, dirnames, filenames in os.walk(inputdir):
            for filename in fnmatch.filter(filenames, '*l1c.nc'):
                matches.append(os.path.join(root, filename))
        
        flist = sorted([f for f in matches \
                 if (datetime.strptime(os.path.basename(f).split('_')[4][:10],"%Y%m%d%H") > period[0]-timedelta(hours=2) and \
                     datetime.strptime(os.path.basename(f).split('_')[4][:10],"%Y%m%d%H") < period[1])])
                                
        for f in flist:
            
            print(">>> Working on ", os.path.basename(f))
            
            fid = cdms.open(f)
            
            # Check for netcdf format that changed at some time (my current database is mixed)
            # product_netcdf_version attribute did not exist in previous files
            newncformat = False
            if "product_netcdf_version" in fid.listglobal() and fid.product_netcdf_version == "2.0": newncformat = True
            
            # Only processors version later than 4 have avhrr colocated data
            hasavhrr = int(fid.processor_major_version) > 4
            # Always use L2 products for clouds instead of AVHRR
            #hasavhrr = False
            # Dimensions
            alongTrDim = fid['along_track'].shape[0]
            acrossTrDim = fid['across_track'].shape[0]
            snot = fid['snot'].shape[0]
            pn = fid['pn'].shape[0]
            spectral = fid['spectral'].shape[0]
            giadr_length = fid['giadr_length'].shape[0]
            giadr_length2 = fid['giadr_length2'].shape[0]
            # Read positional data
            lats = fid("lat",raw=1)
            lons = fid("lon",raw=1)
            if newncformat:
                lats = lats.reshape(alongTrDim,snot*pn)
                lons = lons.reshape(alongTrDim,snot*pn)
            iasitime = fid("measurement_date",raw=1)
            iasitime = np.array([timedelta(seconds=float(sec))+datetime(2000,1,1,0,0) for sec in iasitime[:,0]],dtype=self.np_typetime)
            time = np.ma.array(np.tile(iasitime,(acrossTrDim,1)).transpose())
                    
            # Compute indexes for further data extraction
            # First do a time and geographical screening
            # isinside = True if data is retained (inverse of np.ma masks)
            isinside = np.full((alongTrDim,acrossTrDim), True, dtype=bool)
        
            if timebnd is not None:
                print("Selecting obs in time between ",timebnd[0]," and ",timebnd[1])
                timemsk = ((time > timebnd[0]) & (time < timebnd[1]))
                isinside = np.logical_and(isinside,timemsk)
            if latbnd is not None: 
                print("Selecting obs latitudes between ",latbnd[0]," and ",latbnd[1])
                latmsk = np.ma.masked_outside(lats,latbnd[0],latbnd[1],copy=False).mask
                isinside = np.logical_and(isinside,~latmsk)
            if lonbnd is not None: 
                print("Selecting obs longitudes between ",lonbnd[0]," and ",lonbnd[1])
                lonmsk = np.ma.masked_outside(lons,lonbnd[0],lonbnd[1],copy=False).mask
                isinside = np.logical_and(isinside,~lonmsk)
                
            # Extract only retained data and extend observation arrays
            if isinside.any():
                self.nobs += isinside.sum()  
                self.time = np.ma.concatenate((self.time,time[isinside.nonzero()]))
                self.lats = np.ma.concatenate((self.lats,lats[isinside.nonzero()]))
                self.lons = np.ma.concatenate((self.lons,lons[isinside.nonzero()]))
                self.qflag = np.ma.concatenate((self.qflag,fid("detailed_flag_quality",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()]))
                # Use genutil picker function to read only the wanted selection 
                #picker = genutil.picker(along_track=isinside.nonzero()[0],across_track=isinside.nonzero()[1],spectral=self.channels)
                # It's extremely slow, keep classical read for now, if I pick only in one dimension is pretty ok, but I keep reading all lats,lons and time
                spectra = fid("gs_1c_spect",genutil.picker(spectral=[n-1 for n in self.channels]),raw=1).astype(self.spectra.dtype)
                # The n-1 channel shift is because IASI netcdf starts counting from zero, whereas self.channels are RTTOV compliant and start counting from 1
                if newncformat: spectra = spectra.reshape(alongTrDim,snot*pn,-1)
                
                IDefNsFirst = fid("i_def_ns_first_1b",raw=1)
                IDefNsLast = fid("i_def_ns_last_1b",raw=1)
                IDefScaleSondNbScale = fid("i_def_scale_sond_nb_scale",raw=1,squeeze=1)
                IDefScaleSondNsFirst = fid("i_def_scale_sond_ns_first",raw=1)
                IDefScaleSondNsLast = fid("i_def_scale_sond_ns_last",raw=1)
                IDefScaleSondScaleFactor = fid("i_def_scale_sond_scale_factor",raw=1)
                 # Convert spectra to physical units (follows IASI L1 Product Format Specification, page 14 of 29)
                for itr in range(alongTrDim):
                    sclfcts = [ -jscl for jsclidx, jscl in enumerate(IDefScaleSondScaleFactor[:IDefScaleSondNbScale]) 
                                for js in self.channels 
                                if ((js >= IDefScaleSondNsFirst[jsclidx]-IDefNsFirst[itr]) and (js < IDefScaleSondNsLast[jsclidx]-IDefNsFirst[itr]))]
                    spectra[itr,:,:] *= np.power(10,np.array(sclfcts).astype(self.spectra.dtype))
                    
                self.spectra = np.ma.concatenate((self.spectra,spectra[isinside.nonzero()]*1e5)) # Convert W/m2/sr/m-1 (IASI) to mW/m2/sr/cm-1 (RTTOV)
                
                if newncformat:
                    self.satzen = np.ma.concatenate((self.satzen,fid("pixel_zenith_angle",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()].astype(self.satzen.dtype)))
                    self.satazim = np.ma.concatenate((self.satazim,fid("pixel_azimuth_angle",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()].astype(self.satazim.dtype)))
                    self.sunzen = np.ma.concatenate((self.sunzen,fid("pixel_solar_zenith_angle",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()].astype(self.sunzen.dtype)))
                    self.sunazim = np.ma.concatenate((self.sunazim,fid("pixel_solar_azimuth_angle",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()].astype(self.sunazim.dtype)))
                else:     
                    self.satzen = np.ma.concatenate((self.satzen,fid("pixel_zenith_angle",raw=1)[isinside.nonzero()].astype(self.satzen.dtype)))
                    self.satazim = np.ma.concatenate((self.satazim,fid("pixel_azimuth_angle",raw=1)[isinside.nonzero()].astype(self.satazim.dtype)))
                    self.sunzen = np.ma.concatenate((self.sunzen,fid("pixel_solar_zenith_angle",raw=1)[isinside.nonzero()].astype(self.sunzen.dtype)))
                    self.sunazim = np.ma.concatenate((self.sunazim,fid("pixel_solar_azimuth_angle",raw=1)[isinside.nonzero()].astype(self.sunazim.dtype)))
                
                if hasavhrr:               
                    # Use AVHRR cloud product
                    self.CloudF = np.ma.concatenate((self.CloudF,fid("geum_avhrr_1b_cloud_fraction",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()]))
                    landfrac = fid("geum_avhrr_1b_land_fraction",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()]
                    self.LandF = np.ma.concatenate((self.LandF,landfrac))
                    # Compute and store also binary landsea mask
                    landfrac[landfrac < 90] = 1
                    landfrac[landfrac >= 90] = 0
                    self.SurfLandSeaMask = np.ma.concatenate((self.SurfLandSeaMask,landfrac.astype(np.int32)))
                    self.CloudQFlag = np.ma.concatenate((self.CloudQFlag,fid("geum_avhrr_1b_quality",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()]))
                else:
                    # Use correspondent eumetsat L2 product
                    print(">>> Reading L2 product from ", os.path.basename(f.replace('L1','L2').replace('l1c.nc','l2.nc')))
                    if not os.path.exists(f.replace('L1','L2').replace('l1c.nc','l2.nc')):
                        print("Warning: could not find correspondent L2 granule. Cloud mask set to missing value for all pixels")
                        cldfac = np.zeros(isinside.sum())-1
                        qflag = np.ones(isinside.sum()) # The actual pixel filtering must be done afterwards 
                        landf = np.zeros(isinside.sum())-1
                        landfrac = np.zeros(isinside.sum())-1
                    else:
                        fidl2 = cdms.open(f.replace('L1','L2').replace('l1c.nc','l2.nc'))
                        alongTrDimL2 = fidl2['along_track'].shape[0]
                        acrossTrDimL2 = fidl2['across_track'].shape[0]
                        
                        if (alongTrDimL2 != alongTrDim) or (acrossTrDimL2 != acrossTrDim):
                            print("Warning: incoherence between L1c and L2 granules. Cloud mask set to missing value for all pixels")
                            cldfac = np.zeros(isinside.sum())-1
                            qflag = np.ones(isinside.sum()) # The actual pixel filtering must be done afterwards 
                            landf = np.zeros(isinside.sum())-1
                            landfrac = np.zeros(isinside.sum())-1
                        else:
                            cldfrm = fidl2("number_cloud_formations",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()]
                            qflag = fidl2("flg_iasibad",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()]
                            #itconv = fidl2("flg_itconv",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()]
                            cldfac1 = fidl2("fractional_cloud_cover",raw=1)[:,:,0].reshape(alongTrDim,snot*pn)[isinside.nonzero()]
                            cldfac2 = fidl2("fractional_cloud_cover",raw=1)[:,:,1].reshape(alongTrDim,snot*pn)[isinside.nonzero()]
                            cldfac3 = fidl2("fractional_cloud_cover",raw=1)[:,:,2].reshape(alongTrDim,snot*pn)[isinside.nonzero()]
                            cldfac = np.zeros(cldfac1.shape,dtype=cldfac1.dtype)-1 # Avoids setting the pixel as clear by default
                            cldfac[cldfrm == 0] = 0.0 # Set zero cloud formations as clear pixel
                            cldfac[cldfrm == 1] = cldfac1[cldfrm == 1]
                            cldfac[cldfrm == 2] = cldfac1[cldfrm == 2]+cldfac2[cldfrm == 2]
                            cldfac[cldfrm == 3] = cldfac1[cldfrm == 3]+cldfac2[cldfrm == 3]+cldfac3[cldfrm == 3]
                            landfrac = fidl2("flg_landsea",raw=1).reshape(alongTrDim,snot*pn)[isinside.nonzero()]
                            landf = np.array(landfrac, copy=True)
                            landf[landfrac > 0] = 0 # Land (includes mixed pixels)
                            landf[landfrac == 0] = 1 # Sea 
                            # Define the acceptance threshold based on the product guide: first 6 bits true accepted 
                            # (this gives much less pixels than with avhrr selection in the south pole)
                            qflag[qflag < 128] = 0
                            #cldfac[qflag > 0] = -1
                            #cldfac[cldfac > 100] = -1
                            # Also exclude non converged retrievals
                            #qflag[itconv < 1] = 2**13+1
                            #cldfac[qflag >= 128] = -1e2
                            #cldfac[itconv < 1] = -1e2
                            
                        fidl2.close()
                        
                    self.CloudF = np.ma.concatenate((self.CloudF,cldfac.astype(self.CloudF.dtype)))
                    self.LandF = np.ma.concatenate((self.LandF,landfrac))
                    self.SurfLandSeaMask = np.ma.concatenate((self.SurfLandSeaMask,landf.astype(np.int32)))
                    self.CloudQFlag = np.ma.concatenate((self.CloudQFlag,qflag))         
                
                self.mask = np.concatenate((self.mask,np.zeros((isinside.sum()),dtype=bool))) # By default all obs are accepted
                
                # Extract ancillary values from atlases or model forecasts
                self.ExtractSurfaceHeight(lats[isinside.nonzero()],lons[isinside.nonzero()])
                
                self.ExtractSurfaceVarsFromIFS(lats[isinside.nonzero()],lons[isinside.nonzero()],time[isinside.nonzero()],\
                                                    varlist=[var for var in self.surfvarlist if var not in ['SurfHeight','SurfLandSeaMask']],\
                                                    inputdir="/archive/pae/emili/ifs_data/0125X0125")
                
            fid.close()
        
    def Select(self,**kwargs):
        
        HighestQFlag, ClearSky, RemoveDust, SelChannels = True, False, False, False
        
        if 'HighestQFlag' in kwargs: HighestQFlag = kwargs['HighestQFlag']
        if 'ClearSky' in kwargs: ClearSky = True
        if 'SelChannels' in kwargs: SelChannels = True
        if 'RemoveDust' in kwargs: RemoveDust = True
        
        if self.nobs > 0:
            
            self.SelectGeo(**kwargs)
            
            if HighestQFlag is True:
                # IASI radiance mask
                qflagmsk = np.ma.masked_not_equal(self.qflag,0).mask
                if qflagmsk.any(): 
                    print(">>> Masking %i pixels over %i with IASI quality flag != 0" % (qflagmsk.sum(),self.nobs))
                    self.mask = np.ma.mask_or(self.mask,qflagmsk)
                # AVHRR L1b quality mask
                qflagmsk = np.ma.masked_not_equal(self.CloudQFlag,0).mask
                if qflagmsk.any(): 
                    print(">>> Masking %i pixels over %i with AVHRR/L2 quality flag != 0" % (qflagmsk.sum(),self.nobs))
                    self.mask = np.ma.mask_or(self.mask,qflagmsk)
                
            if ClearSky is True:
                clftrshold = 25
                if "CloudThreshold" in kwargs["ClearSky"]: clftrshold = kwargs["ClearSky"]["CloudThreshold"]
                #clearskymsk = np.ma.masked_greater(self.CloudF,clftrshold).mask
                clearskymsk = np.ma.masked_outside(self.CloudF,0,clftrshold).mask # Exclude also negative values (missing AVHHR cloud retrieval?)
                if clearskymsk.any():
                    print(">>> Masking %i pixels over %i with cloud fraction > %i percent" % (clearskymsk.sum(),self.nobs,clftrshold))
                    self.mask = np.ma.mask_or(self.mask,clearskymsk)
                    
            if RemoveDust is True:
                dusttrshold = 0.5
                if "DustThreshold" in kwargs["RemoveDust"]: dusttrshold = kwargs["RemoveDust"]["DustThreshold"]
                dustmsk = np.ma.masked_greater(self.ComputeDustIndex(),dusttrshold).mask
                if dustmsk.any():
                    print(">>> Masking %i pixels over %i with dust dTB > %3.1f" % (dustmsk.sum(),self.nobs,dusttrshold))
                    self.mask = np.ma.mask_or(self.mask,dustmsk)
                    
            if SelChannels is True:
                chanmask = np.ma.make_mask([ wl not in kwargs["Channels"] for wl in self.wl])
                chanmask = np.tile(chanmask,(self.nobs,1))
                if chanmask.any():
                    print(">>> Masking unrequested channels")
                    self.spectra.mask = chanmask
            
            # Update the data masks
            if self.mask.any():
                print(">>> Retained %i pixels at the end of the selection" % ((~self.mask).sum()))
                
                self.time.mask = self.mask
                self.lats.mask = self.mask
                self.lons.mask = self.mask
                
                self.sunzen.mask = self.mask
                self.sunazim.mask = self.mask
                self.satzen.mask = self.mask
                self.satazim.mask = self.mask
                
                for var in self.surfvarlist:
                    self.__dict__[var].mask = self.mask
                    
                if self.SurfEmissivity.size > 0: self.SurfEmissivity[self.mask,:] = np.ma.masked
                if self.JacobianMatrix.size > 0: self.JacobianMatrix[self.mask,:,:] = np.ma.masked
                if self.Profile.size > 0: self.Profile[self.mask,:] = np.ma.masked
                if self.P.size > 0: self.P[self.mask,:] = np.ma.masked
                
                self.spectra[self.mask,:] = np.ma.masked
                self.nobs = (~self.mask).sum()
                
    def Rads2BrightnessTemp(self):
        
        from scipy.constants import c, h, k
        
        nu = IASIWavenumber(self.wl)*1e2 # cm-1 > m-1
        
        IASIrads2IS = 1e-5 # mW/m2/sr/cm-1 > W/m2/sr/m-1
        
        bT = h*c*nu/(k*np.ma.log(2*h*c**2*nu**3/(self.spectra*IASIrads2IS)+1))
        
        return bT
        
    def PlotAVHRRL1B(self,outdir="./",plotQFlag=True,**kwargs):
        
        fig = plt.figure(figsize=(15,8),dpi=200)
        
        print(">>> Plotting AVHHR masks on a map")
        
        if plotQFlag is True: 
            nsubplt = 3 
        else:
            nsubplt = 2
            
        ax = fig.add_subplot(1,nsubplt,1)
        
        ax.set_title("Cloud mask (pct) \n Avg: %.1f,Min: %.1f,Max: %.1f" % \
                     (self.CloudF.mean(),self.CloudF.min(),self.CloudF.max()))
        
        m = self.SetBaseMap(ax,**kwargs)
        
        sc = m.scatter(np.array(self.lons),np.array(self.lats),0.5,marker='o', \
                  c=self.CloudF, cmap="YlOrBr",\
                  latlon=True,\
                  vmin=0,vmax=50,\
                  zorder=10)
        
        plt.colorbar(sc,orientation='horizontal')

        ax = fig.add_subplot(1,nsubplt,2)
        
        ax.set_title("Land mask (%)")
        
        m = self.SetBaseMap(ax,**kwargs)
        
        m.scatter(np.array(self.lons),np.array(self.lats),0.5,marker='o', \
                  c=self.LandF, cmap="gist_earth",\
                  latlon=True,zorder=10)
        
        if plotQFlag is True:
            
            ax = fig.add_subplot(1,nsubplt,3)
            ax.set_title("qflag")
            
            m = self.SetBaseMap(ax,**kwargs)

            m.scatter(np.array(self.lons),np.array(self.lats),0.5,marker='o', \
                      c=self.CloudQFlag, cmap="afmhot_r",\
                      latlon=True,zorder=10)
            
        fig.savefig(os.path.join(outdir,"%s_AvhrrL1B_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
    def PlotQFlag(self,outdir="./",detailedQFlag=False,**kwargs):
        
        fig = plt.figure(figsize=(10,10),dpi=200)
        
        print(">>> Plotting quality flag")
        
        ax = fig.add_subplot(1,1,1)
        m = self.SetBaseMap(ax,**kwargs)
    
        m.scatter(np.array(self.lons),np.array(self.lats),2,marker='o', \
                  c=self.qflag, cmap="afmhot_r",\
                  latlon=True,zorder=10)
        
        ax.set_title('%s Qflag' % (self.name))
        fig.savefig(os.path.join(outdir,"%s_Qflag_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
        if detailedQFlag:
        
            qflagbit = [np.binary_repr(qf,width=16) for qf in self.qflag]
            
            for qbit in range(13):
                
                fig = plt.figure(figsize=(10,10),dpi=200)
                
                print(">>> Plotting quality flag %2i " % (qbit))
                
                ax = fig.gca()
                m = self.SetBaseMap(ax,**kwargs)
    
                tmpflag = [ True if qf[-qbit] is '1' else False for qf in qflagbit ]
            
                m.scatter(np.array(self.lons),np.array(self.lats),2,marker='o', \
                          c=np.array(tmpflag,dtype=np.int16), cmap="afmhot_r",\
                                        latlon=True,zorder=10)

                ax.set_title('%s QFlag bit %2i' % (self.name,qbit))
                fig.savefig(os.path.join(outdir,"%s_Qflag_bit%2i_%s_%s" % \
                            (self.name,qbit,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
                plt.close(fig)
                
    def PlotDiagsHisto(self,outdir="./",**kwargs):
        
        fig = plt.figure(figsize=(20,20),dpi=200)
        
        print(">>> Plotting diagnostics variables histograms")
        
        varlist = ["SurfEmissivity"]
        npltsxpage = min(self.nwl,10)
        isubplot = 0
        
        if 'DiagsHistoOpt' in kwargs:
            if 'variables' in kwargs["DiagsHistoOpt"]: varlist = kwargs["DiagsHistoOpt"]['variables']
        
        for ip,var in enumerate(varlist):
            
            if var == "SurfEmissivity" :
                for iw,ich in enumerate(range(0,self.nwl,self.nwl/npltsxpage)):
                    isubplot += 1 
                    ax = fig.add_subplot(4,4,isubplot)
                    fld = self.__dict__[var][:,ich]
                    ax.set_title(r'Emissivity at $\lambda$: %4i' % (self.wl[ich]))
                    if fld.count() > 0:
                        ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                                  (fld.mean(),np.ma.std(fld)))
                        ax.hist(fld.compressed(),bins=35,normed=True)
            else:
                isubplot += 1
                ax = fig.add_subplot(4,4,isubplot)
                fld = self.__dict__[var]
                ax.set_title('%s' % (var))
                ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                              (fld.mean(),np.ma.std(fld)))
                ax.hist(fld.compressed(),bins=35,normed=True)
    
            
        fig.savefig(os.path.join(outdir,"%s_DiagsHisto_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
                
    def PlotDiagsValues(self,outdir="./",**kwargs):
        
        print(">>> Plotting diagnostics variables on a map")
        lkwargs = {}
        figsize=(10,10)
        markersize = 1.5
        colormap = 'viridis'
        varlist = ["SurfEmissivity"]
        
        if 'ObsDiagsOpt' in kwargs: lkwargs = kwargs["ObsDiagsOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
        if 'variables' in lkwargs: varlist = lkwargs['variables']
        
        fig = plt.figure(figsize=figsize,dpi=200)
        nsubplt = int(np.ceil(np.sqrt(len(varlist))))
        
        for ip,var in enumerate(varlist):
            
            if self.__dict__[var].size == 0: continue
        
            ax = fig.add_subplot(nsubplt,nsubplt,ip+1)
            
            m = self.SetBaseMap(ax,**kwargs)
    
            if var == "SurfEmissivity" :
                fld = self.__dict__[var].mean(axis=1).compressed()
            else:
                fld = self.__dict__[var].compressed()
                
            vmin = fld.min()
            vmax = fld.max()
                # Select the range based on kwargs
            if "range" in lkwargs:
                vmin = lkwargs["range"][var]["min"]
                vmax = lkwargs["range"][var]["max"]
                
            sc = m.scatter(self.lons.compressed(),self.lats.compressed(),markersize,marker='o', \
                           c=fld, cmap=colormap,\
                           latlon=True,zorder=10,vmin=vmin,vmax=vmax)
            plt.colorbar(sc,orientation='horizontal')
            
            ax.set_title('%s N=%i \nAvg: %.2f Min: %.2f Max: %.2f' % (var,self.nobs,\
                         fld.mean(),fld.min(),fld.max()))
            
        fig.savefig(os.path.join(outdir,"%s_DiagsMap_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-",""))))
        plt.close(fig)
        
    def PlotAvgBTSpectra(self,outdir="./",**kwargs):
        
        figsize=(10,6)
        lkwargs, llkwargs = {}, {}
        savedata = False
        
        if 'AvgBTSpectraOpt' in kwargs: lkwargs = kwargs["AvgBTSpectraOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if "savedata" in lkwargs: savedata = lkwargs["savedata"]
         
        var = "spectra"
        
        fig = plt.figure(figsize=figsize,dpi=300)
        filename = os.path.join(outdir,"%s_AvgBTemp%s_%s_%s" % \
                        (self.name,var.title(),self.time.compressed().min().astype(str)[:13].replace("-",""), \
                         self.time.compressed().max().astype(str)[:13].replace("-","")))
    
        print('>>> Plotting BT %s' % (var.title()))
        
        ax = fig.add_subplot(111)
        
        ax.set_title(r' Average %s' % (var.title()))
        
        y = np.ma.mean(self.Rads2BrightnessTemp(),axis=0)
        y.fill_value = self.missing_value # Numpy bug?
        
        ax.set_xlabel("Avg:%4.2e Std:%4.2e" % \
                      (y.mean(),np.ma.std(y)))
        ax.set_ylabel("%s" % ("BT (K)"))

        if 'LogScale' in lkwargs:
            if lkwargs['LogScale'] == 'symlog' and 'linthreshy' in lkwargs: 
                llkwargs['linthreshy']=lkwargs['linthreshy']
            ax.set_yscale(lkwargs['LogScale'],**llkwargs)
        if 'range' in lkwargs:
            if var in lkwargs['range']:
                vmin, vmax = lkwargs["range"][var]["min"], lkwargs["range"][var]["max"]
            else:
                vmin, vmax = lkwargs['range']['ymin'],lkwargs['range']['ymax']
            ax.set_ylim(vmin,vmax)
        ax.yaxis.grid(True)
        ax.grid(True,which='minor',linestyle='--')
            
        ax.plot(IASIWavenumber([int(wl) for wl in self.wl]),y)
                    
        fig.savefig(filename)
        plt.close(fig)
        
        if savedata:
            np.savez(filename,\
            wl=self.wl,wltype=self.wltype,\
            avgspct=y.filled(),\
            missing_value=self.missing_value)
            
    def ComputeDustIndex(self):
        
        idx = []
        for ch in (737,1311,1804,2230): 
            try:
                idx.append(list(self.wl).index(ch))
            except:
                print("Error: channel %i needed for dust index not found" % (ch))
        
        BTspectra = self.Rads2BrightnessTemp()
        
        return (BTspectra[:,idx[0]] - BTspectra[:,idx[1]]) + \
                (BTspectra[:,idx[3]] - BTspectra[:,idx[2]])        
        
    def PlotDustIndex(self,outdir="./",**kwargs):
        
        lkwargs = {}
        figsize = (12,12)
        markersize = 2
        colormap = "YlOrBr"
            
        if 'DustIndexOpt' in kwargs: lkwargs = kwargs["DustIndexOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
            
        fig = plt.figure(figsize=figsize,dpi=200)
        
        filename = os.path.join(outdir,"%s_DustIndex_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-","")))
        
        ax = fig.add_subplot(1,1,1)
        
        fld = self.ComputeDustIndex()
        
        if fld is None:
            print("Warning: %s has no valid values")
            return
        
        # Plot only values corresponded to retained obs
        lmask = self.mask | fld.mask
        
        # Plot all values 
        #fld.fill_value = 999
        #fld = fld.filled() 
        #lmask = np.zeros((fld.size),dtype=bool)
        
        vmin = fld[~lmask].min()
        vmax = fld[~lmask].max()
        
        ax.set_title("%s \n Avg: %.2f,Min: %.2f,Max: %.2f" % ("Dust Index",fld[~lmask].mean(),vmin,vmax))
        
        if "range" in lkwargs:
            vmin = lkwargs["range"]["min"]
            vmax = lkwargs["range"]["max"]
        
        m = self.SetBaseMap(ax,**kwargs)
        
        sc = m.scatter(self.lons[~lmask],self.lats[~lmask],markersize,marker='o', \
                   c=fld[~lmask],\
                   cmap=colormap,vmin=vmin,vmax=vmax,\
                   latlon=True,zorder=10)

        plt.colorbar(sc,orientation='horizontal')
                
        fig.savefig(filename)
        plt.close(fig)
        
    def PlotRTMJacobians(self,outdir="./",**kwargs):
        
        lkwargs = {}
        figsize = (12,12)
        markersize = 2
        colormap = "YlOrBr"
            
        if 'RTMJacobiansOpt' in kwargs: lkwargs = kwargs["RTMJacobiansOpt"]
        if "figsize" in lkwargs: figsize = lkwargs["figsize"]
        if 'markersize' in lkwargs: markersize = lkwargs['markersize']
        if 'colormap' in lkwargs: colormap = lkwargs['colormap']
        
        if self.JacobianMatrix.size == 0 : return
        
        print(">>> Plotting Jacobians statistics")
        
        fig = plt.figure(figsize=figsize,dpi=200)
        
        filename = os.path.join(outdir,"%s_O3JacobiansTrace_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-","")))
        
        ax = fig.add_subplot(1,1,1)
        
        fld = np.ma.sum(np.ma.abs(self.JacobianMatrix),axis=(1,2))/self.nwl
                
        vmin = fld.min()
        vmax = fld.max()
        
        ax.set_title("%s \n Avg: %.2f,Min: %.2f,Max: %.2f" % ("O3 Jacobians Tr",fld.mean(),vmin,vmax))       
        
        if "range" in lkwargs and "Trace" in lkwargs["range"]:
            vmin = lkwargs["range"]["Trace"]["min"]
            vmax = lkwargs["range"]["Trace"]["max"]
        
        m = self.SetBaseMap(ax,**kwargs)
        
        sc = m.scatter(self.lons,self.lats,markersize,marker='o', \
                   c=fld,\
                   cmap=colormap,vmin=vmin,vmax=vmax,\
                   latlon=True,zorder=10)

        plt.colorbar(sc,orientation='horizontal')
                
        fig.savefig(filename)
        plt.close(fig)
        
        filename = os.path.join(outdir,"%s_O3JacobiansProfile_%s_%s" % \
                            (self.name,self.time.compressed().min().astype(str)[:13].replace("-",""), \
                             self.time.compressed().max().astype(str)[:13].replace("-","")))
        
        fig = plt.figure(figsize=(8,12),dpi=300)
        # Some kind of aggregated Jacobian metrics
        ax = fig.add_subplot(1,1,1)
        ax.plot(np.ma.mean(np.ma.sum(np.ma.abs(self.JacobianMatrix),axis=1)/self.nwl,axis=0),np.ma.mean(self.P,axis=0)/100,lw=1.5)
        ax.set_ylim(1000,10)
        ax.set_yscale('log')
        ax.yaxis.set_minor_locator(plt.FixedLocator([0.2,0.5,2,5,8,20,30,50,70,90,150,250,300,400,500,600,750,900]))
        ax.yaxis.set_major_formatter(plt.FormatStrFormatter('%.2f'))
        ax.yaxis.set_minor_formatter(plt.FormatStrFormatter('%.2f'))
        ax.grid(True,which='minor',linestyle='--')
        ax.tick_params(which='minor',labelsize=7)
        ax.set_xlim(0,6)
        ax.set_title("Ozone Jacobian sum")
        ax.set_ylabel('z (p)')
        ax.set_xlabel(r'(rads vmr$^{-1}$)')
        ax.grid(True)
        ax.xaxis.grid(True,which='minor',linestyle='--')
        
        fig.savefig(filename)
        plt.close(fig)    
        
    # A comprehensive printer for IASI stats
    def Stats(self,**kwargs):
        
        if self.nobs > 0:
            self.PrintGeoStats()
            self.PrintGeometryStats()
            self.PrintAuxStats()
            self.PrintObsStats(**kwargs)
        
    # A comprehensive plotter for IASI data
    def Plot(self,outdir="./",**kwargs):
        
        sns.set() # Apply seaborn beautificator
        if self.nobs > 0:
            if 'makeplots' in kwargs:
                if 'ObsLoc' in kwargs['makeplots']: self.PlotObsLocation(outdir=outdir,**kwargs)
                if 'NObsTimeSer' in kwargs['makeplots']: self.PlotNobsTSer(outdir=outdir,**kwargs)
                if 'ObsDensity' in kwargs['makeplots']: self.PlotObsDensity(outdir=outdir,**kwargs)
                if 'ObsHisto' in kwargs['makeplots']: self.PlotObsHisto(outdir=outdir,**kwargs)
                if 'ObsOnMap' in kwargs['makeplots']: self.PlotObsValues(outdir=outdir,**kwargs)
                if 'QualityFlags' in kwargs['makeplots']:
                    self.PlotAVHRRL1B(outdir=outdir,**kwargs)
                    self.PlotQFlag(outdir=outdir,**kwargs)   
                if 'ObsAux' in kwargs['makeplots']: self.PlotAuxValues(outdir=outdir,**kwargs)
                if 'GeometryHisto' in kwargs['makeplots']: self.PlotGeometryHisto(outdir=outdir,**kwargs)
                if 'AvgSpectra' in kwargs['makeplots']: self.PlotAvgSpectra(outdir=outdir,**kwargs)
                if 'AvgBTSpectra' in kwargs['makeplots']: self.PlotAvgBTSpectra(outdir=outdir,**kwargs)
                if 'DustIndex' in kwargs['makeplots']: self.PlotDustIndex(outdir=outdir,**kwargs)
                if 'RTMJacobians' in kwargs['makeplots']: self.PlotRTMJacobians(outdir=outdir,**kwargs)
                if 'ObsDiags' in kwargs['makeplots']: self.PlotDiagsValues(outdir=outdir,**kwargs)
                if 'DiagsHisto' in kwargs['makeplots']: self.PlotDiagsHisto(outdir=outdir,**kwargs)
                if 'MaskOnMap' in kwargs['makeplots']: self.PlotMaskValues(outdir=outdir,**kwargs)
        else:
            print('>> No obs to be plotted')
        
#    def MakeSuperObs(self):
#        pass

# Return the IASI wavelengths corresponding to some channel selection
def IASIWavelengths(channels):
    
    iasiwn = np.linspace(645,2760,8461)
    # The minus 1 is because of pyhon zero indexing
    # Wavelengths are given in nm
    return 1.0/iasiwn[np.array(channels,dtype=int)-1]*1e7

# Return the IASI wavenumbers corresponding to some channel selection
def IASIWavenumber(channels):
    
    iasiwn = np.linspace(645,2760,8461)
    
    return iasiwn[np.array(channels,dtype=int)-1]
    
    
    
