# coding: utf8
""" @author: plum """

import os
import fnmatch
import time
from datetime import datetime, time, timedelta

import h5py
from pyhdf.SD import *
import numpy.ma as ma
import numpy as np

from AOD_instr import AODInstrument
from obsdb import daterange


class VIIRS_AOD(AODInstrument):
    """ Class for VIIRS AOD data """

    def __init__(self, name, **kwargs):
        AODInstrument.__init__(self, name, **kwargs)

    def Convert(self, period, inputdirs, **kwargs):
        """ Reads necessary data from files

        Qflag: quality flag selection (can be 0,1,2 or 3 ; default=3)
        Works for a single wavelength

        """
        if isinstance(inputdirs, str):
            inputdirs = [inputdirs]  
            # to make sure that inputdir is iterable
            # as a sequence in a for loop (that it's a list, and not a string)
        if "Qflag" in kwargs:
            Qflag = kwargs["Qflag"]
        else:
            Qflag = 3

        wv = self.wllist[0]  # Single wavelength

        for dir in inputdirs:
            for d in daterange(period[0], period[1]):
                # Search for file name
                yyyy = str(d.timetuple().tm_year)
                mm = str(d.timetuple().tm_mon).zfill(2)
                dd = str(d.timetuple().tm_mday).zfill(2)
                juld = str(d.timetuple().tm_yday).zfill(3)  # Julian day
                # print('M?D04_L2.' + yyyy + juld + '*.hdf')
                for filename in find_files(dir, 'M?D04_L2.A' + yyyy + juld +\
                        '*.hdf'):
                    tabName = os.path.basename(filename).split(".")
                    HHMN = tabName[2]
                    obsdate = yyyy+mm+dd+HHMN
                    print('file: ', filename, " -- Instant: ", obsdate)

                    # open and read filename
                    f = SD(filename)
                    dsets = f.datasets()
                    # convert dict_keys to list
                    dsNames = [k for k in dsets.keys()]
                    dsNames.sort()

                    # list of variables
                    print("Extract " + dsNames[0])
                    print("Extract " + dsNames[2])
                    print("... should be " + wv + "nm wavelength")

                    AOD = f.select(dsNames[0])
                    QAFLAG = f.select(dsNames[2])
                    LONGITUDES = f.select('Longitude')
                    LATITUDES = f.select('Latitude')
                    # -9999 # Valeur missing values AOD                    
                    miss_value = AOD.getfillvalue()  
                    # flag quality  0 1 2 3
                    qlt_land = QAFLAG.get()
                    # we keep flag >= Qflag (3 by default)
                    QA = ma.masked_less(qlt_land, Qflag)

                    latr = LATITUDES.get()
                    lonr = LONGITUDES.get()
                    dataAOD = AOD.get()

                    # Mask application
                    # indices such as good quality flag

                    maskIndicesQA = QA.nonzero()
                    latm = latr[maskIndicesQA[0], maskIndicesQA[1]]
                    lonm = lonr[maskIndicesQA[0], maskIndicesQA[1]]
                    AODm = dataAOD[maskIndicesQA[0], maskIndicesQA[1]]\
                            * AOD.scale_factor

                    # remove negative values
                    idx = AODm > 0
                    latm = latm[idx]
                    lonm = lonm[idx]
                    AODm = AODm[idx]

                    # count number of observations kept
                    Nobs = len(latm.flatten().tolist())

                    # Time values
                    listdate = [datetime.strptime(obsdate+'00', 
                        "%Y%m%d%H%M%S%f") for i in range(Nobs)]

                    print("Instant ", obsdate, " ", Nobs, 
                            "values kept after applying QAflag =", Qflag)

                    # Output variables
                    self.nobs = self.nobs + Nobs
                    self.aod[wv] = np.ma.concatenate((self.aod[wv],
                        np.ma.array(AODm.flatten().tolist(), dtype=np.float32)))
                    self.lats = np.ma.concatenate((self.lats, np.ma.array(
                        latm.flatten().tolist(), dtype=self.lats.dtype)))
                    self.lons = np.ma.concatenate((self.lons, np.ma.array(
                        lonm.flatten().tolist(), dtype=self.lons.dtype)))
                    self.mask = np.concatenate((self.mask, np.zeros(Nobs,
                        dtype=bool)))  # keep all
                    self.aod[wv].mask = self.mask
                    self.time = np.ma.concatenate((self.time, np.ma.array(
                        listdate, dtype=self.np_typetime)))


def find_files(directory, pattern):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename
