# coding: utf8

"""

Read MSG ash from EUMETSAT data center (operational)
and convert it into IntQty HDF5 format
archive in /cnrm/comets/Data/EUMETSAT_ash/ (F. Cheroux)

@author: plum

"""

import os
import fnmatch
import time
from datetime import datetime, time, timedelta

import h5py
import numpy.ma as ma
import numpy as np

from IntQty_instr import IntQtyInstrument
from obsdb import daterange

from netCDF4 import Dataset, MFDataset, num2date, date2num

class MSGashEUMETSAT(IntQtyInstrument):
    """ Class for ash retrievals from MSG downloaded from EUMETSAT"""

    def __init__(self, name, **kwargs):

        IntQtyInstrument.__init__(self, name, species=["ASHVOL"],nmaxlevs=2, **kwargs)

    def Convert(self, period, inputdirs, **kwargs):
        """ Reads necessary data from files

        Option : THRmask : keep only the pixels for which the ash column is > thr
        By default, thr is 0.001 (to avoid files full of zeros)

        """
        if isinstance(inputdirs, basestring):
            inputdirs = [inputdirs]  # to make sure that inputdir is iterable as a sequence in a for loop (that it's a list, and not a string)

        self.pcol["ASHVOL"]=[]
        self.avk["ASHVOL"]= np.ma.empty((0,2),dtype=np.float32)
        self.units["ASHVOL"]="g/m2"
        self.vert= np.ma.empty((0,2),dtype=np.float32)
        self.vertgridcnst = 0
        self.vertunit="km"

        #Define threshold to keep the pixels - otherwise the data is full of zeros!
        if "THRmask" in kwargs:
            thr=float(kwargs["THRmask"]["thr"])
        else:
            thr = 0.001

        for dir in inputdirs:
            for d in daterange(period[0], period[1]):
                date = d.strftime('%Y%m%d')

                #Lecture des latitudes et longitudes du disque MSG
                #Depuis le fichier MSGDISK-LATLON.nc
                #qui doit être dans lldir ou dans le meme repertoire que les fichiers d'observation
                lldir = dir
                if "lldir" in kwargs:
                    lldir=kwargs["lldir"]["lldir"]
                fgeo = Dataset (lldir + "MSGDISK-LATLON.nc")
                for var_name, var_values in fgeo.variables.iteritems():
                    if var_name=="latitude":
                        lat=var_values[:].flatten().tolist()
                    elif var_name=="longitude":
                        lon=var_values[:].flatten().tolist()
                fgeo.close()

                for filename in find_files(dir, '*' + date + '*.nc'):
                    indexdate = filename.find("-"+date)
                    obsdate = filename[indexdate+1:indexdate+13]
                    print(obsdate)
                    f = Dataset(filename)
                    for var_name, var_values in f.variables.iteritems():
                        if var_name=="ash_loading":
                            val=var_values[:].flatten().tolist()
                        elif var_name=="ash_height":
                            ha=var_values[:].flatten().tolist()
                    f.close()

                    #keep only pixels for which ash_loading is above thr (to avoid void pixels)
                    latm=[lat[i] for i in range(len(val)) if val[i]>thr]
                    lonm=[lon[i] for i in range(len(val)) if val[i]>thr]
                    hm=[ha[i] for i in range(len(val)) if val[i]>thr]
                    valm=[val[i] for i in range(len(val)) if val[i]>thr]

                    Nobs=len(valm)
                    print(str(Nobs)+ " pixels kept above "+str(thr))
                    listdate = [datetime.strptime(obsdate+'00', "%Y%m%d%H%M%S%f") for i in range(Nobs)]
                    hm2=np.zeros((len(hm),2))
                    hm2[:,1] = hm
                    av2=np.zeros((len(hm),2))

                    # Output variables
                    self.nobs = self.nobs + Nobs
                    self.pcol["ASHVOL"] = np.ma.concatenate((self.pcol["ASHVOL"], np.ma.array(valm, dtype=np.float32)))
                    self.time = np.ma.concatenate((self.time, np.ma.array(listdate, dtype=self.np_typetime)))
                    self.vert = np.ma.concatenate((self.vert, hm2))
                    self.avk["ASHVOL"] = np.ma.concatenate((self.avk["ASHVOL"], av2))
                    self.lats = np.ma.concatenate((self.lats, np.ma.array(latm, dtype=self.lats.dtype)))
                    self.lons = np.ma.concatenate((self.lons, np.ma.array(lonm, dtype=self.lons.dtype)))
                    self.mask = np.concatenate((self.mask,np.zeros((Nobs),dtype=bool))) # By default all obs are accepted


def find_files(directory, pattern):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename


