# coding=utf-8

#@author: bsic

import os
import fnmatch
import time
from datetime import datetime, time, timedelta

import h5py
import numpy as np
import math

from IntQty_instr import IntQtyInstrument


class SACS_SO2(IntQtyInstrument):
    """
    Class for IntQty data retrieved in SACS for GOME2 and OMI
    """
    def __init__(self, name, **kwargs):

        IntQtyInstrument.__init__(self, name, **kwargs)

        self.ccf = np.ma.empty((0), dtype=np.float32)
        self.clp = np.ma.empty((0), dtype=np.float32)

    def Convert(self, period, inputdirs, **kwargs):
        """
        Reads necessary data from files
        """
        if isinstance(self.species, (list, tuple)):
            species = self.species[0]  # make an string; the list has to contain only one species
        self.vertunit = 'km'  # unit in input files

        if isinstance(inputdirs, basestring):
            inputdirs = [inputdirs] #to make sure that inputdir is iterable as a sequence in a for loop (that it's a list, and not a string)

        for dir in inputdirs:
            for d in daterange(period[0], period[1]):
                date = d.strftime('%Y%m%d')
                for filename in find_files(dir, '*' + date + '*.he5'):
                    print('file:', filename)
                    f = h5py.File(filename, 'r')

                    # List all groups
                    #print("Keys: %s" % f.keys())
                    keys = list(f.keys())

                    #check so2 dataset name in h5
                    so2_h5name = [x for x in keys if 'SO2 vcd' in x][0]

                    cloudfrac = f["CCF"][:]
                    cloudpress = f["Cloud pressure"][:]
                    alt = f["SO2 altitude grid"][:]
                    avk = f["SO2 averaging kernel"][:]
                    so2 = f[so2_h5name][:]
                    time = f["Time"][:].astype(int).astype('|S9')  # Time (HHMMSSmmm, UTC time HH: hours, MM: minutes, SS: seconds, mmm: milliseconds)
                    date = filename.split("-")[-1].split("_")[0]
                    time = np.char.zfill(time, 9)  # from np.char module, left-fill all strings with zeros up to 9 characters
                    obs_time = np.array([datetime.strptime(date + hms, "%Y%m%d%H%M%S%f") for hms in time[:]], dtype=self.np_typetime)
                    lat = f["latitude"][:]
                    lon = f["longitude"][:]

                    nobsfile = so2.size
                    self.nobs += nobsfile
                    self.time = np.ma.concatenate((self.time, obs_time))
                    self.lats = np.ma.concatenate((self.lats, lat.astype(self.lats.dtype)))
                    self.lons = np.ma.concatenate((self.lons, lon.astype(self.lons.dtype)))
                    self.avk[species] = np.ma.concatenate((self.avk[species], np.swapaxes(avk, 0, 1)))
                    self.vert = np.ma.concatenate((self.vert, np.swapaxes(alt, 0, 1)))
                    self.pcol[species] = np.ma.concatenate((self.pcol[species], so2))

                    self.ccf = np.ma.concatenate((self.ccf, cloudfrac))
                    self.clp = np.ma.concatenate((self.clp, cloudpress))

                    self.mask = np.concatenate(
                        (self.mask, np.zeros((nobsfile), dtype=bool)))  # By default all obs are accepted (mask=false)

    def AVKPreProcess(self):
        #add a conversion factor to avk to have Ax[vmr] = x[Dobson]

        Mair = 0.02897  # kg/mol
        for i in range(len(self.vert)):

            dair = np.zeros(len(self.vert[i]))

            for j in range(len(self.vert[i])):
                #density of air
                p, t, dair[j] = standardatmosphere(self.vert[i][j]*1000.0) #kg/m3
                dair[j] /= Mair #mol/m3

            #dz
            heights = np.append(2*self.vert[i][0] - self.vert[i][1], self.vert[i].data) # add  one top level
            heights = np.append(heights, 0.0) #add surface
            heights = (heights[1:] + heights[:-1]) / 2 #midlevels
            heights[-1] = 0.0 #back to surface
            dz = np.abs(np.diff(heights) * 1000) #m

            #air partial columns [mol m-2]
            aircol = dair * dz[:]

            aircol = aircol / 4.4615e-4 #dobson = 0.4462 mmol m−2
            for spc in self.species:
                #add to avk, so that Ax will give us Dobsons
                self.avk[spc][i] *= aircol

    def Select(self, **kwargs):

        MaxCCF = None  # cloud cover fraction
        MinClpP = None  # cloud pressure
        CombCCFCldP = None
        if 'MaxCCF' in kwargs: MaxCCF = kwargs['MaxCCF']
        if 'MinClpP' in kwargs: MinClpP = kwargs['MinClpP']
        if 'CombCCFCldP' in kwargs: CombCCFCldP = kwargs['CombCCFCldP']

        self.SelectGeo(**kwargs)

        if CombCCFCldP is True:
            clmsk = (((self.clp < MinClpP) & (self.ccf > MaxCCF)) | (
            self.ccf < 0.0))  # mask if clp negative or higher clouds with a big cover fraction
            if clmsk.any():
                self.mask = self.mask | clmsk
                print(">>> Masking %i pixels with cloud pressure < %i mbar and cloud fraction > %i percent" % (
                      self.nobs - (~self.mask).sum(), MinClpP, int(MaxCCF * 100)))
                self.nobs = (~self.mask).sum()
        else:
            if MaxCCF is not None:
                ccfmsk = ((self.ccf > MaxCCF) | (self.ccf < 0.0))  # true means it is masked
                if ccfmsk.any():
                    self.mask = self.mask | ccfmsk
                    print(">>> Masking %i pixels with cloud fraction > %i percent" % (
                          self.nobs - (~self.mask).sum(), int(MaxCCF * 100)))
                    self.nobs = (~self.mask).sum()

            if MinClpP is not None:
                clpmsk = (self.clp < MinClpP)  # true means it is masked
                if clpmsk.any():
                    self.mask = self.mask | clpmsk
                    print(">>> Masking %i pixels with cloud pressure < %i mbar" % (
                          self.nobs - (~self.mask).sum(), MinClpP))
                    self.nobs = (~self.mask).sum()

        # remove negative and zero values
        for spc in self.species:
            negmsk = (self.pcol[spc] <= 0.0) #true means it is masked
            if negmsk.any():
                self.mask = self.mask | negmsk #selection
                print(">>> Masking %i pixels with negative/zero values for species %s" % (self.nobs-(~self.mask).sum(), spc))
                self.nobs = (~self.mask).sum()  # no of obs after the selection

        # Update the data : discard masked elements
        if self.mask.any():

            self.time = self.time[~self.mask]
            self.lats = self.lats[~self.mask]
            self.lons = self.lons[~self.mask]

            # multidim array
            self.vert = self.vert[~self.mask, :]

            # multispecies arrays
            for spc in self.species:
                self.pcol[spc] = self.pcol[spc][~self.mask]
                # multidim arrays
                self.avk[spc] = self.avk[spc][~self.mask, :]
                if self.hascov:
                    self.cov[spc] = self.cov[spc][~self.mask, :, :]

            self.nobs = (~self.mask).sum()

            self.mask = self.mask[~self.mask]

        # Data thinning
        if 'DataThinning' in kwargs: self.DataThinning(**kwargs)

        # # Update the data masks
        # if self.mask.any():
        #     print(">>> Retained %i pixels at the end of the selection" % ((~self.mask).sum()))
        #
        #     self.time.mask = self.mask
        #     self.lats.mask = self.mask
        #     self.lons.mask = self.mask
        #
        #     # multidim array
        #     self.vert[self.mask, :] = np.ma.masked
        #
        #     # multispecies arrays
        #     for spc in self.species:
        #         self.pcol[spc].mask = self.mask
        #         # multidim arrays
        #         self.avk[spc][self.mask, :] = np.ma.masked
        #         if self.hascov:
        #             self.cov[spc][self.mask, :, :] = np.ma.masked
        #
        #     self.nobs = (~self.mask).sum()

        # Update the data : discard masked elements
        if self.mask.any():

            self.time = self.time[~self.mask]
            self.lats = self.lats[~self.mask]
            self.lons = self.lons[~self.mask]

            # multidim array
            self.vert = self.vert[~self.mask, :]

            # multispecies arrays
            for spc in self.species:
                self.pcol[spc] = self.pcol[spc][~self.mask]
                # multidim arrays
                self.avk[spc] = self.avk[spc][~self.mask, :]
                if self.hascov:
                    self.cov[spc] = self.cov[spc][~self.mask, :, :]

            self.nobs = (~self.mask).sum()

            self.mask = self.mask[~self.mask]

def find_files(directory, pattern):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename

def daterange(d1, d2):
    d1 = datetime.combine(d1, time.min) #truncate to the midnight
    d2 = enddtrange(d2)
    return (d1 + timedelta(days=i) for i in range((d2 - d1).days)) #in daterange, as in range, the last date is not inclusive

def enddtrange(dt):
    if dt == datetime.combine(dt, time.min):
        return dt # just return it - it's already midnight
    else:
        return datetime.combine(dt + timedelta(days=1), time.min) #find a forthcoming midnight


def isa(p0, t0, a, h0, h1):
    g = 9.80665
    R = 287.00
    if a != 0:
        t1 = t0 + a * (h1 - h0)
        p1 = p0 * (t1 / t0) ** (-g / a / R)
    else:
        t1 = t0
        p1 = p0 * math.exp(-g / R / t0 * (h1 - h0))
    return t1, p1


def standardatmosphere(altitude):
    g = 9.80665
    R = 287.00
    a = [-0.0065, 0, 0.001, 0.0028, 0, -0.0028, 0.002, 0]
    h = [11000, 20000, 32000, 47000, 51000, 71000, 84852]
    p0 = 101325
    t0 = 288.15
    prevh = 0
    if altitude < 0 or altitude > 84852:
        print("altitude must be in [0, 84852]")
        return
    for i in range(0, 7):
        if altitude <= h[i]:
            temperature, pressure = isa(p0, t0, a[i], prevh, altitude)
            break;
        else:
            # sth like dynamic programming
            t0, p0 = isa(p0, t0, a[i], prevh, h[i])
            prevh = h[i]

    density = pressure / (R * temperature)
    # strformat = 'Temperature: {0:.2f} \nPressure: {1:.2f} \nDensity: {2:.4f}'
    # print(strformat.format(temperature, pressure, density))
    return pressure, temperature, density

# np.set_printoptions(precision=3) #works only for numpy float arrays
# np.set_printoptions(suppress=True) # suppress the use of scientific notation for small numbers


#
# timebnd = DefPeriod('2011052208', '2011052408')
# lonbnd = [110,160]
# latbnd = [-40,90]
#
# # Compute indexes for further data extraction
# # First do a time and geographical screening
# # isinside = True if data is retained (inverse of np.ma masks)
# if lon.size == lat.size  and lon.size == time.size:
#     isinside = np.full(time.size, True, dtype=bool)
# else:
#     sys.exit('longitude, latitude and time variables do not have the same size. Exit!')
#
# if timebnd is not None:
#     #print("Selecting obs in time between ", timebnd[0], " and ", timebnd[1])
#     timemsk = (obs_time > timebnd[0]) & (obs_time < timebnd[1])
#     isinside = np.logical_and(isinside, timemsk)
#
# if latbnd is not None:
#     #print("Selecting obs latitudes between ", latbnd[0], " and ", latbnd[1])
#     latmsk = np.ma.masked_outside(lat, latbnd[0], latbnd[1], copy=False).mask
#     isinside = np.logical_and(isinside, ~latmsk) #'~' because a True entry in the mask indicates an invalid data.
#
# if lonbnd is not None:
#     #print("Selecting obs longitudes between ", lonbnd[0], " and ", lonbnd[1])
#     lonmsk = (lon > lonbnd[0]) & (lon < lonbnd[1])
#     isinside = isinside & lonmsk
#     ##slower:
#     #lonmsk = np.ma.masked_outside(lon, lonbnd[0], lonbnd[1], copy=False).mask
#     #isinside = np.logical_and(isinside, ~lonmsk) #'~' because a True entry in the mask indicates an invalid data.
#
# #Select only non-negative values
# nnmask = so2 > 0.0
# isinside = isinside & nnmask


print('')
