# coding: utf8
""" @author: plum """

import os
import fnmatch
import glob
from datetime import datetime

import numpy as np
from netCDF4 import Dataset

from AOD_instr import AODInstrument
from obsdb import daterange
from rotation_routines import rtll

class VIIRS_AOD(AODInstrument):
    """ Class for MODIS AOD data """

    def __init__(self, name, **kwargs):
        AODInstrument.__init__(self, name, **kwargs)

    def Convert(self, period, inputdir, **kwargs):
        """ Reads necessary data from files

        """
        # In most recent Netcdf files from Calum there are always non rotated lat
        # and lon, so this rotated=True should probably be deprecated
        rotated = False
        cenlon, cenlat = None, None
        
        if 'rotated' in kwargs: 
            rotated = kwargs['rotated']
            cenlon, cenlat = kwargs['cenlon'],kwargs['cenlat']
            
        wv = self.wllist[0]  # Single wavelength
        deg_to_rad = np.pi / 180.

        for d in daterange(period[0], period[1]):
            # Search for file name
            yyyy = str(d.timetuple().tm_year)
            mm = str(d.timetuple().tm_mon).zfill(2)
            dd = str(d.timetuple().tm_mday).zfill(2)
                
            for filename in sorted(glob.glob(os.path.join(inputdir, 'obs' + yyyy + mm + dd + \
                    '*.nc'))):
                
                print('Reading %s' % (filename))
                f = Dataset(filename, 'r', format="NETCDF4")
                hh = int(filename[-5:-3])
                # Output variables
                nobs = f['index'].size
                self.nobs = self.nobs + f['index'].size
                self.aod[wv] = np.ma.concatenate((self.aod[wv],
                    np.ma.array(f['obs'][:], dtype=self.aod[wv].dtype)))
                self.cov[wv] = np.ma.concatenate((self.cov[wv],
                    np.ma.array(f['obserr'][:], dtype=self.cov[wv].dtype)))
                if rotated:
                    # Convert back rotated lats lons
                    lons, lats = rtll(f['rlon'][:]*deg_to_rad,f['rlat'][:]*deg_to_rad,cenlon,
                                            np.cos(cenlat*deg_to_rad),np.sin(cenlat*deg_to_rad))
                else:
                    lons, lats = f['lon'][:], f['lat'][:]
                self.lats = np.ma.concatenate((self.lats, np.ma.array(lats,
                                                    dtype=self.lats.dtype)))
                self.lons = np.ma.concatenate((self.lons, np.ma.array(
                    lons, dtype=self.lons.dtype)))
                self.mask = np.concatenate((self.mask, np.zeros(nobs,
                    dtype=bool)))  # keep all
                self.aod[wv].mask = self.mask
                self.time = np.ma.concatenate((self.time, np.ma.array(
                    [datetime(int(yyyy),int(mm),int(dd),int(hh)) for i in range(nobs)], 
                    dtype=self.np_typetime)))
                
                f.close()

def find_files(directory, pattern):
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename

