# coding: utf8
""" ATom (Atmospheric Tomography Mission) module. """

import os
import numpy as np
from datetime import datetime, timedelta
from profconc_instr import ProfConcInstrument
from netCDF4 import Dataset


class ATOM_flights(ProfConcInstrument):
    """ Class designed to process ATOM measurements.

    Parameters
    ----------
    name : str
        Name of the instrument.
    species : list of str, optional
        List of species stored in the object.

    """
    # Dictionnary for the corresponding measurements
    atom_spec = {
        # CFC & Halogènes
        'CH3Cl_GCMSD': ['GCMSD', 'CH3Cl_PMSD'],
        'CH3Br_GCMSD': ['GCMSD', 'CH3Br_PMSD'],
        'CFC113': ['GCECD', 'CFC113_PECD'],
        'CFC11': ['GCECD', 'CFC11_PECD'],
        'CFC12': ['GCECD', 'CFC12_PECD'],
        'H1211': ['GCECD', 'H1211_PECD'],
        'HCFC22_GCMSD': ['GCMSD', 'HCFC22_PMSD'],
        # CH4
        'CH_4': ['GCECD', 'CH4_PECD'],
        'CH_4_GCECD': ['GCECD', 'CH4_PECD'],
        'CH_4_PICARRO': ['NOAA-Picarro', 'CH4_NOAA'],
        'CH_4_QCLS': ['QCLS-CH4-CO-N2O', 'CH4_QCLS'],
        'CH_4_UCATS': ['UCATS-GC', 'CH4_UCATS'],
        # CO
        'CO': ['GCECD', 'CO_PECD'],
        'CO_GCECD': ['GCECD', 'CO_PECD'],
        'CO_PICARRO': ['NOAA-Picarro', 'CO_NOAA'],
        'CO_QCLS': ['QCLS-CH4-CO-N2O', 'CO_QCLS'],
        'CO_UCATS': ['UCATS-GC', 'CO_UCATS'],
        # HCHO
        'HCHO': ['ISAF-H2CO', 'HCHO'],
        # HNO3
        'HNO_3': ['CIT-HNO3', 'HNO3_CIT'],
        # N2O
        'N_2O': ['GCECD', 'N2O_PECD'],
        'N_2O_GCECD': ['GCECD', 'N2O_PECD'],
        'N_2O_QCLS': ['QCLS-CH4-CO-N2O', 'N2O_QCLS'],
        'N_2O_UCATS': ['UCATS-GC', 'N2O_UCATS'],
        # NO
        'NO': ['NOyO3-NO', 'NO_CL'],
        # NO2
        'NO_2': ['NOyO3-NO2', 'NO2_CL'],
        # NOy
        'NOy': ['NOyO3-NOy', 'NOy_CL'],
        # ozone
        'O3': ['UCATS-O3', 'O3_UO3'],
        'O3_UCATS': ['UCATS-O3', 'O3_UO3'],
        'O3_NOyO3': ['NOyO3-O3', 'O3_CL'],
        # PAA
        'PAA': ['CIT-PAA', 'PAA_CIT'],
        # PAN
        'PAN': ['GCECD', 'PAN_PECD'],
        # SO2
        'SO_2': ['CIT-SO2', 'SO2_CIT'],
    }

    def __init__(self, name, species, nmaxlevs=1, **kwargs):
        ProfConcInstrument.__init__(
            self, name, species=species, nmaxlevs=nmaxlevs, **kwargs)
        if species is not None:
            for spec in species:
                if spec not in self.atom_spec:
                    raise Exception(
                        "{} not implemented or not existing".format(spec))

        else:
            raise Exception(
                "list of species to retrieve from ATOM files needed.")

    def Convert(self, period, inputdir, fileATOM, flight, **kwargs):
        """ Read ATom raw input files.

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input file directory.
        fileATOM : str
            Input file name.
        flight : str
            Research flight number.

        """
        f = inputdir + fileATOM

        prof = {spc: np.ma.empty((0, self.nmaxlevs), dtype=np.float32)
                for spc in self.species}
        pres = np.ma.empty((0, self.nmaxlevs), dtype=np.float32)

        print(">>> Working on ", os.path.basename(f))
        fid = Dataset(f, mode='r')
        # Identify indices of the wanted flight
        rf = fid.variables['RF']
        f_ind = np.where(rf[:] == flight)
        group = fid.groups['MMS']
        lat = group.variables['G_LAT'][f_ind]
        lon = group.variables['G_LONG'][f_ind]

        time = fid.variables['time']
        nobsfile = time[f_ind].size
        pres = group.variables['P'][f_ind]
        # reshape and conversion hpa to Pa
        pres = np.reshape(pres, (nobsfile, self.nmaxlevs)) * 100.0

        # Read the date start for the time encoding
        date_init = time.units.split()[2]
        year = int(date_init.split('-')[0])
        month = int(date_init.split('-')[1])
        day = int(date_init.split('-')[2])

        obs_time = np.array(
            [timedelta(seconds=float(sec))+datetime(year, month, day, 0, 0)
             for sec in time[f_ind]],
            dtype=self.np_typetime,
        )

        for spec in self.species:
            print("Retrieving ", spec, "measurments...")
            group = fid.groups[self.atom_spec[spec][0]]
            prof[spec] = group.variables[self.atom_spec[spec][1]][f_ind]
            # prof[spec][prof[spec] == 1.e+20]
            prof[spec] = np.reshape(prof[spec], (nobsfile, self.nmaxlevs))
            self.units[spec] = group.variables[self.atom_spec[spec][1]].units
            self.cnvfctr[spec] = float(self.units[spec])
            self.prof[spec] = np.ma.concatenate((self.prof[spec], prof[spec]))
            self.prof[spec].fill_value = self.missing_value
            self.prof[spec] *= self.cnvfctr[spec]

        self.nobs += nobsfile
        self.time = np.ma.concatenate((self.time, obs_time))
        self.lats = np.ma.concatenate(
            (self.lats, lat.astype(self.lats.dtype))
        )
        self.lats.fill_value = self.missing_value
        self.lons = np.ma.concatenate(
            (self.lons, lon.astype(self.lons.dtype))
        )
        self.lons.fill_value = self.missing_value
        self.press = np.ma.concatenate((self.press, pres))
        self.press.fill_value = self.missing_value
        self.mask = np.concatenate(
            (self.mask, np.zeros((nobsfile), dtype=bool))
        )

        fid.close()
