# coding: utf8

import os
import numpy as np
import datetime
from pyhdf.SD import SD
from lidar_instr import LidarInstrument
import glob
from functools import reduce


class CaliopLidar(LidarInstrument):
    """ Class for Caliop lidar data.

    Parameters
    ----------
    name : str
        Name of the instrument
    wllist : list of str
        List of studied wavelengths.
    nmaxlevs : int
        Maximal number of vertical levels.
    hascov : bool
        If True, covariance errors are computed.
    vert_isconst : bool
        Tells wether the vertical axis is constant or variable for each
        measurement.
    kwargs : dict
        Used to allow unexpected keyword arguments.

    """
    def __init__(self, name, wavelength, nmaxlevs=1, hascov=False,
                 vert_isconst=False, **kwargs):
        LidarInstrument.__init__(
            self, name, wllist=[wavelength], nmaxlevs=nmaxlevs, hascov=hascov,
            upward=False, vert_isconst=vert_isconst, vert_name='Pressure')

    def Convert(self, period, inputdir, prefix, dset_name, wavelength,
                filter_max=1e30, vertical_levels=None, **kwargs):
        """ Read data from a directory.

        File names format must be:
        {prefix}.{year}-{month}-{day}T??-??-??ZD.hdf
        for ex : CAL_LID_L15_Exp-Beta-V3-40.2017-02-22T01-30-00ZD.hdf

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input directory.
        prefix : str
            Input filename prefix.
        dset_name : str
            Name of the hdf5 dataset where to find the coefficients.
        wavelength : str
            Wavelength.
        filter_max : float, optional
            Threshold above which values are considered missing.
        vertical_levels : None or 1D array-like, optional
            If not None, data is interpolated vertically according to given
            levels.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.

        """
        nbdays = (period[1] - period[0]).days
        dates = (period[0] + datetime.timedelta(d) for d in range(nbdays))
        file_list = []
        for d in dates:
            file_list.extend(
                glob.glob(os.path.join(
                    inputdir,
                    "{prefix}.{date}T??-??-??Z?.hdf".format(
                        prefix=prefix, date=d.strftime('%Y-%m-%d')))))
        if len(file_list) > 0:
            self.convert_list_of_files(file_list=file_list,
                                       filter_max=filter_max,
                                       dset_name=dset_name,
                                       wavelength=wavelength,
                                       vertical_levels=vertical_levels)
        else:
            print("No file found for period {}".format(period))

    def convert_list_of_files(self, file_list, dset_name, wavelength,
                              filter_max=1e30, vertical_levels=None):
        """ Reads necessary data from files.

        Works for a single wavelength.

        Parameters
        ----------
        file_list : list of str
            List of input files to read.
        dset_name : str
            Name of the hdf5 dataset where to find the coefficients.
        wavelength : str
            Wavelength.
        filter_max : float, optional
            Threshold above which values are considered missing.
        vertical_levels : None or 1D array-like, optional
            If not None, data is interpolated vertically according to given
            levels.

        """
        units = []
        vert_units = []
        nb_alt = []
        times = []
        lats = []
        lons = []
        vert = []
        prof = []
        stations = []
        nobs = 0
        for f in file_list:
            ds = SD(f)

            # get var unit
            if 'units' in ds.select(dset_name).attributes():
                units.append(ds.select(dset_name).units)
            else:
                units.append('')

            # get vertical axis unit
            if 'units' in ds.select(self.vertName).attributes():
                vert_units.append(ds.select(self.vertName).units)
            else:
                vert_units.append('')

            # get number of observations and number of vertical levels
            file_nobs, nvert = ds.select(dset_name)[:].shape
            nobs += file_nobs
            nb_alt.extend([nvert]*file_nobs)

            # ~~~~~~~~~~#
            # MODIF NE MARCHERA PLUS POUR CERTAINS FICHIERS !! ##

            # get times
            # ds.select('Profile_UTC_Time')[:]#
            file_times = [datetime.datetime.strptime(str(int(t)).zfill(6), '%y%m%d') +
                          datetime.timedelta(days=(t % 1))
                          for t in np.squeeze(
                            ds.select('Profile_UTC_Time')[:, 1])]
            times.extend(file_times)

            # get coords
            # lats.extend(list(np.squeeze(ds.select('Latitude')[:])))
            # lons.extend(list(np.squeeze(ds.select('Longitude')[:])))
            lats.extend(list(np.squeeze(ds.select('Latitude')[:, 1])))
            lons.extend(list(np.squeeze(ds.select('Longitude')[:, 1])))
            vert.extend(list(np.squeeze(ds.select(self.vertName)[:])))

            # ~~~~~~~~~#
            # get data
            for i in range(file_nobs):
                prof.append(ds.select(dset_name)[i, :])

        # check data integrity
        if len(np.unique(units)) != 1:
            raise Exception('Different units found in files: {}'.format(
                np.unique(units)))
        wv = str(wavelength)
        if len(np.unique(vert_units)) != 1:
            raise Exception('Different altitude units found in ' +
                            'files: {}'.format(np.unique(vert_units)))

        # give all measures the same number of levels
        nmaxlevs = np.max(nb_alt)
        vert_padded = [np.append(p, np.full(nmaxlevs-n, self.missing_value))
                       for n, p in zip(nb_alt, vert)]
        vert_padded = np.array(vert_padded)
        prof_padded = [np.append(p, np.full(nmaxlevs-n, self.missing_value))
                       for n, p in zip(nb_alt, prof)]
        prof_padded = np.array(prof_padded)

        # define instrument
        self.__init__(self.name, wavelength=wv, nmaxlevs=nmaxlevs,
                      vert_name=self.vertName)
        self.vert_unit = vert_units[0]
        self.units = {wv: units[0]}
        self.nobs = nobs
        self.time = np.ma.concatenate(
            (self.time, np.ma.array(times, dtype=self.np_typetime)))
        self.lats = np.ma.concatenate(
            (self.lats, np.ma.array(lats, dtype=self.lats.dtype)))
        self.lats.fill_value = self.missing_value
        self.lons = np.ma.concatenate(
            (self.lons, np.ma.array(lons, dtype=self.lons.dtype)))
        self.lons.fill_value = self.missing_value
        self.mask = np.concatenate(
            (self.mask, np.zeros((nobs), dtype=bool)))
        self.vert = np.ma.concatenate(
            (self.vert, np.ma.array(vert_padded, dtype=self.vert.dtype)))
        self.vert.fill_value = self.missing_value
        self.prof[wv] = np.ma.concatenate(
            (self.prof[wv],
             np.ma.array(prof_padded,
                         dtype=self.prof[wv].dtype,
                         fill_value=self.missing_value)))
        self.prof[wv].fill_value = self.missing_value

        # millibars to Pa
        if self.vert_unit in ['hPA', 'millibars']:
            self.vert *= 1e2
            self.vert_unit = 'Pa'
        elif self.vert_unit == 'Pa':
            pass
        else:
            raise Exception('Vertical unit {} not supported.'.format(
                self.vert_unit))

        # km-1 to m-1
        if self.units[wv] in ['km^-1', 'per kilometer']:
            self.prof[wv] *= 1e-3
            self.units[wv] = 'm^-1'
        elif self.units[wv] in ['km^-1 sr^-1', 'per kilometer per steradian']:
            self.prof[wv] *= 1e-3
            self.units[wv] = 'm^-1 sr^-1'
        else :
            raise Exception('Data unit {} not supported.'.format(
                self.units[wv]))

        # mask negative values and values above filter_max
        idx = reduce(np.logical_or, [self.prof[wv] < 0,
                                     self.prof[wv] > filter_max])
        self.prof[wv][idx] = np.ma.masked
        self.vert[idx] = np.ma.masked

        # vertical interpolation
        if vertical_levels is not None:
            self.verticalRegrid(vertical_levels, interp='layerAverage')
