"""
Ash total column observation class
from EUMETSAT CAP dataset
Created on Thu May 12 2021

@author: frebourg
"""

import os
import numpy as np
import netCDF4
import glob
import datetime
from colconc_instr import ColConcInstrument
#from datetime import datetime,timedelta
import numpy.ma as ma
import os.path

class CAP_ASH(ColConcInstrument):
	# Class for ASH loading total column
	# CAP ASH is a geostationnary product
	# Datas are set on a square matrix with a mask cutting a circle
	# in this matrix.
    def __init__(self, name, **kwargs):
    	#nmaxlevs =
        kwargs['species'] = ['Totash']
        kwargs['nmaxlevs'] = 1
        ColConcInstrument.__init__(self, name, **kwargs)

    def Convert(self, period, inputdir, meta_inputfile, dset_name,
                **kwargs):
        """ Read data from a directory.

        File names format must be:
        {prefix}.{year}-{month}-{day}T??-??-??ZD.hdf
        for ex : CAL_LID_L15_Exp-Beta-V3-40.2017-02-22T01-30-00ZD.hdf

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input directory.
        meta_inputfile : str
        	Inpur directory and file for metadata file
        dset_name : str
            Name of the netcdf4 variable where to find the coefficients.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.    
		"""

        # -- Find Files -- #
        nbdays = (period[1] - period[0]).days
        dates = (period[0] + datetime.timedelta(i) for i in range(nbdays))
        file_list = []
        file_name = "MSG4-SEVI-MSGVOLE-0101-0101-{date}????00.000000000Z-"+\
                        "{date}??????-???????.nc"
        for d in dates :
            file_list.extend(
                sorted(glob.glob(os.path.join(
                    inputdir,
                    file_name.format(date=d.strftime('%Y%m%d'))))))
        if not os.path.isfile(meta_inputfile):
            print("No file found for metadata")
        if len(file_list) > 0:
            self.convert_list_of_files(file_list=file_list,
                                       meta_inputfile=meta_inputfile,
                                       dset_name=dset_name,
                                       **kwargs)
        else:
            print("No file found for period {}".format(period))


    def convert_list_of_files(self, file_list, meta_inputfile, dset_name,
                              **kwargs):
        """ Reads necessary data from files.

        Parameters
        ----------
        file_list : list of str
            List of input files to read.
        meta_inputfile : str
        	Inpur directory and file for metadata file
        dset_name : str
            Name of the netcdf4 variable where to find the coefficients.
        """
        # get Coordinates from metafile
        ds_geo = netCDF4.Dataset(meta_inputfile)
        lat = ds_geo.variables["latitude"][:]
        lon = ds_geo.variables["longitude"][:]

        # -- Read Files -- #
        for f in file_list:
            ds = netCDF4.Dataset(f)
            # Time
            strtime = ds.measurement_time
            time = datetime.datetime.strptime(strtime,'%Y-%m-%dT%H:%M:%SZ')
            print('Time: ', time)
            # Extract data
            data = ds.variables[dset_name][:]
            qflag = ds.variables['ash_height_quality_flag'][:]
            # Get mask
            mask = data.mask

            # Convertion to kg/m2 (if necessary)
            units = ds.variables[dset_name].units
            if units in ["g/m2","g.m-2"]:
                units = "kg/m2"
                data = data/1000.

            # Define threshold to keep the pixels.
            # Take a define threshold or by defaut 0
            # otherwise the data is full of zeros!
            if "THRmin" in kwargs:
                thrmin = float(kwargs["THRmin"]["thr"])
                thr_mask_min = (data <= thrmin)
                mask = np.ma.mask_or(mask, thr_mask_min)
            else:
                thrmin = 0
                thr_mask_min = (data <= thrmin)
                mask = np.ma.mask_or(mask, thr_mask_min)
            # Define high value threshold if considered as noise.
            if "THRmax" in kwargs:
                thrmax = float(kwargs["THRmax"]["thr"])
                thr_mask_max = (data >= thrmax)
                mask = np.ma.mask_or(mask, thr_mask_max)
            # mask height quality flag
            # 0 : ash_and_height / 1 : ash_no_height / 2 : no_ash
            # 3 : not_processed
            if 'QFLAG_max' in kwargs :
                filter_qflag = int(kwargs['QLAG_max'])
                mask_value = (qflag >= filter_qflag)
                mask = np.ma.mask_or(mask, mask_value)

            # Apply mask on dataset 
            # Select only unmasked data
            data.mask = mask
            lat.mask = mask
            lon.mask = mask 
            vect_data = data.compressed()
            vect_data = vect_data.reshape(vect_data.size,1)
            vect_lat = lat.compressed()
            vect_lon = lon.compressed()
            # get number of observations
            nobs = vect_data.size
            # define time 
            listdate = [time for i in range(nobs)]
            # set pressure bottom and top of the column.
            if self.press.size == 0:
                self.press_isconst = True
                self.press = np.ma.concatenate(
                    (self.press, np.ma.array(np.array([[530, 101300]]),
                                             dtype=np.float32)))
                self.press.fill_value = self.missing_value
            # Output variables
            #self.species = ['ash']
            #self.nmaxlevs = 1
            self.nobs = self.nobs + nobs
            self.units = {self.species[0]:units}
            self.col[self.species[0]] = np.ma.concatenate(
                    (self.col[self.species[0]], 
                     np.ma.array(vect_data,
				 dtype=self.col[self.species[0]].dtype)))
            self.col[self.species[0]].fill_value = self.missing_value
            self.lats = np.ma.concatenate(
                    (self.lats, np.ma.array(vect_lat, dtype=self.lats.dtype)))
            self.lats.fill_value = self.missing_value
            self.lons = np.ma.concatenate(
                    (self.lons, np.ma.array(vect_lon, dtype=self.lons.dtype)))
            self.lons.fill_value = self.missing_value
            self.mask = np.concatenate(
                    (self.mask, np.zeros((nobs), dtype=bool)))
            self.time = np.ma.concatenate(
                    (self.time, np.ma.array(listdate, dtype=self.np_typetime)))
            ds.close()
