# coding: utf8

import os
import numpy as np
import datetime
import netCDF4
from lidar_instr import LidarInstrument
import glob
from functools import reduce


class EarlinetLidar(LidarInstrument):
    """ Class for EARLINET lidar data.

    Parameters
    ----------
    name : str
        Name of the instrument
    wllist : list of str
        List of studied wavelengths.
    nmaxlevs : int
        Maximal number of vertical levels.
    hascov : bool
        If True, covariance errors are computed.
    vert_isconst : bool
        Tells wether the vertical axis is constant or variable for each
        measurement.
    kwargs : dict
        Used to allow unexpected keyword arguments.

    """
    def __init__(self, name, wllist=None, nmaxlevs=0, hascov=False,
                 vert_isconst=False, **kwargs):
        LidarInstrument.__init__(
            self, name, wllist=wllist, nmaxlevs=nmaxlevs, hascov=hascov,
            upward=True, vert_isconst=vert_isconst)
        self.station_names = np.ma.empty((0), dtype=np.str)

    def Convert(self, period, inputdir, wavelength, nmaxlevs, data_type='b',
                filter_max=1e30, vertical_levels=None, **kwargs):
        """ Read data from a directory.

        File names format must be: *yyddhhmmss.{data_type}{wavelength}

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input directory.
        wavelength : str
            Wavelength.
        nmaxlevs : int
            Number of vertical levels. This number must be greater than
            the maximal number of vertical levels found in input files.
        data_type : str, optional
            Type of coefficient as specified in the filename (e.g. 'b'
            for backscatter or 'e' for extinction).
        filter_max : float, optional
            Threshold above which values are considered missing.
        vertical_levels : None or 1D array-like, optional
            If not None, data is interpolated vertically according to given
            levels.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.

        """
        nbdays = (period[1] - period[0]).days
        dates = (period[0] + datetime.timedelta(d) for d in range(nbdays))
        file_list = []
        for d in dates:
            file_list.extend(
                glob.glob(os.path.join(
                    inputdir,
                    "??{date}????.{data_type}{wavelength}".format(
                        date=d.strftime('%y%m%d'),
                        data_type=data_type,
                        wavelength=wavelength))))
        if len(file_list) == 0:
            print("No file found for period {}".format(period))
        self.convert_list_of_files(
            file_list,
            wavelength=str(wavelength),
            nmaxlevs=nmaxlevs,
            filter_max=filter_max,
            vertical_levels=vertical_levels)

    def convert_list_of_files(self, file_list, wavelength, nmaxlevs,
                              filter_max=1e30, vertical_levels=None):
        """ Reads necessary data from files.

        Works for a single wavelength.

        Parameters
        ----------
        file_list : list of str
            List of input files to read.
        wavelength : str
            Wavelength.
        nmaxlevs : int
            Number of vertical levels. This number must be greater than
            the maximal number of vertical levels found in input files.
        filter_max : float, optional
            Threshold above which values are considered missing.
        vertical_levels : None or 1D array-like, optional
            If not None, data is interpolated vertically according to given
            levels.

        """
        units = []
        wvs = []
        vert_units = []
        nb_alt = []
        times = []
        lats = []
        lons = []
        vert = []
        prof = []
        stations = []
        vertName = 'Altitude'
        for f in file_list:
            ds = netCDF4.Dataset(f, 'r')

            # get backscatter unit
            if 'units' in ds['Backscatter'].ncattrs():
                units.append(ds['Backscatter'].units)
            else:
                units.append('')

            # get wavelength
            wvs.append(ds.EmissionWavelength_nm)

            # get vertical axis unit
            if 'units' in ds[vertName].ncattrs():
                vert_units.append(ds[vertName].units)
            else:
                vert_units.append('')

            # get number of vertical levels
            nb_alt.append(ds['Backscatter'].shape[0])

            # get times
            times.append(datetime.datetime.strptime(
                str(ds.StartDate).split('.')[0].zfill(6) +
                str(ds.StartTime_UT).split('.')[0].zfill(6),
                '%Y%m%d%H%M%S'))

            # get coords
            lats.append(ds.Latitude_degrees_north)
            lons.append(ds.Longitude_degrees_east)
            vert.append(ds[vertName][:])

            # get data
            prof.append(ds['Backscatter'][:])

            # get station names
            stations.append(ds.Location)

        # check data integrity
        if len(np.unique(units)) > 1:
            raise Exception('Different units found in files: {}'.format(
                np.unique(units)))
        if len(np.unique(wvs)) > 1:
            raise Exception('Different wavelengths found in files: {}'.format(
                np.unique(wvs)))
        if len(np.unique(vert_units)) > 1:
            raise Exception('Different altitude units found in ' +
                            'files: {}'.format(np.unique(vert_units)))

        nobs = len(file_list)
        if nobs > 0:
            # give all measures the same number of levels
            if nmaxlevs < np.max(nb_alt):
                raise Exception("nmaxlevs arguments must be >= {}".format(
                    np.max(nb_alt)))
            vert_padded = [np.append(p, np.full(nmaxlevs-n, self.missing_value))
                           for n, p in zip(nb_alt, vert)]
            vert_padded = np.array(vert_padded)
            vert_padded = np.where(vert_padded < filter_max,
                                   vert_padded,
                                   self.missing_value)
            prof_padded = [np.append(p, np.full(nmaxlevs-n, self.missing_value))
                           for n, p in zip(nb_alt, prof)]
            prof_padded = np.array(prof_padded)
            prof_padded = np.where(prof_padded < filter_max,
                                   prof_padded,
                                   self.missing_value)

        # define instrument
        self.__init__(self.name, wllist=[wavelength], nmaxlevs=nmaxlevs)
        self.vertName = vertName
        if nobs > 0:
            self.vert_unit = vert_units[0]
            self.units = {wavelength: units[0]}
            self.nobs += nobs
            self.time = np.ma.concatenate(
                (self.time, np.ma.array(times, dtype=self.np_typetime)))
            self.lats = np.ma.concatenate(
                (self.lats, np.ma.array(lats, dtype=self.lats.dtype)))
            self.lats.fill_value = self.missing_value
            self.lons = np.ma.concatenate(
                (self.lons, np.ma.array(lons, dtype=self.lons.dtype)))
            self.lons.fill_value = self.missing_value
            self.mask = np.concatenate(
                (self.mask, np.zeros((nobs), dtype=bool)))
            self.vert = np.ma.concatenate(
                (self.vert, np.ma.array(vert_padded, dtype=self.vert.dtype)))
            self.vert.fill_value = self.missing_value
            self.prof[wavelength] = np.ma.concatenate(
                (self.prof[wavelength],
                 np.ma.array(prof_padded,
                             dtype=self.prof[wavelength].dtype,
                             fill_value=self.missing_value)))
            self.prof[wavelength].fill_value = self.missing_value
            self.station_names = np.ma.concatenate(
                (self.station_names,
                 np.ma.array(stations)))
            self.station_names.fill_value = self.missing_value

            # mask negative values and values above filter_max
            idx = reduce(np.logical_or, [self.prof[wavelength] < 0,
                                         self.prof[wavelength] > filter_max])
            self.prof[wavelength][idx] = np.ma.masked
            self.vert[idx] = np.ma.masked

            # vertical interpolation
            if vertical_levels is not None:
                self.verticalRegrid(vertical_levels, interp='layerAverage')
        else:
            self.vert_unit = ''
            self.units = {wavelength: ''}
