# coding: utf8

from __future__ import print_function
import os
import numpy as np
import datetime
from surfconc_instr import SurfConcInstrument
import glob
from functools import reduce
import pandas as pd


class EeaSurfaceObservations(SurfConcInstrument):
    """ Class for EEA surface observation data.

    One object of this class can handle only one species.

    Parameters
    ----------
    name : str
        Name of the instrument
    species : str
        Species to process.
    kwargs : dict
        Used to allow unexpected keyword arguments.

    """
    # species names are in lower case in input csv files
    csv_species_names = {'O3': 'o3', 'NO2': 'no2', 'PM10': 'pm10',
                         'PM2.5': 'pm2p5','CO': 'co','SO2': 'so2'}

    def __init__(self, name, species, **kwargs):
        if species not in self.csv_species_names:
            raise Exception('Unsupported species {}'.format(species))
        SurfConcInstrument.__init__(self, name, species=[species])
        if species == 'NO2':
            self.hascov = True

    def Convert(self, period, inputdir, filter_max=1e30,
                minutesToSubtractToTime=0, station_types_file=None, **kwargs):
        """ Read data from a directory.

        File names format must be:
        obsmacc4ana_{date}.csv

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input directory.
        filter_max : float, optional
            Threshold above which values are considered missing.
        minutesToSubtractToTime : int
            Number of minutes to substract to each times.
        station_types_file : str
            Path of the station listing. This listing is a text file
            with two columns separated by a space. The first column
            contains the station codes and the second one contains
            the corresponding station types.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.

        """
        nbdays = (period[1] - period[0]).days
        dates = (period[0] + datetime.timedelta(d) for d in range(nbdays))
        file_list = []
        for d in dates:
            file_list.extend(
                glob.glob(os.path.join(
                    inputdir,
                    "obsmacc4ana_{date}.csv".format(
                        date=d.strftime('%Y%m%d')))))
        if len(file_list) > 0:
            self.convert_list_of_files(
                file_list=file_list,
                filter_max=filter_max,
                minutesToSubtractToTime=minutesToSubtractToTime,
                station_types_file=station_types_file)
        else:
            print("No file found for period {}".format(period))

    def convert_list_of_files(self, file_list, filter_max=500*1e-9,
                              minutesToSubtractToTime=0,
                              station_types_file=None):
        """ Reads necessary data from files.

        The columns of the csv files must be:
        STATION;LAT;LON;ALT(m);PARAMETER;YEAR;MONTH;DAY;HOUR;
        AVERAGING_PERIOD(h);CONCENTRATION(kg/m3)

        Parameters
        ----------
        file_list : list of str
            List of input files to read.
        filter_max : float, optional
            Threshold (in kg/m3) above which values are dropped.
        minutesToSubtractToTime : int
            Number of minutes to substract to each times.
        station_types_file : str
            Path of the station listing. This listing is a text file
            with two columns separated by a space. The first column
            contains the station codes and the second one contains
            the corresponding station types.

        """
        # conversion factors from volume mixing ratio (ppb) to kg/m3
        convert_unit = {'o3': 1.99, 'co': 1.16, 'so2': 2.66, 'no2': 1.91,
                        'pm10': 1., 'pm2p5': 1., 'no': 1.25}
        self.units[self.species[0]] = 'ppb'
        spe = self.csv_species_names[self.species[0]]

        conc = []
        cov = []
        times = []
        lats = []
        lons = []
        for f in file_list:
            # read csv
            df = pd.read_table(f, sep=';', header=0, index_col=False)
            df = df.loc[df.PARAMETER == spe]
            # remove negative values and values above filter_max
            idx = ~np.logical_or(df['CONCENTRATION(kg/m3)'] < 0,
                                 df['CONCENTRATION(kg/m3)'] > filter_max)
            df = df.loc[idx]
            nobs = df.shape[0]
            print("{spe} total number of observations in {f}: {nobs}".format(
                spe=self.species[0], f=os.path.basename(f), nobs=nobs))

            # convert dates to datetime.date objects
            dateh = [datetime.datetime(yy, mm, dd, hh) -
                     datetime.timedelta(minutes=minutesToSubtractToTime)
                     for yy, mm, dd, hh
                     in zip(df.YEAR, df.MONTH, df.DAY, df.HOUR)]

            # compute VMR
            vmr = df['CONCENTRATION(kg/m3)']/convert_unit[spe]

            # covariance calculation
            # parameters specific to station types (MF values)
            # abserror = {'rural': 1.*1e-9, 'urban': 6.*1e-9,
            #             'suburban': 4.*1e-9,
            #             'unknown': 20.*1e-9}  # sigma in vmr
            # relerror = {'rural': 0., 'urban': 0., 'suburban': 0.,
            #             'unknown': 0.}  # percentage
            # reprerror = {'rural': 0., 'urban': 0., 'suburban': 0.,
            #              'unknown': 0.}  # sigma in vmr
            
            abserror = {'rural': 1.*1e-9, 'urban': 6.*1e-9,
                        'suburban': 4.*1e-9,
                        'unknown': 4.*1e-9}  # sigma in vmr
            relerror = {'rural': 0., 'urban': 0., 'suburban': 0.,
                        'unknown': 0.}  # percentage
            reprerror = {'rural': 0., 'urban': 0., 'suburban': 0.,
                         'unknown': 0.}  # sigma in vmr
            
            if self.hascov:
                cov = []
                dict_sta = {}
                if station_types_file:
                    with open(station_types_file) as f:
                        for line in f:
                            (key, val) = line.split()
                            if key not in dict_sta:
                                dict_sta[key] = val
                else:
                    print('No listing found for covariance calculation, ' +
                          'all station types are considered unknown.')
                for s, val in zip(df.STATION, vmr):
                    if s in dict_sta:
                        area_type = dict_sta[s]
                    else:
                        area_type = 'unknown'
                    cov.append(float(np.maximum(val*relerror[area_type],
                                                abserror[area_type]) +
                                     reprerror[area_type]))

            # store data in attributes
            self.nobs += nobs  # increase the instrument counter
            if nobs > 0:
                self.conc[self.species[0]] = np.ma.concatenate(
                    (self.conc[self.species[0]],
                     np.ma.array(vmr, dtype=self.conc[self.species[0]].dtype)))
                self.lats = np.ma.concatenate(
                    (self.lats, np.ma.array(df.LAT, dtype=self.lats.dtype)))
                self.lons = np.ma.concatenate(
                    (self.lons, np.ma.array(df.LON, dtype=self.lons.dtype)))
                self.time = np.ma.concatenate(
                    (self.time, np.ma.array(dateh, dtype=self.np_typetime)))
                if self.hascov:
                    self.cov[self.species[0]] = np.ma.concatenate(
                        (self.cov[self.species[0]],
                         np.ma.array(cov,
                                     dtype=self.cov[self.species[0]].dtype)))
                self.mask = np.concatenate((self.mask,
                                            np.zeros((nobs), dtype=bool)))
