# coding: utf8

import os
import numpy as np
import datetime
import netCDF4
from lidar_instr import LidarInstrument
import glob
from functools import reduce


class EprofileLidar(LidarInstrument):
    """ Class for E-profile lidar data.

    Parameters
    ----------
    name : str
        Name of the instrument
    wllist : list of str
        List of studied wavelengths.
    nmaxlevs : int
        Maximal number of vertical levels.
    hascov : bool
        If True, covariance errors are computed.
    vert_isconst : bool
        Tells wether the vertical axis is constant or variable for each
        measurement.
    kwargs : dict
        Used to allow unexpected keyword arguments.

    """
    def __init__(self, name, wllist=None, nmaxlevs=0, hascov=False,
                 **kwargs):
        LidarInstrument.__init__(
            self, name, wllist=wllist, nmaxlevs=nmaxlevs, hascov=hascov,
            upward=True, vert_isconst=False)
        self.station_names = np.ma.empty((0), dtype=np.str)

    def Convert(self, period, inputdir, file_name, wavelength, dset_name,
                vert_name='altitude', time_name='time',
                lat_name='station_latitude', lon_name='station_longitude',
                date_format='%Y%m%d', filter_max=1e30,
                vertical_levels=None, quality_flag=[0], cloud_flag=[0],
                SNR_flag=[0], **kwargs):
        """ Read data from a directory.

        File names format must be: 

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input directory.
        file_name : str
            Generic file name to be read. This name is read by
            python module glob and can contain {date} instead of the
            required date.
        wavelength : str
            Wavelength.
        dset_name : str
            Name of the hdf5 dataset where to find the coefficients.
        vert_name : str
            Name of the vertical coordinate variable to retrieve from the file.
        time_name, lat_name, lon_name : str
            Name of the variable corresponding to time, latitude and longitude
            coordinates to retrieve from the file.
        date_format : str
            Date string format used in the file name. This argument is used
            only if file_name argument contain {date}.
        filter_max : float, optional
            Threshold above which values are considered missing.
        vertical_levels : None or 1D array-like, optional
            If not None, data is interpolated vertically according to given
            levels.
        quality_flag, cloud_flag, SNR_flag : None or list of int
            List of values allowed for these flags.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.
            
        """
        nbdays = (period[1] - period[0]).days
        dates = (period[0] + datetime.timedelta(d) for d in range(nbdays))
        file_list = []
        for d in dates:
            file_list.extend(
                glob.glob(os.path.join(
                    inputdir,
                    file_name.format(date=d.strftime(date_format)))))
        if len(file_list) == 0:
            print("No file found for period {}".format(period))
        self.convert_list_of_files(file_list=file_list,
                                   dset_name=dset_name,
                                   vert_name=vert_name,
                                   time_name=time_name,
                                   lat_name=lat_name,
                                   lon_name=lon_name,
                                   wavelength=str(wavelength),
                                   filter_max=filter_max,
                                   vertical_levels=vertical_levels,
                                   quality_flag=quality_flag,
                                   cloud_flag=cloud_flag,
                                   SNR_flag=SNR_flag)

    def convert_list_of_files(
        self, file_list, wavelength, dset_name, vert_name, time_name,
        lat_name, lon_name, filter_max, vertical_levels,
        quality_flag, cloud_flag, SNR_flag):
        """ Read necessary data from files.

        Works for a single wavelength.

        Parameters
        ----------
        file_list : list of str
            List of input files to read.
        wavelength : str
            Wavelength.
        dset_name : str
            Name of the hdf5 dataset where to find the coefficients.
        vert_name : str
            Name of the vertical coordinate variable to retrieve from the file.
        time_name, lat_name, lon_name : str
            Name of the variable corresponding to time, latitude and longitude
            coordinates to retrieve from the file.
        filter_max : float, optional
            Threshold above which values are considered missing.
        vertical_levels : None or 1D array-like, optional
            If not None, data is interpolated vertically according to given
            levels.
        quality_flag, cloud_flag, SNR_flag : None or list of int
            List of values allowed for these flags.

        """
        units = []
        wvs = []
        vert_units = []
        nb_alt = []
        times = []
        lats = []
        lons = []
        vert = []
        prof = []
        stations = []
        nobs = 0        
        for f in file_list:
            ds = netCDF4.Dataset(f, 'r')
                
            # get var unit
            if 'units' in ds[dset_name].ncattrs():
                units.append(ds[dset_name].units)
            else:
                units.append('')

            # get vertical axis unit
            if 'units' in ds[vert_name].ncattrs():
                vert_units.append(ds[vert_name].units)
            else:
                vert_units.append('')

            # get number of observations and number of vertical levels
            # file_nobs, nvert = ds[dset_name][:].shape
            file_nobs = ds[time_name].shape[0]
            nvert = ds[vert_name].shape[0]
            nobs += file_nobs
            nb_alt.extend([nvert]*file_nobs)

            # find dataset dimensions order
            if ds[dset_name].dimensions == (time_name, vert_name):
                order = 'time_first'
            elif ds[dset_name].dimensions == (vert_name, time_name):
                order = 'vert_first'
            else:
                raise Exception("Error when reading dataset dimensions: " +
                                "dimensions do not match vert_name and " + 
                                "time_name.")

            # get times
            times.extend(netCDF4.num2date(ds[time_name][:],
                                          ds[time_name].units))

            # get coords
            if ds[lat_name].shape == ():
                lats.extend([ds[lat_name][0]]*file_nobs)
                lons.extend([ds[lon_name][0]]*file_nobs)
            else:
                lats.extend(list(np.squeeze(ds[lat_name][:])))
                lons.extend(list(np.squeeze(ds[lon_name][:])))

            # get data
            data = ds[dset_name][:]

            # quality flags
            if quality_flag is not None:
                if 'quality_flag' in ds.variables.keys():
                    data[~np.isin(ds['quality_flag'],
                                  quality_flag)] = self.missing_value
                else:
                    print("quality_flag not found in file {}".format(f))
            if cloud_flag is not None:
                if 'cloud_flag' in ds.variables.keys():
                    data[~np.isin(ds['cloud_flag'],
                                  cloud_flag)] = self.missing_value
                else:
                    print("cloud_flag not found in file {}".format(f))
            if SNR_flag is not None:
                if 'SNR_flag' in ds.variables.keys():
                    data[~np.isin(ds['SNR_flag'],
                                  SNR_flag)] = self.missing_value
                else:
                    print("SNR_flag not found in file {}".format(f))

            for i in range(file_nobs):
                vert.append(ds[vert_name][:])
                if order == 'time_first':
                    prof.append(data[i, :])
                else:
                    prof.append(data[:, i])

            # check wavelength matching
            if 'laser_wavelength' in ds.variables.keys():
                assert(int(wavelength) ==
                       int(ds.variables['laser_wavelength'][0]))
            elif 'lidar_wavelength' in ds.ncattrs():
                assert(int(wavelength) ==
                       int(ds.lidar_wavelength.split()[0]))
            else:
                raise Exception("No wavelength found in file.")

        # check data integrity
        if len(np.unique(units)) > 1:
            raise Exception('Different units found in files: {}'.format(
                np.unique(units)))
        if len(np.unique(vert_units)) > 1:
            raise Exception('Different altitude units found in ' +
                            'files: {}'.format(np.unique(vert_units)))

        # give all measures the same number of levels
        if nobs > 0:
            nmaxlevs = np.max(nb_alt)
            vert_padded = [np.append(p, np.full(nmaxlevs-n, self.missing_value))
                           for n, p in zip(nb_alt, vert)]
            vert_padded = np.array(vert_padded)
            prof_padded = [np.append(p, np.full(nmaxlevs-n, self.missing_value))
                           for n, p in zip(nb_alt, prof)]
            prof_padded = np.array(prof_padded)
        else:
            nmaxlevs = 1

        # define instrument
        self.__init__(self.name, wllist=[wavelength], nmaxlevs=nmaxlevs)
        self.vertName = 'Altitude'
        if nobs > 0:
            self.vert_unit = vert_units[0]
            self.units = {wavelength: units[0]}
            self.nobs += nobs
            self.time = np.ma.concatenate(
                (self.time, np.ma.array(times, dtype=self.np_typetime)))
            self.lats = np.ma.concatenate(
                (self.lats, np.ma.array(lats, dtype=self.lats.dtype)))
            self.lats.fill_value = self.missing_value
            self.lons = np.ma.concatenate(
                (self.lons, np.ma.array(lons, dtype=self.lons.dtype)))
            self.lons.fill_value = self.missing_value
            self.mask = np.concatenate(
                (self.mask, np.zeros((nobs), dtype=bool)))
            self.vert = np.ma.concatenate(
                (self.vert, np.ma.array(vert_padded, dtype=self.vert.dtype)))
            self.vert.fill_value = self.missing_value
            self.prof[wavelength] = np.ma.concatenate(
                (self.prof[wavelength],
                 np.ma.array(prof_padded,
                             dtype=self.prof[wavelength].dtype,
                             fill_value=self.missing_value)))
            self.prof[wavelength].fill_value = self.missing_value
            self.station_names = np.ma.concatenate(
                (self.station_names,
                 np.ma.array(stations)))
            self.station_names.fill_value = self.missing_value
            
            # mask negative values and values above filter_max
            idx = reduce(np.logical_or, 
                         [self.prof[wavelength] < 0,
                          self.prof[wavelength] > filter_max,
                          self.prof[wavelength] == self.missing_value])
            self.prof[wavelength][idx] = np.ma.masked
            #self.vert[idx] = np.ma.masked
            
            # km-1 to m-1
            if self.units[wavelength] in ['km^-1', 'km-1', 'per kilometer']:
                self.prof[wavelength] *= 1e-3
                self.units[wavelength] = 'm^-1'
            elif self.units[wavelength] in ['Mm^-1.sr^-1']:
                self.prof[wavelength] *= 1e-6
                self.units[wavelength] = 'm^-1.sr^-1'
            else:
                raise Exception('Data unit {} not supported.'.format(
                    self.units[wavelength]))

            # vertical interpolation
            if vertical_levels is not None:
                self.verticalRegrid(vertical_levels, interp='layerAverage')
        else:
            self.vert_unit = ''
            self.units = {wavelength: ''}
        
