""" IAGOS measurements processing module """

import os
import sys
import numpy as np
from datetime import datetime, timedelta
from profconc_instr import ProfConcInstrument
from netCDF4 import Dataset, num2date
import glob
from functools import reduce
import pandas as pd


class IagosFlight(ProfConcInstrument):
    """ Class designed to process IAGOS measurements.

    Parameters
    ----------
    name : str
        Name of the instrument.
    species : list of str, optional
        List of species stored in the object.

    """
    def __init__(self, name, species, **kwargs):
        ProfConcInstrument.__init__(
            self, name=name, nmaxlevs=1, species=species, **kwargs)

        # Auxiliary info from IAGOS intruments
        # They may latter be used for filtering/masking data development
        self.statflag = np.ma.empty((0), dtype=np.int16)  # status flag
        self.valflag = np.ma.empty((0), dtype=np.int16)  # validation flag

    def Convert(self, period=None, inputdir='./', freq=None, **kwargs):
        """ Read IAGOS raw input files.

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input file directory.

        """
        print(inputdir)
        prof = {spc: np.ma.empty((0, self.nmaxlevs), dtype=np.float32)
                for spc in self.species}
        pres = np.ma.empty((0, self.nmaxlevs), dtype=np.float32)

        if period is not None:
            nbdays = (period[1] - period[0]).days
            dates = (period[0] + timedelta(d-1)
                     for d in range(nbdays + 1))
            file_list = []
            for d in dates:
                file_list.extend(
                    glob.glob(os.path.join(
                        inputdir,
                        "IAGOS_timeseries_{}????????.nc".format(
                            d.strftime('%Y%m%d'))))
                )
        else:
            file_list = glob.glob(inputdir)

        for f in file_list:
            print(">>> Working on ", os.path.basename(f))
            fid = Dataset(f, mode='r')
            lat = fid.variables['lat'][:]
            lon = fid.variables['lon'][:]
            intr_type = fid.source
            if (intr_type[-4:] == 'CORE'):
                suffix = '_P1'
            elif (intr_type[-6:] == 'MOZAIC'):
                suffix = '_PM'
            elif(intr_type[-7:] == 'CARIBIC'):
                number = fid.title.split(", ")[1][-1]
                suffix = '_PC'+number
            else:
                raise Exception("Error: instrument type not supported")

            time = num2date(
                fid.variables['UTC_time'][:],
                fid.variables['UTC_time'].units,
                fid.variables['UTC_time'].calendar,
            )
            if freq is not None:
                time = pd.DatetimeIndex(time).round(freq)

            press = fid.variables['air_press_AC'][:]
            # press = np.reshape(press, (nobsfile, self.nmaxlevs))

            for spec in self.species:
                print("Retrieving ", spec, "measurements...")
                prof[spec] = fid.variables[spec + suffix][:]
                # prof[spec] = np.reshape(prof[spec], (nobsfile, self.nmaxlevs))
                cnvfctr = float(fid.variables[spec + suffix].units)
                prof[spec] *= cnvfctr

            df = pd.DataFrame(
                dict(
                    time=time,
                    lat=lat.filled(np.nan),
                    lon=lon.filled(np.nan),
                    press=press.filled(np.nan),
                    **{spec: prof[spec].filled(np.nan) for spec in self.species}
                )
            )
            if freq is not None:
                df = df.groupby('time').median()

            self.nobs += df.shape[0]
            self.time = np.ma.concatenate(
                (self.time, df.index.values.astype(self.np_typetime))
            )
            self.lats = np.ma.concatenate(
                (self.lats, df['lat'].values.astype(self.lats.dtype)))
            self.lats.fill_value = self.missing_value
            self.lons = np.ma.concatenate(
                (self.lons, df['lon'].values.astype(self.lons.dtype)))
            self.lons.fill_value = self.missing_value

            self.press = np.ma.concatenate(
                (self.press, df['press'].values[:, np.newaxis])
            )
            self.press.fill_value = self.missing_value

            for spec in self.species:
                self.prof[spec] = np.ma.concatenate(
                    (self.prof[spec], df[spec].values[:, np.newaxis]))
                self.prof[spec].fill_value = self.missing_value
                self.prof[spec][np.isnan(self.prof[spec])] = np.ma.masked

            mask = reduce(
                np.logical_or,
                (
                    np.isnan(self.lats),
                    np.isnan(self.lons),
                ),
            )
            self.mask = np.concatenate((self.mask, mask))

            fid.close()

        if period is not None:
            self.Select(timebnd=period)
