# -*- coding: utf-8 -*-
"""
Created on Tue Jan 16 17:02:31 2018

@author: emili
"""
from __future__ import print_function
import os
import sys
import numpy as np
from datetime import datetime, timedelta
import time
try:
    import utils_Hobs_mini as UtilsASCII # needed for conversion of old MOCAGE obs files
except ImportError:
    pass
import csv
from profconc_instr import ProfConcInstrument
import fnmatch
import re
try:
    import cdms2 as cdms # needed for vertical regridding of raw radiosoundings data
    import MV2 as MV 
except ImportError:
    pass
try:
    from netCDF4 import Dataset # needed for netcdf storage
except ImportError:
    pass

class RadioSondeInstrument(ProfConcInstrument):
    
    def __init__(self,name,**kwargs):

        ProfConcInstrument.__init__(self,name,**kwargs)
        self.press_coorddir = 'downward'
        self.press_isconst = True
        self.SiteID = np.empty((0),dtype='S6')
        self.hascov = False # Switch to treat or not error covariances
        self.hasavk = False # Switch to treat or not averaging kernels

    def ConvertMocageASCII(self,date,inputdir):
        """
        Reads monospecies observations from old MOCAGE ASCII
        """                          
        spc = self.species[0] # Old ASCII are monospecies, so take only one if multiple species are present (the first)!
        
        # Use the old MOCAGE HDAT reader for now
        obstype, obsdate, obsname, obslat, obslon, obsval, obspress = \
            UtilsASCII.LecHOBS(os.path.join(inputdir, \
            'HDAT+'+date.isoformat()[:10].replace("-","")),obstype='profile')
            
        nobs = len(obsdate)
            
        obsname = [n[0] for n in obsname] # Reduce old strings found in HDATs to a single site code
        obsdate = [datetime.strptime(t,"%Y%m%d%H%M") for t in obsdate] # Datetime conversion
        
        # Convert observation lists to numpy arrays
        nppress = np.full((nobs,self.nmaxlevs),self.missing_value,dtype=self.press.dtype)
        npval = np.full((nobs,self.nmaxlevs),self.missing_value,dtype=self.prof[spc].dtype)
        
        for ip in range(nobs):
            nppress[ip,:len(obspress[ip])] = obspress[ip]
            npval[ip,:len(obspress[ip])] = obsval[ip]
        
        self.nobs += nobs # Increase the instrument counter
        
        if nobs > 0:
            self.press = np.ma.array(np.ma.concatenate((self.press,np.ma.array(nppress))),fill_value=self.missing_value)
            self.prof[spc] = np.ma.array(np.ma.concatenate((self.prof[spc],np.ma.array(npval))),fill_value=self.missing_value)
            self.lats = np.ma.concatenate((self.lats,np.ma.array(obslat,dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate((self.lons,np.ma.array(obslon,dtype=self.lons.dtype)))
            self.time = np.ma.concatenate((self.time,np.ma.array(obsdate,dtype=self.time.dtype)))
            self.SiteID = np.ma.concatenate((self.SiteID,np.ma.array(obsname,dtype=self.SiteID.dtype)))
            self.mask = np.concatenate((self.mask,np.zeros((len(obsdate)),dtype=bool))) # By default all obs are accepted
            
    def Convert(self,period,inputdir,**kwargs):
        """
        Convert WOUDC database files
        """
        provider = "WOUDC"
        if "provider" in kwargs:  provider = kwargs["provider"]

        spc = 'O3'
        if 'ConvertOpt' in kwargs:
            if 'PressureGrid' in kwargs['ConvertOpt']:
                self.press_isconst = True                 
        
        fcount=0
        pcount=0
        unpfiles=[]

        for date in [period[0] + timedelta(days=i) for i in range((period[1]-period[0]).days)]:
            print('Working on ',date.isoformat().replace('-','')[:8])
            siteslist = []
            if provider == 'WOUDC':
                filetypes = {'WOUDC':date.isoformat().replace('-','')[:8]+'.*.csv'}
            elif provider == 'HEGIFTOM':
                filetypes = {'WOUDC':date.isoformat().replace('-','')[:8]+'.*.csv',
                'SHADOZ':'*_'+date.isoformat().replace('-','')[:8]+'T*_SHADOZV06.dat',
                'NDACC':'*'+date.isoformat().replace('-','')[2:8]+'.b*'}
            matches = []
            fformat = []
            for root, dirnames, filenames in os.walk(inputdir):
                for filetype, pattern in filetypes.items():
                    for filename in fnmatch.filter(filenames, pattern):
                        matches.append(os.path.join(root, filename))
                        fformat.append(filetype)
            nprofiles = len(matches)
            
            print('Found ', nprofiles,' files')
            
            for fn,filename in enumerate(matches):
                
                fcount+=1

                print('Working on: ', filename)

                if fformat[fn] == 'WOUDC':
                    typeinstr = os.path.basename(filename)[9:12]
                    [press, o3p, lat, lon, time, stationid, sitename, pflag] = processWOUDCFile(filename,interpolate=self.press_isconst)
                elif fformat[fn] == 'SHADOZ':
                    typeinstr = ''
                    [press, o3p, lat, lon, time, stationid, sitename, pflag] = processSHADOZFile(filename,interpolate=self.press_isconst)
                elif fformat[fn] == 'NDACC':
                    typeinstr = ''
                    [press, o3p, lat, lon, time, stationid, sitename, pflag] = processNDACCFile(filename,interpolate=self.press_isconst)
                else:
                    print("File format converter for %s not available yet" % (fformat[fn]))

                if not pflag: unpfiles.append(filename)

                if pflag and sitename not in siteslist: # This is done to avoid accidental data duplication (WOUDC messy archive)

                    pcount+=1
                    siteslist.append(sitename)

                    code = stationid+typeinstr
                    print('Profile code: ', code)

                    nppress = np.ma.empty((1,self.nmaxlevs),dtype=self.press.dtype)
                    npval = np.ma.empty((1,self.nmaxlevs),dtype=self.prof[spc].dtype)
                    
                    # Fill only with found measurements
                    if not self.press_isconst:
                        nppress[0,:press.count()] = press.compressed()
                        nppress[0,press.count():] = np.ma.masked
                        npval[0,:o3p.count()] = o3p.compressed()
                        npval[0,o3p.count():] = np.ma.masked
                    else:
                        nppress[0,:] = press
                        npval[0,:] = o3p
                        
                    if time.split(':')[3][0] == '+': 
                        utchour = timedelta(hours=int(time.split(':')[0]),minutes=int(time.split(':')[1]))+timedelta(hours=int(time.split(':')[3]))
                    elif time.split(':')[3][0] == '-': 
                        utchour = timedelta(hours=int(time.split(':')[0]),minutes=int(time.split(':')[1]))-timedelta(hours=int(time.split(':')[3]))
                    else:
                        utchour = timedelta(hours=int(time.split(':')[0]),minutes=int(time.split(':')[1]))

                    nptime = [datetime(date.year,date.month,date.day,0,0)+utchour]

                    self.nobs += 1
                    if self.press_isconst and self.press.size == 0: self.press = np.ma.array(np.ma.concatenate((self.press,nppress)),fill_value=self.missing_value)
                    self.prof[spc] = np.ma.array(np.ma.concatenate((self.prof[spc],npval)),fill_value=self.missing_value)
                    self.lats = np.ma.concatenate((self.lats,np.ma.array([lat],dtype=self.lats.dtype)))
                    self.lons = np.ma.concatenate((self.lons,np.ma.array([lon],dtype=self.lons.dtype)))
                    self.time = np.ma.concatenate((self.time,np.ma.array(nptime,dtype=self.time.dtype)))
                    self.SiteID = np.ma.concatenate((self.SiteID,np.ma.array([code],dtype=self.SiteID.dtype)))
                    self.mask = np.concatenate((self.mask,np.zeros((1),dtype=bool))) # By default all obs are accepted

        print("Extracted %i profiles from %i files" % (pcount,fcount))
        print("The following files could not be processed for some reason:")
        print(unpfiles)

    def Read(self,filename,**kwargs):

        ProfConcInstrument.Read(self, filename, closefile=False, **kwargs)

        if 'ReadAssim' not in kwargs: 
            geogrp = self.f[self.name+"/GEOLOCALIZATION"]
            self.SiteID = np.ma.concatenate((self.SiteID,np.ma.array(geogrp['SoundingID'][:],dtype=self.SiteID.dtype)))

        self.CloseFile()
        
    def Write(self,filename,**kwargs):
        """
        Writes multispecies observations in MOCAGE HDF5
        """          
        writeNC = False       
        if "writeNC" in kwargs: writeNC = kwargs["writeNC"]

        if writeNC:
            self.WriteNC(filename)
        else:
            # Write profile data
            ProfConcInstrument.Write(self,filename,closefile=False)
            
            # Add radiosonde site ID
            dset = self.f[self.name+"/GEOLOCALIZATION"].create_dataset('SoundingID',(self.nobs,),dtype=self.SiteID.dtype,maxshape=(None,))
            if self.nobs > 0:  dset[...] = self.SiteID.compressed()[self.sorted_idx]
            
            self.CloseFile()
            print(">>> HDF5 wrote ")

    def WriteNC(self,filename,**kwargs):

        if self.nobs > 0:
            sorted_idx = np.lexsort(
                (self.lons.compressed(),
                 self.lats.compressed(),
                 (self.time-self.time_orig).compressed()))

        # Headers
        rootgrp = Dataset(filename+".nc", "w", format="NETCDF4")
        rootgrp.CreationDate = np.string_(time.strftime("%c"))
        instrgrp = rootgrp.createGroup(self.name)
        geoloc = instrgrp.createGroup("GEOLOCALIZATION")

        # Dimensions
        iobs = instrgrp.createDimension("iobs", None)
        level = instrgrp.createDimension("level", self.nmaxlevs)

        indexes = instrgrp.createVariable("iobs","i4",("iobs",))
        indexes[:] = np.arange(self.nobs)
        
        if self.press_isconst:
            levels = instrgrp.createVariable("level","f4",("level",))
            levels.Units = "Pa"
            levels.Orientation = np.string_(self.press_coorddir)
            levels.IsGridConstant = np.int32(self.press_isconst)
            if self.nobs > 0: levels[:] = self.press.compressed()

        # Geolocalization
        lats = geoloc.createVariable("Latitude","f4",("iobs",))
        lats.Units = np.string_('Degrees North')
        if self.nobs > 0: lats[:] = self.lats.compressed()[sorted_idx]
        lons = geoloc.createVariable("Longitude","f4",("iobs",))
        lons.Units = np.string_('Degrees East')
        if self.nobs > 0: lons[:] = self.lons.compressed()[sorted_idx]
        timestamps = geoloc.createVariable("Timestamp","i4",("iobs",))
        timestamps.Units = np.string_('seconds UTC since 1970 01 01')
        if self.nobs > 0: timestamps[:] = (self.time-self.time_orig).compressed()[sorted_idx]
        timehuman = geoloc.createVariable("TimeHuman",str,("iobs",))
        timehuman.Units = np.string_('date UTC in ISO 8601 format, YYYY-MM-DDTHH:MM:SS')
        if self.nobs > 0: timehuman[:] = self.time.astype(np.dtype("S")).compressed()[sorted_idx]
        siteid = geoloc.createVariable("SoundingID",str,("iobs",))
        if self.nobs > 0: siteid[:] = self.SiteID[~self.mask].compressed()[sorted_idx]
        # Data
        obsgrp = instrgrp.createGroup("OBSERVATIONS")
        for spc in self.species:
            obs = obsgrp.createVariable(spc,"f4",("iobs","level",),fill_value=self.missing_value)
            obs.Units = np.string_("Volume Mixing Ratio")
            if self.nobs > 0: obs[:,:] = self.prof[spc][~self.mask,...][sorted_idx]

        rootgrp.close()
        
def processWOUDCFile(filename,**kwargs):	
    """
    Extractor for WOUDC csv sondes files
    """      
    np.set_printoptions(threshold=sys.maxsize)
    import codecs
    interp = False
    pflag = True
    if 'interpolate' in kwargs and kwargs['interpolate']:
        interp = True
        if 'pressgrid' in kwargs:
            interpgrid = kwargs['pressgrid']
        else:
            interpgrid = np.array([ 101325., 100543., 98588., 95744., 92246., 88280., 83995., 79509., 74912., 70273., \
                          65643., 61060., 56554., 52146., 47854., 43695., \
                          39681., 35828., 32150., 28660., 25371., 22294., \
                          19436., 16795., 14384., 12204., 10205.,  8518., \
                          6997.,  5673.,  4529.,  3551.,  2726.,  2040.,  \
                          1481.,  1037., 695., 440.7, 261.1, 142., 69., 29., 10.])[::-1]
    
    fid = codecs.open(filename,"r",encoding='utf-8',errors='ignore')
    cr = csv.reader(fid)
    rownum = 0
    nlevels = 0
    press_sonde = []
    ozone_sonde = []
    
    for row in cr:
        #LOOP OVER ROWS SKIPPING EMPTY ROWS AND MEASUREMENT UNITS ROWS (*)
        #READ STATION COORDINATES   
        if len(row) > 0 and row[0] == "#PLATFORM":
            dum=next(cr)
            print(dum)
            platform = next(cr)
            siteid = platform[1]
            sitename = platform[2]
        if len(row) > 0 and row[0] == "#LOCATION":
            dum=next(cr)
            print(dum)
            geoloc=next(cr)
            lat_sonde=float(geoloc[0])
            lon_sonde=float(geoloc[1])
            alt_sonde=float(geoloc[2])
            print(geoloc)
        if len(row) > 0 and row[0] == "#TIMESTAMP":
            dum=next(cr)
            print(dum)
            sondetime=next(cr)
            utcoffset=sondetime[0]
            if len(sondetime) == 3:
                hour_sonde=sondetime[2]
            else:
                print("!!! Warning, ozonesonde hour not found, setting to 12 UTC")
                hour_sonde = "12:00:00"
            print(sondetime)
        #READ PROFILE LINES SKIPPING MISSING OZONE VALUES, * LINES
        # print(row)
        if nlevels > 0 and len(row) > 2 and row[columns.index('O3PartialPressure')] not in ('','nan'):
            if row[0][0] == '*': continue
            if 'SondeCurrent' in columns:
                if row[columns.index('SondeCurrent')] == '-999.99': continue # A patch for Lauder bad ozone values
            press_sonde.append(row[columns.index('Pressure')])
            ozone_sonde.append(row[columns.index('O3PartialPressure')])
            nlevels+=1
        #BREAK CYCLE WHEN PROFILE IS OVER
        if nlevels > 0 and len(row) == 0: break     
        #READ THE COLUMN DESCRIPTION AND JUMP ONE LINE IF STARTING POINT FOUND  
        if nlevels == -1:
            nlevels = 1
            columns = [name.strip() for name in row]
            print(columns)           
        #CHECK PROFILE STARTING POINT
        if len(row) > 0 and "#PROFILE" in row[0]:
            print(row)
            nlevels = -1
    rownum += 1

    if utcoffset[0] == '+': 
        time = hour_sonde+':+'+utcoffset[1:3]
    elif utcoffset[0] == '-': 
        time = hour_sonde+':-'+utcoffset[1:3]
    elif utcoffset[0] == '0': 
        time = hour_sonde+':+'+utcoffset[0:2]
    else:
        print('Warning: could not read UTC offset. Applying no offset')
        time = hour_sonde+':+00'
        
    fid.close()
                        
    print(lat_sonde,lon_sonde,time,siteid,sitename)

    press_sonde = np.ma.array(press_sonde,np.float)*100
    ozone_sonde = np.ma.array(ozone_sonde,np.float)

    dp = press_sonde[:-1]-press_sonde[1:]
    dp = np.ma.append(dp,0.0)
    if (dp < 0.0).any():
        print('>>> WARNING: ozonesounding with non monotonical pressure array. Cut at the firt occurrence')
        idxinv = (dp < 0.0).nonzero()[0]
        press_sonde[idxinv] = np.ma.masked
        ozone_sonde[idxinv] = np.ma.masked
    
    # Remove other potential bad ozone values
    ozone_sonde = np.ma.masked_less(ozone_sonde,1e-10)
    mask = np.ma.mask_or(press_sonde.mask, ozone_sonde.mask)         
    press_sonde.mask = mask
    ozone_sonde.mask = mask

    press_sonde = press_sonde.compressed()
    ozone_sonde = ozone_sonde.compressed()

    print('Found ', press_sonde.size, ' levels')

    if press_sonde.size < 40:
        return None, None, None, None, None, None, None, False

    press_sonde = press_sonde[::-1]

    if interp:
        sonde_presslev=cdms.createAxis(MV.array(press_sonde),id='plev')
        sonde_presslev.designateLevel()
        sonde_ozone=MV.zeros((1,1,len(ozone_sonde)))
        sonde_ozone[0][0][::-1] = np.array(ozone_sonde,np.float)
        sonde_ozone.getAxis(0).id = "lat"
        sonde_ozone.getAxis(1).id = "lon"		
        sonde_ozone.setAxis(2,sonde_presslev)
        sonde_ozone=sonde_ozone/(press_sonde*1000) # To vmr
        # for i in range(sonde_ozone.size): print(sonde_ozone[0,0,i])
        model_presslev = cdms.createAxis(MV.array(interpgrid),id='plev')
        model_presslev.designateLevel()
        new_ozone = sonde_ozone.pressureRegrid(model_presslev,method='log')
        # Fill missing values
        indout = np.nonzero(interpgrid < sonde_presslev[0])[0]
        if len(indout) > 0: new_ozone[0,0,indout] = -999.
        indout = np.nonzero(interpgrid > sonde_presslev[-1])[0]
        if len(indout) > 0: new_ozone[0,0,indout] = -999.
        new_ozone = np.ma.masked_values(new_ozone[0,0,:],-999.)
        #new_press = np.ma.array(interpgrid,mask=new_ozone.mask)
        new_press = interpgrid
    else:
        new_ozone = np.ma.array(ozone_sonde,np.float)
        new_press = np.ma.array(press_sonde,np.float)
        new_ozone = new_ozone/(new_press*1000) # To vmr

    # if siteid == '435': sys.exit()
    
    return new_press, new_ozone, lat_sonde, lon_sonde, time, siteid, sitename, pflag

def processSHADOZFile(filename,**kwargs):    
    """
    Extractor for WOUDC csv sondes files
    """      
    np.set_printoptions(threshold=sys.maxsize)
    interp = False
    pflag = True
    if 'interpolate' in kwargs and kwargs['interpolate']:
        interp = True
        if 'pressgrid' in kwargs:
            interpgrid = kwargs['pressgrid']
        else:
            interpgrid = np.array([ 101325., 100543., 98588., 95744., 92246., 88280., 83995., 79509., 74912., 70273., \
                          65643., 61060., 56554., 52146., 47854., 43695., \
                          39681., 35828., 32150., 28660., 25371., 22294., \
                          19436., 16795., 14384., 12204., 10205.,  8518., \
                          6997.,  5673.,  4529.,  3551.,  2726.,  2040.,  \
                          1481.,  1037., 695., 440.7, 261.1, 142., 69., 29., 10.])[::-1]
    
    fid = open(filename,"r",encoding='utf-8')
    headersize = int(fid.readline())
    header = {}
    for line in range(headersize-3):
        line = fid.readline()
        if ':' in line: header[line.split(':',1)[0]] = line.split(':',1)[1].strip('\n')

    datadescr = fid.readline()
    units = fid.readline()

    fid.close()

    # print(header) 
    # print(datadescr)
    # print(units)
    # Find necessary info
    lat_sonde = [float(item) for key,item in header.items() if "Latitude" in key][0]
    lon_sonde = [float(item) for key,item in header.items() if "Longitude" in key][0]
    time = [item for key,item in header.items() if "Launch Time" in key][0]+':+00'
    siteid = [item.strip() for key,item in header.items() if "Ozonesonde Instrument" in key][0]
    sitename = [item for key,item in header.items() if "STATION" in key][0]
    missing_val = [float(item) for key,item in header.items() if "Missing" in key][0]

    data = np.genfromtxt(filename, skip_header=headersize)

    print(lat_sonde,lon_sonde,time,siteid,sitename,missing_val)

    press_sonde = np.ma.masked_values(data[:,1],missing_val)
    ozone_sonde = np.ma.masked_values(data[:,6],missing_val)
               
    mask = np.ma.mask_or(press_sonde.mask, ozone_sonde.mask)         
    press_sonde.mask = mask
    ozone_sonde.mask = mask

    dp = press_sonde[:-1]-press_sonde[1:]
    dp = np.ma.append(dp,0.0)
    if (dp < 0.0).any():
        print('>>> WARNING: ozonesounding with non monotonical pressure array. Cut at the firt occurrence')
        idxinv = (dp < 0.0).nonzero()[0]
        press_sonde[idxinv] = np.ma.masked
        ozone_sonde[idxinv] = np.ma.masked
    
    press_sonde = press_sonde.compressed()
    ozone_sonde = ozone_sonde.compressed()

    print('Found ', press_sonde.size, ' levels')

    if press_sonde.size < 40:
        return None, None, None, None, None, None, None, False

    press_sonde = press_sonde[::-1]*100

    if interp:
        sonde_presslev=cdms.createAxis(MV.array(press_sonde),id='plev')
        sonde_presslev.designateLevel()
        sonde_ozone=MV.zeros((1,1,len(ozone_sonde)))
        sonde_ozone[0][0][::-1] = ozone_sonde
        sonde_ozone.getAxis(0).id = "lat"
        sonde_ozone.getAxis(1).id = "lon"       
        sonde_ozone.setAxis(2,sonde_presslev)
        sonde_ozone=sonde_ozone/1.e6 # To vmr
        model_presslev = cdms.createAxis(MV.array(interpgrid),id='plev')
        model_presslev.designateLevel()
        new_ozone = sonde_ozone.pressureRegrid(model_presslev,method='log')
        # Fill missing values
        indout = np.nonzero(interpgrid < sonde_presslev[0])[0]
        if len(indout) > 0: new_ozone[0,0,indout] = -999.
        indout = np.nonzero(interpgrid > sonde_presslev[-1])[0]
        if len(indout) > 0: new_ozone[0,0,indout] = -999.
        new_ozone = np.ma.masked_values(new_ozone[0,0,:],-999.)
        #new_press = np.ma.array(interpgrid,mask=new_ozone.mask)
        new_press = interpgrid
    else:
        new_ozone = np.ma.array(ozone_sonde,np.float)
        new_press = np.ma.array(press_sonde,np.float)
        new_ozone = new_ozone/1.e6 # To vmr
    
    return new_press, new_ozone, lat_sonde, lon_sonde, time, siteid, sitename, pflag

def processNDACCFile(filename,**kwargs):    
    """
    Extractor for WOUDC csv sondes files
    """      
    np.set_printoptions(threshold=sys.maxsize)
    interp = False
    pflag = True
    if 'interpolate' in kwargs and kwargs['interpolate']:
        interp = True
        if 'pressgrid' in kwargs:
            interpgrid = kwargs['pressgrid']
        else:
            interpgrid = np.array([ 101325., 100543., 98588., 95744., 92246., 88280., 83995., 79509., 74912., 70273., \
                          65643., 61060., 56554., 52146., 47854., 43695., \
                          39681., 35828., 32150., 28660., 25371., 22294., \
                          19436., 16795., 14384., 12204., 10205.,  8518., \
                          6997.,  5673.,  4529.,  3551.,  2726.,  2040.,  \
                          1481.,  1037., 695., 440.7, 261.1, 142., 69., 29., 10.])[::-1]
    
    fid = open(filename,"r",encoding='utf-8')
    lines = fid.readlines()
    fid.close()

    sitename = lines[0].split()[-5]

    foundinfo, founddata = False, False
    idx = 0
    for line in lines[1:]:
        idx+=1
        if re.search(sitename,line,re.IGNORECASE): 
            foundinfo = True
            idxinfo = idx
        if re.search('PO3',line,re.IGNORECASE):
            founddata = True
            idxdata = idx
            break

    if not foundinfo or not founddata: 
        print('>>> WARNING: skipping ozonesounding with uknown NDACC data format')
        return None, None, None, None, None, None, None, False

    info = lines[idxinfo+1].split()

    # Find necessary info
    lat_sonde = float(info[4])
    lon_sonde = float(info[3])
    time = filename[-2:]+':00:00:+00'
    siteid = 'UNK'

    data = np.genfromtxt(filename, skip_header=idxdata+2)

    colnames = lines[idxdata].split()

    print(lat_sonde,lon_sonde,time,siteid,sitename)

    # Read the first row and then 
    press_sonde = np.ma.masked_values(data[:,colnames.index('Press')],99999)
    ozone_sonde = np.ma.masked_values(data[:,colnames.index('PO3')],99999)

    mask = np.ma.mask_or(press_sonde.mask, ozone_sonde.mask)         
    press_sonde.mask = mask
    ozone_sonde.mask = mask
                    
    dp = press_sonde[:-1]-press_sonde[1:]
    dp = np.ma.append(dp,0.0)
    if (dp < 0.0).any():
        print('>>> WARNING: ozonesounding with non monotonical pressure array. Cut at the firt occurrence')
        idxinv = (dp < 0.0).nonzero()[0]
        press_sonde[idxinv] = np.ma.masked
        ozone_sonde[idxinv] = np.ma.masked
    
    press_sonde = press_sonde.compressed()
    ozone_sonde = ozone_sonde.compressed()

    print('Found ', press_sonde.size, ' levels')

    if press_sonde.size < 40:
        return None, None, None, None, None, None, None, False

    press_sonde = press_sonde[::-1]*100

    if interp:
        sonde_presslev=cdms.createAxis(MV.array(press_sonde),id='plev')
        sonde_presslev.designateLevel()
        sonde_ozone=MV.zeros((1,1,len(ozone_sonde)))
        sonde_ozone[0][0][::-1] = ozone_sonde
        sonde_ozone.getAxis(0).id = "lat"
        sonde_ozone.getAxis(1).id = "lon"       
        sonde_ozone.setAxis(2,sonde_presslev)
        sonde_ozone=sonde_ozone/(press_sonde*1000) # To vmr
        model_presslev = cdms.createAxis(MV.array(interpgrid),id='plev')
        model_presslev.designateLevel()
        new_ozone = sonde_ozone.pressureRegrid(model_presslev,method='log')
        # Fill missing values
        indout = np.nonzero(interpgrid < sonde_presslev[0])[0]
        if len(indout) > 0: new_ozone[0,0,indout] = -999.
        indout = np.nonzero(interpgrid > sonde_presslev[-1])[0]
        if len(indout) > 0: new_ozone[0,0,indout] = -999.
        new_ozone = np.ma.masked_values(new_ozone[0,0,:],-999.)
        #new_press = np.ma.array(interpgrid,mask=new_ozone.mask)
        new_press = interpgrid
    else:
        new_ozone = np.ma.array(ozone_sonde,np.float)
        new_press = np.ma.array(press_sonde,np.float)
        new_ozone = new_ozone/(new_press*1000) # To vmr
    
    return new_press, new_ozone, lat_sonde, lon_sonde, time, siteid, sitename, pflag