# coding: utf8

import os
import numpy as np
import datetime
import netCDF4
from IntQty_instr import IntQtyInstrument
import glob
from functools import reduce


class TropomiInstr(IntQtyInstrument):
    """ Class for TROPOMI spectrometer data.

    This class in specific to SO2 data.

    Parameters
    ----------
    name : str
        Name of the instrument
    nmaxlevs : int
        Maximal number of vertical levels.
    kwargs : dict
        Used to allow unexpected keyword arguments.

    """
    species_name = 'SO2'
    species_long_name = 'sulfurdioxide'

    def __init__(self, name, nmaxlevs=34, **kwargs):
        IntQtyInstrument.__init__(
            self, name, species=[self.species_name], nmaxlevs=nmaxlevs,
            VertGridCnst=0)

    def Convert(self, period, inputdir, qflag=.5, detect_flag=1,
                representative_height=None, filter_max=1e30, **kwargs):
        """ Read data from a directory.

        File names format must be:
        S5P_OFFL_L2__*{year}{month}{day}T*.nc

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (fisrt day, last day).
        inputdir : str
            Input directory.
        qflag : float, optional
            Quality flag between 0 and 1 below which observations are
            rejected.
        detect_flag : int, optional
            Integer between 0 and 4 (default is 1). Only observations with
            detection flag equal to detect_flag are kept. The flag is 0 for
            no detection, 1 SO2 detection, 2 for clear volcanic detection,
            3 for detection close to known anthropogenic source, 4 for
            detection at high SZA (potential false-positive).
        representative_height : int or None, optional
            Integer choosen among [1, 7, 15]. It corresponds to the
            representative height in km. If None, retrieved data is the
            total vertical column.
        filter_max : float, optional
            Threshold above which values are considered missing.
        kwargs : dict, optional
            Used to allow unexpected keyword arguments.

        """
        if representative_height is not None:
            data_path = ("PRODUCT/SUPPORT_DATA/DETAILED_RESULTS/" +
                             "{species}_total_vertical_column_{h}km").format(
                                species=self.species_long_name,
                                h=representative_height)
            amf_ratio_path = ("PRODUCT/SUPPORT_DATA/DETAILED_RESULTS/" +
                              "{species}_averaging_kernel_scaling_" +
                              "box_{h}km").format(
                                species=self.species_long_name,
                                h=representative_height)
        else:
            data_path = "PRODUCT/{}_total_vertical_column".format(
                self.species_long_name)

        spe = self.species_name
        nbdays = (period[1] - period[0]).days
        dates = (period[0] + datetime.timedelta(d) for d in range(nbdays))
        file_list = []
        for current_date in dates:
            file_list.extend(
                glob.glob(os.path.join(
                    inputdir,
                    "S5P_OFFL_L2__*{date}T*.nc".format(
                        date=current_date.strftime('%Y%m%d')))))
        if len(file_list) == 0:
            print("No file found for period {}".format(period))
            return

        self.units = {spe: "mol m-2"}
        self.VertDsetName = 'Pressure'
        self.vertunit = 'Pa'
        self.vertorient = 'down'
        for f in file_list:
            print(f)
            ds = netCDF4.Dataset(f, 'r')
            product_vars = ds['PRODUCT'].variables
            results_vars = ds[
                'PRODUCT/SUPPORT_DATA/DETAILED_RESULTS'].variables
            input_data_vars = ds['PRODUCT/SUPPORT_DATA/INPUT_DATA'].variables

            # get times
            time_ref = datetime.datetime.strptime(ds.time_reference,
                                                  "%Y-%m-%dT%H:%M:%S")
            assert(product_vars['delta_time'].units == 'milliseconds')
            times = np.array(
                [time_ref + datetime.timedelta(milliseconds=int(dt))
                 for dt in product_vars['delta_time'][:].flatten()])
            kept_times = np.array([t.date() == current_date.date()
                                   for t in times])

            # index of accepted obs
            kept_obs = reduce(
                np.logical_and,
                [ds['PRODUCT/qa_value'][:].flatten() > qflag] +
                [ds[data_path][:].flatten().filled(filter_max) < filter_max] +
                [kept_times] +
                [results_vars["{}_detection_flag".format(
                    self.species_long_name)][:].flatten().data == detect_flag])
            nobs = np.sum(kept_obs)

            # select times
            times = times[kept_obs]

            # get coords
            lats = product_vars['latitude'][:].flatten()[kept_obs]
            lons = product_vars['longitude'][:].flatten()[kept_obs]

            # compute vertical pressure = a + b*surf_press
            nlevs = product_vars['layer'].shape[0]
            assert(nlevs == self.nmaxlevs)
            assert(input_data_vars['surface_pressure'].units == 'Pa')
            press = np.array(
                [a + b*input_data_vars[
                    'surface_pressure'][:].flatten()[kept_obs]
                 for a, b in
                 zip(input_data_vars['tm5_constant_a'][:].squeeze(),
                     input_data_vars['tm5_constant_b'][:].squeeze())])

            # fill geolocalization attributes
            self.nobs += nobs
            self.time = np.ma.concatenate(
                (self.time, np.ma.array(times, dtype=self.np_typetime)))
            self.lats = np.ma.concatenate(
                (self.lats, np.ma.array(lats, dtype=self.lats.dtype)))
            self.lons = np.ma.concatenate(
                (self.lons, np.ma.array(lons, dtype=self.lons.dtype)))
            self.mask = np.concatenate(
                (self.mask, np.zeros((nobs), dtype=bool)))
            self.vert = np.ma.concatenate(
                (self.vert, np.ma.array(press.T, dtype=self.vert.dtype)))

            assert(ds[data_path].units == 'mol m-2')
            pcol = ds[data_path][:].flatten()[kept_obs]
            avk = results_vars['averaging_kernel'][:].reshape(
                (len(kept_obs), nlevs))[kept_obs, :]
            if representative_height is not None:
                avk = avk*ds[amf_ratio_path][:].flatten()[kept_obs, None]
            self.avk[spe] = np.ma.concatenate(
                (self.avk[spe],
                 np.ma.array(avk, dtype=self.avk[spe].dtype)))
            self.pcol[spe] = np.ma.concatenate(
                (self.pcol[spe],
                 np.ma.array(pcol, dtype=self.pcol[spe].dtype)))
