# coding: utf8
""" WOUDC (World Ozone and Ultraviolet Radiation Data Centre) module.

WOUDC ozone sonde data processing module. Raw data can be downloaded at
https://www.woudc.org/.

"""
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from io import StringIO
import json
import glob
from profconc_instr import ProfConcInstrument
import h5py
from functools import reduce


class WoudcInstrument(ProfConcInstrument):
    """ WOUDC ozone sonde observation class

    Parameters
    ----------
    name : str
        Name of the instrument
    nmaxlevs : int
        Maximal number of vertical levels.
    kwargs : dict, optional
        Used to allow unexpected keyword arguments.

    """
    def __init__(self, name, nmaxlevs=None, **kwargs):
        ProfConcInstrument.__init__(self, name, species=['O3'],
                                    vert_isconst=False, nmaxlevs=nmaxlevs)
        self.id = np.empty((0), dtype=str)

    def Convert(self, period, input_path, vertical_levels=None,
                filter_max=1e-4, **kwargs):
        """ Read WOUDC raw JSON files

        Parameters
        ----------
        period : couple of datetime.datetime
            Time period defined as (start, end), end is excluded.
        input_path : str
            Generic file path to be read. This path is read by
            python module glob and therefore can contain special caracters
            "*" and "?".
        vertical_levels : None or 1D array-like, optional
            If not None, data is interpolated vertically according to given
            levels.
        filter_max : float
            Threshold above which values are considered missing.

        """
        assert(not self.press_isconst)

        # when reading raw WOUDC file with no vertical interpolation, vertical
        # direction is 'upward'
        self.press_coorddir = 'upward'

        # only O3 is supported
        spc = self.species[0]

        nobs = 0
        time_list = []
        latitudes = []
        longitudes = []
        prof_list = []
        vert_list = []
        code_list = []

        matches = glob.glob(input_path)
        for filename in matches:
            with open(filename) as f:
                data = json.load(f)['features']
            for i, prof in enumerate(data):
                # read and round launch time
                time_str = prof['properties']['instance_datetime'].split('+')
                if int(time_str[1]) != 0.:
                    raise Exception("Not UTC")
                time = datetime.strptime(time_str[0], "%Y/%m/%d %H:%M:%S")
                h_round = int(round(
                    time.hour + time.minute/60. + time.second/3600., 0))
                if h_round == 24:
                    time = (datetime(time.year, time.month, time.day, 0) +
                            timedelta(days=1))
                else:
                    time = datetime(time.year, time.month, time.day, h_round)

                # check period
                if time < period[0] or time >= period[1]:
                    continue

                try:
                    # read metadata
                    # code = prof['properties']['gaw_id']
                    code = prof['properties']['platform_name'].replace(" ", "")
                    # print('Profile code: ', code)
                    lon = prof['geometry']['coordinates'][0]
                    lat = prof['geometry']['coordinates'][1]

                    # read data
                    df = pd.read_csv(
                        StringIO(prof['properties']['data_block']),
                        skipinitialspace=True,
                    )

                    # convert to VMR
                    # O3PartialPressure is in mPa and Pressure in hPa -> 1e5
                    prof = df['O3PartialPressure']/(df['Pressure']*1e5)

                    nobs += 1
                    time_list.append(time)
                    latitudes.append(lat)
                    longitudes.append(lon)
                    prof_list.append(prof)
                    vert_list.append(df['Pressure']*1e2)
                    code_list.append(code)
                except Exception as err:
                    print("Error in observation file for {}:".format(time))
                    print(type(err), err)

        print("{n} profile{plural} found".format(
            n=nobs,  plural=('s' if nobs > 1 else '')))
        if nobs > 0:
            # give all measures the same number of levels
            nb_levs = [p.shape[0] for p in prof_list]
            nmaxlevs = np.max(nb_levs)
            if not hasattr(self, 'nmaxlevs'):
                self.nmaxlevs = nmaxlevs
                self.InitArrays()
            else:
                if self.nmaxlevs < nmaxlevs:
                    raise Exception("nmaxlevs must be >= {}".format(
                        nmaxlevs))
                nmaxlevs = self.nmaxlevs
            vert_pad = [np.append(p, np.full(nmaxlevs-n, self.missing_value))
                        for n, p in zip(nb_levs, vert_list)]
            vert_pad = np.array(vert_pad)
            prof_pad = [np.append(p, np.full(nmaxlevs-n, self.missing_value))
                        for n, p in zip(nb_levs, prof_list)]
            prof_pad = np.array(prof_pad)

            # append data to the intrument object attributes
            self.nobs += nobs
            self.time = np.ma.concatenate(
                (self.time, np.ma.array(time_list, dtype=self.np_typetime)))
            self.lats = np.ma.concatenate(
                (self.lats, np.ma.array(latitudes, dtype=self.lats.dtype)))
            self.lats.fill_value = self.missing_value
            self.lons = np.ma.concatenate(
                (self.lons, np.ma.array(longitudes, dtype=self.lons.dtype)))
            self.lons.fill_value = self.missing_value
            self.mask = np.concatenate(
                (self.mask, np.zeros((nobs), dtype=bool)))
            self.press = np.ma.concatenate(
                (self.press, np.ma.array(vert_pad, dtype=self.press.dtype)))
            self.press.fill_value = self.missing_value
            self.prof[spc] = np.ma.concatenate(
                (self.prof[spc],
                 np.ma.array(prof_pad,
                             dtype=self.prof[spc].dtype,
                             fill_value=self.missing_value)))
            self.prof[spc].fill_value = self.missing_value
            self.id = np.concatenate(
                (self.id,
                 np.array(code_list)))

            # mask negative values and values above filter_max
            with np.errstate(invalid='ignore'):
                idx = reduce(np.logical_or, [self.prof[spc] < 0,
                                             self.prof[spc] > filter_max])
            self.prof[spc][idx] = np.ma.masked
            self.press[idx] = np.ma.masked

            # vertical interpolation
            if vertical_levels is not None:
                self.vertical_regrid(vertical_levels, interp='layerAverage')
        else:
            self.nmaxlevs = 1
