# coding: utf8
from __future__ import print_function
import numpy
import cdms2 as cdms
import MV2 as MV
import genutil
import cdutil
import datetime
import sys
import os
#
# -------------------------------------------------------------
# CREATION D'UNE VARIABLE AU FORMAT CDMS
# -------------------------------------------------------------
#    
def CreateCdmsVariable(ntime, nlon, nlat, nlev, var_liste ) : 
  """ Creation d'une variable au format cdms """
#
#  1. Initialisation des variables CDMS
#
  var_cdms = numpy.zeros((ntime, nlon, nlat, nlev), float)
#
#  2. Remplissage des variables CDMS avec les variables "liste"
#
  for itime in range(ntime) : 
    var_cdms[itime][0][0][0:nlev] = var_liste[itime][0][0][0:nlev]
#
#  3. Creation des variables CDMS 
#
  var_cdms = cdms.createVariable(var_cdms)
  var_cdms.getAxis(0).id = "time"
  var_cdms.getAxis(1).id = "lon"
  var_cdms.getAxis(2).id = "lat"
  var_cdms.getAxis(3).id = "level"
#
# 4. Retour des variables calculees
#
  return var_cdms
#  
# -------------------------------------------------------------
# INTERPOLATION VERTICALE AVEC LA BIBLIOTHEQUE CDAT
# -------------------------------------------------------------
#    
def interpole(pres_ini, field, pres_fin, method='log', pad=True) :
  """ Interpolation des donnees HPALM sur une grille definie """
#
#  1. Initilisation du tableau de travail avec le champ a interpoler 
#  
  field_array = numpy.zeros((1,1,len(field)), float)
  field_array[0][0] = field
#
#  2. Conversion du tableau de travail en tableau CDMS
#  
  field_cdms = cdms.createVariable(field_array)
  field_cdms.getAxis(0).id = "lat"
  field_cdms.getAxis(1).id = "lon"
  field_cdms.getAxis(2).id = "level"
  field_cdms.getAxis(2)[:] = pres_ini
#
#  3. Interpolation
#  
  new_axis = cdms.createAxis(pres_fin)
  new_field_cdms = field_cdms.pressureRegrid(new_axis,method=method)
#
#  4. Ajout de la constante -999.hors des bornes d'interpolation
#  
  if pad : 
    if pres_ini[0] < pres_fin[0] :
      iniv = 0 
      while (iniv < len(pres_fin)) and (pres_ini[0] < pres_fin[iniv]) : 
          new_field_cdms[0,0,iniv] = -999.
          iniv = iniv + 1
#
# 5. Retour des variables calculees
#
  return new_field_cdms[0,0]

def BuildMocageGrid (regcode) : 
    """  
    Création des axes pour un des domaines MOCAGE
    """
    if regcode == 'GLOB22':
      grid=cdms.createUniformGrid(-89,90,2.,-179,180,2.,order='yx')
    if regcode == 'GEMS02':
      grid=cdms.createUniformGrid(32.1,200,0.2,-15.9,260,0.2,order='yx')
    if regcode == 'MACC02':
      grid=cdms.createUniformGrid(28.1,220,0.2,-25.9,360,0.2,order='yx')
    if regcode == 'GEMS01':
      grid=cdms.createUniformGrid(32.1,399,0.1,-15.9,519,0.1,order='yx')
    if regcode == 'GEMS04':
      grid=cdms.createUniformGrid(32.1,101,0.4,-15.9,131,0.4,order='yx')
    if regcode == 'MEDI02':
      grid=cdms.createUniformGrid(16.1,51.9,0.2,-19.9,39.9,0.2,order='yx')
    return grid

def BuildMocagePlevels(nlev=47):

  lev = numpy.arange(nlev,dtype='i')+1
  lev = cdms.createAxis(lev)
  lev.designateLevel()
  lev.id = 'lev'
  lev.standard_name = 'atmosphere_hybrid_sigma_pressure_coordinate'
  lev.units = 'lev'
  lev.positive = 'down'

  a, b = ReadFileVtable(nlev,path='/wkdir/pae/emili/wkdir_pae1/DATA_MCG')
  
  psurf = MV.array(101325.)
  a = numpy.array((a[1:]+a[:-1])*0.5)
  b = numpy.array((b[1:]+b[:-1])*0.5)
  #Reconvert to MV arrays
  a, b = MV.array(a,axes=(lev,)), MV.array(b,axes=(lev,))
  #Compute pressure levels (p0=1 because we want p=a+b*ps) 
  plev=cdutil.vertical.reconstructPressureFromHybrid(psurf,a,b,1.)
  return cdms.createAxis(plev,id='pressure')
    
def CreateVar2dMV (lat, lon, vtype, varid = None, varname = None, varunit = None) :
    """
    Création d'une variable 2D (lat,lon) au format MV
    """
    #
    # Calcul de la dimension des tableaux
    #
    nlat = len(lat)
    nlon = len(lon)
    #
    # Initialisation de la variable
    #
    var = MV.zeros([nlat,nlon], vtype)
    #
    # Ajout des info
    #
    if varid != None : var.id = varid
    if varname != None : var.long_name = varname
    if varunit != None : var.units = varunit
    #
    # Ajout des axes
    #
    lonAxis = cdms.createAxis(lon); lonAxis.designateLongitude()
    latAxis = cdms.createAxis(lat); latAxis.designateLatitude()
    var.getAxis(0).id = "lat"
    var.getAxis(1).id = "lon"
    var.getAxis(0)[:] = latAxis
    var.getAxis(1)[:] = lonAxis
    #
    # Fin de la routine
    #
    return var
#
# -------------------------------------------------------------
# SET CDMS TO WRITE NETCDF 3 CLASSIC FILES
# -------------------------------------------------------------
#
def switchToNetCDF3ClassicMode():
  
  print(">>> Sets CDMS/NetCDF settings to mirror classic NetCDF mode.")
  value = 0
  cdms.setNetcdfShuffleFlag(value)
  cdms.setNetcdfDeflateFlag(value)
  cdms.setNetcdfDeflateLevelFlag(value)
#
# -------------------------------------------------------------
# CONSTRUCTION OF PRESSURE LEVELS FROM MOCAGE HYBRID SIGMA LEVELS
# -------------------------------------------------------------
#  
def compute_press_lev(psurf,a,b):
  #Compute pressure grid center values (!!MOCAGE netcdf values are for the interfaces!!)
  vaxis = a.getAxisList()
  #Roof at the top of the atmosphere (P=0)
  p0 = 0
  a, b = numpy.array(a), numpy.array(b)
  a = numpy.insert(a,0,p0)
  b = numpy.insert(b,0,0)
  a = numpy.array((a[1:]+a[:-1])*0.5)
  b = numpy.array((b[1:]+b[:-1])*0.5)
  #Reconvert to MV arrays
  a, b = MV.array(a,axes=vaxis), MV.array(b,axes=vaxis)
  #Compute pressure levels (p0=1 because we want p=a+b*ps) 
  plev = cdutil.vertical.reconstructPressureFromHybrid(psurf,a,b,1.)
  plev = plev(squeeze=1)
  return plev

def compute_press_bounds(psurf,a,b):
  #Compute pressure grid interfaces (!!MOCAGE netcdf values are for the interfaces!!)
  # Create a new vertical axe
  lev = numpy.arange(a.size+1,dtype='i')+1
  lev = cdms.createAxis(lev)
  lev.designateLevel()
  lev.id = 'lev'
  lev.standard_name = 'atmosphere_hybrid_sigma_pressure_coordinate'
  lev.units = 'lev'
  lev.positive = 'down'
  #Roof at the top of the atmosphere (P=0)
  p0 = 0
  a, b = numpy.array(a), numpy.array(b)
  a = numpy.insert(a,0,p0)
  b = numpy.insert(b,0,0)
  #Reconvert to MV arrays
  a, b = MV.array(a,axes=(lev,)), MV.array(b,axes=(lev,))
  #Compute pressure levels (p0=1 because we want p=a+b*ps) 
  plev = cdutil.vertical.reconstructPressureFromHybrid(psurf,a,b,1.)
  plev = plev(squeeze=1)
  return plev    
#
# -------------------------------------------------------------
# CONSTRUCTION OF PRESSURE INTERVALS FOR VERTICAL INTEGRATION
# -------------------------------------------------------------
# Considers the given pressure array given for grid center values
def compute_press_interv(pressin,silent=True):
    
  # Check if pressin is masked array and if it has missing values, count the number of masked values
  # and work on the compressed array
  masked = False  
  if isinstance(pressin,numpy.ma.MaskedArray) is True and pressin.count() < pressin.size:
    # Reduce array to valid values for dp calculations
    masked = True
    mask = pressin.mask
    headcount = numpy.argmin(mask)
    tailcount = numpy.argmin(mask[::-1])
    if not silent: print('Masked values (head,tail): ', headcount, tailcount)
    # Return error if masked values are inside the array (to do...)
    pressintmp = pressin.compressed()[:]
  else:
    pressintmp = numpy.array(pressin)
    
  if (numpy.sort(pressintmp[::-1]) == pressintmp[::-1]).all(): 
    if not silent: print('>>> Decreasing pressure array:')
    dp = -(pressintmp[2:]-pressintmp[:-2])/2
    dp = numpy.insert(dp,0,(pressintmp[0]-pressintmp[1])/2)
    dp = numpy.append(dp,pressintmp[-1]+(pressintmp[-2]-pressintmp[-1])/2)
  elif (numpy.sort(pressintmp) == pressintmp).all():
    if not silent: print('>>> Increasing pressure array')
    dp = (pressintmp[2:]-pressintmp[:-2])/2
    dp = numpy.append(dp,(pressintmp[-1]-pressintmp[-2])/2)
    dp = numpy.insert(dp,0,pressintmp[0]+(pressintmp[1]-pressintmp[0])/2)
  else:
    print('>> Error: non monotone pressure array!')
    sys.exit()
  
  if masked is True:
    # Set dp=0 for originally masked values 
    if headcount > 0: dp = numpy.insert(dp,0,numpy.zeros(headcount))  
    if tailcount > 0: dp = numpy.append(dp,numpy.zeros(tailcount))
    
  if not silent:
    print('>> Compute press interv diags')  
    if numpy.min(pressin) >= 100 or numpy.max(pressin) <= 95000:
      print('>>> Warning: first/last pressure level supposed to represent surface/top of atmosphere')
      print('>>> Do not use the dp later if the vertical grid is not really covering the entire atmosphere')
    print(pressin)
    print(dp)
    print(len(dp))
  return dp
#
# -------------------------------------------------------------
# CALCULATION OF PARTIAL COLUMNS
# -------------------------------------------------------------
# Works on pressure levels (vertical regridding shoud be done before)
def compute_partial_columns(x,plevels):
  #Prepare pressure interval array (first level excluded, it is the top of the atmosphere)
  interpol_dp = compute_press_interv(plevels)[1:]
  print('>>> Pressure intervals: ', interpol_dp)
  #Repeat the dp in space and time
  if len(x.shape) == 4:
    interpol_dp = numpy.repeat(interpol_dp[None,:],x.shape[0],axis=0)
    interpol_dp = numpy.repeat(interpol_dp[:,:,None],x.shape[2],axis=2)
    interpol_dp = numpy.repeat(interpol_dp[:,:,:,None],x.shape[3],axis=3)
    interpol_dp = MV.array(interpol_dp,axes=x.getAxisList())
    xpcol = MV.sum(x*interpol_dp,axis=1)*1e6
  elif len(x.shape) == 3:  
    interpol_dp = numpy.repeat(interpol_dp[:,None],x.shape[1],axis=1)
    interpol_dp = numpy.repeat(interpol_dp[:,:,None],x.shape[2],axis=2)
    interpol_dp = MV.array(interpol_dp,axes=x.getAxisList())
    xpcol = MV.sum(x*interpol_dp,axis=0)*1e6
  else:
    print('>>> Shape of input array not recognized (must be 3d or 4d)')
    print(x.shape)
    sys.exit()
  #print xpcol.shape
  #print xpcol.getAxisList()
  return xpcol  
#
# -------------------------------------------------------------
# EXTRACTION DES VALEURS 1D D'UNE VARIABLE 2/3/4-D MV
# -------------------------------------------------------------
#    
def ExtractValueLatLon(cdmsvar,lats,lons,dates,interpolation = None):
  # Python list containing extracted values
  values = []

  initdate = dates[0]
  finaldate = dates[1]

  print('>>> Extracting values for the period:')
  print(initdate, finaldate)
  for lat,lon in zip(lats,lons):
    print('>>> Extracting at ', lat, lon)
    tmp = numpy.ma.array(cdmsvar(time=(initdate,finaldate),latitude=(lat, lat, 'cob'), \
                    longitude=(lon, lon, 'cob'), squeeze=1 ))
    values.append(tmp)

  return values

# Extract time series of model profiles on a costant pressure grid (much faster than 
# first reprojecting all the grid and then extract profiles over few locations)
def ExtractProfileLatLon(cdmsvar,levels,lats,lons,levelsout,dates,levtype='mlev'):
  # Python list containing extracted values
  values = []

  initdate = dates[0]
  finaldate = dates[1]

  print('>>> Extracting profiles for the period:')
  print(initdate, finaldate)

  for lat,lon in zip(lats,lons):
    print('>>> Extracting at ', lat, lon)
      # Lat lon selection
    tmp = cdmsvar(time=(initdate,finaldate),latitude=(lat, lat, 'cob'), \
                      longitude=(lon, lon, 'cob'), squeeze=1 )
    if levtype == 'plev':
        # Pressure grid selection
        tmplevels = levels(time=(initdate,finaldate),latitude=(lat, lat, 'cob'), \
                          longitude=(lon, lon, 'cob'), squeeze=1)
        # Vertical interpolation
        tmp = cdutil.vertical.logLinearInterpolation(tmp,tmplevels,levels=levelsout)
        
    values.append(numpy.ma.array(tmp))
  return values
#
# -------------------------------------------------------------
# EXTRACTION DES VALEURS 0D D'UNE VARIABLE 3/4-D MV (must be in plev)
# -------------------------------------------------------------
#    
# Extract time series of model values at a certain pressure level, which can be a function of time
def ExtractValueLatLonPlev(cdmsvar,lats,lons,pressout,plev,dates,interpolation = None):
  
  values = []
  initdate = dates[0]
  finaldate = dates[1]
  print('>>> Extracting values for the period:')
  print(initdate, finaldate)
  tmp = cdmsvar(time=(initdate,finaldate),squeeze=1)
  ptmp = plev(time=(initdate,finaldate),squeeze=1)
  axlist = tmp.getAxisList(omit='level')
  print('>>> Read ',len(axlist[0]), ' columns')
  if len(axlist[0]) != len(pressout):
    print('>>> Error: output pressure array not compatible with data array')
    return values
  for lat,lon in zip(lats,lons):
    print('Extracting at ', lat, lon)
    try:
      #Loop over dates to interpolate to the wanted pressure
      tmp2 = MV.zeros(tmp[:,0,:,:].shape,axes=axlist)
      for time in range(len(axlist[0])):
        tmp2[time] = cdutil.vertical.linearInterpolation(tmp[time],ptmp[time],levels=pressout[time])[0]
      #Extract single position time series
      tmp2 = tmp2(latitude=(lat, lat, 'cob'), \
                    longitude=(lon, lon, 'cob'),squeeze=1)
      values.append(tmp2)
    except StandardError:
      print('>>> WARNING: Error!')
  return values  
#
# -------------------------------------------------------------
# COMPUTATION OF GRID HORIZONTAL SURFACE FOR A MV 2-3-4D VARIABLE
# -------------------------------------------------------------
#    
def ComputeGridSurface(cdmsvar):
  re = 6.37122e06
  # Compute the column horizontal surface 
  grid = cdmsvar.getGrid()
  latwts, lonwts = grid.getWeights()
  weights = MV.outerproduct(latwts, lonwts)
  area = MV.array(weights*4*numpy.pi*re**2,axes=(grid.getLatitude(),grid.getLongitude()),id='GridArea',\
                    attributes={'units':'m^2','standard_name':'Horizontal area of grid cells'})
  return area
#
# -------------------------------------------------------------
# COMPUTATION OF THE HORIZONTAL FLUX THROUGH A VERTICAL SURFACE 
# WITH ORIENTATION THETA (THETA = 0 -> NORD) AT A GIVEN PRESSURE LEV
# INPUT: x (MASS MIX RATIO, 2D), u,v (WIND COMPONENTS, m/s), theta (ANGLE
# WITH RESPECT TO NORTH, radians), dp (THICKNESS OF THE DISCRETI
# ZED PRESSURE LAYER, Pa > 0)
# OUTPUT: MASS FLUX (kg/s)
# -------------------------------------------------------------
#    
def ComputeHorizFluxPlev(x,u,v,theta):

  # Compute the wind component on the wanted direction
  wind = MV.sqrt(u**2+v**2)
  thetawind = MV.arctan2(u,v)
  projwind = wind*MV.cos(thetawind-theta)
  flux = x*projwind 
  return flux
#
# -------------------------------------------------------------
# FIND THE INTEGRATION INTERVAL IN AN ORDERED PRESSURE ARRAY
# -------------------------------------------------------------
# It selects the closest pressure value in the given array   
def find_pressure_band(band_out,pressin,silent=True):
    
    # In case of masked array it finds the proper index
    if isinstance(pressin[:],numpy.ma.MaskedArray) is True:
        pressintmp = pressin.filled()[:]
    else:
        pressintmp = pressin[:]
        
    index = [-1,-1]
    #Check array orientation
    if (numpy.sort(pressintmp[::-1]) == pressintmp[::-1]).all(): 
      if not silent: print('>>> Decreasing pressure array')
      index[1] = numpy.ma.argmin(numpy.abs(band_out[0]-pressin))
      #index[0] = numpy.argmin(numpy.abs(band_out[1]-pressintmp))+1 This was used for SOFRID paper!!!
      index[0] = numpy.ma.argmin(numpy.abs(band_out[1]-pressin))
    elif (numpy.sort(pressintmp) == pressintmp).all():
      if not silent: print('>>> Increasing pressure array')
      #index[0] = numpy.argmin(numpy.abs(band_out[0]-pressintmp))+1 This was used for SOFRID paper!!!
      index[0] = numpy.ma.argmin(numpy.abs(band_out[0]-pressin))
      index[1] = numpy.ma.argmin(numpy.abs(band_out[1]-pressin))
    else:
      print('>>> ERROR: Pressure array not ordered or not recognized!!!')

    if not silent:
      print('>>> Using this pressure bands')
      print(index)
      print(pressin[index[0]:index[1]])
      print(len(pressin[index[0]:index[1]]))
    return index
#
# -------------------------------------------------------------
# PREPARE A TIME SERIES BASED ON INITIAL DATE, N DAYS AND TIME STEP
# -------------------------------------------------------------
#    
def Create_Time_Series(ini_date,n_days,step=3600,returnstring=False):
  nsteps = n_days*int(3600.*24./step)+1
  timeseries = [datetime.datetime(int(ini_date[:4]),int(ini_date[4:6]),int(ini_date[6:8]))+datetime.timedelta(seconds=step*i) for i in range(nsteps)]
  timeseries_str = ['%04d%02d%02d%02d' % (timeseries[i].year,timeseries[i].month,timeseries[i].day,timeseries[i].hour) for i in range(len(timeseries))]
  fin_time = timeseries_str[-1]
  if returnstring is True: 
    return timeseries_str, ini_date, fin_time
  else:
    return timeseries, ini_date, fin_time

def Create_Date_Series(ini_date,n_days,returnstring=False):
  datein_dt = datetime.date(int(ini_date[:4]),int(ini_date[4:6]),int(ini_date[6:8]))
  dateseries = [datein_dt+datetime.timedelta(days=i) for i in range(n_days)]
  fin_date = dateseries[-1].isoformat().replace("-","")
  if returnstring is True: 
    return [date.isoformat().replace("-","") for date in dateseries], ini_date, fin_date
  else:
    return dateseries, ini_date, fin_date
#
# -------------------------------------------------------------
# CALCUL DES STATISTIQUES 2D POUR UN CHAMP MODEL/SATELLITE OU POINT DATA
# -------------------------------------------------------------
#    
def ComputeScatteredStatistics(x,detailed=False,newline='~C~'):
  # x is a numpy (masked) array
  xtmp = MV.array(x)
  avg = MV.average(xtmp)
  max = MV.max(xtmp)
  min = MV.min(xtmp)
  
  print('Points stats (N,avg,max,min): ',xtmp.count(),avg,max,min)
  if detailed is True:
    perc05, median, perc95 = genutil.statistics.percentiles(xtmp,[5.,50.,95.])
    print('Field percentiles (5,50,95): ', perc05, median, perc95)
    textlabel =  " N: %6d %s Avg (Med) : %.1f (%.1f) %s Max (95prc): %.1f (%.1f) %s Min (5prc) : %.1f (%.1f)" \
                 % (xtmp.count(),newline,avg,median,newline,max,perc95,newline,min,perc05)
  else:
    textlabel =  " N: %6d %s Avg: %.1f %s Max: %.1f %s Min: %.1f" \
                 % (xtmp.count(),newline,avg,newline,max,newline,min)
  
  return textlabel

def ComputeAreaStatistics(x,detailed=False,newline='~C~'):
  # x is a MV variable 
  xtmp = MV.array(x)

  avg = cdutil.averager(xtmp,axis='xy')
  max = MV.max(xtmp)
  min = MV.min(xtmp)

  print('Field stats (avg,max,min): ',avg,max,min)
  if detailed is True:
    perc05, median, perc95 = genutil.statistics.percentiles(xtmp,[5.,50.,95.],axis='xy')
    textlabel =  " Avg (Med) : %.2f (%.2f) %s Max (95prc): %.1f (%.1f) %s Min (5prc) : %.1f (%.1f)" \
                 % (avg,median,newline,max,perc95,newline,min,perc05)
  else:
    textlabel =  " Avg: %.2f %s Max: %.1f %s Min: %.1f" \
                 % (avg,newline,max,newline,min)
  return textlabel

def ComputeDifferenceStatistics(x,y,newline='~C~'):
  
  units = '(ppbv)'
  line = 'Statistics for N=%i' % (x.count())
  line = line+newline+'R: %.2f' % (genutil.statistics.correlation(x,y))
  line = line+newline+'Bias '+units+': %.1f' % (numpy.ma.average(y-x))
  line = line+newline+'Bias (%%): %.1f' % (numpy.ma.average(y-x)/numpy.ma.average(x)*100)
  line = line+newline+'Std '+units+': %.1f' % (genutil.statistics.std(x-y))
  line = line+newline+'Std (%%): %.1f' % (genutil.statistics.std(x-y)/numpy.ma.average(x)*100)
  line = line+newline+'Rmse '+units+': %.1f' % (genutil.statistics.rms(x,y))
  line = line+newline+'Rmse (%%): %.1f' % (genutil.statistics.rms(x,y)/numpy.ma.average(x)*100)
  return line
#
# -------------------------------------------------------------
# Computation of the distance between two lat/lon points on the sphere   
# -------------------------------------------------------------
# 
def GreatCircleDistance(lat1,lat2,lon1,lon2,r):
          
    dlat = numpy.radians(lat2-lat1)
    dlon = numpy.radians(lon2-lon1)
    a = numpy.sin(dlat/2) * numpy.sin(dlat/2) + numpy.cos(numpy.radians(lat1)) \
        * numpy.cos(numpy.radians(lat2)) * numpy.sin(dlon/2) * numpy.sin(dlon/2)
    c = 2 * numpy.arctan2(numpy.sqrt(a), numpy.sqrt(1-a))
    d = r * c
    return d

def ReadFileVtable(nlev, path='./') :
  """
  Read a file of the vertical coefficients  with a format similar to 
  /vol/rdx_dir/data/ifs/vtable_L* (ECMWF)
  [nlev] is the number of levels
  [path] is the path where the file can be found
  """  
  #
  # Read the lines of the file
  #
  filename = 'vtable_L%i' % (nlev)
  filename = os.path.join(path,filename)
  if not os.path.exists(filename) : sys.exit("File %s not found" % filename) 
  fid = open(filename,'r')
  lines = fid.readlines()
  fid.close()
  #
  # Read the coeficients
  #
  dvalh=[0]
  dvblh=[0]
  for iline in range(2,2+nlev) : 
    val = float(lines[iline].split(',')[0])
    dvalh.append(val)
  for iline in range(3+nlev,3+nlev*2) : 
    val = float(lines[iline].split(',')[0])
    dvblh.append(val)
  #
  # Routine end
  #  
  return numpy.array(dvalh), numpy.array(dvblh)
#
# -------------------------------------------------------------
# Computation of hourly climatology for a MV variable   
# -------------------------------------------------------------
# 
def compute_hourly_clim(x):

  time = x.getTime().asComponentTime()[:]

  hourax=cdms.createAxis(numpy.arange(1,25,dtype='f')*3600)
  hourax.designateTime()
  hourax.id = 'time'
  datestring = "%4i-%02i-%02i"\
      % (int(time[0].year), int(time[0].month), int(time[0].day))
  hourax.units = 'seconds since '+datestring+' 00:00:00'
  hourax.time_origin = datestring+' 00:00:00'
  hourax.calendar = 'noleap'

  newaxes = x.getAxisList(omit='time')
  newaxes.insert(0,hourax)
  newshape = tuple([len(a) for a in newaxes]) 

  mean = MV.zeros(newshape,axes=newaxes,id=x.id)
  
  # The climatology is for each hour
  for s in range(1,25):
    if s == 24 : s = 0
    timeidx = [i for i,t in enumerate(time) if t.hour == s]
    for idx in timeidx:
      mean[s] += x[idx]
    mean[s] = mean[s]/len(timeidx)
    
  print(mean.shape)
  return mean
