# -*- coding: utf-8 -*-
"""
Library containing plot procedures for MOCAGE-DAIMON fields and obs

@author: emili
"""
from __future__ import print_function
try:
    import Ngl
except ImportError:
    pass

#import cartopy.crs as ccrs
import os
import numpy
import MV2 as MV
from cdms2.selectors import Selector
#import seaborn as sns # Some incompatibiliy with cartopy
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, FormatStrFormatter, NullFormatter
from matplotlib import rcParams
from matplotlib import dates as mpdates
import matplotlib.colors as mc
from mpl_toolkits.basemap import Basemap
import itertools

###################################################################
def PlotUnits(species,plttype,plotconfig,differences=False,std=False,vertlev=None,silent=True):
    
    #Some constant unit conversion factors
    convfactors = { \
                   # Generic conversion factors
                   "pappmv2moleccm2":0.0078914*2.69*10, \
                   "one":1., \
                   "pappmv2du":0.0078914, \
                   "vmr2ppmv":1.e6, \
                   "vmr2ppbv":1.e9, \
                   "pa2mbar":1./100, \
                   # MOCAGE specific (Mocage tot cols in netcdf outputs are in molec/m2)
                   "mocage2moleccm2":1./1.e4/1.e15, \
                   "mocage2du":1./2.6868e+20, \
                   "mocage2mugm3":1e9}

    # Dictionary of equivalence between the species described in the user json config 
    # and the routine caller defined species
    # Can be useful to avoid adding species with similar scales to the json config file or 
    # for species with names that cannot be freely defined in the calling routine
    species_dict = {\
                    # Chemicals
                    'O_3':'O3','O3':'O3','O_x':'O3','NO_2':'NO2','NO2':'NO2','CO':'CO','HCHO':'HCHO',\
                    'NO3':'NO2','NO':'NO1','PM10':'PM10','PM2.5':'PM2.5','PM10-PM2.5':'PM10',\
                    'CH_4':'CH4','CO2':'CO2','NH_3':'NH3','SO_2':'SO2','ISO':'ISO','HNO_3':'HNO3',\
                    # Covariance stats
                    'Lx':'CorrelLength','Ly':'CorrelLength','Lz':'CorrelLength','Ave':'CO','Std':'CO',\
                    'O_x_Lx_3D':'CorrelLength','O_x_Ly_3D':'CorrelLength','O_x_Lz_3D':'CorrelLength',\
                    'O_x_var_3d':'O3','O_x_ave_3d':'O3','O_x_v_cor_3d':'CorrelLength',\
                    # Meteorology
                    'x_wind':'Wind','y_wind':'Wind','z_wind':'Wind','uv':'Wind',\
                    'air_temperature':'Temp','air_pressure_at_surface':'Press','pv':'PV','psi':'PV',\
                    'u':'Wind','v':'Wind',\
                    # Aerosols
                    'DESDUS':'PM','PMANTH':'PM','DYNSAL':'PM','BLACKC':'PM','TOTAM':'TOTAM','AOD':'AOD',\
                    # Others
                    'percentage':'Percent','ratio':'Ratio'}
                         
    cfg_species = species_dict[species] if species in species_dict else species

    if silent is False: 
        print( '>>> Setting plot levels for ', cfg_species)
        print('>>> Plot type: ', plttype)

    # Define the type of values to be found in the json namelist
    valtype = 'values'
    if differences is True: 
        if silent is False: print('>>> Bias plot')
        valtype = 'bias'
    if std is True: 
        if silent is False: print('>>> Rmse plot')
        valtype = 'rmse'

    if cfg_species in plotconfig['PlotLevels']:
        title = str(plotconfig['PlotLevels'][cfg_species]['title'])
        # If a specific plot type (level,column,profile,totcol or emissions)
        # is found select the proper object
        if plttype in plotconfig['PlotLevels'][cfg_species]:
            # If we plot horizontal levels and the vertlev key is found use the proper object
            # Do the vertical selection again
            if (plttype == 'level') and ('layers' \
             in plotconfig['PlotLevels'][cfg_species][plttype]) \
               and(vertlev is not None):
                # Locate the atmospheric layer
                layers = plotconfig['PlotLevels'][cfg_species][plttype]['layers']
                usedlayer = None
                for layer in layers.keys():
                    limits = sorted(layers[layer])
                    if vertlev in range(limits[0],limits[1]+1):
                        usedlayer = layer
                if usedlayer is None: print('Error! Cannot locate this level in the namelist layers: ' ,vertlev)
                if silent is False: print('>>> Selecting levels for '+plttype+' in '+usedlayer)
                cfg = plotconfig['PlotLevels'][cfg_species][plttype][usedlayer]
            else:
                if silent is False: print('>>> Selecting levels for ',plttype)
                cfg = plotconfig['PlotLevels'][cfg_species][plttype]
        else:
            # Use general config otherwise
            cfg = plotconfig['PlotLevels'][cfg_species]

        # If multiple type of levels are found use the correct one
        if 'LevelsType' in cfg.keys():
            if silent is False: print('Using '+cfg['LevelsType']+' levels')
            plotlev = cfg[cfg['LevelsType']][valtype]
            if cfg['LevelsType'] == "LinearRange":
                levstyle = 'linear'
                plotrange = plotlev[:2]
            elif cfg['LevelsType'] == "Explicit":
                levstyle = 'explicit'
                plotrange = [plotlev[0],plotlev[-1]]
            else:
                print('!!! Problem with cfg: expecting LinearRange or Explicit levels')
        else:
            # Set linear range as default otherwise
            if silent is False: print('Using linear levels')
            plotlev = cfg['LinearRange'][valtype]
            levstyle = 'linear'
            plotrange = plotlev[:2]
    else:
        # Use automatic levels if the namelist does not contain levels information
        if silent is False: print('Using automatic levels')
        cfg = {'units':'(ppbv)','convfactor':'vmr2ppbv'}
        title = cfg_species
        levstyle = 'auto'
        plotlev = None
        plotrange = [None,None]
    
    # Returns a dictionary, which contains complete information for
    # producing chemical plots with the functions that follow   
    plotdict = {'units':cfg['units'],\
                'species':title,\
                'convfact':convfactors[cfg['convfactor']] if cfg['convfactor'] in convfactors else cfg['convfactor'],\
                'plotlev':[levstyle,plotlev,plotrange] \
                }

    return plotdict
    # Should switch to this when all scripts are adapted (for improved clarity of usage)    
    # return {'units':plotunits,'species':species_title,'convfact':convfactor,\
    #        'plotlev':{'levstyle':levstyle,'levels':plotlev,'range':{'min':plotrange[0],'max':plotrange[1]}},\
    #        'region':region,'mapinfo':mapinfo}

def SelectRegion(regcode,plotconfig):
    
    selector = Selector(latitude=(plotconfig['PlotRegions'][regcode]['LatMin'],\
                                  plotconfig['PlotRegions'][regcode]['LatMax']),\
                        longitude=(plotconfig['PlotRegions'][regcode]['LonMin'],\
                                   plotconfig['PlotRegions'][regcode]['LonMax']))
            
    return selector,plotconfig['PlotRegions'][regcode]

def DefineRegion(plotdict):
    
    if 'region' in plotdict.keys():
        # If not given the default longitude axis is not cyclic for the plotter
        if 'LonIsCyclic' not in plotdict['region']: plotdict['region']['LonIsCyclic'] = False
        return plotdict['region']
    else:
        return {'name':'GLOB22','LatMin':-90,'LatMax':90,'LonMin':-180,'LonMax':180,'LonIsCyclic':True,\
            'legend':{'TextLat':-65,'TextLon':-170}}

def kill_outsiders(x,xrange):
    # Set outsiders to the outer elements of color table
    # It speeds up the color plots in presence of bizarre values
    x[(x < xrange[0]).nonzero()] = xrange[0]-numpy.abs(xrange[0])*0.01
    x[(x > xrange[1]).nonzero()] = xrange[1]+numpy.abs(xrange[1])*0.01
    return x

def matplotlib_colors(i,palette='default',cfgfile='default'):
    if palette == 'manual':
#        with open(jsonConfigFile(cfgfile)) as file:    
#            plotconfig = json.load(file)
        # Set the colors manually by name (the number of plotted lines is limited to the length of the color list)
        #colorlist = plotconfig['PlotOptions']['Colors']['mplibPalette']
        colorlist = ['royalblue','salmon','brown','black','darkcyan','violet', \
                     'crimson','teal','chartreuse','brown']
        color = colorlist[i]
    elif palette == 'default':
        # Generate seaborn palette with the iterator and take the corresponding color
        palette = itertools.cycle(sns.color_palette('hls'))
        colorlist = [next(palette) for n in range(i+1)]
        color = colorlist[-1]
    elif palette == 'husl':
        # Generate seaborn palette with the iterator and take the corresponding color
        palette = itertools.cycle(sns.color_palette("husl"))
        colorlist = [next(palette) for n in range(i+1)]
        color = colorlist[-1]
    else:
        # Use standard matplotlib colors
        colorlist = colors.cnames.keys()
        color = colorlist[i]
    return color

def SetMatplotlibStyle():
    ### rcParams are the default parameters for matplotlib
    rcParams['font.size'] = 13.
    rcParams['axes.labelsize'] = 14.
    rcParams['xtick.labelsize'] = 13.
    rcParams['ytick.labelsize'] = 14.
    rcParams['legend.fontsize'] = 12
    rcParams['savefig.dpi'] = 300
    #rcParams['savefig.format'] = 'pdf'

def define_levels(res,x,dictionary):
    
    if dictionary['lev'][0] == 'linear':
        levels = dictionary['lev'][1]
        res.cnLevelSelectionMode = "ManualLevels"
        res.cnMinLevelValF       = levels[0]
        res.cnMaxLevelValF       = levels[1]
        res.cnLevelSpacingF      = levels[2]
        x = kill_outsiders(x,[levels[0],levels[1]])
    if dictionary['lev'][0] == 'explicit':
        levels = dictionary['lev'][1]
        res.cnLevelSelectionMode = "ExplicitLevels"
        res.cnLevels = levels
        x = kill_outsiders(x,[levels[0],levels[-1]])
    if dictionary['lev'][0] == 'auto':
        res.cnLevelSelectionMode = "AutomaticLevels"

    return x

def define_map_projection(res,dictionary):
    
    # Equivalence between basemap and ncl projections
    nclmaptypes = {'ortho':'Orthographic','nsper':'Orthographic','cyl':'CylindricalEquidistant','robin':'Robinson'}
    
    if "MapProjection" in dictionary:
        res.mpProjection = nclmaptypes[str(dictionary["MapProjection"]["type"])]
        if res.mpProjection == 'Orthographic':
            res.mpCenterLatF      =  dictionary["MapProjection"]["LatCenter"]
            res.mpCenterLonF      =  dictionary["MapProjection"]["LonCenter"]
    else:
        res.mpProjection = "CylindricalEquidistant"
        
    return None

def define_map_limits(res,region):
    res.mpLimitMode                 = "LatLon"
    res.mpMinLonF                   =  region['LonMin']
    res.mpMaxLonF                   =  region['LonMax']
    res.mpMinLatF                   =  region['LatMin']
    res.mpMaxLatF                   =  region['LatMax']
    res.mpGridAndLimbOn             = True
    res.mpGridLineDashPattern       = 2
    res.mpGridSpacingF = 20.
    res.mpGeophysicalLineThicknessF = 2.
    res.pmTitleDisplayMode          = "Always"
    res.mpGeophysicalLineThicknessF = 4.
    res.mpGeophysicalLineColor = "Black"
    return None

def define_vertical_grid(res,dictionary,zonal=False):
    if 'levtype' in dictionary.keys():
        if dictionary['levtype'] in ('plev','plevhPa'):
            res.trYMinF     =  0.1
            res.trYMaxF     =  1000
            if zonal is True: 
                res.nglYAxisType = "LogAxis"
            else:
                res.trYAxisType = "LogAxis"
            res.tiYAxisString    = "z (hPa)"
            res.tmYLMode   = "Explicit"
            res.tmYLValues = [900.,700.,500.,300.,200.,100.,50.,20.,10.,5.,1.,0.5,0.1]
            res.tmYLLabels = ['900','700','500','300','200','100','50','20','10','5','1','0.5','0.1']
            #res.tmYLValues = [900.,800.,700.,600.,500.,400,300.,200.]
            #res.tmYLLabels = ['900','800','700','600','500','400','300','200']
        if dictionary['levtype'] == 'mocage47':
            res.tiYAxisString    = "z (Mocage Lev)"
            res.tmYUseLeft = False
            res.tmYRLabelsOn = True
            res.tmYRMode   = "Explicit"
            res.tmYRValues = [10,20,30,40]
            res.tmYRLabels = ['40 hPa','130 hPa','450 hPa','870 hPa']
        if dictionary['levtype'] == 'mocage60':
            res.tiYAxisString    = "z (Mocage Lev)"
            res.tmYUseLeft = False
            res.tmYRLabelsOn = True
            res.tmYRMode   = "Explicit"
            res.tmYRValues = [11,22,26,30,35,39,46,49]
            res.tmYRLabels = ['5 hPa','50 hPa','100 hPa','200 hPa','350 hPa','500 hPa','750 hPa','850 hPa']            
    else:
        res.tiYAxisString    = "z (Model Lev)"
        res.tmYUseLeft = False
        res.tmYRLabelsOn = True           
                    
    res.trYReverse   = True
    return None

def add_text(wks,map,dictionary,position):
    txres               = Ngl.Resources()
    txres.txAngleF              = 0.
    txres.txJust                = "CenterLeft"
    txres.txPerimOn             = True
    txres.txFontColor           = "black"
    txres.txFontHeightF = 0.015
    txres.txBackgroundFillColor = "White"
    text  = Ngl.text(wks,map,dictionary['text'],position[0],position[1],txres)
    return None

def add_signature(wks,map,region):
    txres               = Ngl.Resources()
    txres.txAngleF              = 0.
    txres.txPerimOn             = True
    txres.txFontColor           = "black"
    txres.txFontHeightF = 0.013
    txres.txBackgroundFillColor = "turquoise1"
    text  = Ngl.text(wks,map,'CERFACS',region['logolon'],region['logolat'],txres)
    return None
    
###################################################################

def plot_2DmapNGL(field,plotdict):
    
    #Copy the variable to avoid modifications in calling routine
    x = MV.array(field,copy=1)

            #Plot map of profile number
    rlist            = Ngl.Resources()
    if 'colormap' in plotdict.keys(): 
        rlist.wkColorMap = plotdict['colormap']
    else:
        rlist.wkColorMap = 'BlAqGrYeOrReVi200'
    wks_type = "png"
    #wks_type = "newpdf"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']),rlist)
    #
    #  Map resources.
    #
    region = DefineRegion(plotdict)

    res            = Ngl.Resources()   # map resources
    res.sfYArray = x.getLatitude()[:]
    if region['LonIsCyclic']:
        x, res.sfXArray = Ngl.add_cyclic(x,lon_coord=x.getLongitude()[:])
    else:
        res.sfXArray = x.getLongitude()[:]

    res.nglFrame = False

    define_map_projection(res,plotdict)

    define_map_limits(res,region)
    
    res.cnFillOn             = True
    res.cnLinesOn            = False
    res.cnLineLabelsOn       = False
    
    if 'satdata' in plotdict.keys():
        res.cnFillMode = "RasterFill"
    if 'clines' in plotdict.keys() and plotdict['clines'] is True:
        res.cnLinesOn            = True
        res.cnLineLabelsOn       = True
        res.cnLineLabelPlacementMode = 'Constant'
        res.cnLineLabelBackgroundColor = -1
        #res.cnLineDashPattern = 1
        #res.cnLineLabelFontHeightF = 0.006
        res.cnLineLabelFontHeightF = 0.009
        res.cnLineThicknessF = 1
        #res.mpGeophysicalLineThicknessF = 1.
    if 'lev' in plotdict.keys(): x=define_levels(res,x,plotdict)
    #res.lbBoxLinesOn  = False
    #res.lbLabelStride = 6
    res.lbAutoManage = False
    res.lbOrientation = "Horizontal"
    res.lbLabelFontHeightF = .022
    
    if 'title' in plotdict.keys(): res.tiMainString = plotdict['title']
    res.tiMainFontHeightF = 0.02    
    
    print('printing plot:', plotdict['filename'])
    map = Ngl.contour_map(wks,x,res)
    #Add text subtitle if keyword is given
    if 'text' in plotdict.keys(): add_text(wks,map,plotdict,\
        [region['legend']['TextLon'],region['legend']['TextLat']])
    #add_signature(wks,map,region)
    Ngl.frame(wks)
    Ngl.destroy(wks) 

###################################################################

def plot_2Dmap_and_wind(field,u,v,plotdict):

    #Copy the variable to avoid modifications in calling routine
    x = MV.array(field,copy=1)
    
    #Plot map of profile number
    rlist            = Ngl.Resources()
    if 'colormap' in plotdict.keys():
        rlist.wkColorMap = plotdict['colormap']
    else:
        rlist.wkColorMap = 'BlAqGrYeOrReVi200'
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']),rlist)
    #
    #  Map resources.
    #
    mpres            = Ngl.Resources()   # map resources
    mpres.nglDraw  = False
    mpres.nglFrame = False

    region = DefineRegion(plotdict)

    define_map_projection(mpres,plotdict)

    define_map_limits(mpres,region)

    cres            = Ngl.Resources()    # contour resources
    cres.sfYArray = x.getLatitude()[:]
    if region['LonIsCyclic']:
        x, cres.sfXArray = Ngl.add_cyclic(x,lon_coord=x.getLongitude()[:])
    else:
        cres.sfXArray = x.getLongitude()[:]

    cres.nglFrame = False
    cres.nglDraw  = False
    cres.cnFillOn             = True
    cres.cnLinesOn            = False
    cres.cnLineLabelsOn       = False
    cres.lbAutoManage = False
    cres.lbOrientation = "Horizontal"
    cres.lbLabelFontHeightF = .022

    if 'clines' in plotdict.keys() and plotdict['clines'] is True:
        cres.cnLinesOn            = True
        cres.cnLineLabelsOn       = True
        cres.cnLineLabelPlacementMode = 'Constant'
        cres.cnLineLabelBackgroundColor = -1
        #cres.cnLineDashPattern = 1
        cres.cnLineLabelFontHeightF = 0.009
        cres.cnLineThicknessF = 1
        #mpres.mpGeophysicalLineThicknessF = 1.
    if 'lev' in plotdict.keys(): x = define_levels(cres,x,plotdict)
    if 'title' in plotdict.keys(): cres.tiMainString = plotdict['title']
    cres.tiMainFontHeightF = 0.02
    
    print('printing plot:', plotdict['filename'])
    
    map = Ngl.map(wks,mpres)
    cmap = Ngl.contour(wks,x,cres)

    #Vector resource
    vcres = Ngl.Resources()
    vcres.nglFrame = False
    vcres.nglDraw  = False
    vcres.vfXArray = x.getLongitude()[:]
    vcres.vfYArray = x.getLatitude()[:]
    #vcres.vcGlyphStyle       = "CurlyVector"
    vcres.vcMinDistanceF = 0.02
    vcres.vcRefMagnitudeF = plotdict['RefMagnitudeF']
    vcres.vcRefLengthF = 0.01
    #vcres.vcMinFracLengthF   = 0.3
    vcres.vcRefAnnoArrowLineColor = "Black"
    #vcres.vcMaxLevelCount      = 23
    #vcres.vcMonoLineArrowColor = False
    vcres.vcLineArrowThicknessF = 2

    vmap = Ngl.vector(wks,u,v,vcres)
    #
    # Overlay everything on the map plot.
    #
    Ngl.overlay(map,cmap)
    Ngl.overlay(map,vmap)
    Ngl.maximize_plot(wks,map)
    Ngl.draw(map)
    #Add text subtitle if keyword is given
    if 'text' in plotdict.keys(): add_text(wks,map,plotdict,\
        [region['legend']['TextLon'],region['legend']['TextLat']])
    Ngl.frame(wks)
    Ngl.destroy(wks)
    
##################################################################

def plot_2DmapCartopy(field,plotdict):
    
    if not os.path.exists(os.path.split(plotdict['filename'])[0]):
        os.makedirs(os.path.split(plotdict['filename'])[0])
    
    # Don't plot cause seg fault with cartopy on my laptop install
    # return
    #Copy the variable to avoid modifications in calling routine
    x = MV.array(field,copy=1)
    
    lats = x.getLatitude()[:]
    lons = x.getLongitude()[:]

    lons, lats = numpy.meshgrid(lons, lats)
    
    fig = plt.figure()
    # Setup a global EckertIII map with faint coastlines.
    #ax = fig.add_subplot(1, 1, 1, projection=ccrs.EckertIII())
    print('add subplot')
    ax = fig.add_subplot(1, 1, 1, projection=ccrs.PlateCarree())
    #ax.set_global()
    print('set coastlines')
    ax.coastlines('110m', alpha=0.1)

    # if 'colormap' in plotdict.keys(): 
    #     rlist.wkColorMap = plotdict['colormap']
    # else:
    #     rlist.wkColorMap = 'BlAqGrYeOrReVi200'
    
    filled_c = ax.contourf(lons, lats, x, transform=ccrs.PlateCarree())
    
    # And black line contours.
    line_c = ax.contour(lons, lats, x, levels=filled_c.levels,
                    colors=['black'],
                    transform=ccrs.PlateCarree())

    # Uncomment to make the line contours invisible.
    # plt.setp(line_c.collections, visible=False)
    
    # Add a colorbar for the filled contour.
    fig.colorbar(filled_c, orientation='horizontal')
    
    # Use the line contours to place contour labels.
    ax.clabel(
        line_c,  # Typically best results when labelling line contours.
        colors=['black'],
        manual=False,  # Automatic placement vs manual placement.
        inline=True,  # Cut the line where the label will be placed.
        fmt=' {:.0f} '.format,  # Labes as integers, with some extra space.
    )
    
    if 'title' in plotdict.keys(): ax.set_title(plotdict['title'])
#    if 'text' in plotdict.keys(): add_text(wks,map,plotdict,\
#        [region['legend']['TextLon'],region['legend']['TextLat']])
    #plt.show()
    
    plt.savefig(plotdict['filename'])
    # if region['LonIsCyclic']:
    #     x, res.sfXArray = Ngl.add_cyclic(x,lon_coord=x.getLongitude()[:])
    # else:
    #     res.sfXArray = x.getLongitude()[:]

###################################################################

def plot_2Dmap(field,plotdict):

    addgrid = True
    resol = 'l'
    dlat = 10.
    dlon = 20.
    cmap = 'jet'

    if 'colormap' in plotdict.keys(): cmap = plotdict['colormap']

    print('printing plot:', plotdict['filename'])

    x = MV.array(field,copy=1)
    
    lats = x.getLatitude()[:]
    lons = x.getLongitude()[:]

    lons, lats = numpy.meshgrid(lons, lats)
    fig = plt.figure(figsize=(10,8),dpi=200)
        
    ax = fig.gca()

    m = Basemap(projection='cyl',resolution=resol)

    m.drawmapboundary()
    #m.fillcontinents()

    m.drawcoastlines()

    if addgrid: 
        m.drawparallels(numpy.arange(-90.,90.,dlat))
        m.drawmeridians(numpy.arange(-180.,180.,dlon))

    if 'region' in plotdict and (m.projection not in ('nsper','ortho','geos','robin')):
        # your extent in lat/lon (dec degrees)
        ulx = plotdict['region']["LonMin"]
        uly = plotdict['region']["LatMax"]
        lrx = plotdict['region']["LonMax"]
        lry = plotdict['region']["LatMin"]
        # transform coordinates to map projection
        xmin, ymin = m(ulx, lry)
        xmax, ymax = m(lrx, uly)
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

    vmin, vmax = x.min(), x.max()
    
    if 'lev' in plotdict.keys():
        if plotdict['lev'][0] == 'linear':
            levels = plotdict['lev'][1]
            vmin       = levels[0]
            vmax       = levels[1]
            x = kill_outsiders(x,[levels[0],levels[1]])
        if plotdict['lev'][0] == 'explicit':
            levels = plotdict['lev'][1]
            x = kill_outsiders(x,[levels[0],levels[-1]])

    levels=numpy.linspace(vmin,vmax,11)
    norm = mc.BoundaryNorm(levels, 256)

    plot = m.contourf(lons, lats, x, latlon=True, cmap=cmap, levels=levels,
                           norm=norm,
                           extend='both')

    if 'title' in plotdict.keys(): ax.set_title(plotdict['title'])
    
    cbar = fig.colorbar(plot, orientation='horizontal')
    plt.savefig(plotdict['filename'])
    plt.close(fig)

def plot_2Dmap_and_obs(field,lon,lat,plotdict):

    #Copy the variable to avoid modifications in calling routine
    x = MV.array(field,copy=1)

            #Plot map of profile number
    rlist            = Ngl.Resources()
    if 'colormap' in plotdict.keys(): 
        rlist.wkColorMap = plotdict['colormap']
    else:
        rlist.wkColorMap = 'BlAqGrYeOrReVi200'
    wks_type = "png"

    print('printing plot:', plotdict['filename'])
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']),rlist)
    #
    #  Map resources.
    #
    mpres            = Ngl.Resources()   # map resources
    mpres.nglFrame = False
    mpres.nglDraw  = False

    region = DefineRegion(plotdict)
        
    define_map_projection(mpres,plotdict)

    define_map_limits(mpres,region)

    if 'title' in plotdict.keys(): mpres.tiMainString = plotdict['title']
    mpres.tiMainFontHeightF = 0.02    

    cres            = Ngl.Resources()    # contour resources
    cres.sfYArray = x.getLatitude()[:]
    if region['LonIsCyclic']:
        x, cres.sfXArray = Ngl.add_cyclic(x,lon_coord=x.getLongitude()[:])
    else:
        cres.sfXArray = x.getLongitude()[:]
    cres.nglFrame = False
    cres.nglDraw  = False
    cres.cnFillOn             = True
    cres.cnLinesOn            = False
    cres.cnLineLabelsOn       = False
    cres.lbAutoManage = False
    cres.lbOrientation = "Horizontal"
    cres.lbLabelFontHeightF = .022
    if 'satdata' in plotdict.keys() and plotdict['satdata'] is True:
        cres.cnFillMode = "RasterFill"

    if 'clines' in plotdict.keys() and plotdict['clines'] is True:
        cres.cnLinesOn            = True
        cres.cnLineLabelsOn       = True
        cres.cnLineLabelPlacementMode = 'Constant'
        cres.cnLineLabelBackgroundColor = -1
        #res.cnLineDashPattern = 1
        cres.cnLineLabelFontHeightF = 0.006
        cres.cnLineThicknessF = 1
        #res.mpGeophysicalLineThicknessF = 1.
    if 'lev' in plotdict.keys(): x = define_levels(cres,x,plotdict)
    
    cres.lbAutoManage = False
    cres.lbOrientation = "Horizontal"
    cres.lbLabelFontHeightF = .022
    
    #Prepare plots
    map = Ngl.map(wks,mpres)
    cmap = Ngl.contour(wks,x,cres)
    #Add contours
    Ngl.overlay(map,cmap)
    Ngl.maximize_plot(wks,map)
    Ngl.draw(map)

    #Prepare symbol resources
    mres  = Ngl.Resources() # marker resources
    
    mres.gsMarkerIndex = 16
    mres.gsMarkerSizeF  = 3.        # marker size
    mres.gsMarkerColor  = "black"     # marker color
    mres.gsMarkerThicknessF = 1.5
    mres.gsMarkerOpacityF = 1

    Ngl.polymarker(wks,map,lon,lat,mres)
    
    mres.gsMarkerIndex = 16
    mres.gsMarkerSizeF  = 2.        # marker size
    mres.gsMarkerColor  = "red"     # marker color
    mres.gsMarkerThicknessF = 1.5
    mres.gsMarkerOpacityF = 0.8
    
    Ngl.polymarker(wks,map,lon,lat,mres)

    if 'text' in plotdict.keys(): add_text(wks,map,plotdict,\
        [region['legend']['TextLon'],region['legend']['TextLat']])
    
    Ngl.frame(wks)    
    Ngl.destroy(wks) 
    
##################################################################

def plot_obs_position(lon,lat,plotdict):

            #Plot map of profile number
    rlist            = Ngl.Resources()
    if 'colormap' in plotdict.keys(): 
        rlist.wkColorMap = plotdict['colormap']
    else:
        rlist.wkColorMap = 'BlAqGrYeOrReVi200'
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']),rlist)
    print('printing plot:', plotdict['filename'])
    #
    #  Map resources.
    #
    res            = Ngl.Resources()   # map resources
    res.nglFrame = False
    
    region = DefineRegion(plotdict)
        
    define_map_limits(res,region)

    define_map_projection(res,plotdict)

    if 'title' in plotdict.keys(): res.tiMainString      = plotdict['title']  

    map = Ngl.map(wks,res)
    
    mres  = Ngl.Resources() # marker resources
    mres.gsMarkerIndex = 0
    mres.gsMarkerSizeF  = 12.        # marker size
    mres.gsMarkerColor  = "black"     # marker color
    mres.gsMarkerThicknessF = 2.5

    Ngl.polymarker(wks,map,lon,lat,mres)

    if 'text' in plotdict.keys(): add_text(wks,map,plotdict,\
        [region['legend']['TextLon'],region['legend']['TextLat']])
        
    Ngl.frame(wks)    
    Ngl.destroy(wks)

###################################################################

def plot_obs_values(obs,lon,lat,plotdict):

    x = MV.array(obs,copy=1)
    
    rlist            = Ngl.Resources()
    if 'colormap' in plotdict.keys(): 
        rlist.wkColorMap = plotdict['colormap']
    else:
        rlist.wkColorMap = 'BlAqGrYeOrReVi200'
    
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']),rlist)
    print('printing plot:', plotdict['filename'])
    cmap = Ngl.get_MDfloat_array(wks,"wkColorMap")
    cmap_ncolors = cmap.shape[0]
    #
    #  Map resources.
    #
    res            = Ngl.Resources()   # map resources
    res.nglFrame = False
    res.nglDraw  = False

    region = DefineRegion(plotdict)
        
    define_map_projection(res,plotdict)

    define_map_limits(res,region)
       
    #res.mpGridAndLimbOn        = True

    if 'title' in plotdict.keys(): res.tiMainString = plotdict['title']
    res.tiMainFontHeightF = 0.02    
    
    map = Ngl.map(wks,res)

    mres  = Ngl.Resources() # marker resources
    mres.nglDraw = False
    mres.nglFrame = False
    mres.gsMarkerThicknessF = 2.5
            
    #Eliminate extreme values
    x = kill_outsiders(x,[plotdict['range'][0],plotdict['range'][1]])
   
    colors = range(cmap_ncolors)
    
    #Select the number of bins based on the color table
    if cmap_ncolors == 22: 
        nbins = 10 
    elif cmap_ncolors == 202 :
        nbins = 12
    elif cmap_ncolors == 190 :
        nbins = 14
    else:
        print('The selected table is not tuned for this type of plot')
        return

    #Prepare color binning
    bins = numpy.linspace(plotdict['range'][0],plotdict['range'][1],num=nbins)
    bin_labels = (bins[:-1]+bins[1:])/2.
    bin_labels = ['%3.2f' % (value) for value in bin_labels]
    col_lenbins = cmap_ncolors/nbins        
    #Set colors 
    stride = [col_lenbins*i+(cmap_ncolors % nbins)/2 for i in range(nbins+1)] 
    
    if 'typedata' in plotdict.keys() and plotdict['typedata'] == 'surf':
        size1 = 7.0
        size2 = 5.0
    if 'typedata' in plotdict.keys() and plotdict['typedata'] == 'sat':
        size1 = 5.0
        size2 = 4.0
        if res.mpProjection == 'Orthographic':
            size1 = 8.0
            size2 = 7.0
    if 'typedata' not in plotdict.keys():
        size1 = 9.0
        size2 = 7.0

    for st_ind in range(len(x)):
        indx = numpy.argmin(numpy.abs(x[st_ind]-bins))
        mres.gsMarkerColor  = stride[indx]
        mres.gsMarkerIndex = 16 
        mres.gsMarkerSizeF  = size2        
        Ngl.add_polymarker(wks,map,lon[st_ind],lat[st_ind],mres)

    lbres = Ngl.Resources()
    lbres.lbAutoManage = False
    lbres.nglDraw              = False
    lbres.lbPerimOn            = False            # turn off perimeter.

    lbres.vpWidthF             = 0.60             # width and height
    lbres.vpHeightF            = 0.15

    lbres.lbOrientation        = "Horizontal"
    lbres.lbLabelPosition      = "Bottom"         # label position
    lbres.lbFillColors         = colors[col_lenbins/2:(cmap_ncolors-1):col_lenbins]
    lbres.lbMonoFillPattern    = True             # solid color fill

#---Create the labelbar
    lbid = Ngl.labelbar_ndc(wks,len(bin_labels),bin_labels,0.,0.,lbres)
    amres                  = Ngl.Resources()
    amres.amOrthogonalPosF =  0.6
    annoid = Ngl.add_annotation(map,lbid,amres)
    if 'text' in plotdict.keys(): add_text(wks,map,plotdict,\
        [region['legend']['TextLon'],region['legend']['TextLat']])
    
    Ngl.maximize_plot(wks,map)
    Ngl.draw(wks)
    Ngl.frame(wks)
    Ngl.destroy(wks)
    #Ngl.end()   

###################################################################
    
def plot_zonal(field,plotdict):
    
    #Copy the variable to avoid modifications in calling routine
    x = MV.array(field,copy=1)

    wk_res            = Ngl.Resources()
    if 'colormap' in plotdict.keys(): 
        wk_res.wkColorMap = plotdict['colormap']
    else:
        wk_res.wkColorMap = 'BlAqGrYeOrReVi200'
    print('printing plot:', plotdict['filename'])
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']),wk_res)
#
# Set up a resource list for plot options.
#
    resources = Ngl.Resources()
    resources.sfXArray        = x.getLatitude()[:]
    resources.tiXAxisString    = "Latitude (degrees)"
    
# Set the vertical levels
    if 'levtype' in plotdict.keys() and plotdict['levtype'] == 'plev': 
        resources.sfYArray = x.getLevel()[:]/100 #Plot is in hPa scale
    elif 'levtype' in plotdict.keys() and plotdict['levtype'] == 'plevhPa':
        resources.sfYArray = x.getLevel()[:]
    else:
        resources.sfYArray = x.getLevel()[:]

    define_vertical_grid(resources,plotdict,zonal=True)

    resources.cnFillOn        = True
    #resources.cnLineLabelsOn  = False
    resources.cnLineLabelsOn  = True
    resources.cnLineLabelPlacementMode = 'Constant'
    resources.cnLineLabelBackgroundColor = -1
    if 'lev' in plotdict.keys(): x = define_levels(resources,x,plotdict)
        
    if 'title' in plotdict.keys(): resources.tiMainString     = plotdict['title']        # main title 
     #Add text subtitle if keyword is given       
    plot = Ngl.contour(wks,x,resources)
    
    Ngl.destroy(wks) 
    
###################################################################

def plot_vprofile(prof,lev,plotdict):
    
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']))
    print('printing plot:', plotdict['filename'])
    res1 = Ngl.Resources()
    
    res1.nglDraw  = False
    res1.nglFrame = False
    #
    # Set some axis strings, including the main title.
    #
    if 'title' in plotdict.keys(): res1.tiMainString     = plotdict['title']

    if 'xtitle' in plotdict.keys(): res1.tiXAxisString    = plotdict['xtitle']
    #
    # Don't maximize the plot in the frame, because we are controlling the
    # size and location of both plots with the profpXXX resources.
    #
    res1.nglMaximize = False
    res1.vpHeightF   = 0.70
    res1.vpWidthF    = 0.65
    res1.vpXF        = 0.20
    res1.vpYF        = 0.90
    #
    # Change line color and thickness.
    #
    #res1.xyLineColor      = "red"
    res1.xyLineThicknessF = 10.

    # Set the vertical levels
    define_vertical_grid(res1,plotdict)
    #
    # Set the actual limits of the X and Y axes.
    #
    if 'range' in plotdict.keys():
        levels = plotdict['range']
        res1.trXMinF     =  levels[0]
        res1.trXMaxF     =  levels[1]
        
    if 'log' in plotdict.keys() and plotdict['log'] == True:
        res1.trXAxisType = "LogAxis"
        res1.tmXBMinorOn = True
        res1.tmXMajorGrid = True
        res1.tmXMajorGridLineDashPattern = 1
        res1.tmXMajorGridThicknessF = 0.7
        res1.tmYUseLeft = True
        res1.tmYRMinorOn  = False
        res1.tmXTFormat   = "f"
              
    if isinstance(prof, list) is True:
        # Draw the first line to set axis etc...
        if isinstance(lev,list) is True:
            y = lev[i]
        else:
            y = lev
                
        if 'levtype' in plotdict.keys() and plotdict['levtype'] == 'plev':
            y = y/100

        x = prof[0]     

        plot = Ngl.xy(wks,x,y,res1)
        
        # Then plot the full list of profiles
        for i, x in enumerate(prof):
            if isinstance(lev,list) is True:
                y = lev[i]
            else:
                y = lev
            if 'levtype' in plotdict.keys() and plotdict['levtype'] == 'plev': 
                y = lev/100 #Plot is in hPa scale

            Ngl.add_polyline(wks,plot,x,y,res1)
            
    else:
        # Make only one plot if the profiles are not a list
        if 'levtype' in plotdict.keys() and plotdict['levtype'] == 'plev': 
            lev = lev/100 #Plot is in hPa scale
        plot = Ngl.xy(wks,prof,lev,res1)

    Ngl.draw(plot)
    Ngl.frame(wks)
    Ngl.destroy(wks)

###################################################################

def plot_2vprofile(v1,v2,z,plotdict):
    
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']))
    print('printing plot:', plotdict['filename'])
    res1 = Ngl.Resources()
    res2 = Ngl.Resources()
    res1.nglDraw  = False
    res1.nglFrame = False
    res2.nglDraw  = False
    res2.nglFrame = False
    #
    # Set some axis strings, including the main title.
    #
    if 'xtitle1' in plotdict.keys(): res1.tiXAxisString    = plotdict['xtitle1']
    #res1.tiXAxisFontColor = "red"
    #
    # Don't maximize the plot in the frame, because we are controlling the
    # size and location of both plots with the vpXXX resources.
    #
    res1.nglMaximize = False
    res1.vpHeightF   = 0.70
    res1.vpWidthF    = 0.65
    res1.vpXF        = 0.20
    res1.vpYF        = 0.85
    #
    # Change line color and thickness.
    #
    res1.xyLineColor      = "red"
    res1.xyLineThicknessF = 5.0
    #
    # Set the actual limits of the X and Y axes.
    #
    if 'range' in plotdict.keys():
        levels = plotdict['range']
        res1.trXMinF     =  levels[0]
        res1.trXMaxF     =  levels[1]
        res2.trXMinF     =  levels[0]
        res2.trXMaxF     =  levels[1]
        
    # Set the vertical levels
    if 'levtype' in plotdict.keys() and plotdict['levtype'] == 'plev': 
        z = z/100 #Plot is in hPa scale
    define_vertical_grid(res1,plotdict)
    
    res1.tmXTOn       = False
    res1.tmXTLabelsOn = False
    res1.tmXTMinorOn  = False
    res1.tmYLMinorOn  = True
    res1.tmXBFormat   = "f"  # Gets rid of unnecessary trailing zeros
    res1.xyLineThicknessF = 6.0

    plot1 = Ngl.xy(wks,v1,z,res1)

    #Put the plot one over the other
    res2.nglMaximize = False
    res2.vpHeightF   = res1.vpHeightF
    res2.vpWidthF    = res1.vpWidthF
    res2.vpXF        = res1.vpXF
    res2.vpYF        = res1.vpYF
    # Set the vertical levels for the second profile
    define_vertical_grid(res2,plotdict)
    res2.tmYUseLeft = True
    res2.tiYAxisOn = False
    #
    # Turn off the bottom and left tickmarks, since we will use the ones from
    # the first plot.
    #
    res2.tmXBOn       = False
    res2.tmXBLabelsOn = False
    res2.tmXBMinorOn  = False
    res2.tmYLOn       = False
    res2.tmYLLabelsOn = False
    res2.tmYLMinorOn  = False
    #
    # Turn on the right Y labels and tickmarks and move the axis string to
    # the right.
    #
    res2.tmXTLabelsOn = True
    res2.tmXTOn       = True
    res2.tmXUseBottom   = False  
    res2.tmYRMinorOn  = True
    res2.tmXTFormat   = "f"      # Gets rid of unnecessary trailing zeros
    #
    # Move the Y axis string to the right.
    #
    if 'xtitle2' in plotdict.keys(): res2.tiXAxisString    = plotdict['xtitle2']
    res2.tiXAxisSide        = "Top"
    res2.tiXAxisFontColor   = "black"
    #res2.tiYAxisFontHeightF = Ngl.get_float(plot1,"tiXAxisFontHeightF")
    
    #pattern = "$_____$_____$$_____$$_____$$_____$$_____$$___"
    #res2.xyDashPattern    = Ngl.new_dash_pattern(wks,pattern)
    res2.xyLineColor      = "black"
    res2.xyLineThicknessF = 6.0
    
    plot2 = Ngl.xy(wks,v2,z,res2)    

    Ngl.draw(plot1)
    Ngl.draw(plot2)
    Ngl.frame(wks)
    Ngl.destroy(wks)

###################################################################

def plot_histogram(x,plotdict):
    xtmp = numpy.ma.array(x).compressed()
    SetMatplotlibStyle()
    fig = plt.figure()
    ax = fig.add_subplot(111)
    #This puts all array values in a line and plot the histogram
    if 'title' in plotdict.keys(): ax.set_title(plotdict['title'])
    if 'xtitle' in plotdict.keys(): ax.set_xlabel(plotdict['xtitle'])
    if 'range' in plotdict.keys(): 
        hist_range=plotdict['range'] 
    else:
        hist_range = [numpy.ma.min(xtmp),numpy.ma.max(xtmp)]
    #stats = utils_cdms.ComputeScatteredStatistics(xtmp,detailed=True,newline='\n')
    ax.hist(xtmp,bins=35,range=hist_range,normed=True)#,label=stats)
    ax.legend()
    print('printing plot:', plotdict['filename'])
    plt.savefig(plotdict['filename'])
    plt.close()
    return

###################################################################

def plot_time_series(s,t,plotdict):

    # Time series plot which takes multiple series
    # t must be a list of python datetime objects
    # If only one datetime list is provided all series must be 
    # defined for the same time ticks 
    # (missing data is allowed in the series)

    SetMatplotlibStyle()
    #fig = plt.figure(figsize=(13,5))
    fig = plt.figure(figsize=(10,5))
    ax1 = fig.add_subplot(111)               
    
    if 'title' in plotdict.keys(): ax1.set_title(plotdict['title'])
    if 'ytitle' in plotdict.keys(): ax1.set_ylabel(plotdict['ytitle'])

    fig.autofmt_xdate()

    if 'colors' not in plotdict.keys() and 'palette' in plotdict.keys():
        palette = plotdict['palette']
    else:
        palette = 'default'
    
    for i, series in enumerate(s):
        if 'colors' in plotdict.keys(): 
            color = plotdict['colors'][i]
        else:
            color = matplotlib_colors(i,palette=palette)
        if 'linestyles' in plotdict.keys(): 
            linestyle = plotdict['linestyles'][i]
        else:
            linestyle = '-'
        if 'labels' in plotdict.keys() and plotdict['labels'] is not None: 
            label = plotdict['labels'][i]
        elif 'labels' in plotdict.keys() and plotdict['labels'] is None:
            label = None
        else:
            label = str(i)
        if 'markers' in plotdict.keys():
            marker = plotdict['markers'][i]
        else:
            marker = None
        if 'linewidth' in plotdict.keys(): 
            linewidth = plotdict['linewidth'][i]
        else:
            linewidth = 2
        if isinstance(t[0],list):
            ax1.plot(t[i],series,color=color,ls=linestyle,lw=linewidth,marker=marker,label=label,markersize=4.0)
        else:
            ax1.plot(t,series,color=color,ls=linestyle,lw=linewidth,marker=marker,label=label,markersize=4.0)
    
    if 'yrange' in plotdict.keys():
        majorLocator   = MultipleLocator(int((plotdict['yrange'][1]-plotdict['yrange'][0])/5))
        minorLocator   = MultipleLocator(int((plotdict['yrange'][1]-plotdict['yrange'][0])/10))
        ax1.yaxis.set_major_locator(majorLocator)
        ax1.yaxis.set_minor_locator(minorLocator)

    if 'tickformat' in plotdict.keys():
        majorFormatter = FormatStrFormatter(plotdict['tickformat'])
        ax1.yaxis.set_major_formatter(majorFormatter)
    #for the minor ticks, use no labels; default NullFormatter
    
    ax1.grid(True)

    if 'yrange' in plotdict.keys(): ax1.set_ylim(plotdict['yrange'][0],plotdict['yrange'][1])   
  
    if 'cycle' in plotdict.keys():
        if 'xrange' in plotdict.keys(): ax1.set_xlim(plotdict['xrange'][0],plotdict['xrange'][1])
    else:
        ax1.xaxis.set_major_formatter(mpdates.DateFormatter('%d %b, %H'))
        
    ax1.set_xlabel('Time (UTC)')
    ax1.legend(loc=2)

    print('printing plot:', plotdict['filename'])
    plt.savefig(plotdict['filename'])
    plt.close()

    return

###################################################################

def plot_2time_series(s1,s2,t,plotdict):
    # Time plot but with 2 different y axis, multiple series allowed
    # for both axis, series could be numpy, MA or MV array
    # t can be one or multiple lists of python datetime objects 
    # If only one datetime list is provided all series must be 
    # defined for the same time ticks 
    # (missing data is allowed)

    SetMatplotlibStyle()
    fig = plt.figure(figsize=(10,5))
    ax1 = fig.add_subplot(111)               

    if 'title' in plotdict.keys(): ax1.set_title(plotdict['title'])
    if 'ytitle' in plotdict.keys(): ax1.set_ylabel(plotdict['ytitle'])

    fig.autofmt_xdate()
    
    if 'colors' not in plotdict.keys() and 'palette' in plotdict.keys():
        palette = plotdict['palette']
    else:
        palette = 'default'

    for i, series in enumerate(s1):
        if 'colors' in plotdict.keys(): 
            color = plotdict['colors'][i]
        else:
            color = matplotlib_colors(i,palette=palette)
        if 'linestyles' in plotdict.keys(): 
            linestyle = plotdict['linestyles'][i]
        else:
            linestyle = '-'
        if 'linewidth' in plotdict.keys(): 
            linewidth = plotdict['linewidth'][i]
        else:
            linewidth = 2
        if 'labels' in plotdict.keys() and plotdict['labels'] is not None: 
            label = plotdict['labels'][i]
        elif 'labels' in plotdict.keys() and plotdict['labels'] is None:
            label = None
        else:
           label = str(i)
        if 'markers' in plotdict.keys():
            marker = plotdict['markers'][i]
        else:
            marker = None
        if isinstance(t[0],list):
            ax1.plot(t[i],series,color=color,ls=linestyle,lw=linewidth,marker=marker,label=label,markersize=4.0)
        else:
            ax1.plot(t,series,color=color,ls=linestyle,lw=linewidth,marker=marker,label=label,markersize=4.0)

    ax2 = ax1.twinx()
    ax2.xaxis.set_visible(False)   
    if 'ytitler' in plotdict.keys(): ax2.set_ylabel(plotdict['ytitler'])
    for i, series in enumerate(s2):
        if 'colorsr' in plotdict.keys(): 
            color = plotdict['colorsr'][i]
        else:
            color = matplotlib_colors(i,palette=palette)
        if 'linestylesr' in plotdict.keys(): 
            linestyle = plotdict['linestylesr'][i]
        else:
            linestyle = '-'
        if 'linewidthr' in plotdict.keys(): 
            linewidth = plotdict['linewidthr'][i]
        else:
            linewidth = 2
        if 'labelsr' in plotdict.keys() and plotdict['labelsr'] is not None: 
            label = plotdict['labelsr'][i]
        elif 'labelsr' in plotdict.keys() and plotdict['labelsr'] is None:
            label = None
        else:
            label = str(i)
        if 'markersr' in plotdict.keys():
            marker = plotdict['markersr'][i]
        else:
            marker = None
        if isinstance(t[0],list):
            ax2.plot(t[i],series,color=color,ls=linestyle,lw=linewidth,marker=marker,label=label,markersize=4.0) 
        else:
            ax2.plot(t,series,color=color,ls=linestyle,lw=linewidth,marker=marker,label=label,markersize=4.0) 
    
    ax1.legend(loc=2)
    ax2.legend(loc=1)
    
    if 'yrange' in plotdict.keys(): ax1.set_ylim(plotdict['yrange'][0],plotdict['yrange'][1])
    if 'yranger' in plotdict.keys(): ax2.set_ylim(plotdict['yranger'][0],plotdict['yranger'][1])
    if 'cycle' in plotdict.keys():
        if 'xrange' in plotdict.keys(): ax1.set_xlim(plotdict['xrange'][0],plotdict['xrange'][1])
    else:
        ax1.xaxis.set_major_formatter(matplotlib.dates.DateFormatter('%d %b, %H'))
    ax1.set_xlabel('Time (UTC)')
    ax1.grid(True) 
    ax2.grid(False)

    if 'yrange' in plotdict.keys():
        majorLocator   = MultipleLocator(int((plotdict['yrange'][1]-plotdict['yrange'][0])/5))
        minorLocator   = MultipleLocator(int((plotdict['yrange'][1]-plotdict['yrange'][0])/10))
        ax1.yaxis.set_major_locator(majorLocator)
        ax1.yaxis.set_minor_locator(minorLocator)
        
    if 'tickformat' in plotdict.keys():
        majorFormatter = FormatStrFormatter(plotdict['tickformat'])
        ax1.yaxis.set_major_formatter(majorFormatter)

    #for the minor ticks, use no labels; default NullFormatter

    print('printing plot:', plotdict['filename'])
    plt.savefig(plotdict['filename'])
    plt.close()

    return
 
###################################################################

def plot_2D_vector_map(u,v,plotdict):
    
           #Plot map of profile number
    rlist            = Ngl.Resources()
    if 'colormap' in plotdict.keys(): 
        rlist.wkColorMap = plotdict['colormap']
    else:
        rlist.wkColorMap = 'BlAqGrYeOrReVi200'
    rlist.wkWidth = 1600
    rlist.wkHeight = 1600
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']),rlist)

    region = DefineRegion(plotdict)
    
    vcres = Ngl.Resources()

    define_map_limits(vcres,region)
    
    vcres.mpProjection = "CylindricalEquidistant"
    
    vcres.nglFrame = False
    #vcres.nglDraw  = False
    vcres.sfYArray = u.getLatitude()[:]
    if region['LonIsCyclic']:
        u, vcres.sfXArray = Ngl.add_cyclic(u,lon_coord=u.getLongitude()[:])
    else:
        vcres.sfXArray = x.getLongitude()[:]

    if 'vectype' in plotdict.keys():
        vcres.vcGlyphStyle = plotdict['vectype']
    #else:
    #    vcres.vcGlyphStyle = "CurlyVector"
    #vcres.vcMinDistanceF = 0.015
    vcres.vcMinDistanceF = 0.025
    vcres.vcRefMagnitudeF = plotdict['RefMagnitudeF']
    vcres.vcRefLengthF = 0.01
    #vcres.vcMinFracLengthF   = 0.3
    #vcres.vcRefAnnoArrowLineColor = "Black"
    vcres.vcMaxLevelCount      = 23
    vcres.vcMonoLineArrowColor = False
    vcres.vcMonoFillArrowFillColor = False
    vcres.vcLineArrowThicknessF = 6
    if 'title' in plotdict.keys(): vcres.tiMainString = plotdict['title']
    
    print('printing plot:', plotdict['filename'])
    #map = Ngl.map(wks,mpres)
    map = Ngl.vector_map(wks,u,v,vcres)
    #Ngl.overlay(map,vmap)
    #Add text subtitle if keyword is given
    if 'text' in plotdict.keys(): add_text(wks,map,plotdict,\
        [region['legend']['TextLon'],region['legend']['TextLat']])

    Ngl.frame(wks)
    Ngl.destroy(wks) 
    
    #Ngl.end()

###################################################################

def plot_xy(x,y,plotdict):
    
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']))    
    print('printing plot:', plotdict['filename'])

    res1 = Ngl.Resources()
#
# Don't maximize the plot in the frame, because we are controlling the
# size and location of both plots with the vpXXX resources.
#
    res1.nglMaximize = False
    res1.vpHeightF   = 0.75
    res1.vpWidthF    = 0.65
    res1.vpXF        = 0.20
    res1.vpYF        = 0.90
#
# Set some axis strings, including the main title.
#
    if 'title' in plotdict.keys(): res1.tiMainString = plotdict['title']
    if 'xtitle' in plotdict.keys(): res1.tiXAxisString = plotdict['xtitle']
    if 'ytitle' in plotdict.keys(): res1.tiYAxisString = plotdict['ytitle']
#    res1.tiYAxisFontColor = "red"
#
# Change line color and thickness.
#
    res1.xyLineColor      = "red"
    res1.xyLineThicknessF = 4.0
    if 'ptype' in plotdict.keys():
        if plotdict['ptype'] == 'ylog':
            res1.trYAxisType = "LogAxis"

    plot1 = Ngl.xy(wks,y,x,res1)

###################################################################

def plot_scatter(x,y,plotdict):
    
    wks_type = "png"
    wks = Ngl.open_wks(wks_type,str(plotdict['filename']))    
    print('printing plot:', plotdict['filename'])

    res = Ngl.Resources()
#
# Don't maximize the plot in the frame, because we are controlling the
# size and location of both plots with the vpXXX resources.
#
    res.nglMaximize = False
    res.vpHeightF   = 0.75
    res.vpWidthF    = 0.65
    res.vpXF        = 0.20
    res.vpYF        = 0.90
    if 'xrange' in plotdict.keys():
        res.trXMinF = plotdict['xrange'][0]
        res.trXMaxF = plotdict['xrange'][1]
    else:
        res.trXMinF = numpy.min(x)
        res.trXMaxF = numpy.max(x)
    if 'yrange' in plotdict.keys():
        res.trYMinF = plotdict['yrange'][0]
        res.trYMaxF = plotdict['yrange'][1]
    else:
        res.trYMinF = numpy.min(y)
        res.trYMaxF = numpy.max(y)
#
# Set some axis strings, including the main title.
#
    if 'title' in plotdict.keys(): res.tiMainString = plotdict['title']
    if 'xtitle' in plotdict.keys(): res.tiXAxisString = plotdict['xtitle']
    if 'ytitle' in plotdict.keys(): res.tiYAxisString = plotdict['ytitle']

    if 'ptype' in plotdict.keys():
        if plotdict['ptype'] == 'ylog':
            res.trYAxisType = "LogAxis"

    res.nglFrame             = False    
    
    #Plot diagonal line
    if 'validation' in plotdict.keys():
        res.xyLineColor      = "red"
        res.xyLineThicknessF = 4.0
        x_diag = numpy.linspace(res.trXMinF,res.trXMaxF,100)     
        plot = Ngl.xy(wks,x_diag,x_diag,res)
        #Plot points
        mres = Ngl.Resources()
        mres.gsMarkerIndex = 1
        mres.gsMarkerSizeF =  0.015
        Ngl.polymarker(wks,plot,x,y,mres)
    else:
        res.xyMarkLineMode      = "Markers"
        res.xyMonoMarker = True
        res.xyMarker = 1
        res.xyMonoMarkerSize = 0.015
        plot = Ngl.xy(wks,x,y,res)

    #Add text
    if 'text' in plotdict.keys():
        txres               = Ngl.Resources()
        txres.txJust                = "CenterLeft"
        txres.txPerimOn             = False
        txres.txFontColor           = "black"
        txres.txFontHeightF = 0.019
        txres.txBackgroundFillColor = "White"
        text  = Ngl.text(wks,plot,plotdict['text'],res.trXMaxF*1./10.,res.trXMaxF*8./10.,txres)

    Ngl.frame(wks)
    Ngl.destroy(wks)
    
def plot_scatter_histo(x,y,plotdict):
    
    print('printing plot:', plotdict['filename'])
    # x and y are possibly lists of x,y pairs, to be plotted in different colors
    nullfmt   = NullFormatter()         # no labels
    left, width = 0.1, 0.65
    bottom, height = 0.1, 0.65
    bottom_h = left_h = left+width+0.02
    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom_h, width, 0.2]
    rect_histy = [left_h, bottom, 0.2, height]
    plt.figure(1, figsize=(8,8))
    axScatter = plt.axes(rect_scatter)
    axHistx = plt.axes(rect_histx)
    axHisty = plt.axes(rect_histy)
    axHistx.xaxis.set_major_formatter(nullfmt)
    axHisty.yaxis.set_major_formatter(nullfmt)
    limxmin, limxmax, limymin, limymax = 0., 0., 0., 0.
    binsx = 20
    binsy = 20
    if isinstance(x,list) is False: x = [x]
        
    for i,xx in enumerate(x):
        if isinstance(y,list): 
            yy = y[i]
        else:
            yy = y
            
        axScatter.scatter(xx, yy, s=30,c=matplotlib_colors(i))

        limxxmax = numpy.max(xx)
        limxxmin = numpy.min(xx)
        if limxxmax > limxmax: limxmax = limxxmax
        if limxxmin < limxmin: limxmin = limxxmin        
        limyymax = numpy.max(yy)
        limyymin = numpy.min(yy)
        if limyymax > limymax: limymax = limyymax
        if limyymin < limymin: limymin = limyymin
            
        axHistx.hist(xx, bins=binsx, color=matplotlib_colors(i))
        axHisty.hist(yy, bins=binsy, orientation='horizontal',color=matplotlib_colors(i))
    
    if 'xrange' in plotdict.keys():
        limx = (plotdict['xrange'][0],plotdict['xrange'][1])
    else:
        limx = (limxmin-(limxmax-limxmin)/binsx,limxmax+(limxmax-limxmin)/binsx)
    if 'yrange' in plotdict.keys():
        limy = (plotdict['yrange'][0],plotdict['yrange'][1])
    else:
        limy = (limymin-(limymax-limymin)/binsy,limymax+(limymax-limymin)/binsy)
        
    if 'xtitle' in plotdict.keys(): axScatter.set_xlabel(plotdict['xtitle'])
    if 'ytitle' in plotdict.keys(): axScatter.set_ylabel(plotdict['ytitle'])
    
    axScatter.set_xlim( limx )
    axScatter.set_ylim( limy )    
    axHistx.set_xlim( axScatter.get_xlim() )
    axHisty.set_ylim( axScatter.get_ylim() )
    
    if 'addtext' in plotdict.keys(): axScatter.text(0.80,0.87,plotdict['addtext'],transform=axScatter.transAxes)
        
    if 'addfit' in plotdict.keys():
        plotx = numpy.linspace(limxmin,limxmax,num=50)
        ploty = numpy.linspace(limymin,limymax,num=50)
        for curve in plotdict['addfit']:
            if curve['type'] is 'px': 
                fity = plotx**2*curve['coeff'][0]+plotx*curve['coeff'][1]+curve['coeff'][2]
                axScatter.plot(plotx,fity,label='y=$a$x$^2$ Res:%4.3f' % curve['res'])
            if curve['type'] is 'py': 
                fitx = ploty**2*curve['coeff'][0]+ploty*curve['coeff'][1]+curve['coeff'][2]
                axScatter.plot(fitx,ploty,label='x=$b$y$^2$ Res:%4.3f' % curve['res'])
            if curve['type'] is 'lx': 
                fity = plotx*curve['coeff'][0]+curve['coeff'][1]
                axScatter.plot(plotx,fity,label='y=$l$x Res:%4.3f' % curve['res'])
            if curve['type'] is 'ly': 
                fitx = ploty*curve['coeff'][0]+curve['coeff'][1]
                axScatter.plot(fitx,ploty,label='x=$m$y Res:%4.3f' % curve['res'])
                
        axScatter.legend(loc=2)
                
    plt.savefig(plotdict['filename'])
    plt.close()
